
/******************************************************************************
*******************************************************************************
***                                                                         ***
***                           ANY-MODULE-1.0                                ***
***                                                                         ***
***         Copyright 1997-2000 by Georgia Institute of Technology          ***
***                          Atlanta, Georgia                               ***
***                                                                         ***
***                        All Rights Reserved                              ***


     Permission   to  use, copy   and   modify   this  software  and  its 
     documentation for any research purpose and without  fee  is   hereby
     granted, provided that the above copyright  notice  appears  in  all
     copies  and  that  both  that  copyright  notice and this permission
     notice  appear in supporting documentation.  The use or inclusion of
     this   software  or  its  documentation in any commercial product or
     distribution of  this  software to any other party without specific,
     written   prior  permission  is prohibited.

     THE SOFTWARE IS PROVIDED  AS  IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.  IN  NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE  LIABLE  FOR  ANY  SPECIAL,
     INDIRECT  OR  CONSEQUENTIAL  DAMAGES   OR  ANY  DAMAGES   WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER  TORTIOUS  ACTION,  ARISING  OUT OF OR
     IN   CONNECTION  WITH  THE  USE  OR  PERFORMANCE OF  THIS  SOFTWARE.


     ====================================================================
                                                                       
     FILE:      	am_arch.h 
     INITIAL AUTHOR:    Maria Hybinette                                      

     ====================================================================

	Supported Architectures:
		SUN4SOL2
		SUNMP
		SGIMP64

***.........................................................................***
*******************************************************************************
******************************************************************************/
#ifndef AM_ARCH_H
#define AM_ARCH_H

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>

#include "am_utility.h"

/*****************************************************************************/
/***************************** GENERAL ***************************************/
/*****************************************************************************/


#define	AM_NEGINFINITY	 -FLT_MAX
#define AM_EPSILON       DBL_EPSILON

#ifndef _MAXFLOAT
#define _MAXFLOAT
#endif 

#ifndef MAXFLOAT
#define MAXFLOAT        FLT_MAX
#endif 

/*****************************************************************************/
/**************************** SUN4SOL2 ***************************************/
/*****************************************************************************/
#ifdef SUN4SOL2

#define AM_LOCK_TYPE 		mutex_t
#define AM_LOCK(lck) ;
#define AM_UNLOCK(lck) ;
#define AM_INIT_LOCK(lck) ;

#endif /* SUN4SOL2 */

/*****************************************************************************/
/**************************** SUNMP ******************************************/
/*****************************************************************************/
#ifdef SUNMP

#define AM_LOCK_TYPE 		mutex_t
#define AM_LOCK(lck)  		mutex_lock(&(lck));
#define AM_UNLOCK(lck) 		mutex_unlock(&(lck));
#define AM_INIT_LOCK(lck) 	mutex_init(&(lck), USYNC_THREAD, (void *) 0);


#endif /* SUNMP */


/*****************************************************************************/
/********************** SUNMP or SUN4SOL *************************************/
/*****************************************************************************/
#if(defined(SUNMP) || defined(SUN4SOL2))
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <synch.h>
#include <sys/time.h>
#include <thread.h>
#include <sys/resource.h>
#include <values.h>

typedef char *  AM_ADDR_TYPE;

#define AM_INIT_SYNCH
#define AM_CLOSE_SYNCH

/*
#define MAXINT          ULONG_MAX
*/

#endif /* SUNMP SUN4SOL */


/*****************************************************************************/
/**************************** SGIMP64 ****************************************/
/*****************************************************************************/
#if(defined(SGI5) || defined(SGI64) || defined(SGIMP64))

#include <ulocks.h>
#include <task.h>
#include <unistd.h>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

extern usptr_t          *am_shm_handle;
extern char             am_shared_area[80];
typedef char *  AM_ADDR_TYPE;

#define AM_LOCK_TYPE    	ulock_t
#define AM_LOCK(lck)    	ussetlock(lck);                 
#define AM_UNLOCK(lck)  	usunsetlock(lck);
#define AM_INIT_LOCK(lck) 	lck = usnewlock(am_shm_handle);  

#define AM_INIT_SYNCH			  		  \
   if ( am_shm_handle != NULL )			  	  \
	{						  \
	perror("only allocate am_shared_area  once");	  \
        perror("usinit");				  \
	}						  \
   strcpy(am_shared_area, mktemp("/tmp/am_XXXXXX")); 	  \
   if( (am_shm_handle = usinit(am_shared_area)) == NULL ) \
        {						  \
        printf("usinit");				  \
        exit(-1);					  \
        } \

#define AM_CLOSE_SYNCH			  		  \
   usdetach ( am_shm_handle ); 			  	  \
   if ( unlink ( am_shared_area ) == -1 )		  \
        {				  		  \
        perror("usinit");				  \
        exit (-1);				  	  \
        } 

#define MAXINT  INT_MAX /* ULONG_MAX*/

#endif /* SGI5 SGI64 SGIMP64 */

/*****************************************************************************/
/**************************** LINUX ******************************************/
/*****************************************************************************/
#if(defined(LINUX))

#include <pthread.h>

#define AM_LOCK_TYPE    	pthread_mutex_t *
#define AM_LOCK(lck)    	pthread_mutex_lock(lck);                 
#define AM_UNLOCK(lck)  	pthread_mutex_unlock(lck);
#define AM_INIT_LOCK(lck) 	pthread_mutex_init(lck,NULL);  
#define AM_INIT_SYNCH
#define AM_CLOSE_SYNCH

typedef char *  AM_ADDR_TYPE;

/* MAXINT is already defined */

#endif /* LINUX */


/*---------------------------------------------------------------------------*/
#endif /* AM_ARCH_H */

