/******************************************************************************
*******************************************************************************
***                                                                         ***
***                           CLONE-SIM-1.0                                 ***
***                                                                         ***
***         Copyright 1997-2000 by Georgia Institute of Technology          ***
***                          Atlanta, Georgia                               ***
***                                                                         ***
***                        All Rights Reserved                              ***


     Permission   to  use, copy   and   modify   this  software  and  its 
     documentation for any research purpose and without  fee  is   hereby
     granted, provided that the above copyright  notice  appears  in  all
     copies  and  that  both  that  copyright  notice and this permission
     notice  appear in supporting documentation.  The use or inclusion of
     this   software  or  its  documentation in any commercial product or
     distribution of  this  software to any other party without specific,
     written   prior  permission  is prohibited.

     THE SOFTWARE IS PROVIDED  AS  IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.  IN  NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE  LIABLE  FOR  ANY  SPECIAL,
     INDIRECT  OR  CONSEQUENTIAL  DAMAGES   OR  ANY  DAMAGES   WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER  TORTIOUS  ACTION,  ARISING  OUT OF OR
     IN   CONNECTION  WITH  THE  USE  OR  PERFORMANCE OF  THIS  SOFTWARE.


     ====================================================================
                                                                       
     FILE:      	cs_api.c 
     INITIAL AUTHOR:    Maria Hybinette                                      

     ====================================================================

     DESCRIPTION:
	- no dependencies on particular simulator

***.........................................................................***
*******************************************************************************
******************************************************************************/

#include "am_arch.h"
#include "cs_internal.h"
#include "is_kernel.h"



/* ---------------------------------------------------------------------------
** CloneSim_Create( )
**
** DESCRIPTION:  New API, this function is called from a blocking
**	conservative event. This function (1) creates a new simulation (if
**	the LP that envokes it is the first LP) and (2) clones an LP.
**
** RETURNS:
**	the identification number of the newly clones simulation
**
** NEED:
**	A copy LP function from kernel, or application?
**
**  int *LPs, 
**  int num_lps_in_set,  
**---------------------------------------------------------------------------*/
int CloneSim_Create( 
	int uid, 		/* parent LP that clones (unique ID) */
	double current_sim_time
	)
{
  char 	fn[]="[CloneSim_Create]";
  int	cloneID=0;
  int	to;

  if( cs_do_debug > 0 )
  	printf("%s: (LP=%d, ts=%lf) begin*****\n", fn, uid, current_sim_time );

  cloneID = CS_SimCreate( uid, current_sim_time );

  if( cloneID < 0 ) return cloneID; /* MH_HERE */

  to = cs_GIDtoID[cs_IDtoGID[uid]].clone_id[cloneID];
  IS_CopyLPState( to, uid );

  if( cs_do_debug > 0 )
  	printf("*cloning: %s child_uid=%d parent=%d clone=%d at %lf\n", 
		fn, to, uid, cs_IDtoClone[to], current_sim_time );
  return cloneID;
}



/* ---------------------------------------------------------------------------
** CloneSim_CreateOptimistic( )
**
** DESCRIPTION:  
**	This function is similar to above exept that the incremental update
**	scheme is cloned optimistically instead of conservatively.
**
**
**---------------------------------------------------------------------------*/
int CloneSim_CreateOptimistic( 
	int uid, 
	double current_sim_time 
	)
{
  char 	fn[]="[CloneSim_CreateOptimistic]";
  int	cloneID=0;
  int	to;

  if( cs_do_debug > 0 )
  	printf("%s: (LP=%d, ts=%lf) begin*****\n", fn, uid, current_sim_time );

  cs_optimistic = 1;
  cloneID = CloneSim_Create( uid, current_sim_time );

  if( cs_do_debug > 0 )
  	printf("*cloning: %s child_uid=%d parent=%d clone=%d at %lf\n", 
		fn, to, uid, cs_IDtoClone[to], current_sim_time );

  return cloneID;
}



/* ---------------------------------------------------------------------------
** CloneSim_Delete ()
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
int CloneSim_Delete( 
	int cloneID,
	double startTS,		/* start time to monitore a trigger */
	double endTS,		/* end   time to monitore a trigger */
	CSFunc_p TriggerToPrune
	)
{
  char 	fn[]="[CloneSim_Delete]";
  printf( "not done\n" );
  return cloneID;
}




/* ---------------------------------------------------------------------------
** CloneSim_GetCloneID ()
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
int CloneSim_GetCloneID( int LP )
{
  char 	fn[]="[CloneSim_GetCloneID]";
  return cs_IDtoClone[LP];
}




/* ---------------------------------------------------------------------------
** CloneSim_GlobalMe ()
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
int CloneSim_GID( int ULP )
{
  char fn[]="[CloneSim_GlobalMe]";
  int GID;

  return cs_IDtoGID[ULP];
}



#if 0 /* implemented in Interactive-Sim as a macro */

/* ---------------------------------------------------------------------------
** CloneSim_UniqueMe ()
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
int CloneSim_UID( void )
{
  char 	fn[]="[CloneSim_UniqueMe]";
  int 	UID;

  return UID;
}
#endif



/* ---------------------------------------------------------------------------
** CloneSim_GetMsg ()
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void CloneSim_GetMsg( int src, int dst, char *Msg, double ts, double rcv_ts )
{
  AM_ADDR_TYPE 		msg_begin_ptr;
  char			fn[]		= "[CloneSim_GetMsg]";
  struct MsgData 	*my_msg 	= (struct MsgData *) Msg;


  if( cs_do_debug > 10 )
  	printf("%s: (frm=%d, to=%d) sndTS=%lf rcvTS=%lf\n", 
		fn, src, dst, ts, rcv_ts );

  if( dst > cs_vlp )
            {
            AM_Error("** %s exiting invalid LP=%d\n", fn, dst );
            }

  if( do_cloning == 0 ) 
	return;

  CS_PiggyBackMSG(&(my_msg->Admn_Data), src, dst, ts );

  IS_SetMsg( src, dst );  			/* sets pointer */

  if( cs_do_debug > 100 )
  	printf("%s: (LP=%d) end (%d)\n", fn, src, CSMsgSize );
}



/* ---------------------------------------------------------------------------
** CloneSim_SetMsg ()
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
int CloneSim_SetMsg( int source, int destination, char *Msg )
{
  char 	fn[]="[CloneSim_SetMsg]";

  return 1;
}




/* ---------------------------------------------------------------------------
** CloneSim_InitAppl ()
**
** DESCRIPTION: 
**	called AFTER detetermining number of LPs and its mapping
**
**---------------------------------------------------------------------------*/
void CloneSim_InitAppl( int argc, char ** argv )
{
  char 	fn[]="[CloneSim_InitAppl]";

  /* printf("%s: begin\n", fn ); */

  /* commandline processing */
  CS_GetCommandLine(argc,argv);

  if( do_cloning == 0 )
	{
	cs_InitLockUpTables();
	return;
	}

  /* XXX must be called before InitKernel */
  cs_InitCSMessage();

  /* XXXX set number of Clones */

  /* intialize Interactive-Sim */
  IS_InitKernel();

  /* initialize Cloning-Sim */
  CS_InitCloning();

  /* printf("%s: done\n", fn ); */
}



/* ---------------------------------------------------------------------------
** CloneSim_CloseAppl ()
**
** DESCRIPTION: 
**	called AFTER detetermining number of LPs and its mapping
**
**---------------------------------------------------------------------------*/
void CloneSim_CloseAppl( void )
  {
  if( do_cloning == 0 )
	return;
  CS_CloseCloning();
  IS_CloseAppl();
  }

