
/******************************************************************************
*******************************************************************************
***                                                                         ***
***                           CLONE-SIM-1.0                                 ***
***                                                                         ***
***         Copyright 1997-2000 by Georgia Institute of Technology          ***
***                          Atlanta, Georgia                               ***
***                                                                         ***
***                        All Rights Reserved                              ***


     Permission   to  use, copy   and   modify   this  software  and  its 
     documentation for any research purpose and without  fee  is   hereby
     granted, provided that the above copyright  notice  appears  in  all
     copies  and  that  both  that  copyright  notice and this permission
     notice  appear in supporting documentation.  The use or inclusion of
     this   software  or  its  documentation in any commercial product or
     distribution of  this  software to any other party without specific,
     written   prior  permission  is prohibited.

     THE SOFTWARE IS PROVIDED  AS  IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.  IN  NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE  LIABLE  FOR  ANY  SPECIAL,
     INDIRECT  OR  CONSEQUENTIAL  DAMAGES   OR  ANY  DAMAGES   WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER  TORTIOUS  ACTION,  ARISING  OUT OF OR
     IN   CONNECTION  WITH  THE  USE  OR  PERFORMANCE OF  THIS  SOFTWARE.


     ====================================================================
                                                                       
     FILE:      	cs_internal.c
     INITIAL AUTHOR:    Maria Hybinette                                      

     ====================================================================

     DESCRIPTION:	data structure manipulatation


***.........................................................................***
*******************************************************************************
******************************************************************************/

#include "am_arch.h"
#include "cs_internal_t.h"

cs_rmap_t                *cs_clones;
cs_idmap_t               *cs_GIDtoID;           
cs_vlpmap_t             **cs_vlpmapping;

int                      *cs_IDtoGID;
int                      *cs_IDtoClone;
int                       CSMsgSize=0;

int			  cs_vlp=0;
int			  cs_nlp=0;

int                	  cs_clone_cnt=0;
int                	  cs_un_clone_cnt=0;
AM_LOCK_TYPE           	  cs_clone_cnt_lck;

int           		  cs_num_clones=1;

int           		  cs_optimistic=2;
int			  cs_do_checkpointing=0;

/* command line */
int                	  do_cloning=1;
int                	  cs_do_debug=-1000;
int			  cs_clone_parent=1;


/* moving from GTW */
double 			  CloneSim_CLONETIME[CS_MAX_CLONES];  
double 			  CloneSim_CLONE_RT[CS_MAX_CLONES];
int			  CloneSim_NumClones=1;
int			  CloneSim_setVersions=0;
int                       clone_cnt=0;

double			  CloneSim_CloningPoint=MAXFLOAT;
double			  CloneSim_CloningSpread=MAXFLOAT;

/* command line: DPAT -- need to remove */
int                       clone_lp=0;
double             	  do_preschedule=0.0;
double                    MITconfigPercent=0.0;
int                       Sisforked=0;
double                    BAcloning_pt1=0;
double                    do_prune=0.0;



/* ---------------------------------------------------------------------------
** cs_PrintBits()
**
** DESCRIPTION:
**
**---------------------------------------------------------------------------*/
void cs_PrintBits( unsigned long int map, char *instring, int num,
                int num_clones)
{
   char map_string[100];
   int i;

   for ( i = 0; i < num_clones; i ++ )
        {
        if ( get_bits( map,i,1) & (01) )
                map_string[i] = '1';
        else
                map_string[i] = '0';
        }
   map_string[i] = '\0';
   if ( num )
      printf ("           : %s\n", "012345678901234567890123456789012345678");
   printf ("       %s: %s\n", instring, map_string );
}



/* ---------------------------------------------------------------------------
** CS_SetNumClones ()
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
int CS_SetNumClones( int new_num_clones )
  {
  char fn[]	= "[CS_SetNumClones]";

  cs_num_clones=new_num_clones;

  if ( cs_do_debug > 10 )
  	printf("%s: done %d\n", fn, cs_num_clones );
  return cs_num_clones;
  }



/* ---------------------------------------------------------------------------
** CS_SetVLPs ()
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
int CS_SetVLPs( int new_vlps )
  {
  char fn[]	= "[CS_SetVLPs]";

  cs_vlp = new_vlps;
 
  if ( cs_do_debug > 10 )
  	printf("%s: done (%d)\n", fn, cs_vlp );

  return cs_vlp;
  }



/* ---------------------------------------------------------------------------
** CS_SetNLPs ()
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
int CS_SetNLPs( int new_nlps )
  {
  char fn[]	= "[CS_SetNLPs]";

  cs_nlp = new_nlps;

  if ( cs_do_debug > 10 )
 	printf("%s: done (%d)\n", fn, cs_nlp );

  return cs_nlp;
  }



/* ---------------------------------------------------------------------------
** CS_PiggyBackMSG ()
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void CS_PiggyBackMSG( struct CSMsgData *admn_data, int src, int dst, double tss )
  {
  int cloneID;
  int gid;
  char fn[]	= "[CS_PiggyBackMSG]";


  cloneID = cs_IDtoClone[src];
  gid 	  = cs_IDtoGID[src];

  if ( cs_do_debug > 100 )
  	 printf("%s: begin (snd=%d, rcv=%d cloneID=%d) at %lf\n", fn, 
		gid,
		cs_IDtoGID[dst],
		cloneID,
		tss
		);

  admn_data->type 	   = MSG_USER;
  admn_data->a_forward_msg = 0;

  /* original sender mapping */
  admn_data->snd_unmapped   = cs_vlpmapping[cloneID][gid].unmapped;
  admn_data->snd_siblings   = cs_vlpmapping[cloneID][gid].siblings;
  admn_data->forwarded 	    = 0L;

  admn_data->snd_uid   	   = src;
  admn_data->rcv_uid   	   = dst;
  admn_data->rcv_gid   	   = cs_IDtoGID[dst];
  admn_data->cloneID       = cloneID;
  admn_data->snd_time      = tss;

  if ( cs_do_debug > 100 )
  	 printf("%s: done\n", fn );
  }



/* ---------------------------------------------------------------------------
** CS_GetCommandLine ()
**
** DESCRIPTION: 
**	this assumes getopt was not used previously (i.e. brook the command
**	line arguments in the application)
**
**---------------------------------------------------------------------------*/
void CS_GetCommandLine( int argc, char **argv )
  {
  char 		fn[]= "[cs_GetCommandLine]";
  extern char 	*optarg;
  extern int  	optind;
  int 		ch;
  int 		error=-0;
  int		j;

  while ( ( ch = getopt ( argc, argv, "Cc:d:j:M:o:P:p:S:t:V:x:y:z:" ) ) != EOF )
        {
        switch ( ch )
                {
                case 'C':
                        {
                        cs_clone_parent = 0;
			/* printf("%s: cs_clone_parent is false\n", fn ); */
                        break;
                        }
                case 'c':
                        {
                        CloneSim_CLONETIME[clone_cnt] = atof( optarg );
                        clone_cnt++;
                        CloneSim_NumClones = 1 + clone_cnt;
                        break;
                        }
                case 'd':
			{
			cs_do_debug = atoi(optarg);
			/* printf("%s: cs_do_debug %d\n", fn, cs_do_debug ); */
                        break;
			}
                case 'j':
                        {
                        clone_lp = atoi( optarg );
                        break;
                        }
                case 'M':
                        {
                        MITconfigPercent = atof( optarg );
                        break;
                        }
                case 'o':
                        {
                        cs_optimistic =  atoi(optarg);
			/* printf("%s: cs_optimistic %d\n",fn,cs_optimistic);*/
                        break;
			}
                case 'P':
                        {
                        do_preschedule = atof( optarg );
                        break;
                        }
                case 'p':
                        {
                        do_prune = atof( optarg );
                        break;
                        }
                case 'S':
                        {
                        Sisforked = atoi( optarg );
                        break;
                        }
                case 't':
                        {
                        BAcloning_pt1 = atof( optarg );
                        break;
                        }

                case 'V':
                        {
			int j;
			CloneSim_setVersions=1;
                        CloneSim_NumClones = 1 + atoi( optarg );
			for( j=0; j < CloneSim_NumClones; j++ )
				CloneSim_CLONETIME[j] = MAXFLOAT; 
                        break;
                        }

                case 'x':
                        {
                        cs_do_checkpointing  =  atoi(optarg);
			/* printf("%s: cs_do_checkpointing is %s\n", 
				fn, 
				(cs_do_checkpointing) ? "true" : "false"
				);
			*/
                        break;
			}
                case 'y':
                        {
                        CloneSim_CloningPoint = atof( optarg );
                        break;
                        }
                case 'z':
                        {
                        CloneSim_CloningSpread = atof( optarg );
                        break;
                        }
                default:
                        {
                        fprintf ( stderr, "%s: unrecognized option %s\n",
                                argv[0], ch );
                        break ;
                        }
                } /* switch */
        } /* while */

  	/* printf("%s: done\n", fn ); */

  if( (CloneSim_CloningPoint == MAXFLOAT) || (CloneSim_CloningSpread == MAXFLOAT) ) 
     	CloneSim_setVersions = 0;
  else
     	CloneSim_setVersions = 1;
  if( CloneSim_setVersions == 1 )
      	{
      	CloneSim_CLONETIME[0] = CloneSim_CloningPoint;
      	for( j=1; j < (CloneSim_NumClones-1); j++ )
		{
		CloneSim_CLONETIME[j] 
			= CloneSim_CLONETIME[j-1]+CloneSim_CloningSpread;
		}
      	}

  }


/* ---------------------------------------------------------------------------
** cs_InitSynchVariables ()
**
** DESCRIPTION:  
**
**---------------------------------------------------------------------------*/
void cs_InitSynchVariables(void)
{
  char 	fn[]	= "[cs_InitSynchVariables]";

  if ( cs_do_debug > 10 )
  	printf("%s: begin\n", fn );

   AM_INIT_SYNCH;

   AM_INIT_LOCK(cs_clone_cnt_lck);

   if ( cs_do_debug > 10 )
  	printf("%s: done\n", fn );
}


/* ---------------------------------------------------------------------------
** cs_CloseSynchVariables ()
**
** DESCRIPTION:  
**
**---------------------------------------------------------------------------*/
void cs_CloseSynchVariables(void)
{
  char 	fn[]	= "[cs_CloseSynchVariables]";

  AM_CLOSE_SYNCH;

  if ( cs_do_debug > 10 )
  	printf("%s: done\n", fn );
}



/* ---------------------------------------------------------------------------
** cs_UpdateCloneTable()
**
** DESCRIPTION:  
**	This table is the relation ship-table, modification to this
**	data structure must be synchronized
**
**---------------------------------------------------------------------------*/
int  cs_UpdateCloneTable( int parent_cloneID, double birth_ts )
  {
  int 			child_cloneID; 
  cs_rmap_t *		current		= NULL;
  cs_rmap_t *		i		= NULL;
  unsigned long int 	bits		= 0L;
  char  		fn[]		= "[cs_UpdateCloneTable]";

  if ( cs_do_debug > 0 )
  	printf("%s: begin\n", fn );

  AM_LOCK(cs_clone_cnt_lck);

  if ( cs_clones[cs_clone_cnt+1].CloneID != cs_clone_cnt + 1 )
	{
	AM_Error ("DoCloning and CloneID mismatch: %d != %d\n", 
		(cs_clone_cnt+1), (cs_clones[cs_clone_cnt+1].CloneID));
	}

  cs_clone_cnt++;
  cs_un_clone_cnt++;
  child_cloneID = cs_clone_cnt;

  cs_clones[child_cloneID].BirthTS = birth_ts;
  cs_clones[child_cloneID].parent  = &cs_clones[parent_cloneID];

  /* update un-mapping of child "up" the ancestor chain */
  bits = 0L;
  set_bit(bits,child_cloneID);

  for ( current = &cs_clones[child_cloneID] ; 
	(current != NULL) && ((current->CloneID) !=0) ;
	current = current->parent )
	{
        current->children |= bits; /* cloned children */
        }
  cs_clones[0].children |= bits;

  /* set lchild and child mapping from parent clone (lchild=last_child) */
  if ( cs_clones[parent_cloneID].child != NULL )
	cs_clones[parent_cloneID].lchild->sibling = &cs_clones[child_cloneID];
  else
 	cs_clones[parent_cloneID].child = &cs_clones[child_cloneID];
  cs_clones[parent_cloneID].lchild = &cs_clones[child_cloneID];

  AM_UNLOCK(cs_clone_cnt_lck);

  if ( cs_do_debug > 0 )
  	printf("%s: done\n", fn );

  return child_cloneID;
  }



/* ---------------------------------------------------------------------------
** cs_InitCloneTable ()
**
** DESCRIPTION:  
**	This table is the relation ship-table, modification to this
**	data structure must be synchronized
**
**---------------------------------------------------------------------------*/
void  cs_InitCloneTable(void)
{
   int 			i; 
   unsigned long  int	ffmask	= 0L;
   char  		fn[]	= "[cs_InitCloneTable]";

  if ( cs_do_debug > 10 )
  	printf("%s: done\n", fn );

   /* allocate cloning table */
   cs_clones = (cs_rmap_t *) calloc ( cs_num_clones, sizeof (cs_rmap_t) );

   /* original "parent" is initially born, but the children are not */
   cs_clones[0].CloneID = 0;
   cs_clones[0].BirthTS = AM_NEGINFINITY;
   cs_clones[0].parent = &cs_clones[0];

   cs_clones[0].children = 0L;
   set_bit(cs_clones[0].children,0); 		/* turn on my own bit */

   cs_clones[0].mymask = 0L;
   set_bit(cs_clones[0].mymask,0);
   cs_clones[0].mask = cs_clones[0].mymask;

   ffmask=0L;
   set_bit(ffmask,0);
   cs_clones[0].ffmask = ffmask;

   for ( i = 1; i < cs_num_clones; i ++ )
        {
        cs_clones[i].CloneID = i;
        cs_clones[i].BirthTS = MAXFLOAT;

        /* calloc 2*N bit arrays, N = NumClones/sizeof(long) + 1  */
        /* that keep tracks of all its children */

	/* set all pointers */
        cs_clones[i].parent = NULL;
        cs_clones[i].child = NULL;
        cs_clones[i].lchild = NULL;
        cs_clones[i].sibling = NULL;

	/* set all masks */
        cs_clones[i].children = 0L;

        set_bit(ffmask,i);
        cs_clones[i].ffmask = ffmask;

        cs_clones[i].mymask = 0L;
        set_bit(cs_clones[i].mymask,i);
        cs_clones[i].mask = cs_clones[i].mymask;
        }

  if ( cs_do_debug > 10 )
  	printf("%s: done\n", fn );
}


/* ---------------------------------------------------------------------------
** cs_AllocateVLPmappingTable ()
**
** DESCRIPTION:  
**
**---------------------------------------------------------------------------*/
void cs_AllocateVLPmappingTable( int num_clones, int num_lps )
  {
  int 	i;
  char  fn[]	= "[cs_AllocateVLPmappingTable]";


  /* allocate pointers to rows (indexed by cloneID) */
  cs_vlpmapping = (cs_vlpmap_t **) calloc(num_clones, sizeof(cs_vlpmap_t *));
  if ( cs_vlpmapping == NULL )
	{
	printf( "calloc error (%s)\n", fn );
	fflush(stdout);
	exit(1);
	}

  /* allocate whole rows (indexed by LP) */
  for ( i = 0; i < num_clones; i ++ )
	{
	cs_vlpmapping[i] = (cs_vlpmap_t *) calloc(num_lps, sizeof(cs_vlpmap_t));
	if ( cs_vlpmapping[i] == NULL )
		{
		printf("calloc error (%s)\n", fn);
		fflush(stdout);
		exit(1);
		}
	}

  if ( cs_do_debug > 10 )
  	printf("%s: done\n", fn );
  }


/* ---------------------------------------------------------------------------
** cs_AddCloneToVLPmapping ()
**
** DESCRIPTION:  
**
**---------------------------------------------------------------------------*/
void cs_AddCloneToVLPmapping( int prt_uid, int chd_uid, double birth_time, int active )
  {
  int 	gid;
  int	parent_cloneID; /* parent */
  int	chd_cloneID; 	/* child */
  char	fn[]		= "[cs_AddCloneToVLPmapping]";

  if ( cs_do_debug > 0 )
  	printf("%s: begin\n", fn );

  parent_cloneID = cs_IDtoClone[prt_uid];
  chd_cloneID    = cs_IDtoClone[chd_uid];
  gid		 = cs_IDtoGID[prt_uid];
 
  /* for consistency checking */
  cs_vlpmapping[parent_cloneID][gid].CurrChildBirthTS 	= birth_time;

  /* parent updates its own state to exclude cloned child + siblings */
  cs_vlpmapping[parent_cloneID][gid].unmapped |= cs_clones[chd_cloneID].mask;

  /* CHILD: vlp update (PInitVLPs) */
  cs_vlpmapping[chd_cloneID][gid].active 		= active;
  cs_vlpmapping[chd_cloneID][gid].BirthTS		= birth_time;
  cs_vlpmapping[chd_cloneID][gid].CurrChildBirthTS 	= AM_NEGINFINITY;
  cs_vlpmapping[chd_cloneID][gid].parentID 		= prt_uid;
  cs_vlpmapping[chd_cloneID][gid].parentCloneID 	= parent_cloneID;
  cs_vlpmapping[chd_cloneID][gid].siblings 		= 0L;
  cs_vlpmapping[chd_cloneID][gid].unmapped 		= 0L;

  if ( cs_do_debug > 0 )
  	printf("%s: end\n", fn );
  }


/* ---------------------------------------------------------------------------
** cs_InitVLPmapping ()
**
** DESCRIPTION:  
**
**---------------------------------------------------------------------------*/
void cs_InitVLPmapping( void )
  {
  int  	i,j;
  int  	vlpID		= 0;
  int 	initial_cloneID = 0;
  char  fn[]		= "[cs_InitVLPmapping]";

  cs_AllocateVLPmappingTable(cs_num_clones, cs_nlp);

  vlpID=0;
  for ( i = 0; i < cs_num_clones; i++ )
        {
        for ( j = 0; j < cs_nlp; j++ )
                {
                cs_vlpmapping[i][j].active  		= 0;
                cs_vlpmapping[i][j].BirthTS 		= MAXFLOAT;
                cs_vlpmapping[i][j].CurrChildBirthTS 	= AM_NEGINFINITY;

                cs_vlpmapping[i][j].parentID 		= -1;
                cs_vlpmapping[i][j].parentCloneID 	= -1;
                cs_vlpmapping[i][j].MyID 		= vlpID;
                vlpID++;

                cs_vlpmapping[i][j].siblings 		= 0L;
                cs_vlpmapping[i][j].unmapped 		= 0L;
                cs_vlpmapping[i][j].copy_request_src	= 0L;
                cs_vlpmapping[i][j].copy_request_time	= 0.0;
                }
        }

   /* activate original clone */
   initial_cloneID = 0;
   for ( i = 0; i < cs_nlp; i++ )
        {
        cs_vlpmapping[initial_cloneID][i].active     = 1;
        cs_vlpmapping[initial_cloneID][i].BirthTS    = AM_NEGINFINITY;
        }

  if ( cs_do_debug > -10 )
  	printf("%s: done\n", fn );
  }



/* ---------------------------------------------------------------------------
** cs_InitLockUpTables ()
**
** DESCRIPTION:  
**
**---------------------------------------------------------------------------*/
void cs_InitLockUpTables( void )
  {
  int 	i;
  int 	GID	= 0;
  int 	cloneID	= 0;
  char  fn[]	= "[cs_InitLockUpTables]";
  int  	*t;

  if ( cs_do_debug > 10 )
  	printf("%s: begin\n", fn );

  cs_IDtoGID 	= (int *) calloc ( cs_vlp, sizeof(int) );
  cs_IDtoClone 	= (int *) calloc ( cs_vlp, sizeof(int) );
  cs_GIDtoID 	= (cs_idmap_t *) calloc ( cs_nlp, sizeof(cs_idmap_t) );

  /* i is the unique ID */
  for (i = 0 ; i < cs_vlp; i++ )
	{
        GID 	= i % (cs_nlp);
        cloneID = i / (cs_nlp);

	cs_IDtoGID[i]   = GID;
        cs_IDtoClone[i] = cloneID;
        cs_GIDtoID[GID].clone_id[cloneID] = i;
	}

  if ( cs_do_debug > 10 )
  	printf("%s: done\n", fn );
  }



/* ---------------------------------------------------------------------------
** CS_CloseCloning ()
**
** DESCRIPTION:  
**
**---------------------------------------------------------------------------*/
void CS_CloseCloning( void )
  {
  char  fn[]	= "[CS_CloseCloning]";

  cs_CloseSynchVariables();

  if ( cs_do_debug > 10 )
  	printf("%s: done\n", fn );
 }



/* ---------------------------------------------------------------------------
** CS_SimCreate ()
**
** DESCRIPTION:  
**
**---------------------------------------------------------------------------*/
int CS_SimCreate( int uid, double birth_ts )
  {
  int child_cloneID	= 0;
  int child_uid		= 0;
  int gid		= 0;
  int parent_cloneID	= 0;
  char  fn[]		= "[CS_SimCreate]";
  int activate		= 1;

  if ( cs_do_debug > 0 )
  	printf("%s: begin\n", fn );

  parent_cloneID = cs_IDtoClone[uid];
  gid		 = cs_IDtoGID[uid];
  
  if ( cs_vlpmapping[parent_cloneID][gid].CurrChildBirthTS == birth_ts )
        {
  	if( cs_do_debug >= 0 )
		{
       	 	printf ("+**error: can only clone once within the same LP event\n");
        	printf ("+**ignoring request\n");
		}
        return -1;
        }

  if ( cs_clone_cnt >= (cs_num_clones-1) )
        {
  	if( cs_do_debug >= 0 )
		{
        	printf ("**can only make %d clones, have %d clones\n", 
			cs_num_clones, cs_clone_cnt+1 );
        	printf ("**ignoring request\n");
		}
        return -1;
        }

  /* update cloning table */
  child_cloneID = cs_UpdateCloneTable( parent_cloneID, birth_ts );

  /* update vlp mapping */
  child_uid = cs_GIDtoID[gid].clone_id[child_cloneID];
  activate = 1;
  cs_AddCloneToVLPmapping(uid, child_uid, birth_ts, activate );

  if ( cs_do_debug > 0 )
  	printf("%s: done (%d) \n", fn, child_cloneID );
  return child_cloneID;
  }



/* ---------------------------------------------------------------------------
** CS_GetMsgType ()
**
** DESCRIPTION:  
**
**---------------------------------------------------------------------------*/
/*
THIS IS A MACRO
int CS_GetMsgType( struct MsgData *CSMsg )
  {
  char  fn[]	= "[CS_GetMsgType]";

  if ( cs_do_debug > 10 )
  	printf("%s: done\n", fn );
  }
*/



/* ---------------------------------------------------------------------------
** CS_AssertMsgValidity ()
**
** DESCRIPTION:  
**
**---------------------------------------------------------------------------*/
int CS_AssertMsgValidity( struct MsgData *CSMsg, double time_now )
  {
  char  fn[]	= "[CS_AssertMsgValidity]";
  int	cloneID;
  int 	gid;
  int	uid;

  if ( cs_do_debug > 100 ) 
	printf("%s: begin\n", fn );

  uid 		= CSMsg->Admn_Data.rcv_uid;
  gid 		= cs_IDtoGID[uid];
  cloneID 	= cs_IDtoClone[uid];

#if 0
  if ( cs_vlpmapping[cloneID][gid].BirthTS == MAXFLOAT )
	AM_Error( "+** %s: (uid=%d, gid=%d, clone=%d) not born at %4.2lf\n", 
		fn, uid, gid, cloneID, time_now );
  if ( cs_vlpmapping[cloneID][gid].BirthTS > time_now )
	AM_Error( "+**  BirthTS > now: %.4f > %.4f\n (%d, %d, %d))",
		cs_vlpmapping[cloneID][gid].BirthTS, time_now,
		 uid, gid, cloneID
		 );
  if ( cs_vlpmapping[cloneID][gid].active == 0 )
	AM_Error( "not active\n" );
#endif
  if (
       (CSMsg->Admn_Data.a_forward_msg == 1)
       && (CSMsg->Admn_Data.forwarded & cs_clones[cloneID].mask)
       && (uid != CSMsg->Admn_Data.snd_uid)
     )
	{
	printf( "+***** doubly received (%d) %.4f\n", uid, time_now ) ;
        return -1;
	}

  if ( cs_do_debug > 100 ) 
	printf("%s: end\n", fn );

  return 1;
  }


/* ---------------------------------------------------------------------------
** CS_GetClone_t ()
**
** DESCRIPTION:  
**
**---------------------------------------------------------------------------*/
cs_clone_t CS_GetClone_t( struct MsgData *CSMsg, int gid, int uid, int cloneID )
  {
  char  		fn[]	= "[CS_GetClone_t]";
  cs_clone_t  		vmap 	= {0L, 0L, 0L, 0L,  0, 0, 0, 0, 0, 0, 0};
  int			snd_uid, rcv_uid;
  int			snd_cloneID, rcv_cloneID;
  unsigned long int 	snd_map, rcv_map;
  int                   i, start;
  int                   num_cloned;
  unsigned long int     rcv_unmapped    = 0L;
  unsigned long int     snd_unmapped    = 0L;
  int			XX_loop_1=0;
  int			XX_loop_2=0;


  /* MH_HERE make sure that the rcv_uid makes sense: recall that for birth event
   * the myID is set to to the Sender in the "old" SCHEME. */

  if ( cs_do_debug > 100 )
  	printf("%s: begin\n", fn );

  /* rcv_uid 	= CSMsg->Admn_Data.rcv_uid; */
  rcv_uid 	=  uid;
  rcv_cloneID 	=  cs_IDtoClone[rcv_uid];

  snd_uid 	=  CSMsg->Admn_Data.snd_uid;
  snd_cloneID 	=  cs_IDtoClone[snd_uid];

  /* for each un-mapped LP, also unmap all their children according to cs_clones */
  num_cloned      = cs_num_clones;
  snd_unmapped    = CSMsg->Admn_Data.snd_unmapped;
  rcv_unmapped    = cs_vlpmapping[rcv_cloneID][gid].unmapped;

  /* could traverse the cs_clones instead */
  /* safer to do separate loops first, be careful when */
  /* when joining loops, and the start condition suppose it is 0 for instance */
  start = snd_cloneID + 1;
  for( i = start;  (i < cs_clone_cnt) ; i ++ ) 	/* num_cloned */
        {
        if( snd_unmapped & cs_clones[i].mymask )
                snd_unmapped |= cs_clones[i].children;
	XX_loop_1++;
        }

  start = rcv_cloneID + 1;
  for( i = start;  (i < cs_clone_cnt) ; i ++ )
        {
        if( rcv_unmapped & cs_clones[i].mymask )
                rcv_unmapped |= cs_clones[i].children;
	XX_loop_2++;
        }

  snd_map  	=  cs_clones[snd_cloneID].children;
  snd_map 	&= ~(snd_unmapped);
  snd_map 	|= CSMsg->Admn_Data.snd_siblings;

  rcv_map  	=  cs_clones[rcv_cloneID].children;
  rcv_map 	&= ~(rcv_unmapped);
  rcv_map 	|= cs_vlpmapping[rcv_cloneID][gid].siblings;

  vmap.snd_mask 	= snd_map;
  vmap.snd_uid 		= snd_uid;
  vmap.cloneID		= snd_cloneID;

  vmap.rcv_mask 	= rcv_map;
  vmap.rcv_uid 		= rcv_uid;
  vmap.rcv_gid		= gid;

  vmap.a_forward_msg	= CSMsg->Admn_Data.a_forward_msg;
  vmap.done_forwarded	= CSMsg->Admn_Data.forwarded;
  vmap.snd_time		= CSMsg->Admn_Data.snd_time;

  
  cs_CNTR_SAMPLE(gid,CS_SYS_GETCLONE,26,XX_loop_1,XX_loop_2);

  return vmap;
  }


/* ---------------------------------------------------------------------------
** CS_SetForwardMsg ()
**
** DESCRIPTION:  
**
**---------------------------------------------------------------------------*/
/* this is a MACRO
void CS_SetForwardMsg( struct MsgData *to, unsigned long int token )
  {
  char  fn[]	= "[CS_SetForwardMsg]";
  }
*/



/* ---------------------------------------------------------------------------
** CS_GetLPForward ()
**
** DESCRIPTION:  
**
**---------------------------------------------------------------------------*/
int CS_GetLPForward( cs_clone_t vmap, unsigned long int *token )
  {
  char  		fn[]		= "[CS_GetLPForward]";
  unsigned long int	done_forward	= 0L;	
  unsigned long int	forward		= 0L;	
  int 			uid		= -1;
  int 			lowest_cloneID	= -1;
   unsigned long int    ffmask;

  if ( cs_do_debug > 100 )
  	printf("%s: begin\n", fn );

  /* forward to the one in senders map but not already rcvd */
  forward = (vmap.snd_mask) & (~(vmap.rcv_mask));

  /* daisy chaining msg sends among clones -- mask out lower clone ID's */
  ffmask = cs_clones[ cs_IDtoClone[vmap.rcv_uid] ].ffmask;
  forward &= ~(cs_clones[ cs_IDtoClone[vmap.rcv_uid] ].ffmask); 

  /* daisy chaining only */
  if ( vmap.a_forward_msg == 1 )
	{
	forward		&= ~(vmap.done_forwarded);
	done_forward	=   (vmap.done_forwarded);
	}

  /* if this is 0 then there are none to foward to */
  if ( forward <= 0 )
	{
	cs_CNTR_SAMPLE(vmap.rcv_gid,CS_SYS_GETLPFWD,8,0,0);
	return -1;
	}

  done_forward |= vmap.rcv_mask;

  /* set the token */
  *token = done_forward;

  /* determine who to forward to next */
  get_lowest(lowest_cloneID, forward, cs_num_clones); 
  uid = cs_GIDtoID[cs_IDtoGID[vmap.rcv_uid]].clone_id[lowest_cloneID];

  cs_CNTR_SAMPLE(vmap.rcv_gid,CS_SYS_GETLPFWD,14,0,0);

  /* see if it me */
  if ( uid == vmap.rcv_uid ) 
	return -1;

  if ( cs_do_debug > 100 )
  	printf("%s: done\n", fn );

  return uid;
  }


/* ---------------------------------------------------------------------------
** CS_GetLPClone ()
**
** DESCRIPTION:  
**	second paramter only matters when cloning is optimistic, in
**	conservative cloning one alway runs the caller.
**
**---------------------------------------------------------------------------*/
int CS_GetLPClone( cs_clone_t vmap ) 
  {
  char  		fn[]	    = "[CS_GetLPClone]";
  unsigned long int	vps_requested = 0L;
  unsigned long int	vps_unaffected = 0L;

  vps_requested  =   (vmap.snd_mask) & (vmap.rcv_mask);
  vps_unaffected = (~(vmap.snd_mask) & (vmap.rcv_mask));

  printf("IS NOT CALLED HERE REMOVED\n");
  /* cs_CNTR_SAMPLE(vmap.rcv_gid,CS_SYS_GETLPCLN,3,0,0); */

  if ( vps_unaffected == 0L )
	return -1;
  else
	return 1;
  }



/* ---------------------------------------------------------------------------
** CS_SetLPClone ()
**
** DESCRIPTION:  
**	the parent executes this
**
**---------------------------------------------------------------------------*/
int CS_SetLPClone( cs_clone_t vmap, double birth_time, int active )
  {
  char                  fn[]        	= "[CS_SetLPClone]";
  unsigned long int     my_mask     	= 0L;
  unsigned long int     my_children 	= 0L;
  unsigned long int     my_siblings 	= 0L;
  unsigned long int     clones_children	= 0L;
  unsigned long int     clones_siblings	= 0L;

  unsigned long int     move_to_clone 	= 0L;
  unsigned long int	vps_requested 	= 0L;
  unsigned long int	vps_unaffected 	= 0L;
  int			my_cloneID;
  int			gid;
  int			rcv_uid;
  int			do_execute_parent;
  int			clones_cloneID;

  my_cloneID      = cs_IDtoClone[vmap.rcv_uid];
  gid             = cs_IDtoGID[vmap.rcv_uid];
  rcv_uid	  = vmap.rcv_uid;

  my_mask         = cs_clones[my_cloneID].mymask;
  my_children     = cs_clones[my_cloneID].children;
  my_siblings     = cs_vlpmapping[my_cloneID][gid].siblings;

  vps_requested  =   (vmap.snd_mask) & (vmap.rcv_mask);
  vps_unaffected = (~(vmap.snd_mask) & (vmap.rcv_mask));

  /* for conservative cloning below is alway true */
  do_execute_parent = my_mask & vps_requested;

  if ( do_execute_parent )
        move_to_clone = vps_unaffected;
   else
        move_to_clone = vps_requested;

  if ( move_to_clone == 0L )
        return -1;

  /* compute new clone's cloneID */
  get_lowest( clones_cloneID, move_to_clone, cs_num_clones )

  clones_children = cs_clones[clones_cloneID].children;
  clones_siblings = (move_to_clone) & ~(clones_children);

  /* un-map clone from my mapping or siblings */
  if ( move_to_clone & my_children )
     cs_vlpmapping[my_cloneID][gid].unmapped |= move_to_clone & my_children;
  else if ( move_to_clone & my_siblings )
     cs_vlpmapping[my_cloneID][gid].siblings &= ~(move_to_clone);

  /* set child mapping */
  cs_vlpmapping[clones_cloneID][gid].active 		= active;
  cs_vlpmapping[clones_cloneID][gid].BirthTS 		= birth_time;
  cs_vlpmapping[clones_cloneID][gid].CurrChildBirthTS 	= AM_NEGINFINITY;
  cs_vlpmapping[clones_cloneID][gid].parentID 		= vmap.rcv_uid;
  cs_vlpmapping[clones_cloneID][gid].parentCloneID 	= cs_IDtoClone[rcv_uid];
  cs_vlpmapping[clones_cloneID][gid].siblings 		= clones_siblings;
  cs_vlpmapping[clones_cloneID][gid].unmapped 		= 0L;

  if ( cs_do_debug > 10 )
        printf("%s: done\n", fn );
  return clones_cloneID;
  }


/* ---------------------------------------------------------------------------
** CS_SimDelete ()
**
** DESCRIPTION:  
**
**---------------------------------------------------------------------------*/
void CS_SimDelete( void )
  {
  char  fn[]	= "[CS_SimDelete]";

  if ( cs_do_debug > 10 )
  	printf("%s: done\n", fn );
  }


/* ---------------------------------------------------------------------------
** CS_MsgPad ()
**
** DESCRIPTION:  
**
**---------------------------------------------------------------------------*/
int CS_MsgPad( int MyMsgSize )
  {
  char  fn[]	= "[CS_MsgPad]";

   /* padding */
   if (MyMsgSize < sizeof(long))
        MyMsgSize = sizeof(long);
   if ((MyMsgSize % sizeof(long)) != 0)
        MyMsgSize = ((int) (MyMsgSize / sizeof(long)) + 1) * sizeof(long);

  if ( cs_do_debug > 10 )
  	printf("%s: done\n", fn );

   return MyMsgSize;
  }



/* ---------------------------------------------------------------------------
** cs_InitCSMessage ()
**
** DESCRIPTION:  
**
**---------------------------------------------------------------------------*/
void cs_InitCSMessage(void)
  {
  char  fn[]="[cs_InitCSMessage]";

  CSMsgSize = sizeof(struct CSMsgData);
  CSMsgSize = CS_MsgPad( CSMsgSize );

  if ( cs_do_debug > 10 )
  	printf("%s: done\n", fn );
  }


/* ---------------------------------------------------------------------------
** CS_InitCloning ()
**
** DESCRIPTION:  
**
**---------------------------------------------------------------------------*/
void CS_InitCloning(void)
  {
  char  fn[] = "[CS_InitCloning]";

  if ( cs_do_debug > -1 )
  	printf("%s: begin\n", fn );

  /* cs_InitCSMessage(); moved to to be called before csInitKernel */
  cs_InitSynchVariables();
  cs_InitCloneTable();	
  cs_InitVLPmapping();	
  cs_InitLockUpTables();

  if ( cs_do_debug > -1 )
  	printf("%s: done\n", fn );
  }

