/******************************************************************************
*******************************************************************************
***                                                                         ***
***                           CLONE-SIM-1.0                                 ***
***                                                                         ***
***         Copyright 1997-2000 by Georgia Institute of Technology          ***
***                          Atlanta, Georgia                               ***
***                                                                         ***
***                        All Rights Reserved                              ***


     Permission   to  use, copy   and   modify   this  software  and  its 
     documentation for any research purpose and without  fee  is   hereby
     granted, provided that the above copyright  notice  appears  in  all
     copies  and  that  both  that  copyright  notice and this permission
     notice  appear in supporting documentation.  The use or inclusion of
     this   software  or  its  documentation in any commercial product or
     distribution of  this  software to any other party without specific,
     written   prior  permission  is prohibited.

     THE SOFTWARE IS PROVIDED  AS  IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.  IN  NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE  LIABLE  FOR  ANY  SPECIAL,
     INDIRECT  OR  CONSEQUENTIAL  DAMAGES   OR  ANY  DAMAGES   WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER  TORTIOUS  ACTION,  ARISING  OUT OF OR
     IN   CONNECTION  WITH  THE  USE  OR  PERFORMANCE OF  THIS  SOFTWARE.

     ====================================================================
                                                                       
     FILE:      	cs_internal_t.h 
     INITIAL AUTHOR:    Maria Hybinette                                      

     ====================================================================

     	ADDRESSING typedef structures

***.........................................................................***
*******************************************************************************
******************************************************************************/
#ifndef CLONING_SIM_INTERNAL_T_H
#define CLONING_SIM_INTERNAL_T_H

#include	"am_arch.h"
#include        "cs_count.h"

#define CS_MAX_CLONES           32
#define CS_MAX_LP               100000

#define MSG_NULL                0
#define MSG_USER                1
#define MSG_SYSTEM              2
#define MSG_USR_SIM_CLONING     3
#define MSG_SYS_SIM_CLONING     4
#define MSG_LPCLONING           5
#define MSG_BIRTH_DUMMY         6
#define MSG_BIRTH_LP            7
#define MSG_BIRTH_SIM           8

#define MSG_CP_REQUEST_AND_ACTIVATE     8
#define MSG_WAIT_LOCAL_AND_ACTIVATE     9
#define MSG_WAIT_REMOTE           	10
#define MSG_WAIT_TO_PROCESS           	11

#define DO_EXECUTE_PARENT       1
#define DO_NOT_EXECUTE_PARENT   0

/* ---------------------------------------------------------------------------
** MACROS and DEFINES
**
** DESCRIPTION:
**
**---------------------------------------------------------------------------*/
#define VP                      0
#define set_bit(n,bit)          ( (n) |= 1 << (bit) )
#define unset_bit(n,bit)        ( (n) &= ~(1 << ( (bit) )) )
#define get_bits(x,lfrom, lto)  (( x >> (lfrom + 1 - lto) )  &  ~ (~0 << lto))

#define get_lowest(m_lowest_in,m_bits_in,m_num_clones) \
   { \
   unsigned long int m_tmp_bit=0L; \
   int m_lowest=m_num_clones; \
   m_tmp_bit = m_bits_in; \
   for ( m_lowest = 0;  m_tmp_bit != 0; m_tmp_bit >>= 1 ) \
        if ( m_tmp_bit & 0001 )  \
                break; \
        else  \
                m_lowest ++; \
   m_lowest_in=m_lowest; \
   }


#define CS_GetMsgType(msg) 		(msg->Admn_Data.type)

#define CS_SetForwardMsg(msg,token) 	\
  {				    	\
  (msg)->Admn_Data.a_forward_msg = 1; 	\
  (msg)->Admn_Data.forwarded = token;	\
  }


/* ---------------------------------------------------------------------------
** CSFuncPtr
**
** DESCRIPTION: function types
**
**---------------------------------------------------------------------------*/
typedef void  (*CSFunc_p)  ( AM_ADDR_TYPE, AM_ADDR_TYPE, int );
typedef void  (*CSFunc0_p) ( void );
typedef void  (*CSFunc1_p) ( int );
typedef void  (*CSFunc2_p) ( AM_ADDR_TYPE, int );
typedef void  (*CSFunc3_p) ( AM_ADDR_TYPE, AM_ADDR_TYPE, int );


/* ---------------------------------------------------------------------------
** _cs_rmap
**
** DESCRIPTION: Relationship table, used for addressing
**
**---------------------------------------------------------------------------*/
typedef struct _cs_rmap cs_rmap_t;
typedef struct _cs_rmap
        {
        cs_rmap_t 	*parent;
        cs_rmap_t 	*sibling;
        cs_rmap_t 	*child;
        cs_rmap_t 	*lchild;

        unsigned long int 	children;
        unsigned long int 	ffmask;
        unsigned long int 	mask;
        unsigned long int 	mymask;

        int 		CloneID;
        double 		BirthTS;
        } *cs_rmap_p;


/* ---------------------------------------------------------------------------
** _cs_vlpmap
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
typedef struct _cs_vlpmap cs_vlpmap_t;
typedef struct _cs_vlpmap
        { /* should add pointers to parent in here! */
        int 		active;
        double 		BirthTS;
        double 		CurrChildBirthTS;           /* used in TWClone */


        int 		parentID;
        int 		parentCloneID;
        int 		MyID;

        unsigned long int 	siblings;   
        unsigned long int 	unmapped;   

	unsigned long int	copy_request_src;
	double			copy_request_time;
        } *cs_vlpmap_p;


/* ---------------------------------------------------------------------------
** CSMsgData
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
struct CSMsgData
        {
        int type;                    /* message type (query etc) */
        int a_forward_msg;           /* a forward message */

        unsigned long int snd_unmapped;
        unsigned long int snd_siblings;  
        unsigned long int forwarded;    
        unsigned long int rb_siblings;	/* only optimistic cloning */    

        int 	snd_uid;                       
	double 	snd_time;

        int 	rcv_uid;                       
        int 	rcv_gid;                       
	int	cloneID;

	int 	copy_to_uid;

	CSFunc_p UpdateProc; /*IO4*/
        };


/* ---------------------------------------------------------------------------
** IOMsgData
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
struct IOMsgData
        {
        int action;
        int counter;
        int numself;
        };

/* ---------------------------------------------------------------------------
** MsgData
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
struct MsgData
        {
        struct CSMsgData Admn_Data;     /* later a union cloning, ipdes */
        AM_ADDR_TYPE 	 Appl_Data;     /* has to be below, b/c size differs */
        };

/* ---------------------------------------------------------------------------
** look-up table
**
** DESCRIPTION:  Temporary lookup table -- can compute this
**
**---------------------------------------------------------------------------*/
struct _cs_idmap
        {
        int clone_id[CS_MAX_CLONES];
        };
typedef struct _cs_idmap cs_idmap_t;

/* ---------------------------------------------------------------------------
** cs_clone_t
**
** DESCRIPTION:  Temporary lookup table -- can compute this
**
**---------------------------------------------------------------------------*/
struct _cs_clone
        {
	unsigned long int	snd_mask;
	unsigned long int	rcv_mask;

	unsigned long int 	done_forwarded;		/* forwarding */
	unsigned long int	clones_siblings; 	/* cloning */
	int			childs_cloneID; 	/* cloning */

	int        		a_forward_msg;

	int       		snd_uid;
	double       		snd_time;

	int        		rcv_uid;
	int        		rcv_gid;
	int			cloneID;
	int 			copy_to_uid;
        };
typedef struct _cs_clone cs_clone_t;
#endif /* CLONING_SIM_INTERNAL_T_H */
