/******************************************************************************
*******************************************************************************
***                                                                         ***
***                         INTERACTIVE-SIM-1.0                             ***
***                                                                         ***
***         Copyright 1997-2000 by Georgia Institute of Technology          ***
***                          Atlanta, Georgia                               ***
***                                                                         ***
***                        All Rights Reserved                              ***


     Permission   to  use, copy   and   modify   this  software  and  its 
     documentation for any research purpose and without  fee  is   hereby
     granted, provided that the above copyright  notice  appears  in  all
     copies  and  that  both  that  copyright  notice and this permission
     notice  appear in supporting documentation.  The use or inclusion of
     this   software  or  its  documentation in any commercial product or
     distribution of  this  software to any other party without specific,
     written   prior  permission  is prohibited.

     THE SOFTWARE IS PROVIDED  AS  IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.  IN  NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE  LIABLE  FOR  ANY  SPECIAL,
     INDIRECT  OR  CONSEQUENTIAL  DAMAGES   OR  ANY  DAMAGES   WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER  TORTIOUS  ACTION,  ARISING  OUT OF OR
     IN   CONNECTION  WITH  THE  USE  OR  PERFORMANCE OF  THIS  SOFTWARE.


     ====================================================================
                                                                       
     FILE:      	is_gtw.c 
     INITIAL AUTHOR:    Maria Hybinette                                      

     ====================================================================

     DESCRIPTION:
	* application layer
	* depends on kernel implementation

	* function calls cannot be called directly by user

***.........................................................................***
*******************************************************************************
******************************************************************************/

#include "am_arch.h"
#include "am_utility.h"
#include "cs_internal.h"
#include "is_kernel.h"
#include "is_gtw.h"
#include "op_gtw.h"

#include "phold.h" 	/* for debugging */

extern void mh_all_mem_init_calloc(int);	/* application defined */
extern void mh_clo_mem_init_calloc(int); 	/* application defined */

int 		ISnlp=0;
int		ISvlp=0;
int 		is_num_clones=1;
static char  	is_gtw[]="[is_gtw]";

/* from ipdes.c XXX */
int 		special=0;
int		is_phold=0; /* debugging */


/* ---------------------------------------------------------------------------
** iPrintIMapping
**
** DESCRIPTION:
**
**---------------------------------------------------------------------------*/
void iPrintIMapping( unsigned long int map, char *instring, int title, 
		int NumClones)
{
   char map_string[100];
   int i;

   for( i = 0; i < NumClones; i ++ )
        {
        if( get_bits( map,i,1) & (01) )
                map_string[i] = '1';
        else
                map_string[i] = '0';
        }
   map_string[i] = '\0';
   if( title )
      printf("              : %s\n", "012345678901234567890123456789012345678");
   printf("       %s: %s\n", instring, map_string );
   fflush( stdout );
}


/* ---------------------------------------------------------------------------
** XPrintIMapping
**
** DESCRIPTION:
**
**---------------------------------------------------------------------------*/
void XPrintIMapping( unsigned long int map, int uid, char *instring, 
	int title, int NumClones)
{
   char map_string[100];
   int i;

   for( i = 0; i < NumClones; i ++ )
        {
        if( get_bits( map,i,1) & (01) )
                map_string[i] = '1';
        else
                map_string[i] = '0';
        }
   map_string[i] = '\0';
   if( title )
      printf("              : %s\n", "012345678901234567890123456789012345678");
   printf("  %4d    %s: %s\n", uid, instring, map_string );
   fflush( stdout );
}


/* ---------------------------------------------------------------------------
** IS_TestState()
**
** DESCRIPTION:
**
**---------------------------------------------------------------------------*/
void IS_TestState( char *cn, struct MyLPState *SV, 
	int uid, int gid, int cloneID, double time_now )
  {
  int i;
  struct MyLPState *MV = (struct MyLPState *) SV;

  if( (MV->CVars.Seeds.S1==0) && (MV->CVars.Seeds.S1==0) )
	{
	int test_uid;
	int test_clone;

	for( test_clone=0;test_clone<=cloneID;test_clone++ )
		{
	        test_uid = cs_GIDtoID[gid].clone_id[test_clone];
		MV = (struct MyLPState *) TWLP[test_uid].State;
 		  	printf( "%s:** cloning Seed error TEST "
				"test_uid=%3d "
				"test_clone=%3d "
				"uid=%3d "
				"gid=%3d "
				"at=%lf "
				"(%s, %s)\n", 
				cn, 
				test_uid,
				test_clone,
				uid, 
				gid, 
				time_now,
			((MV->CVars.Seeds.S1==0) && (MV->CVars.Seeds.S1==0)) ?
			"BAD":" OK",
 			(cs_vlpmapping[cloneID][gid].active==0) ?
			"INACTIVE":"ACTIVE"
		             );
		}
	}
   }




/* ---------------------------------------------------------------------------
** IS_SetMsg()
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void IS_SetMsg( int src, int dst )
{
   struct MsgData 	*TWMsg;
   AM_ADDR_TYPE 	msg_begin_ptr; 
   int  		MyPE=LP2PE(src);		

   set_GTWState();
	
   /* re-set to user message data */
   msg_begin_ptr = MyState->NewBuf->Msg.Data;
   TWMsg = (struct MsgData *) msg_begin_ptr;

   /* re-set to user message data */
   msg_begin_ptr += CSMsgSize;
   TWMsg->Appl_Data = msg_begin_ptr;
   MyState->TWMsg = TWMsg->Appl_Data;
}



/* ---------------------------------------------------------------------------
** IS_ISysLP()
**
** DESCRIPTION: init each LP
**
**	
**---------------------------------------------------------------------------*/
void IS_ISysLP( void *SV, int MyPE )
  {  
  int 			myLP;
  AM_ADDR_TYPE 		state_begin_ptr;
  unsigned long int 	offset;
  char  	fn[]	= "[IS_ISysLP]";

  set_GTWState()

  myLP 	  = CloneSim_UID();
   
  mh_all_mem_init_calloc(MyPE); 			/* XXX HERE */

  if( cs_do_debug > 10 )
  	printf("%s: begin (MyPE=%d, MyLP=%d)\n", fn, MyPE, myLP );

  /* calls user IProc for this LP */
  if( myLP < ISnlp )
	{
   	is_ui[myLP].IProc( SV, MyPE );
	}
  else
	{
	/* ASSUME's parent echek point state have been set */
   	struct LPState 	*PState;
	int 		pLP;	

	pLP = 	myLP % ISnlp; 
	PState = LPST( pLP );

	/* need to know where to auto-check clone */
   	offset = PState->auto_chkpt_state - TWLP[pLP].State;

	/* CopySize is the size of the state that is auto check pointed */
	CurState->CopySize = PState->CopySize;
   	CurState->state    = TWLP[myLP].State;
	state_begin_ptr    = (char *) CurState->state;
	CurState->auto_chkpt_state = state_begin_ptr + offset;

	if( special > 0 )
	   {
	   mh_clo_mem_init_calloc(MyPE); 		/* XXX  HERE */
		
	   /* XXXX probably no need to copy here */
	   TWLP[pLP].CProc( TWLP[myLP].State, TWLP[pLP].State, 0, MyPE );
	   /*dst,frm*/
	   }
	}
}



/* ---------------------------------------------------------------------------
** IS_FSysLP ()
**
** DESCRIPTION: 
**	all FProc's are called, even Clones
**
**---------------------------------------------------------------------------*/
void IS_FSysLP( void * SV, int MyPE )
{  
  set_GTWState()
  is_ui[CurState->ID].FProc( SV, MyPE );
}



/* ---------------------------------------------------------------------------
**  IS_SetNumClones()
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void IS_SetNumClones( void )
  {
  char  fn[] = "[IS_SetNumClones]";

  is_num_clones = CS_SetNumClones(CloneSim_NumClones);

   if( cs_do_debug > 10 )
  	printf("%s: done %d\n", fn, is_num_clones );
  }



/* ---------------------------------------------------------------------------
**  IS_SetLPs ()
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void IS_SetLPs( void )
  {
  char  fn[]="[IS_SetLPs]";

  /* regular LPs */
  ISnlp = CS_SetNLPs(TWnlp);

  /* virtual LPs */
  ISvlp = CS_SetVLPs(is_num_clones*TWnlp);

  /* check for _out-of_ bounds */
  if( (ISvlp < 1) || (ISvlp > MAXLP) )
	{
	AM_Error("Illegal number of ISvlp=%d (Max=%d)\n", ISvlp, MAXLP );
	}

  /* set LPs at kernel level */
  TWnlp = ISvlp;

  if( cs_do_debug > 10 )
  	printf("%s: done (%d, %d)\n", fn, ISnlp, ISvlp );
  }



/* ---------------------------------------------------------------------------
**  IS_AllocateHandlers ()
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void IS_AllocateHandlers( void )
  {
  char  fn[]	= "[IS_AllocateHandlers]";

  is_ui = IS_AllocateISUI(ISvlp); 

  if( cs_do_debug > 10 )
  	printf("%s: done\n", fn );
  }



/* ---------------------------------------------------------------------------
**  IS_CopyLPState()
**
** DESCRIPTION: 
**
**	generic pointers?
**
**	may want pointer instead of integer addresses have the to and from be 
**	generic pointer then the middleware programmer can decide on how to do 
**	this. THIS procedure assume that CLONE and parent is on same PE when 
**	initially! Recall, cloning assumes static resources
**
**---------------------------------------------------------------------------*/
void IS_CopyLPState( int to, int from )
  {
  char  fn[]	= "[IS_CopyLPState]";
  int	MyPE	= TWLP[from].Map;

  if( cs_do_debug > 0 )
  	printf("%s: begin\n", fn );

  if( special > 0 )
       TWLP[cs_IDtoGID[to]].CProc( TWLP[to].State, TWLP[from].State, 1, MyPE );
  else
       AM_MCpSt((TWLP[from].State), (TWLP[to].State), (TWLP[to].LPStateSize)); 
  if( is_phold )
	{
  	IS_TestState(fn, (struct MyLPState *) TWLP[from].State, 
		to, cs_IDtoGID[from], cs_IDtoClone[to],00);
	}

  if( cs_do_debug > 0 )
  	printf("%s: end\n", fn );
  }



/* ---------------------------------------------------------------------------
**  IS_SetHandlers()
**
** DESCRIPTION: 
**	called after TWInitAppl()	
**	gtw dependent
**
**---------------------------------------------------------------------------*/
void IS_SetHandlers( void )
{
  int i, j;
  char  fn[]="[IS_SetHandlers]";

   /* reset event handler so that we can monitor all LPs */
   for( i = 0; i < ISnlp; i++ ) 
	{
	/* save user invoked handlers */
	for( j = 0; j < is_num_clones; j++ )
		{
		is_ui[i+j*ISnlp].Map 	= TWLP[i].Map;
  		is_ui[i+j*ISnlp].IProc 	= TWLP[i].IProc;
  		is_ui[i+j*ISnlp].Proc 	= TWLP[i].Proc;
  		is_ui[i+j*ISnlp].FProc 	= TWLP[i].FProc;

  		is_ui[i+j*ISnlp].UpdateProc = NULL;
		}

	/* reset so kernel channels handling to IS instead of user app */
 	TWLP[i].IProc 	= (FuncPtr2) IS_ISysLP;

  	if( cs_optimistic == 1 )
 		TWLP[i].Proc 	= (FuncPtr)  OP_SysLP;
	else if( cs_optimistic == 2 )
 		TWLP[i].Proc 	= (FuncPtr)  OP_SysLP;
	else
		{
		printf("ERROR no IS_SysLP anymore\n");
		exit(1);
 		/* TWLP[i].Proc 	= (FuncPtr)  IS_SysLP; */
		}

 	TWLP[i].FProc 	= (FuncPtr2) IS_FSysLP;
	
        /* do the same for all possible clones, not j=0 */
	for( j = 1; j < is_num_clones; j++ )
		{ 
		TWLP[i+j*ISnlp].Map 		= TWLP[i].Map;
 		TWLP[i+j*ISnlp].IProc 		= (FuncPtr2) IS_ISysLP;
  		if( cs_optimistic == 1 )
 		   	TWLP[i+j*ISnlp].Proc 	= (FuncPtr)  OP_SysLP;
		else if( cs_optimistic == 2 )
 			TWLP[i+j*ISnlp].Proc 	= (FuncPtr)  OP_SysLP;
		else
			{
			printf("ERROR no IS_SysLP anymore\n");
 			/* TWLP[i+j*ISnlp].Proc 	= (FuncPtr)  IS_SysLP; */
			}

 		TWLP[i+j*ISnlp].FProc 	= (FuncPtr2) IS_FSysLP;

		TWLP[i+j*ISnlp].State = TWMalloc(TWLP[i].LPStateSize);

		TWLP[i+j*ISnlp].CopySize    = TWLP[i].CopySize;
		TWLP[i+j*ISnlp].LPStateSize = TWLP[i].LPStateSize;

		/* make sure we do incremental state saving */
 		TWLP[i+j*ISnlp].IncrSave = TRUE; /* may not be necessary */
		}
      	}

  if( cs_do_debug > 10 )
  	printf("%s: done\n", fn );

}



/* ---------------------------------------------------------------------------
** IS_SetMessageSize()
**
** DESCRIPTION: 
**
**
**---------------------------------------------------------------------------*/
void IS_SetMsgSize( void ) 
  {
  char  fn[] = "[Is_SetMsgSize]";

  TWMsgSize = max( TWMsgSize, sizeof(struct IOMsgData) );

  if( cs_do_debug > 10 )
  	printf("%s: TWMsgSize=%d\n", fn, TWMsgSize );

  TWMsgSize = CS_MsgPad(TWMsgSize+CSMsgSize );

  if( cs_do_debug > 10 )
	{
  	printf("%s: CSMsgSize=%d\n", fn, CSMsgSize );
  	printf("%s: TWMsgSize=%d\n", fn, TWMsgSize );
  	printf("%s: done\n", fn );
	}
  }


/* ---------------------------------------------------------------------------
** IS_CloseAppl()
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void IS_CloseAppl( void ) 
  {
  extern void printMHTimers(void);
  printMHTimers();
  }
