/******************************************************************************
*******************************************************************************
***                                                                         ***
***                           CLONE-SIM-1.0                                 ***
***                                                                         ***
***         Copyright 1997-2000 by Georgia Institute of Technology          ***
***                          Atlanta, Georgia                               ***
***                                                                         ***
***                        All Rights Reserved                              ***


     Permission   to  use, copy   and   modify   this  software  and  its
     documentation for any research purpose and without  fee  is   hereby
     granted, provided that the above copyright  notice  appears  in  all
     copies  and  that  both  that  copyright  notice and this permission
     notice  appear in supporting documentation.  The use or inclusion of
     this   software  or  its  documentation in any commercial product or
     distribution of  this  software to any other party without specific,
     written   prior  permission  is prohibited.

     THE SOFTWARE IS PROVIDED  AS  IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.  IN  NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE  LIABLE  FOR  ANY  SPECIAL,
     INDIRECT  OR  CONSEQUENTIAL  DAMAGES   OR  ANY  DAMAGES   WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER  TORTIOUS  ACTION,  ARISING  OUT OF OR
     IN   CONNECTION  WITH  THE  USE  OR  PERFORMANCE OF  THIS  SOFTWARE.


     ====================================================================

     FILE:              is_gtw_stat.c
     INITIAL AUTHOR:    Maria Hybinette

     ====================================================================

     DESCRIPTION:       prints out time statistics, kernel dependent


***.........................................................................***
*******************************************************************************
******************************************************************************/

#include "gtw_insim.h"
#include "gtw_timer.h"
#include "op_gtw.h" 
#include "cs_internal.h" 

#define MH_INDEX 40
extern TWtimer		mh_init[MH_INDEX];
extern ClockTicks 	mh_secs[MH_INDEX];

/* DEBUG from config.c */
extern int 	clone_multiple;
extern int	print_tables[10];
extern int 	do_mh_memory;
extern int 	do_headers;


/*-----------------------------------------------------------------------------
**  mh_secs_current()
**
**  DESCRIPTION:
**
**---------------------------------------------------------------------------*/
ClockTicks
mh_secs_current( char *str, int index1, ClockTicks whole )
  {
  ClockTicks current;

  current = mh_secs[index1];
  fprintf( stdout,
        "%16s %6.2lf",
        str,
        current
        );
  if ( whole <= 0 ) whole = current;
  fprintf(stdout, " (%6.2lf)", current/whole * 100 );

  fprintf(stdout, "    %6.2lf", current * clone_multiple );

  fprintf( stdout, "\n");
  return current;
  }


/*-----------------------------------------------------------------------------
**  mh_current()
**
**  DESCRIPTION:
**
**---------------------------------------------------------------------------*/
ClockTicks
mh_current( char *str, int index1, int index2, ClockTicks whole )
  {
  ClockTicks current;

  current = TWseconds (mh_init[index2]) - TWseconds( mh_init[index1] );
  fprintf( stdout,
        "%16s %6.2lf",
        str,
        current
        );
  if ( whole <= 0 ) whole = current;
  fprintf(stdout, " (%6.2lf)", current/whole * 100 );

  fprintf(stdout, "    %6.2lf", current * clone_multiple );

  fprintf( stdout, "\n");
  return current;
  }


/*-----------------------------------------------------------------------------
**  printMHTimers()
**
**  DESCRIPTION:
**
**---------------------------------------------------------------------------*/
void
printMHTimers(void)
  {
  ClockTicks current;
  ClockTicks whole;
  ClockTicks init;
  ClockTicks initPE;
  ClockTicks dummy;

  if ( print_tables[5] == 1 )
	{
  	printf("\n");
  	whole   = mh_current ( "TotalTime:", 0, 13,  0 );
  	printf("\n");
  	init = mh_current ( "Initialization:", 0, 10,  whole );
  	dummy = mh_current ( "ExecTime:", 10, 11,  whole );
  	dummy = mh_current ( "WrapUp:", 11, 13,  whole );
  	printf("\n");

 	/* --- */

  	dummy = mh_current( "InitAppl:",  1, 2,  init );
  	dummy = mh_current( "InitIPDES:", 2, 3,  init );
  	dummy = mh_current( "InitPE:", 5, 6,  init );
  	initPE = dummy;
  	dummy = mh_current( "InitGIDTable:", 6, 7,  init );
  	dummy = mh_current( "IProc:", 7, 8,  init );
  	dummy = mh_current( "AcStateVector:", 8, 9,  init );
  	dummy = mh_current( "getFirstBuffer:", 9, 10,  init );

 	/* --- */
  	printf("\n");
  	dummy = mh_current( "RunSim:", 4, 12,  whole );
  	dummy = mh_current( "InitSched:", 5, 10,  init );
  	printf("\n");

  	dummy = mh_current( "InitPE:", 5, 6,  initPE );
  	dummy = mh_current( "AllEvMem:",  5, 14,  initPE );
  	dummy = mh_current( "InitPE bulk:",  14, 15,  initPE );
  	dummy = mh_current( "InitLPs:",  15, 16,  initPE );
  	dummy = mh_current( "InitQ:",  16, 17,  initPE );
  	dummy = mh_current( "InitBuf:",  17, 18,  initPE );
  	dummy = mh_current( "waiting:", 18, 6,  initPE );

  	printf("\n");
  	current = mh_current( "InitLPs:",  15, 16,  initPE );
  	dummy = mh_current( "AllLPState:",  15, 19,  current );
  	init = mh_current( "loop:",  19, 20,  current );
  	mh_secs_current( "IncrLog:", 21, init );
  	mh_secs_current( "Excl IncrLog:", 22, init );
  	printf("\n");
	}

 /* --- */

  if ( print_tables[6] == 1 )
	{
	if ( do_headers == 1 )
		{
  		printf("# -----------------------------------------------"
			"---------------"
        		"---------------\n");
  		printf("#PEs Total   Init    Exc WrapUp InitPE  IProc vLPs"
			"  LPs\n");
  		printf("#------------------------------------------------"
			"---------------"
       	 		"---------------\n");
		}

  	fprintf( stdout, " %2d", TWnpe );

  	/* TotalTime */
  	fprintf( stdout, " %6.2lf", clone_multiple * 
		(TWseconds (mh_init[13])  - TWseconds( mh_init[0]  ) ) );

  	/* Initialization */
  	fprintf( stdout, " %6.2lf",  clone_multiple * 
		(TWseconds(mh_init[10])  - TWseconds( mh_init[0]  ) ) );

  	/* ExecTime */
  	fprintf( stdout, " %6.2lf", clone_multiple * 
		(TWseconds(mh_init[11])  - TWseconds( mh_init[10] ) ) );

  	/* Wrapup */
  	fprintf( stdout, " %6.2lf", clone_multiple * 
		(TWseconds(mh_init[13])  - TWseconds( mh_init[11] ) ) );

  	/* InitPE */
  	fprintf( stdout, " %6.2lf", clone_multiple * 
		(TWseconds(mh_init[6])   - TWseconds( mh_init[5]  ) ) );

  	/* IProc */
  	fprintf( stdout, " %6.2lf", clone_multiple * 
		(TWseconds(mh_init[8])   - TWseconds( mh_init[7]  ) ) );

  	fprintf( stdout, " %2d", TWnlp );
  	fprintf( stdout, " %2d", TWnlp );
  	fprintf( stdout, "\n");

	if ( do_headers == 1 )
		{
  		fprintf( stdout, "\n");
		}
  	fflush (stdout);
	}

  if ( print_tables[2] == 1 )
	{
	if ( do_headers == 1 )
		{
  		fprintf( stdout, 
			"# -----------------------------------------------"
			"---------------"
        		"---------------\n");
		fprintf(stdout, 
			"# P CL ACL vLPs LPs"
			"  ExcTime  TotTime    IProc   InitPE"
			"     SimTS  Pt1CL  CLprd\n"
			);
 	 	fprintf( stdout,
			"# -----------------------------------------------"
			"---------------"
       		 	"---------------\n");
		}

  	fprintf( stdout, " %2d", TWnpe );

	{
	extern int CloneSim_scale_size;
	if( CloneSim_scale_size > 1 ) 
		{
  		fprintf( stdout, " %2d",  CloneSim_scale_size * CloneSim_NumClones );
  		fprintf( stdout, " %2d",  CloneSim_scale_size * (cs_clone_cnt+1) );
		}
	else
		{
  		fprintf( stdout, " %2d",  clone_multiple * CloneSim_NumClones );
  		fprintf( stdout, " %2d",  clone_multiple * (cs_clone_cnt+1) );
		}
	}
  	fprintf( stdout, " %4d", cs_vlp );
  	fprintf( stdout, " %4d", cs_nlp );

 	/* ExecTime */
  	fprintf( stdout, " %8.2lf", clone_multiple * 
		(TWseconds (mh_init[11])  - TWseconds( mh_init[10] ) ) );
	/* TotalTime */
        fprintf( stdout, " %8.2lf", clone_multiple *
                (TWseconds (mh_init[13])  - TWseconds( mh_init[0]  ) ) );
        /* IProc */
        fprintf( stdout, " %8.2lf", clone_multiple *
                (TWseconds (mh_init[8])   - TWseconds( mh_init[7]  ) ) );
        /* InitPE */
        fprintf( stdout, " %8.2lf", clone_multiple *
                (TWseconds (mh_init[6])   - TWseconds( mh_init[5]  ) ) );

        fprintf( stdout, " %9.2lf", ENDTIME );
  	if ( (CloneSim_NumClones) > 1 )
		{
        	fprintf( stdout, " %6.2lf", CloneSim_CLONE_RT[0] );
		}
	else
		{
        	fprintf( stdout, " %6.2lf", 0 ); /* BAcloning_pt1 */
		}

        if ( (CloneSim_NumClones) > 2 )
		{
        	fprintf( stdout, " %6.0lf", CloneSim_CLONE_RT[1] 
						- CloneSim_CLONE_RT[0] );
		}
	else
		{
        	fprintf( stdout, " %6.0lf", BAcloning_pt1 );
		}

  	fprintf( stdout, "\n");
	if ( do_headers == 1 )
  		fprintf( stdout, "\n");
  	fflush (stdout);
	}
  }

