/******************************************************************************
*******************************************************************************
***                                                                         ***
***                         INTERACTIVE-SIM-1.0                             ***
***                                                                         ***
***         Copyright 1997-2000 by Georgia Institute of Technology          ***
***                          Atlanta, Georgia                               ***
***                                                                         ***
***                        All Rights Reserved                              ***


     Permission   to  use, copy   and   modify   this  software  and  its 
     documentation for any research purpose and without  fee  is   hereby
     granted, provided that the above copyright  notice  appears  in  all
     copies  and  that  both  that  copyright  notice and this permission
     notice  appear in supporting documentation.  The use or inclusion of
     this   software  or  its  documentation in any commercial product or
     distribution of  this  software to any other party without specific,
     written   prior  permission  is prohibited.

     THE SOFTWARE IS PROVIDED  AS  IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.  IN  NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE  LIABLE  FOR  ANY  SPECIAL,
     INDIRECT  OR  CONSEQUENTIAL  DAMAGES   OR  ANY  DAMAGES   WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER  TORTIOUS  ACTION,  ARISING  OUT OF OR
     IN   CONNECTION  WITH  THE  USE  OR  PERFORMANCE OF  THIS  SOFTWARE.


     ====================================================================
                                                                       
     FILE:      	opt_gtw.c 
     INITIAL AUTHOR:    Maria Hybinette                                      

     ====================================================================

     DESCRIPTION:
	* application layer
	* depends on kernel implementation

	* function calls cannot be called directly by user

***.........................................................................***
*******************************************************************************
******************************************************************************/

#include "am_arch.h"
#include "am_utility.h"
#include "cs_internal.h"
#include "is_kernel.h"
#include "is_gtw.h"

/* for debugging */
#include "phold.h"
extern int is_phold;

extern int special;
extern int is_num_clones;

extern void iPrintIMapping( unsigned long int, char *, int, int );
extern void XPrintIMapping( unsigned long int, int, char *, int, int );
extern void IS_TestState( char *, struct MyLPState *, int, int,  int, double );



/* ---------------------------------------------------------------------------
** OP_DoSendEnQueue ()
**
** DESCRIPTION:	generic pointer for MyMsgData,
**
**---------------------------------------------------------------------------*/
int OP_DoEnQueueOnSend( int rcv_LP, double rcv_time )
{  
   char  	fn[]	= "[OP_DoEnQueueOnSend]";
   int		cloneID;
   int		gid;
   int		do_enqueue=1;

   if( cs_optimistic == 0 )
   	return do_enqueue;

   if( cs_do_debug > 100 )
        printf("%s: begin\n", fn );

   cloneID 	= cs_IDtoClone[rcv_LP];
   gid 		= cs_IDtoGID[rcv_LP];
   return ((cs_vlpmapping[cloneID][gid].BirthTS < rcv_time) 
		&& (cs_vlpmapping[cloneID][gid].active == 1));
}


/* ---------------------------------------------------------------------------
** OP_CancelPendingEvent()
**
** DESCRIPTION:	
**
**---------------------------------------------------------------------------*/
void OP_CancelPendingEvent( int rcv_LP, char *reason )
{  
   char  	fn[]	= "[OP_CancelPendingEvent]";
   int		cloneID;
   int		gid,uid;
   double	time_now;
   
   if( cs_optimistic == 0 )
   	return;


   cloneID 	= cs_IDtoClone[rcv_LP];
   gid 		= cs_IDtoGID[rcv_LP];
   uid		= rcv_LP;
   time_now	= cs_vlpmapping[cloneID][gid].BirthTS;
   

   if( cs_do_debug > 1 )
	printf("%s: *cloning rolling back uid=%3d gid=%3d at=%lf reason=%s\n", 
		fn, uid, gid, time_now, reason );

   cs_vlpmapping[cloneID][gid].BirthTS = MAXFLOAT;
   cs_vlpmapping[cloneID][gid].parentCloneID = -1;
   cs_vlpmapping[cloneID][gid].parentID = -1;
   cs_vlpmapping[cloneID][gid].siblings = 0L;

   if( cs_do_debug > 10 )
        printf("%s: end\n", fn );

   cs_CNTR_SAMPLE(gid,CS_CAN_PENDING,11,0,0);
}


/* ---------------------------------------------------------------------------
** CS_GetParentRBLP ()
**
** DESCRIPTION:	
**
**---------------------------------------------------------------------------*/
int CS_GetParentRBLP( int rcv_LP, double rcv_ts )
   {
   int		parent_uid;

   cs_vlpmap_t  *parent_vlp;
   int 		i;
   int 		cloneID;
   int 		gid;
   double	birth_ts;
   int 		pcloneID	= -1;
   char  	fn[]		= "[CS_GetParentRBLP]";

   cloneID 	= cs_IDtoClone[rcv_LP];
   gid 		= cs_IDtoGID[rcv_LP];

   if( cs_do_debug > 100 )
        printf("%s: begin\n", fn );

   /* has to find the sibling first */
   for( i = cloneID; i >= 0; i-- )
	{
        if( cs_vlpmapping[i][gid].siblings & cs_clones[cloneID].mask )
		{
                /* it is a sibling */
                pcloneID = i;
                break;
                }

	else if( (cs_clones[i].children & cs_clones[cloneID].mask) /*MH_HERE*/
		&& (cs_vlpmapping[i][gid].BirthTS != MAXFLOAT) )
        	{
                /* it is a child */
                pcloneID = i;
                break;
                }
	} /* for */

   if( pcloneID < 0 )
	{
	AM_Error("+**%s: Invalid Parent\n", fn );
	}

   parent_vlp = &cs_vlpmapping[pcloneID][gid];
   while( (parent_vlp != NULL) && (rcv_ts < parent_vlp->BirthTS) ) /*MH_HERE*/
        {
        if( parent_vlp->parentCloneID < 0 )
                {
                AM_Error("+** %s: unborn parent\n", fn );
                }
        parent_vlp = &(cs_vlpmapping[parent_vlp->parentCloneID][gid]);
        }

   if( parent_vlp == NULL )
	AM_Error("+** %s: parent is null\n", fn );
   parent_uid = parent_vlp->MyID;

   if( cs_do_debug > 10 )
        printf("%s: end\n", fn );

   return parent_uid;
   }



/* ---------------------------------------------------------------------------
** OP_GetRBSibling ()
**
** DESCRIPTION:	generic pointer for MyMsgData,
**
**---------------------------------------------------------------------------*/
int OP_GetRBSibling( int CloneID, int GID, TWTime RcvTS, int *XX_loop_1 )
{
   char                 fn[]    = "[OP_GetRBSibling]";
   int 			i;
   int 			id_found=-1;
   int 			cntr=0;
   cs_vlpmap_t  	*parent;
   int 			CurrID;

   if( cs_do_debug > 100 )
        printf("%s: begin\n", fn );

   CurrID = cs_GIDtoID[GID].clone_id[CloneID];

   if( cs_vlpmapping[CloneID][GID].BirthTS < MAXFLOAT )
        {
        id_found = CloneID;
        }
   else
        {
        for( i = 0; i < CloneID; i++ )
                { /* see if I am in a sibling mapping */
             if( cs_vlpmapping[i][GID].siblings & cs_clones[CloneID].mask )
                        {
                        if( id_found < 0 )
                                id_found = i;
                        cntr++;
                        }
                }
        }
   *XX_loop_1=cntr;

   if( id_found < 0 ) return -1;

   parent = &cs_vlpmapping[id_found][GID];

   cntr=0;
   while( (parent != NULL) && (RcvTS < parent->BirthTS) )
        {
	cntr++;
        if( parent->parentCloneID < 0 )
                {
		if( cs_do_debug > -10 ) 
                	printf("+== %s: sibling has been unborn\n", fn );
		
		*XX_loop_1 += cntr;
                return -1;
                /* AM_Error("+** %s: sibling has been unborn\n", fn ); */
                }
        parent = &(cs_vlpmapping[parent->parentCloneID][GID]);
        }
   *XX_loop_1 += cntr;

   if( parent == NULL )
        AM_Error("+** %s: *** parent mapping is messed up\n", fn );
   return parent->MyID;
}



/* ---------------------------------------------------------------------------
** OP_GetRBParent ()
**
** DESCRIPTION:
**
**---------------------------------------------------------------------------*/
int OP_GetRBParent( int CloneID, int GID, double RcvTS, int *XX_loop_2 )
{
   char                 fn[]    = "[OP_GetRBParent]";
   cs_rmap_t            *first;
   int                  cntr=0;
   int                  CurrID=-1;
   cs_vlpmap_t          *find_vlp;

  if( cs_do_debug > 100 )
        printf("%s: begin\n", fn );

   /* find a ancestor that is born before me */
   first        = &cs_clones[CloneID];
   CurrID       = cs_GIDtoID[GID].clone_id[first->CloneID];
   find_vlp       = &cs_vlpmapping[first->CloneID][GID];

   while( (first != NULL) && (RcvTS < find_vlp->BirthTS) )
        {
        first = first->parent; /* NEGINFINITE should not get in here */
        if( first == NULL )
                AM_Error("+** %s NULL\n", fn );
        CurrID  = cs_GIDtoID[GID].clone_id[first->CloneID];
        find_vlp  = &cs_vlpmapping[first->CloneID][GID];

        cntr++;
        if( cntr > CloneSim_NumClones )
                AM_Error("+** %s: infinite loop\n", fn );
        }
   *XX_loop_2=cntr;
   return CurrID;
}


/* ---------------------------------------------------------------------------
** OP_GetRBLP ()
**
** DESCRIPTION:	generic pointer for MyMsgData,
**
**---------------------------------------------------------------------------*/
int OP_GetRBLP( int rcv_LP, double rcv_ts )
{
   char  	fn[]	= "[OP_GetRBLP]";
   int		cloneID;
   int		gid;
   double	birth_ts;
   int		rollback_LP;
   int          XX_loop_1=0;
   int          XX_loop_2=0;


   if( cs_optimistic == 0 )
   	return rcv_LP;

   cloneID      = cs_IDtoClone[rcv_LP];
   gid          = cs_IDtoGID[rcv_LP];
   birth_ts     = cs_vlpmapping[cloneID][gid].BirthTS;

   rollback_LP = rcv_LP;
   if( ((rcv_ts < birth_ts) && (birth_ts < MAXFLOAT))
        || (birth_ts == MAXFLOAT) )
	{
	if(  (rollback_LP 
		= OP_GetRBSibling( cloneID, gid, rcv_ts, &XX_loop_1 ) ) < 0 )
		{
		rollback_LP = OP_GetRBParent( cloneID, gid, rcv_ts, &XX_loop_2);
		}
       	}

   cs_CNTR_SAMPLE(gid,CS_MSG_Q_FINDLP,18,XX_loop_1,XX_loop_2);

   if( cs_do_debug > 10 )
        printf("%s: end\n", fn );
   return rollback_LP;
}


/* ---------------------------------------------------------------------------
** OP_GetParent ()
**
** DESCRIPTION:	
**
**---------------------------------------------------------------------------*/
int OP_GetParent( int rcv_LP, int gid, int cloneID, double rcv_time, 
	int *XX_cnt )
{
   char  	fn[]	   	= "[OP_GetParent]";
   cs_rmap_t 	*first		=  NULL;
   cs_vlpmap_t  *first_vlp	=  NULL;

   int 		cntr		=  0;
   int 		currID		= -1;

   if( cs_do_debug > 100 )
        printf("%s: begin\n", fn );

   /* find closest born ancestor */
   first 	= &cs_clones[cloneID];
   currID 	= rcv_LP;
   first_vlp 	= &cs_vlpmapping[cloneID][gid];

   /* start at bottom of cs_clones and checks if vlp is born */
   while( (first != NULL) && (MAXFLOAT == first_vlp->BirthTS) )
        {
        first = first->parent;
        if( first == NULL )
                AM_Error( "++** %s NULL\n", fn );
        first_vlp = &cs_vlpmapping[first->CloneID][gid];

        cntr++;

        if( cntr > cs_num_clones )
                AM_Error( "++** %s: infinite loop\n", fn );
        }
   *XX_cnt=cntr;

   if( first == NULL )
        AM_Error( "+** %s: no nearest parent\n", fn );

   if( first_vlp == NULL )
        AM_Error("+** %s: first_vlp is NULL \n", fn );

   if( first_vlp->BirthTS > rcv_time )
        AM_Error("+** %s: parent(gid=%3d cloneID=%3d) of (cloneID=%3d) "
		"not rb'd(parent_birth=%f, rcv_time=%f)\n",
		fn,
		gid,
		cs_IDtoClone[first_vlp->MyID],
                cs_IDtoClone[rcv_LP],
                first_vlp->BirthTS, rcv_time
		);

   currID = first_vlp->MyID;
   if( cs_do_debug > 10 )
        printf("%s: end\n", fn );
   return currID;
}


/* ---------------------------------------------------------------------------
** OP_GetSibling ()
**
** DESCRIPTION:	
**
**---------------------------------------------------------------------------*/
int OP_GetSibling( int gid, int cloneID )
{
   char  	fn[]	   = "[OP_GetSibling]";
   int 		i, cntr	   = 0;
   int 		id_found   = -1;
   cs_vlpmap_t 	*sibling;

   if( cs_do_debug > 100 )
        printf("%s: begin\n", fn );

   /* find out if I am mapped to a sibling */
   for( i = 0; i < cloneID; i++ )
        {
        /* see if I am in a sibling mapping */
        if( cs_vlpmapping[i][gid].siblings & cs_clones[cloneID].mask )
                {
                id_found = i;
                cntr++;
                }
        }

   if( cntr > 1 )
        AM_Error("++** NewGetSibling: multiple sibling map of %d exiting\n",
                cs_GIDtoID[gid].clone_id[cloneID] );

   if( cntr == 0 )
	{
   	if( cs_do_debug > 10 )
        	printf("%s: end\n", fn );
        return -1;
	}

   sibling = &cs_vlpmapping[id_found][gid];

   if( cs_do_debug > 10 )
       	printf("%s: end\n", fn );
   return sibling->MyID;
}



/* ---------------------------------------------------------------------------
** OP_ProcessLP ()
**
** DESCRIPTION:	SAME as GetCareTaker except returns -1 if non valid
**
**---------------------------------------------------------------------------*/
int OP_ProcessLP( int rcv_LP, double rcv_ts, int priority_type )
{
   char  	fn[]	= "[CS_ProcessLP]";
   int		cloneID;
   int		p_cloneID; /* process clone ID */
   int		process_lp;
   int		gid;
   cs_vlpmap_t  *c_vlp;

   int                   XX_loop_2=0;
   int                   XX_loop_1=0;

   if( rcv_LP < cs_nlp ) 
	return rcv_LP;
   else if( cs_optimistic == 0 )
	return rcv_LP;

   if( cs_do_debug > 100 )
        printf("%s: begin rcv_LP=%d, rcv_ts=%f, btype\n", fn, 
		rcv_LP, rcv_ts, priority_type );

   gid 		= cs_IDtoGID[rcv_LP];
   cloneID 	= cs_IDtoClone[rcv_LP];
   c_vlp 	= &(cs_vlpmapping[cloneID][gid]);

   if( cs_vlpmapping[cloneID][gid].active == 1 )
	{
	if( cs_do_debug > 50 )
                printf("%s: already active=%d\n", fn, rcv_LP );
	cs_CNTR_SAMPLE(gid,CS_PROC_FINDLP,1,XX_loop_1,XX_loop_2);
	return rcv_LP;
	}

   XX_loop_1 = cloneID;
   if( (process_lp = OP_GetSibling( gid, cloneID )) < 0 )
	{
       	process_lp = OP_GetParent( rcv_LP, gid, cloneID, rcv_ts, &XX_loop_2 );
	}
   cs_CNTR_SAMPLE(gid,CS_PROC_FINDLP,14,XX_loop_1,XX_loop_2);

   p_cloneID = cs_IDtoClone[process_lp];

        	/* child lp has been born and is a clone */
                /* the LP was not rolled back and reset to MAXFLOAT */
                /* the LP was not rolled back nor its parent */
                /* check child rolled back, rb parent further */
   if(  ( (process_lp != rcv_LP) && !(c_vlp->BirthTS == MAXFLOAT) ) 
      && (
           ((rcv_ts < c_vlp->BirthTS) && (c_vlp->BirthTS < MAXFLOAT))
           || ((rcv_ts < c_vlp->BirthTS) && (rcv_ts < GetLPsLVT(process_lp)))
         )
      )
        {
        AM_Error(" +** missed a case in ParenCare\n");
        }

   if( (cs_vlpmapping[p_cloneID][gid].active == 0) && (priority_type != -2) )
	{
   	if( cs_do_debug > 10 )
       		printf("%s: end\n", fn );
	return -1;
        }
   if( cs_do_debug > 100 )
   	printf("%s: end\n", fn );
   return process_lp;
}



/* ---------------------------------------------------------------------------
** OP_GetLPClone()
**
** DESCRIPTION:	generic pointer for MyMsgData,
**
**---------------------------------------------------------------------------*/
int OP_GetLPClone( cs_clone_t *vmap, 
		double birth_time, int *do_execute_me, int activate )
{  
  char                  fn[]            = "[OP_GetLPClone]";
  unsigned long int     my_mask         = 0L;
  unsigned long int     my_children     = 0L;
  unsigned long int     my_siblings     = 0L;
  unsigned long int     clones_children = 0L;
  unsigned long int     clones_siblings = 0L;

  unsigned long int     move_to_clone   = 0L;
  unsigned long int     vps_requested   = 0L;
  unsigned long int     vps_unaffected  = 0L;
  int                   my_cloneID;
  int                   gid, uid;
  int                   rcv_uid;
  int                   clones_cloneID;

  int                   do_execute_parent;
  int			clone_uid;


  if( cs_do_debug > 100 )
        printf("%s: begin\n", fn );

  clone_uid	  = -1;

  my_cloneID      = cs_IDtoClone[vmap->rcv_uid];
  gid             = cs_IDtoGID[vmap->rcv_uid];
  rcv_uid         = vmap->rcv_uid;
  uid         	  = vmap->rcv_uid;

  my_mask         = cs_clones[my_cloneID].mymask;
  my_children     = cs_clones[my_cloneID].children;
  my_siblings     = cs_vlpmapping[my_cloneID][gid].siblings;

  vps_requested  =   (vmap->snd_mask) & (vmap->rcv_mask);
  vps_unaffected = (~(vmap->snd_mask) & (vmap->rcv_mask));

  /* for conservative cloning below is alway true */
  do_execute_parent = my_mask & vps_requested;

  if( do_execute_parent )
        move_to_clone = vps_unaffected;
   else
        move_to_clone = vps_requested;

  if( move_to_clone == 0L )
	{
	*do_execute_me  = 1;
	clone_uid 	=-1;
   	if( cs_do_debug > 10 )
   		printf("%s: end (no cloning=%d)\n", fn, uid );
  	cs_CNTR_SAMPLE(vmap->rcv_gid,CS_SYS_GETLPCLN,14,0,0);
        return clone_uid; 	/* return do_execute parent = TRUE */
	}

  /* compute new clone's cloneID */
  get_lowest( clones_cloneID, move_to_clone, cs_num_clones )
  clone_uid = cs_GIDtoID[gid].clone_id[clones_cloneID];

  if( clones_cloneID < my_cloneID )
    AM_Error("+**%s:(my_cloneID=%d, ch_cloneID=%d\n", my_cloneID, clones_cloneID);

  clones_children = cs_clones[clones_cloneID].children;
  clones_siblings = (move_to_clone) & ~(clones_children);

  /* un-map clone from my mapping or siblings */
  if( move_to_clone & my_children )
     cs_vlpmapping[my_cloneID][gid].unmapped |= move_to_clone & my_children;
  else if( move_to_clone & my_siblings )
     {
     cs_vlpmapping[my_cloneID][gid].siblings &= ~(move_to_clone);
     if( cs_do_debug > 100 )
		{
     		printf("*%s: at %lf changing uid=%d\n", fn, birth_time, rcv_uid );
     		XPrintIMapping( cs_vlpmapping[my_cloneID][gid].siblings, uid,
		"pvlp\0", 1, is_num_clones);
		}
     }

  /* set child mapping */
  cs_vlpmapping[clones_cloneID][gid].active             = 0; 
  cs_vlpmapping[clones_cloneID][gid].BirthTS            = birth_time;
  cs_vlpmapping[clones_cloneID][gid].CurrChildBirthTS   = AM_NEGINFINITY;
  cs_vlpmapping[clones_cloneID][gid].parentID           = vmap->rcv_uid;
  cs_vlpmapping[clones_cloneID][gid].parentCloneID      = cs_IDtoClone[rcv_uid];
  cs_vlpmapping[clones_cloneID][gid].siblings           = clones_siblings;
  cs_vlpmapping[clones_cloneID][gid].unmapped           = 0L;

  /* set vmap */
  vmap->clones_siblings	= clones_siblings;
  vmap->childs_cloneID	= clones_cloneID;

  if( cs_do_debug > 1 )
	{
  	printf("*%s: %s cloning at rcvTS=%lf sndTS=%lf gid=%3d uid=%3d clone_uid=%3d "
			"cloneID=%d ch_cloneID=%d\n", 
		fn, 
		(do_execute_parent) ? "BACK_CL": "CARE_CL",
		birth_time, vmap->snd_time, gid, uid, clone_uid, my_cloneID, clones_cloneID );
  	XPrintIMapping( vmap->snd_mask, uid, "send\0", 1, is_num_clones);
  	XPrintIMapping( vmap->rcv_mask, uid, " rcv\0", 1, is_num_clones);
  	XPrintIMapping( move_to_clone, uid, "move\0", 1, is_num_clones);
  	XPrintIMapping( clones_children, uid, "kids\0", 0, is_num_clones);
  	XPrintIMapping( vmap->clones_siblings, uid, "sibs\0", 0, is_num_clones);
	}

  *do_execute_me  = do_execute_parent;
  if( cs_do_debug > 10 )
 	printf("%s: end (do cloning) uid=%d\n", fn, uid );

  cs_CNTR_SAMPLE(vmap->rcv_gid,CS_SYS_GETLPCLN,33,0,0);

  return clone_uid;
}



/* ---------------------------------------------------------------------------
** RBBirthEventSnd()
**
** DESCRIPTION: cancellation messages should not access SV
**      dont work with restore
**---------------------------------------------------------------------------*/
void OP_RBBirthEventSnd( void *SV, struct MsgData *CSMsg, int MyPE )
{
   unsigned long int 	child_map;
   unsigned long int	rb_sibs;
   unsigned long int 	iivps_sibs;
   int 			alive		= 1;
   cs_vlpmap_t 		*p_vlp, *c_vlp;
   int			child_uid, child_gid, child_cloneID;
   double		time_now;

   int			uid, gid, cloneID;
   char  		fn[]	= "[OP_RBBirthEventSnd]";

   set_GTWState()

   if( cs_do_debug > 100 )
        printf("%s: begin\n", fn );

   if( CloneSim_UID() != CurEvent->LP )
	{ /* MH_HERE */
        AM_Error( "+** %s: State->ID(%d) != CurEvent->LP(%d)\n", fn,  
		CloneSim_UID(), CurEvent->LP );
	}
   if( CurEvent->BirthType == -2 )
        AM_Error( "+** %s: BirthType=%d for a Snd Event!\n", 
		fn, CurEvent->BirthType );

   /* get identity */
   uid 		= CloneSim_UID();
   gid 		= cs_IDtoGID[uid];
   cloneID 	= cs_IDtoClone[uid];
   p_vlp	= &cs_vlpmapping[cloneID][gid];
   time_now	= TWNow();

   if( cs_do_debug > 1 )
	printf("%s: *cloning rolling back uid=%3d gid=%3d at=%lf\n", fn, uid, gid, time_now );
    
   /* error checking */
   if( CurEvent->BirthType < 0 )
        AM_Error( "+** %s: event should not be unprocessed\n", fn );

   if( p_vlp->BirthTS == MAXFLOAT )
	alive = 0;

   /* get clone info */
   child_uid  		= CurEvent->BirthType;
   child_gid 		= cs_IDtoGID[child_uid];
   child_cloneID 	= cs_IDtoClone[child_uid];
   child_map 		= cs_clones[cloneID].children;

   /* get states */
   p_vlp		= &cs_vlpmapping[cloneID][gid];
   c_vlp		= &cs_vlpmapping[child_cloneID][gid];

   /* re-set child clone */
   c_vlp->BirthTS 	= MAXFLOAT;
   c_vlp->active  	= 0;
   c_vlp->siblings 	= 0L;
   c_vlp->unmapped 	= 0L;

   c_vlp->parentCloneID	= -1; /* 06/16/99 */
   c_vlp->parentID	= -1; /* 06/16/99 */
   

   /* restore my mappings */
   if( alive )
	{
	rb_sibs = CSMsg->Admn_Data.rb_siblings;
	if ( child_map & cs_clones[child_cloneID].mask )
		p_vlp->unmapped &= ~(cs_clones[child_cloneID].mask);
	else
		p_vlp->siblings |=  (cs_clones[child_cloneID].mask);

	p_vlp->unmapped &= ~( (rb_sibs) &  (child_map) );
	p_vlp->siblings |=  ( (rb_sibs) & ~(child_map) );
	}

   /* set event */
   CurEvent->LP 	= CurEvent->oldLP; /* RBBirthEventSnd */
   CurEvent->oldLP 	= -1;              /* RBBirthEventSnd */
   CurEvent->BirthType 	= -1;
   CurEvent->RBFunc 	= NULL;

   if( CurEvent->LP > cs_vlp ) 
	{
	AM_Error("** %s exiting invalid LP=%d\n", fn, CurEvent->LP );
	}

   if( cs_do_debug > 10 )
 	printf("%s: end\n", fn );

   cs_CNTR_SAMPLE(gid,CS_RBFUNC_SND,32,0,0);
}



/* ---------------------------------------------------------------------------
** OP_RBBirthEventRcv()
**
** DESCRIPTION: cancellation messages should not access SV
**      dont work with restore
**---------------------------------------------------------------------------*/
void OP_RBBirthEventRcv( void *SV, struct MsgData *CSMsg, int MyPE )
{
   int			uid, gid, cloneID;
   cs_vlpmap_t 		*p_vlp;
   char  		fn[]	= "[OP_RBBirthEventRcv]";
   double		time_now;

   set_GTWState()

   if( cs_do_debug > 100 )
        printf("%s: begin\n", fn );

   /* get identity */
   uid 			= CloneSim_UID();
   gid 			= cs_IDtoGID[uid];
   cloneID 		= cs_IDtoClone[uid];
   time_now		= TWNow();

   if( cs_do_debug > 1 )
	printf("%s: *cloning rolling back uid=%3d gid=%3d at=%lf\n", fn, uid, gid, time_now );

   p_vlp		= &cs_vlpmapping[cloneID][gid];

#if 1
   p_vlp->active	= 0;
#endif
#if 0
   /* set states */
   p_vlp->siblings	= 0L;
   p_vlp->unmapped	= 0L;
   p_vlp->BirthTS	= MAXFLOAT;

   p_vlp->parentCloneID	= -1; /* 06/16/99 */
   p_vlp->parentID	= -1; /* 06/16/99 */

   /* set event */
   CurEvent->RBFunc 	= NULL;
   CurEvent->BirthType 	= -1;

   if( cs_do_debug > 10 )
 	printf("%s: end\n", fn );
#endif
  

   cs_CNTR_SAMPLE(gid,CS_RBFUNC_RCV,15,0,0);
}



/* ---------------------------------------------------------------------------
** OP_InitVLPs()
**
** DESCRIPTION:  could probably remove some of these
**
**---------------------------------------------------------------------------*/
void OP_InitVLPs( struct MsgData *CSMsg, int MyPE )
{  
   char 		fn[] = "[OP_InitVLPs]";
   struct Event 	*CurEvent; 
   struct LPState 	*CurState;
   double		time_now;

   int cloneID, gid, uid ;
   int i,errflag=0;

  if( cs_do_debug > 100 )
        printf("%s: begin\n", fn );

   CurState = GState[MyPE].CurState;
   CurEvent = GState[MyPE].CurEvent;

   uid 		= CurState->ID;
   gid          = cs_IDtoGID[uid];
   cloneID      = cs_IDtoClone[uid];
   time_now	= TWNow();

  if( cs_do_debug > 1 )
        printf("%s: *cloning activating uid=%3d gid=%3d at=%lf\n", fn, uid, gid, time_now );

   cs_vlpmapping[cloneID][gid].active = 1;
   cs_vlpmapping[cloneID][gid].CurrChildBirthTS = NEGINFINITY;
   if( cs_vlpmapping[cloneID][gid].parentCloneID != CSMsg->Admn_Data.cloneID )
	{
   	errflag=1;
		printf("+== 1. awry (uid=%3d gid=%3d):( vlp=%3d msg=%3d ) "
			"(parentCloneID)\n",
			uid, gid, cs_vlpmapping[cloneID][gid].parentCloneID,
			CSMsg->Admn_Data.cloneID);
	}
   cs_vlpmapping[cloneID][gid].parentCloneID 		= CSMsg->Admn_Data.cloneID;

   if( cs_vlpmapping[cloneID][gid].parentID != CSMsg->Admn_Data.rcv_uid )
	{
   	errflag=1;
		printf("+== 2. awry (uid=%3d gid=%3d):(vlp=%3d msg=%3d)"
		  " (parentID) \n",
		  uid, gid, cs_vlpmapping[cloneID][gid].parentID,
		  CSMsg->Admn_Data.rcv_uid
		  );
	}
   cs_vlpmapping[cloneID][gid].parentID = CSMsg->Admn_Data.rcv_uid;

   if( cs_vlpmapping[cloneID][gid].BirthTS != TWNow() )
	{
   	errflag=1;
		printf("+== 3. awry (uid=%3d gid=%d):(vlp=%lf msg=%lf)"
			" (BirthTS) \n",
			uid, gid, cs_vlpmapping[cloneID][gid].BirthTS, TWNow()
			);
	}
   cs_vlpmapping[cloneID][gid].BirthTS = TWNow();
   cs_vlpmapping[cloneID][gid].unmapped = 0L;
   if( cs_vlpmapping[cloneID][gid].siblings != CSMsg->Admn_Data.snd_siblings ) 
	{
	AM_Error( "+** %s uid=%d sibling mapping", fn, uid );
	}
   cs_vlpmapping[cloneID][gid].siblings = 0L;
   cs_vlpmapping[cloneID][gid].siblings |= CSMsg->Admn_Data.snd_siblings; 

   if ( errflag == 1 )
	AM_Error("+** %s uid=%d gid=%3d awry, degrades performance...", 
	fn, uid, gid );
   }



/* ---------------------------------------------------------------------------
** OP_SysLP()
**
** DESCRIPTION:	generic pointer for MsgData,
**
**---------------------------------------------------------------------------*/
void OP_SysLP( void *SV, void *M, int MyPE )
{  
   struct MsgData 	*TWMsg;
   char                 fn[]    = "[OP_SysLP]";
   cs_vlpmap_t  	*cur_vlp;
   cs_clone_t           vmap;

   int 			cloneID, gid, uid ;
   double		time_now;

   AM_ADDR_TYPE         msg_ptr;
   AM_ADDR_TYPE         user_msg_ptr;
   struct MsgData       *CSMsg;
   extern		double GVT;

   /* access TWNow(), and CloneSim_UID(), TWMe() */
   set_GTWState()

   /* get identity */
   uid 		= CurState->ID;
   gid 		= cs_IDtoGID[uid];
   cloneID	= cs_IDtoClone[uid];
   cur_vlp	= &cs_vlpmapping[cloneID][gid];

   /* get time */
   time_now 	= TWNow();

   msg_ptr 	 = CurEvent->Msg.Data;
   CSMsg = (struct MsgData *) msg_ptr;
   user_msg_ptr = CSMsgSize + msg_ptr;

   /* get cloning structure: macro */
   vmap = CS_GetClone_t( CSMsg, gid, uid, cloneID );

   if( cs_do_debug > 10 )
        printf("*%s: *(uid=%3d gid=%3d clone=%3d at %lf)\n", fn,
                uid, gid, cloneID, time_now );

   /* forward message to appropriate function handler */
   switch ( CSMsg->Admn_Data.type )
     {
     case MSG_USER:
	{
   	if( cur_vlp->BirthTS == MAXFLOAT )
		AM_Error( "+**  %s: (%d) not born\n", fn, uid );
   	if( cur_vlp->BirthTS > time_now )
		{
		AM_Error("+** cur_vlp->BirthTS > time_now %.4f > %.4f\n",
			cur_vlp->BirthTS, time_now );
		}
	if( cs_vlpmapping[cloneID][gid].active == 0 )
		{
		AM_Error("+*** (%d) not active %.4f (%d) from %d\n",
			uid, time_now, CurEvent->BirthType,
			CurEvent->SrcLP );
		}
   	if( CurEvent->BirthType != -1 )
		{
		AM_Error("+** BirthType(%d) is set wrong on (%d, %d) at %.4f\n",
			CurEvent->BirthType, uid, gid, time_now  );
		}

	/* make sure we havent received this message already */
	if( (CSMsg->Admn_Data.a_forward_msg == 1)
	  	&& (CSMsg->Admn_Data.forwarded & cs_clones[cloneID].mask )
		)
		{
		if( cs_do_debug > -10 )
                    printf( "+== uid(%d) != CurEvent->LP(%d) for a forwarded "
                            " msg\n", uid, CurEvent->LP); 
		if( uid != CurEvent->LP )
			CurEvent->LP = uid;
		if( cs_do_debug > -10 )
		    printf("+== doubly received (%d) %.4f\n", uid, time_now );
		}
	else
		{
        	int                     rcv_forward;
        	unsigned long int       token;
        	char                    cn[]    = "[O_MSG_USER]";
        	double                  birth_ts;
        	int                     clone_uid;
        	int                     do_execute=1;

        	birth_ts = cs_vlpmapping[cloneID][gid].BirthTS;

        	if( cs_do_debug > 100 )
               	 	{
               	 	printf("*        %s: (snd=%4d rcv=%4d, clone=%4d) "
                       	 	"at %lf\n",
                        	cn,
                        	vmap.snd_uid, vmap.rcv_uid, vmap.cloneID, time_now );
                	}


        	/* check for forwarding */
        	if( (rcv_forward = CS_GetLPForward( vmap, &token )) >= 0 )
                	{
                	unsigned long int dfrw = vmap.done_forwarded;
                	if( cs_do_debug > 100 )
                       		 {
                       		 printf("*clonin forward: to=%3d frm=%3d uid=%3d "
                               	 	"(snd_time=%lf rcv_time=%lf)\n",
                                	rcv_forward,
                                	vmap.snd_uid,
                                	uid,
                              		vmap.snd_time,
                               		time_now
                                	);
                        	iPrintIMapping(vmap.snd_mask," snd\0",1,is_num_clones);
                        	iPrintIMapping(vmap.rcv_mask," rcv\0",0,is_num_clones);
                        	iPrintIMapping(dfrw,"dfrw\0",0,is_num_clones);
                        	iPrintIMapping(token,"nfrw\0",0,is_num_clones);
                        	}

                	if( rcv_forward < vmap.rcv_uid )
                       	 	{
                        	printf("%s: forwarding to a lower uid (%3d) "
                               		 "than myself(%3d)\n", fn,
                                	 rcv_forward, uid );
                        	fflush(stdout);
                        	exit(1);
                        	}

   			if( rcv_forward > cs_vlp ) 
				{
				AM_Error("** %s forwarding to an invalid LP=%d\n", 
					fn, rcv_forward );
				}

                	_TWGetMsg( 0, rcv_forward, sizeof(struct MsgData), MyPE );
                	TWMsg = (struct MsgData *) MyState->NewBuf->Msg.Data;
                	MCpSt( CSMsg, TWMsg, TWMsgSize );
                	TWMsg->Admn_Data.type           = MSG_USER;
                	TWMsg->Admn_Data.a_forward_msg  = 1;
                	TWMsg->Admn_Data.rcv_uid        = rcv_forward;
                	TWMsg->Admn_Data.forwarded      = token;
                	TWSend();
                	}

        	/* check for cloning */
        	if( (clone_uid = OP_GetLPClone( &vmap, time_now, &do_execute, 0 )) > 0 )
            		{
			if( cs_do_debug > 1 )
			   printf( "%s: *cloning gid=%3d uid=%3d child_uid=%d at %lf\n",
				fn, gid, uid, clone_uid, time_now );
					
            		if( special > 0 )
                	  TWLP[gid].CProc(TWLP[clone_uid].State, SV, 1, MyPE );
            		else
                	  MCpSt( SV,TWLP[clone_uid].State,TWLP[clone_uid].LPStateSize );

            		CSMsg->Admn_Data.rb_siblings = vmap.clones_siblings;
            		CurEvent->oldLP              = CurEvent->LP;    /* cloning */
            		CurEvent->LP                 = uid;		/* cloning */
            		CurEvent->BirthType          = clone_uid;
            		CurEvent->RBFunc             = (IOFuncPtr) OP_RBBirthEventSnd;

   			if( CurEvent->LP > cs_vlp ) 
				{
				AM_Error("** %s exiting invalid LP=%d\n", fn, CurEvent->LP );
				}

            		_TWGetMsg( 0, clone_uid, sizeof(struct MsgData), MyPE );
                	MyState->NewBuf->BirthType = -2;     /* prioritize message */
                	MyState->NewBuf->RBFunc = (IOFuncPtr) OP_RBBirthEventRcv;

                	TWMsg = (struct MsgData *) MyState->NewBuf->Msg.Data;
                	if( do_execute == 0 )
                         	{
                        	AM_MCpSt( CSMsg, TWMsg, TWMsgSize );
                        	TWMsg->Admn_Data.type   = MSG_BIRTH_LP;;
                        	}
                	else
                        	{
                        	TWMsg->Admn_Data.type   = MSG_BIRTH_DUMMY;;
                        	}

                	TWMsg->Admn_Data.a_forward_msg  = 0;
                	TWMsg->Admn_Data.snd_uid        = uid;
                	TWMsg->Admn_Data.rcv_uid        = /*clone_uid;*/ uid;
                	TWMsg->Admn_Data.cloneID        = /*cs_IDtoClone[clone_uid];*/ 
								cloneID;
                	TWMsg->Admn_Data.snd_siblings   = vmap.clones_siblings;

                	if( cs_do_debug > 1 )
                       	 	{
                        	printf("%s: %s at %lf\n", fn, cn, time_now );
                        	XPrintIMapping( vmap.clones_siblings, uid, " sib\0", 1,
                        	is_num_clones);
                        	}

                	TWMsg->Admn_Data.snd_unmapped   = 0L;
            		TWSend();
          		}

        	if( do_execute == 0 ) return;
		if( is_phold )
        		IS_TestState( cn, SV, uid, gid, cloneID, time_now );
        	is_ui[uid].Proc( SV, user_msg_ptr, MyPE );
		}

	break;
	}
     case MSG_BIRTH_LP:
	{
        char                    cn[]    = "[MSG_BIRTH_LP]";

        if( CurEvent->BirthType != -2 )
		{
                AM_Error("+** %s,%s,[LP] BirthType=%d GVT=%lf uid=%d gid=%d cloneID=%d", 
			fn, cn, 
			CurEvent->BirthType, GVT, uid, gid, cloneID );
		return;
		}

	OP_InitVLPs( CSMsg, MyPE );

	if( is_phold )
		IS_TestState( cn, SV, uid, gid, cloneID, time_now );
	is_ui[CurEvent->LP].Proc( SV, user_msg_ptr, MyPE);

	break;
	}
     case MSG_BIRTH_DUMMY:
	{
        char                    cn[]    = "[MSG_BIRTH_DUMMY]";

        if( CurEvent->BirthType != -2 )
		{
                AM_Error("+** %s,%s,[DUMMY] BirthType=%d GVT=%lf uid=%d gid=%d cloneID=%d", 
			fn, cn, 
			CurEvent->BirthType, GVT, uid, gid, cloneID );
		return;
		}

	OP_InitVLPs( CSMsg, MyPE );

	break;
	}
     default:
	{
	printf ( "%s:  unrecognized type (%d) \n", fn, CSMsg->Admn_Data.type );
	fflush ( stdout );
	break;
	}
     }  /* switch */
}


