/******************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.5                             ***
***         Copyright 1997 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/
#include "gtw.h"
#include "phold.h"
#include "cs_api.h" 		/* only should include the api header file */

#define APP_REMOTE   1

int   	app_remote=1;
int   	ClonePeriod=10;		/* Global XXX HERE */
int 	MsgPop; 		/* message population */
TWSeed 	Seeds[MAXLPs]; 		/* random number generator seeds */
int	cloneID=0;		/* Global cloneID */

/* external command line available from GTW */
extern  int do_z;
extern  int do_Z;

extern int  CloneSim( void *SV, struct MsgData *M, int MyPE );

void IPHoldLP( struct MyLPState *, int );
void ClonePHoldLP( struct MyLPState *SV, struct MyMsgData *M, int MyPE );

/* ---------------------------------------------------------------------------
** IPHoldLP  ()
**
** DESCRIPTION: init each LP
**
**---------------------------------------------------------------------------*/
void IPHoldLP( struct MyLPState *SV, int MyPE )
{  
   struct MyMsgData 	*TWMsg;
   int 			i, receiver_lp;
   TWTime 		ts;
   TWEachSeed_t 	s1, s2;
   int 			dst;
   int			myLP;
   char          	fn[]    = "[IPHoldLP]";

   struct LPState * CurState;
   CurState = GState[MyPE].CurState;

   if ( cs_do_debug > 100 )
	printf("%s: MyPE=%d\n", fn, MyPE );

   myLP = TWMe();
   SV = (struct MyLPState *) TWLP[TWMe()].State;

   /* specify variables to be automatically checkpointed */
   TWAutoCheck((char*)&(SV->CVars), sizeof(struct MyAutoVars));

   /* REPLICATION TWMe here, on the other hand, do need to use TWMe
    * since Seed array is only of size TWGetNumLPs() */
   /* init random number generator seeds */
   TWRandGetSeeds (&(Seeds[TWMe()]), &s1, &s2);
   TWRandSetSeeds (&(SV->CVars.Seeds), s1, s2);

   /* initialize state variables */
   if ( CurState->IncrSave )
	{
	SV->ISVars.Count = 0;
	for ( i=0; i<MSZ; i++ )
		{
                SV->ISVars.LastTS[i] = 0.0;
		}
	} 

   SV->CVars.Count = 0;
   SV->CVars.CloneCount = 0;
   SV->CVars.cloned = 0;

   /* send initial messages */
   if ( app_remote > 0 ) /* defined by input file to phold */
	{
	for ( receiver_lp = TWMe(); receiver_lp < MsgPop; 
						receiver_lp += TWGetNumLPs())
		{
		ts = TWRandExponential(&(SV->CVars.Seeds), 1.0);
		dst = TWMe();
		TWGetMsg(ts, dst, sizeof(struct MyMsgData));
		if( cs_do_debug > 10 )
			printf("%s: got message (LP=%d)\n", fn, dst );
   		TWMsg->spreading = 0;			/* IPHold HERE */
		TWMsg->counter = 1;
		TWMsg->CloneCount = SV->CVars.CloneCount;
		TWMsg->numself
		     = TWRandInteger(&(SV->CVars.Seeds), 1, (2*app_remote)+1 );

		if ( cs_do_debug > 10 )
			printf("%s: [MyPE=%d, MyLP=%d] ns=%d cnt=%d\n", 
			   fn, MyPE, myLP,  TWMsg->numself, TWMsg->counter );

		TWSend();
		}
        }
   else
	{
	for ( receiver_lp=TWMe(); receiver_lp < MsgPop; receiver_lp += 
			TWGetNumLPs())
	        {
	        ts = TWRandExponential(&(SV->CVars.Seeds), 1.0);
	        dst = TWRandInteger(&(SV->CVars.Seeds), 0, TWGetNumLPs()-1);
	        TWGetMsg(ts, dst, sizeof (struct MyMsgData));
   		TWMsg->spreading = 0;			/* IPHold HERE */
	        TWMsg->counter = 1;
		TWMsg->CloneCount = SV->CVars.CloneCount;
	        TWMsg->numself
	          = TWRandInteger(&(SV->CVars.Seeds), 1, 2*NUM_SENDS_TO_SELF+1);
	        TWSend();
	        }
        }
}


/* ---------------------------------------------------------------------------
** ClonePHoldLP  ()
**
** DESCRIPTION:  THIS event is called conservatively when cloned
**
**---------------------------------------------------------------------------*/
void ClonePHoldLP( struct MyLPState *SV, struct MyMsgData *M, int MyPE )
  {  
   char          	fn[]    = "[ClonePHoldLP]";
   struct MyMsgData 	*TWMsg;

   struct PEState  	*MyState;      
   struct LPState 	*CurState;

   TWTime 		ts;
   int 			dst;
   int			myLP;
   int			uid;
   int			ch_uid=-1;
   int			old_cloneID;

   MyState  = &GState[MyPE];
   CurState = GState[MyPE].CurState;

   uid		= CloneSim_UID();

   CloneSim_CLONE_RT[SV->CVars.CloneCount-1] = TWNow();	 /* IO4 */

   SV->CVars.cloned = cs_clone_parent; 

   old_cloneID=cloneID;
   if( cs_optimistic == 2 )
   	cloneID 	= CloneSim_Create(CloneSim_UID(), TWNow());
   else
   	cloneID 	= CloneSim_Create(CloneSim_UID(), TWNow());

   SV->CVars.cloned = !(cs_clone_parent); 

   if( cloneID > 0 )
   	ch_uid 	= cs_GIDtoID[cs_IDtoGID[uid]].clone_id[cloneID];

   if( cs_do_debug >= 0 )
	{
	char *msg;
	if( cloneID < 0 )
		msg = (char *) strdup("ignored");
	else
		msg = (char *)strdup("successful");
   	printf("************ %s cloning %s: gid=%3d"
		" uid=%3d"
		" cloneID=%2d"

		" ch_uid=%3d"
		" ch_cloneID=%2d"

		" cloned at %lf\n",
		msg,
		fn, TWMe(), 
		CloneSim_UID(), cs_IDtoClone[CloneSim_UID()], 
		ch_uid,  cloneID,  
		TWNow()
		);
	}

	/* ------------------------------------------------------------------- 
	 *
	 * o last clone is the "original simulation"
	 * o can ONLY change original state
	 * o can do MSG sends and these will be independent of the new clone
	 * o there may be simulations
	 *
	 * ------------------------------------------------------------------- 
	 */
  }


/* ---------------------------------------------------------------------------
** PHoldLP  ()
**
** DESCRIPTION: event handler for phold
**
**---------------------------------------------------------------------------*/
void PHoldLP( struct MyLPState *SV, struct MyMsgData *M, int MyPE )
{  
   struct MyMsgData 	*TWMsg;
   struct LPState 	*CurState;
   TWTime 		ts;
   int 			dst;
   char          	fn[]    = "[PHoldLP]";
   int			myLP;
   double 		time_now;

   CurState = GState[MyPE].CurState;
   
   myLP = TWMe();
   time_now = TWNow();

   if ( cs_do_debug > 9 )
       	printf("%s: begin (uid=%d, gid=%d) at %4.2lf\n", 
		fn, CloneSim_UID(), TWMe(), TWNow() );

   if ( M->spreading == 1 ) 				/* PHold XXX HERE */
	return;

   if ( CurState->IncrSave )
	{
     	TWCheckTWTime(&(SV->ISVars.LastTS[SV->ISVars.Count % MSZ]));
     	SV->ISVars.LastTS[SV->CVars.Count % MSZ] = TWNow();

     	TWCheck(&(SV->ISVars.Count));
     	SV->ISVars.Count++;
   	}

   SV->CVars.LastTS[SV->CVars.Count % MSZ] = TWNow();
   SV->CVars.Count++;

  if ( (TWMe() == cs_GIDtoID[0].clone_id[0]) && (SV->CVars.cloned == 0)  ) 
        {
        if ( CloneSim_CLONETIME[SV->CVars.CloneCount] > 0.0 )
                {
                if ( ((SV->CVars.CloneCount) < CloneSim_NumClones)
                        && (CloneSim_CLONETIME[SV->CVars.CloneCount] < TWNow()) )
                        {
                        SV->CVars.CloneCount++;
   			TWGetMsg(0, TWMe(), sizeof(struct MyMsgData)); /*PHERE not CloneSim_UID() */
			TWBlockingIOSend( (IOFuncPtr) ClonePHoldLP );
			if ( 1 ) 
				{
        			printf("2. ************* PHoldLP: CLONING  "
					"uid=%d, gid=%d clone_cnt=%d -- %.4f cloneID=%d\n",
                	   		CloneSim_UID(), TWMe(), 
					SV->CVars.CloneCount, 
					TWNow(),
					cloneID 
					);
				fflush(stdout);
				}
                        }
                }
        }

   /* bounce message */
   if ( app_remote > 0 )
	{
	ts = 1.0;
	if (cs_do_debug > 9 )
        	printf("%s: [MyPE=%d, MyLP=%d] numself=%d count=%d\n", 
                        fn, MyPE, myLP,  M->numself, M->counter );

	if ( M->numself == 0 )
		{
		printf("+** error numself == 0\n" );
		fflush(stdout);
		exit(1);
		}

	if ( (M->counter % M->numself) == 0 )
		dst = (TWMe()+1)% (TWGetNumLPs());
	else
	        dst = TWMe();
	}
   else
	{
	ts = TWRandExponential(&(SV->CVars.Seeds), 1.0);
	dst = TWRandInteger(&(SV->CVars.Seeds), 0, (TWGetNumLPs())-1);
	}



   TWGetMsg(ts, dst, sizeof (struct MyMsgData));
   TWMsg->spreading = 0;			        /* HERE */
   TWMsg->counter = M->counter + 1;
   TWMsg->numself = M->numself;
   TWMsg->CloneCount = SV->CVars.CloneCount;		/* HERE */

   TWSend();

   if ( cs_do_debug > 10 )
       	printf("%s: end (frm=%d, to=%d) at %lf\n", fn, TWMe(),dst,TWNow() +ts );
}


/* ---------------------------------------------------------------------------
** FPHoldLP  ()
**
** DESCRIPTION: final event handler
**
**---------------------------------------------------------------------------*/
void FPHoldLP (struct MyLPState * SV, int MyPE)
{  
   int i;

   struct LPState * CurState;
   CurState = GState[MyPE].CurState;

   if (!CurState->IncrSave) return;
}


/* ---------------------------------------------------------------------------
** TWInitAppl  ()
**
** DESCRIPTION: final event handler
**
**---------------------------------------------------------------------------*/
void TWInitAppl(int argc, char ** argv)
{
   int i,j;
   int my_lps;
   extern int CloneSim_scale_size;
   extern int CloneSim_map_msgs;
   extern int CloneSim_read_files;

   void IPHoldLP();
   void PHoldLP(), FPHoldLP();

   my_lps = 4;
   MsgPop = 256;

   if( CloneSim_read_files == 0 ) 
   	{
   	ClonePeriod=4;
   	app_remote=CloneSim_map_msgs;
   	}
   else
   	{
   	if ( TWGetNumMachines() == 1 ) 
		{
		if ( show_noprint == 0 )
			{
			fprintf ( stderr, "\n");
			fprintf( stderr, "Enter number of LPs      -> ");
			}
		scanf( "%d", &my_lps);
		if ( show_noprint == 0 )
			{
			fprintf ( stderr, "\n");
			fprintf( stderr, "Enter message population -> ");
			}
		scanf("%d", &MsgPop);
		if ( show_noprint == 0 )
			{
			fprintf (stderr, "\n"); fflush(stderr);
			fprintf( stderr, "Enter Cloning Period     -> ");
			}
		scanf( "%d", &ClonePeriod);
		if ( show_noprint == 0 )
			{
			fprintf (stderr, "\n"); fflush(stderr);
			fprintf( stderr, "Enter APP REMOTE (0/1)   -> ");
			}
		scanf( "%d", &app_remote );
		if ( show_noprint == 0 )
			{
			fprintf (stderr, "\n"); fflush(stderr);
			}
		}
   	}

  if( do_z != 0 )
	my_lps = do_z;
  if( do_Z != 0 )
	MsgPop = do_Z;

   /* printf("my_lps=%d MsgPop=%d\n", my_lps, MsgPop ); */

   MsgPop *= CloneSim_scale_size;
   my_lps *= CloneSim_scale_size;

   TWSetNumLPs(my_lps);

   if ( (TWGetNumLPs() < 1) || ( TWGetNumLPs() > MAXLPs ))
	error ("Illegal number of LPs (%d >%d)\n", TWGetNumLPs(), MAXLPs );

   if ( (MsgPop < 1) || (MsgPop > MAXMSGS))
      	error ("Illegal number of messages\n");

   if ( show_noprint == 10 )
   	printf ("TWnpe=%d, TWGetNumGlobalPEs() = %d\n",
		TWnpe, TWGetNumGlobalPEs());

   /* set initialization, event handler, and wrap-up procedures */
   for ( i = 0; i < TWGetNumLPs(); i++ ) 
	{
      	TWLP[i].Map = i%TWGetNumGlobalPEs(); /* LP to PE:  map round robin */
      	TWLP[i].IProc =  (FuncPtr2) IPHoldLP;
      	TWLP[i].Proc  =  (FuncPtr)  PHoldLP;
      	TWLP[i].FProc =  (FuncPtr2) FPHoldLP;
      	TWLP[i].IncrSave = TRUE;

        TWLP[i].State = TWMalloc( sizeof(struct MyLPState) ); /* InitAppl */
        TWLP[i].LPStateSize = sizeof(struct MyLPState);       /* InitAppl */
        TWLP[i].CopySize = sizeof(struct MyAutoVars);         /* InitAppl */
      	}

   /* get initial random number generator seeds */
   for ( i=0; i < TWGetNumLPs(); i++ )
     	TWRandInit (&(Seeds[i]), 0);

   TWMsgSize = sizeof(struct MyMsgData);
   TWAppStatMsgSize = sizeof( struct MyMsgData );

  if  ( TWapp_flags == app_remote )
	{
   	if ( show_noprint == 10 )
		printf("TWapp_flags == app_remote\n");
	}

   for ( i=0; i < TWnpe ; i++ )
	{
	for ( j=0; j < TWnpe; j++ )
		{
		if ( i != TWnpe || j != i ) TWMemMap[i][j] = 1;
		if ( i == j ) TWMemMap[i][j] = 1;
		TWMemMap[i][j] = 1;
		}
	}

  CloneSim_InitAppl( argc, argv );
  cs_Cntr_Initialize( TWGetNumGlobalPEs() );
}


/* ---------------------------------------------------------------------------
** TWWrapAppl  ()
**
** DESCRIPTION: WrapUp for Application
**
**---------------------------------------------------------------------------*/
void TWWrapAppl(void)
{
   int i;
#if 0
   extern void ipdes_stats(void); 	/* HERE XXX TWWrapAppl */
   ipdes_stats(); 			/* HERE XXX TWWrapAppl */
#endif
  CloneSim_CloseAppl();
}

