/******************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.7                             ***
***         Copyright 1997 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/
#include "gtw.h"
#include "pong.h"
#include "cs_api.h" 		/* cloning api header file */

int   	app_remote=1;
int   	ClonePeriod=10;		
int 	MsgPop; 		/* message population */
TWSeed 	Seeds[MAXLPs]; 		/* random number generator seeds */
int	cloneID=0;		/* Global cloneID */
int	version0_clones	= 1;	/* version 0 clones all new clones */

void IPong( struct MyLPState *, int );
void IPing( struct MyLPState *, int );
void ClonePHoldLP( struct MyLPState *SV, struct MyMsgData *M, int MyPE );


/* ---------------------------------------------------------------------------
** IPing  ()
**
** DESCRIPTION: init each LP
**
**---------------------------------------------------------------------------*/
void IPing( struct MyLPState *SV, int MyPE )
{  
   struct MyMsgData 	*TWMsg;
   int 			i, rcv_lp;
   TWTime 		ts;
   TWEachSeed_t 	s1, s2;
   int 			dst;
   char          	fn[]    = "[IPHoldLP]";

   struct LPState * CurState;
   CurState = GState[MyPE].CurState;

   SV = (struct MyLPState *) TWLP[TWMe()].State;

   /* specify variables to be automatically checkpointed */
   TWAutoCheck((char*)&(SV->CVars), sizeof(struct MyAutoVars));

   /* initialize random number generator seeds */
   TWRandGetSeeds (&(Seeds[TWMe()]), &s1, &s2);
   TWRandSetSeeds (&(SV->CVars.Seeds), s1, s2);

   /* initialize state variables */
   if( CurState->IncrSave )
	{
	SV->ISVars.Count = 0;
	for( i=0; i < MSZ; i++ )
		{
                SV->ISVars.LastTS[i] = 0.0;
		}
	} 
      
   if ( TWMe() == TWGetNumLPs() - 1 )
	{
	SV->ROVars.NeighborID = 0;   /* Ping is LP 0 */
	}
   else
	{
	SV->ROVars.NeighborID = TWMe() + 1;
	}


   SV->CVars.Count = 0;
   SV->CVars.CloneCount = 0;
   SV->CVars.cloned = 0;

   /* send initial message */
   ts = TWRandExponential(&(SV->CVars.Seeds), 1.0);
   dst = SV->ROVars.NeighborID;

   TWGetMsg(ts, dst, sizeof(struct MyMsgData));
   TWMsg->counter = 1;
   TWMsg->CloneCount = SV->CVars.CloneCount;
   TWMsg->numself = 1;
   TWSend();
}


/* ---------------------------------------------------------------------------
** IPong  ()
**
** DESCRIPTION: init each LP
**
**---------------------------------------------------------------------------*/
void IPong( struct MyLPState *SV, int MyPE )
{  
   struct MyMsgData 	*TWMsg;
   int 			i, rcv_lp;
   TWTime 		ts;
   TWEachSeed_t 	s1, s2;
   int 			dst;
   char          	fn[]    = "[IPHoldLP]";

   struct LPState * CurState;
   CurState = GState[MyPE].CurState;

   SV = (struct MyLPState *) TWLP[TWMe()].State;

   /* specify variables to be automatically checkpointed */
   TWAutoCheck((char*)&(SV->CVars), sizeof(struct MyAutoVars));

   /* initialize random number generator seeds */
   TWRandGetSeeds (&(Seeds[TWMe()]), &s1, &s2);
   TWRandSetSeeds (&(SV->CVars.Seeds), s1, s2);

   /* initialize state variables */
   if( CurState->IncrSave )
	{
	SV->ISVars.Count = 0;
	for( i=0; i < MSZ; i++ )
		{
                SV->ISVars.LastTS[i] = 0.0;
		}
	} 
      
   if ( TWMe() == TWGetNumLPs() - 1 )
	{
	SV->ROVars.NeighborID = 0;   /* Ping is LP 0 */
	}
   else
	{
	SV->ROVars.NeighborID = TWMe() + 1;
	}


   SV->CVars.Count = 0;
   SV->CVars.CloneCount = 0;
   SV->CVars.cloned = 0;
}





/* ---------------------------------------------------------------------------
** ClonePHoldLP  ()
**
** DESCRIPTION:  THIS event is called conservatively when cloned
**
**---------------------------------------------------------------------------*/
void ClonePHoldLP( struct MyLPState *SV, struct MyMsgData *M, int MyPE )
  {  
   char          	fn[]    = "[ClonePHoldLP]";
   struct MyMsgData 	*TWMsg;

   struct PEState  	*MyState;      
   struct LPState 	*CurState;

   TWTime 		ts;
   int			uid;
   int			ch_uid=-1;
   char 		*msg;

   MyState  = &GState[MyPE];
   CurState = GState[MyPE].CurState;

   uid		= CloneSim_UID();
   CloneSim_CLONE_RT[SV->CVars.CloneCount-1] = TWNow();

   SV->CVars.cloned = version0_clones; 
   cloneID 	= CloneSim_Create(CloneSim_UID(), TWNow());
   SV->CVars.cloned = !(version0_clones); 

   if( cloneID > 0 )
   	ch_uid 	= cs_GIDtoID[cs_IDtoGID[uid]].clone_id[cloneID];

   if( cloneID < 0 )
	msg = (char *) strdup("ignored");
   else
	msg = (char *)strdup("successful");

   printf("*** %s cloning %s: gid=%3d"
		" uid=%3d"
		" cloneID=%2d"

		" ch_uid=%3d"
		" ch_cloneID=%2d"

		" cloned at %lf\n",
		msg,
		fn, TWMe(), 
		CloneSim_UID(), cs_IDtoClone[CloneSim_UID()], 
		ch_uid,  cloneID,  
		TWNow()
		);

  /* ------------------------------------------------------------------- 
   *
   * o the "original simulation" is kept/backed up and has the highest,
   *   clone identification number
   * o ONLY the LP that "clones" can change it state, the "cloned" is a backup
   *	of un-touched state
   * o can do MSG sends and these will be independent of the new clone
   *
   * ------------------------------------------------------------------ */
 }

/* ---------------------------------------------------------------------------
** Ping()
**
** DESCRIPTION: event handler for ping
**
**---------------------------------------------------------------------------*/
void Ping( struct MyLPState *SV, struct MyMsgData *M, int MyPE )
{  
   struct MyMsgData 	*TWMsg;
   struct LPState 	*CurState;
   TWTime 		ts;
   int 			dst;
   char          	fn[]    	= "[PHoldLP]";
   int 			cloning_lp	= 0;

   CurState = GState[MyPE].CurState;
   
   if ( CurState->IncrSave )
	{
     	TWCheckTWTime(&(SV->ISVars.LastTS[SV->ISVars.Count % MSZ]));
     	SV->ISVars.LastTS[SV->CVars.Count % MSZ] = TWNow();

     	TWCheck(&(SV->ISVars.Count));
     	SV->ISVars.Count++;
   	}
 
   SV->CVars.LastTS[SV->CVars.Count % MSZ] = TWNow();
   SV->CVars.Count++;

   clone_lp = cs_GIDtoID[0].clone_id[cloning_lp];
   if( (TWMe() == clone_lp) && (SV->CVars.cloned == 0) )
        {
        if( CloneSim_CLONETIME[SV->CVars.CloneCount] > 0.0 )
                {
                if ( ((SV->CVars.CloneCount) < CloneSim_NumClones)
                        && (CloneSim_CLONETIME[SV->CVars.CloneCount] < TWNow()) )
                        {
                        SV->CVars.CloneCount++;
   			TWGetMsg(0, TWMe(), sizeof(struct MyMsgData));
			TWBlockingIOSend( (IOFuncPtr) ClonePHoldLP );
                        }
                }
        }

   /* bounce message */
   ts = TWRandExponential(&(SV->CVars.Seeds), 1.0);
   dst = SV->ROVars.NeighborID;
   TWGetMsg(ts, dst, sizeof (struct MyMsgData));
   TWMsg->counter = M->counter + 1;
   TWMsg->numself = M->numself;
   TWMsg->CloneCount = SV->CVars.CloneCount;		/* HERE */
   TWSend();
}



/* ---------------------------------------------------------------------------
** Pong  ()
**
** DESCRIPTION: event handler for phold
**
**---------------------------------------------------------------------------*/
void Pong( struct MyLPState *SV, struct MyMsgData *M, int MyPE )
{  
   struct MyMsgData 	*TWMsg;
   struct LPState 	*CurState;
   TWTime 		ts;
   int 			dst;
   char          	fn[]    	= "[Pong]";
   int 			cloning_lp	= 0;

   CurState = GState[MyPE].CurState;
   
   if ( CurState->IncrSave )
	{
     	TWCheckTWTime(&(SV->ISVars.LastTS[SV->ISVars.Count % MSZ]));
     	SV->ISVars.LastTS[SV->CVars.Count % MSZ] = TWNow();

     	TWCheck(&(SV->ISVars.Count));
     	SV->ISVars.Count++;
   	}
 
   SV->CVars.LastTS[SV->CVars.Count % MSZ] = TWNow();
   SV->CVars.Count++;

   clone_lp = cs_GIDtoID[0].clone_id[cloning_lp];
   if( (TWMe() == clone_lp) && (SV->CVars.cloned == 0) )
        {
        if( CloneSim_CLONETIME[SV->CVars.CloneCount] > 0.0 )
                {
                if ( ((SV->CVars.CloneCount) < CloneSim_NumClones)
                        && (CloneSim_CLONETIME[SV->CVars.CloneCount] < TWNow()) )
                        {
                        SV->CVars.CloneCount++;
   			TWGetMsg(0, TWMe(), sizeof(struct MyMsgData));
			TWBlockingIOSend( (IOFuncPtr) ClonePHoldLP );
                        }
                }
        }

   /* bounce message */
   ts = TWRandExponential(&(SV->CVars.Seeds), 1.0);
   dst = SV->ROVars.NeighborID;

   TWGetMsg(ts, dst, sizeof (struct MyMsgData));
   TWMsg->counter = M->counter + 1;
   TWMsg->numself = M->numself;
   TWMsg->CloneCount = SV->CVars.CloneCount;		/* HERE */
   TWSend();
}


/* ---------------------------------------------------------------------------
** FPing  ()
**
** DESCRIPTION: final event handler
**
**---------------------------------------------------------------------------*/
void FPing(struct MyLPState * SV, int MyPE)
{  
   int i;

   struct LPState * CurState;
   CurState = GState[MyPE].CurState;

   if (!CurState->IncrSave) return;
}


/* ---------------------------------------------------------------------------
** TWInitAppl  ()
**
** DESCRIPTION: final event handler
**
**---------------------------------------------------------------------------*/
void TWInitAppl(int argc, char ** argv)
{
   int i,j;
   int my_lps;

   void IPing(), IPong(), Ping(), Pong(), FPing();

   my_lps = 2;
   MsgPop = 256;

   fprintf ( stderr, "\n");
   fprintf( stderr, "Enter number of LPs      -> ");
   scanf( "%d", &my_lps);

   fprintf ( stderr, "\n");
   fprintf( stderr, "Enter message population -> ");
   scanf("%d", &MsgPop);

   fprintf (stderr, "\n"); fflush(stderr);
   fprintf( stderr, "Enter Cloning Period     -> ");
   scanf( "%d", &ClonePeriod);

   fprintf (stderr, "\n"); fflush(stderr);
   fprintf( stderr, "Enter Number Messages Sent to SELF -> ");
   scanf( "%d", &app_remote );
   fprintf (stderr, "\n"); fflush(stderr);

   TWSetNumLPs(my_lps);

   if ( (TWGetNumLPs() < 1) || ( TWGetNumLPs() > MAXLPs ))
	error ("Illegal number of LPs (%d >%d)\n", TWGetNumLPs(), MAXLPs );

   if ( (MsgPop < 1) || (MsgPop > MAXMSGS))
      	error ("Illegal number of messages\n");

   if ( show_noprint == 10 )
   	printf ("TWnpe=%d, TWGetNumGlobalPEs() = %d\n",
		TWnpe, TWGetNumGlobalPEs());

   /* there is only one ping */
   TWLP[0].Map = 0;
   TWLP[0].IProc  = (FuncPtr2) IPing;
   TWLP[0].Proc  = (FuncPtr) Ping;
   TWLP[0].FProc =  (FuncPtr2) FPing;
   TWLP[0].IncrSave = TRUE;
   TWLP[0].State = TWMalloc( sizeof(struct MyLPState) ); /* InitAppl */
   TWLP[0].LPStateSize = sizeof(struct MyLPState);       /* InitAppl */
   TWLP[0].CopySize = sizeof(struct MyAutoVars);         /* InitAppl */

   /* set initialization, event handler, and wrap-up procedures */
   for( i = 1; i < TWGetNumLPs(); i++ ) 
	{
      	TWLP[i].Map = i % TWGetNumGlobalPEs(); /* LP to PE:  map round robin */
      	TWLP[i].IProc =  (FuncPtr2) IPong;
    	TWLP[i].Proc  =  (FuncPtr)  Pong;
      	TWLP[i].FProc =  (FuncPtr2) FPing;
      	TWLP[i].IncrSave = TRUE;

        TWLP[i].State = TWMalloc( sizeof(struct MyLPState) ); /* InitAppl */
        TWLP[i].LPStateSize = sizeof(struct MyLPState);       /* InitAppl */
        TWLP[i].CopySize = sizeof(struct MyAutoVars);         /* InitAppl */
      	}

   /* get initial random number generator seeds */
   for ( i=0; i < TWGetNumLPs(); i++ )
     	TWRandInit (&(Seeds[i]), 0);

   TWMsgSize = sizeof(struct MyMsgData);
   TWAppStatMsgSize = sizeof( struct MyMsgData );

   for( i=0; i < TWnpe ; i++ )
	{
	for ( j=0; j < TWnpe; j++ )
		{
		if ( i != TWnpe || j != i ) TWMemMap[i][j] = 1;
		if ( i == j ) TWMemMap[i][j] = 1;
		TWMemMap[i][j] = 1;
		}
	}

  /* initialize cloning library */
  CloneSim_InitAppl( argc, argv );
  version0_clones	= 0;	/* 0: last cloned version clones all new clones */
  version0_clones	= 1;	/* 1: version 0 clones all new clones, more 
				 * ROBUST way*/
}


/* ---------------------------------------------------------------------------
** TWWrapAppl  ()
**
** DESCRIPTION: WrapUp for Application
**
**---------------------------------------------------------------------------*/
void TWWrapAppl(void)
{
  int i;
  CloneSim_CloseAppl();
}

