/******************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.4                             ***
***         Copyright 1997 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                                                                         ***
***                        All Rights Reserved                              ***


     Permission to use, copy and modify this software and its
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/

/*****************************************************************************
        TIME WARP kernel on Distributed & Shared Memory Architectures.
******************************************************************************/
#include "gtw.h"
#include "pong.h"
#include "cs_api.h"     /* cloning api header file */


void IPong( struct MyLPState *, int );
void IPing( struct MyLPState *, int );



/* ---------------------------------------------------------------------------
** TWInitAppl ()
**
** DESCRIPTION: Initializes Application
**
**---------------------------------------------------------------------------*/
void TWInitAppl( int argc, char **argv )
{
   int my_lps;
   void IPing(), IPong(), Ping(), Pong(), FPing();

   my_lps = 2;
   TWnlp = 2;         /* specify number of LP */

   TWSetNumLPs(my_lps); 

   /* LP to PE mapping */
   TWLP[0].Map = 0;
   if ( TWGetNumGlobalPEs() == 1) TWLP[1].Map = 0;
   else TWLP[1].Map = 1;
   
   /* set initialization, event handler, and wrap-up procedures */
   TWLP[0].IProc = (FuncPtr2) 	IPing;
   TWLP[0].Proc  = (FuncPtr) 	Ping;
   TWLP[0].FProc = (FuncPtr2) 	FPing;
   TWLP[0].IncrSave = TRUE;
   TWLP[0].State = TWMalloc(sizeof(struct MyLPState));

   TWLP[0].LPStateSize = sizeof(struct MyLPState);	
   TWLP[0].CopySize = sizeof(struct MyAutoVars); 	

   TWLP[1].IProc = IPong;
   TWLP[1].Proc  = Pong;
   TWLP[1].FProc = FPing;  
   TWLP[1].IncrSave = TRUE;
   TWLP[1].State = TWMalloc(sizeof(struct MyLPState));

   TWLP[1].LPStateSize = sizeof(struct MyLPState); 	
   TWLP[1].CopySize = sizeof(struct MyAutoVars);  	

   TWMsgSize = sizeof(struct MyMsgData);  

   CloneSim_InitAppl( argc, argv ); 
}

/* ---------------------------------------------------------------------------
** IPing ()
**
** DESCRIPTION: Initializes LP
**
**---------------------------------------------------------------------------*/
void IPing( struct MyLPState *SV, int MyPE)               /* Ping is LP 0 */
{   
   struct MyMsgData *TWMsg;
   int i;
   TWTime ts;

   struct LPState * CurState;
   CurState = GState[MyPE].CurState; 

   /* SV is NULL since TWInit() did not allocate State */
   /* allocate memory for state vector, and set pointer to LP state */

   TWLP[TWMe()].State = TWMalloc(sizeof(struct MyLPState));
   SV = (struct MyLPState *) TWLP[TWMe()].State;

   /* specify variables to be automatically checkpointed */
   TWAutoCheck( (char *) &(SV->CVars), sizeof(struct MyAutoVars));

   /* initialize state variables */
   SV->ROVars.NeighborID = 1;   /* Pong is LP 1 */
   SV->CVars.Count = 0;
   for (i=0; i<1000; i++) SV->ISVars.LastTS[i] = 0.0;
   TWRandInit (&(SV->CVars.Seeds), 0);
 
   /* send initial message */
   ts = TWRandExponential(&(SV->CVars.Seeds), 1.0);
   TWGetMsg(ts, SV->ROVars.NeighborID, sizeof (struct MyMsgData));
   TWMsg->count = 0;
   printf("iping\n");
   fflush(stdout);

   TWSend();

}

/* ---------------------------------------------------------------------------
** IPong ()
**
** DESCRIPTION: Initializes LP
**
**---------------------------------------------------------------------------*/
void IPong( struct MyLPState *SV, int MyPE)               /* Pong is LP 1 */
{   
   struct MyMsgData *TWMsg;
   int i;

   struct LPState * CurState;
   CurState = GState[MyPE].CurState; 
   
   /* allocate memory for state vector, and set pointer to LP state */
   SV = (struct MyLPState *) TWLP[TWMe()].State;

   /* specify variables to be automatically checkpointed */
   TWAutoCheck( (char *) &(SV->CVars), sizeof(struct MyAutoVars));

   /* initialize state variables */
   SV->ROVars.NeighborID = 0;   /* Ping is LP 0 */
   SV->CVars.Count = 0;
   for (i=0; i<1000; i++) SV->ISVars.LastTS[i] = 0.0;
   TWRandInit (&(SV->CVars.Seeds), 0);

   printf("ipong\n");
   fflush(stdout);
}

/* ---------------------------------------------------------------------------
** Ping  ()
**
** DESCRIPTION: Event Handler for Ping
**
**---------------------------------------------------------------------------*/
void Ping ( struct MyLPState *SV, struct MyMsgData *M, int MyPE )
{   
   struct MyMsgData *TWMsg;

   TWTime ts;

   struct LPState * CurState;
   CurState = GState[MyPE].CurState; 

   printf("ping\n");
   fflush(stdout);

   /* update state */
   TWCheckTWTime(&(SV->ISVars.LastTS[SV->CVars.Count % 1000])); 
   SV->ISVars.LastTS[SV->CVars.Count % 1000] = TWNow();

   SV->CVars.Count++;

   /* bounce message back to Pong */
   ts = TWRandExponential(&(SV->CVars.Seeds), 1.0);
   TWGetMsg(ts, SV->ROVars.NeighborID, sizeof(struct MyMsgData));
   TWMsg->count = M->count + 1;
   TWSend();


}

/* ---------------------------------------------------------------------------
** Pong  ()
**
** DESCRIPTION: Event handler for Pong; actually, it is identical to Ping! 
**
**---------------------------------------------------------------------------*/
void Pong ( struct MyLPState *SV, struct MyMsgData *M, int MyPE )
{  
   struct MyMsgData *TWMsg;
   TWTime ts;

   struct LPState * CurState;
   CurState = GState[MyPE].CurState; 


   printf("pong\n");
   fflush(stdout);

   /* update state */
   TWCheckTWTime(&(SV->ISVars.LastTS[SV->CVars.Count % 1000])); 
   SV->ISVars.LastTS[SV->CVars.Count % 1000] = TWNow();

   SV->CVars.Count++;
   ts = TWRandExponential(&(SV->CVars.Seeds), 1.0);
   /* bounce message back to Ping */
   TWGetMsg(ts, SV->ROVars.NeighborID, sizeof (struct MyMsgData));
   TWMsg->count = M->count + 1;
   TWSend();
}

/* ---------------------------------------------------------------------------
** FPing  ()
**
** DESCRIPTION: Wrap up for Ping
**
**---------------------------------------------------------------------------*/
/* void FPing (int LP, struct MyLPState *SV, TW_ADDR_TYPE Data ) */
void FPing ( struct MyLPState *SV, int MyPE )

{  int i;

   if (SV->CVars.Count < 1000) /* less than 1000 messages received */
      for (i=SV->CVars.Count; i>=0; i--)
	/* printf ("%f\n", SV->ISVars.LastTS[i])*/ ;
   else /* 1000 or more messages received */
      for (i=0; i<1000; i++)
	/* printf ("%f\n",SV->ISVars.LastTS[(SV->CVars.Count-i+1000)%1000])*/ ;

   printf("FPing is done with application\n");
}

/* ---------------------------------------------------------------------------
** TWWrapAppl  ()
**
** DESCRIPTION: WrapUp for Application
**
**---------------------------------------------------------------------------*/
void TWWrapAppl( void )
{
   printf("Application done \n");

   CloneSim_CloseAppl();
}




