#ifndef GTW_H
#define GTW_H
/******************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.5                             ***
***         Copyright 1997 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/


/******************************************************************************
**
**	FILE:
**		gtw.h
**
**	DESCRIPTION:
**		To be included by the application program.
**  		Contains some useful macros, plus all externals. 
**
******************************************************************************/

#include "gtw_glob.h"
#include "gtw_incr.h"
#include "gtw_rand.h"
#include "gtw_ds.h"
#include "gtw_ds_glob.h"
#include "cs_internal.h"  /* IO5 */
#include "cs_api.h"  	  /* IO5 */
#include "gtw_io.h"	   /* IO5, IO4 MH_HERE */

/* added temporarily externs from insim.h (CLONING) */
#if( defined(SGI5) || defined(SGIMP64) || defined(SGI64) )
extern usptr_t          *shm_handle;
extern char             shared_area[80];
#endif /* SGI5 || SGIMP64 || SGI64 */
extern TW_SHARED int    LPNum[];

/* cloning and ipdes variables */

extern TW_SHARED int show_gvt;
extern TW_SHARED int show_noprint;
extern TW_SHARED int show_gvt_dump;
extern TW_SHARED int spread_clone;
extern TW_SHARED int limit_output;

/* distributed */
extern TW_SHARED int TWAppStatMsgSize;
extern TW_SHARED int RFthread;
extern TW_SHARED int TWnlp;
extern TW_SHARED int TWnpe;
extern TWLPType  TWLP[MAXLP];
extern TW_SHARED int TWMsgSize;
extern TW_SHARED int TWMemMap[D_MAXPROC*MAXMACHINES][D_MAXPROC*MAXMACHINES];
extern TW_SHARED TWTime ENDTIME;
extern TW_SHARED struct PEState GState[MAXPROC];
extern TW_SHARED int TWapp_flags;

/* extern char *TWMalloc(TW_LONG,int); */
extern char *TWMalloc(TW_LONG); 

extern void error( char *,... );
extern double GetLPsLVT( int LP ); /*IO4*/

/* current local virtual time */
#define TWNow()   (GState[MyPE].CurEvent->Msg.RcvTS)

#define TWSetNumLPs(num_lps)        \
		(TWnlp =  num_lps); \
		(cs_nlp = num_lps); \
		(cs_vlp = num_lps);

#define TWGetNumLPs() 	(cs_nlp)

/* whoami? */
#define TWMe()  (cs_IDtoGID[GState[MyPE].CurState->ID])
#define CloneSim_GID()  (cs_IDtoGID[GState[MyPE].CurState->ID])
#define CloneSim_UID() (GState[MyPE].CurState->ID)

/* End time of simulation */
#define TWEndTime()  (ENDTIME)

/* exit */
#define TWQuit() exit();

#define TWCheckStraggler() ( checkStraggler( MyPE ) )

#define TWGetMsg(ts,gdst,size) 					\
   { 			/*IO4*/					\
   struct PEState 	*MyState;       			\
   struct LPState 	*CurState;      			\
   int 			src; 	        			\
   int 			new_dst;        			\
   int 			cloneID;        			\
   double		tss;					\
								\
   MyState  = &GState[MyPE];             			\
   CurState = MyState->CurState;        			\
   src 	    = CurState->ID;	              			\
   tss 	    = MyState->CurEvent->Msg.RcvTS ;			\
   cloneID  = cs_IDtoClone[src];				\
   new_dst  = cs_GIDtoID[gdst].clone_id[cloneID];		\
								\
  _TWGetMsg((ts),(new_dst),(size),(MyPE)); 			\
  TWMsg = MyState->TWMsg;					\
  CloneSim_GetMsg(src, new_dst,(MyState->TWMsg), tss,(tss+ts)); \
  TWMsg = MyState->TWMsg;					\
   }

#define TWSend() _TWSend(MyPE);
#define TWAutoCheck(ptr,size) _TWAutoCheck((ptr),(size),(MyPE));

extern void _TWAutoCheck(char *, int , int );
extern void _TWGetMsg(TWTime, int, int, int);
extern void _TWSend(int);
extern int  checkStraggler( int );


/* MARIA/FRED #define TWCheckChar(ch) check( (ch) & 0xFFFC, CurState, MyPE); */
#define TWCheckChar(ch) check_char( ch,  CurState, MyPE);

#define TWCheck(ip) check(ip, CurState, MyPE);
#define TWCheckDouble(l) check_double(l, CurState, MyPE);
#define TWCheckFloat(f) check_float(f, CurState, MyPE);
#define TWCheckLong(l) check_long(l, CurState, MyPE);
#define TWCheckTWTime(l) check_time(l, CurState, MyPE);

#define TWGetNumGlobalPEs() TWds.PEs
#define TWGetNumMachines()  TWds.Machines
#define TWGetNumLocalPEs()  TWds.PEsPerHost[TWds.MyMachine]

#endif /* GTW_H */
