#ifndef GTW_ARCH_H
#define GTW_ARCH_H

/******************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.4                             ***
***         Copyright 1995 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/

/******************************************************************************
*         architecture specific macros definitions/includes
******************************************************************************/

#include "gtw_timing.h"


/**********************************************************************/
/**************************** SUN4SOL2 ********************************/
/**********************************************************************/
#ifdef SUN4SOL2
#define TW_SHARED
#define TW_PRIVATE

#define TW_ALIGN
#define TW_SYNCH

#define TW_LOCK_TYPE mutex_t
#define TW_LOCK(lck) ;
#define TW_UNLOCK(lck) ;
#define TW_INIT_LOCK(lck) ;
#define TW_POSTSTORE(lck)

#define TW_BARRIER_TYPE  barrier_t
#define TW_BARRIER_SYNC(bar,id)  ;
#define TW_BARRIER_SYNC_IO(bar,id) ;
#define TW_RT_BARRIER_SYNC(bar,id) ;

#define TW_INIT_BARRIER(bar,nproc) ;
#define TW_INIT_BARRIER_LOCAL(bar,id) ;

typedef char * TW_ADDR_TYPE;

#define TW_ATOMIC_INCR(glob)  glob++;
#define TW_ATOMIC_DCR(glob)  glob--;

#define SP_SIZE 128             /* subpage size in bytes */
#define PG_SIZE (4*1024)           /* page size in bytes */
#define SP_ALIGNED(x) ((0x000000007fL & (long) (x)) == 0) /*subpage aligned? */
#define SP_IN_PAGE (SP_SIZE)/(PG_SIZE) /* number of subpages in page */

typedef unsigned long  TW_LONG ;
typedef unsigned long  TW_ULONG;
#define MAXINT		ULONG_MAX

#endif /* SUN4SOL */

/**********************************************************************/
/**************************** SUNMP ***********************************/
/**********************************************************************/
#ifdef SUNMP
#include <values.h>

#define TW_SHARED  
#define TW_PRIVATE 

#define TW_ALIGN 
#define TW_SYNCH 

#define TW_LOCK_TYPE mutex_t

/* typedef volatile int LOCK; */

#define TW_LOCK(lck)  mutex_lock(&(lck));
#define TW_UNLOCK(lck) mutex_unlock(&(lck));
#define TW_INIT_LOCK(lck) mutex_init(&(lck), USYNC_THREAD, (void *) 0);

#define TW_POSTSTORE(lck) 

#define TW_BARRIER_TYPE struct central_barrier

#define TW_BARRIER_SYNC(bar,id) barrier_wait(&(bar), &GState[(id)].bar,(id));

/* initializing the barrier ensures RFthread/or IOthread is included */
#define TW_BARRIER_SYNC_IO(bar,id) \
	barrier_wait(&(bar),&(GState[(id)].bar),(id));

#define TW_RT_BARRIER_SYNC(bar,id)\
	barrier_wait(&(bar), &(GState[(id)].bar),(id));

#define TW_INIT_BARRIER(bar,nproc) barrier_init(&(bar), (nproc));
#define TW_INIT_BARRIER_LOCAL(bar,id) barrier_init_local(& GState[(id)].bar);

typedef char *  TW_ADDR_TYPE;

#define TW_ATOMIC_INCR(glob)  	glob++; 
#define TW_ATOMIC_DCR(glob)  	glob--;

#define SP_SIZE 128             /* subpage size in bytes */
#define PG_SIZE (4*1024)           /* page size in bytes */
#define SP_ALIGNED(x) ((0x000000007fL & (long) (x)) == 0) /*subpage aligned? */
#define SP_IN_PAGE (SP_SIZE)/(PG_SIZE) /* number of subpages in page */

typedef int  		TW_LONG ;
typedef unsigned long  TW_ULONG;
/*
#define MAXINT		ULONG_MAX
*/

#endif /* SUNMP */

/**********************************************************************/
/********************** SUNMP or SUN4SOL ******************************/
/**********************************************************************/
#if(defined(SUNMP) || defined(SUN4SOL2))
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <synch.h>
#include <sys/time.h>
#include <thread.h>
#include<sys/resource.h>
#include "gtw_central_barrier.h" /* used #defines from above */

struct args
	{
  	int id;
	};
typedef struct pe_private pe_pvt_t;

#endif

/**********************************************************************/
/**************************** SGI5 or SGI64 ***************************/
/**********************************************************************/
#if(defined(SGI5) || defined(SGI64))
#define TW_SHARED
#define TW_PRIVATE
#define TW_ALIGN
#define TW_SYNCH

#define TW_LOCK_TYPE ulock_t
#define TW_LOCK(lck) ;
#define TW_UNLOCK(lck) ;
#define TW_INIT_LOCK(lck) ;

#define TW_POSTSTORE(lck)

#define TW_BARRIER_TYPE  barrier_t *

#define barrier_pvt_t int

#define TW_BARRIER_SYNC(bar,id) ;
#define TW_RT_BARRIER_SYNC(bar,id);
#define TW_BARRIER_SYNC_IO(bar,id);
#define TW_INIT_BARRIER(bar,nproc) ;
#define TW_INIT_BARRIER_LOCAL(bar,id)

typedef char *  TW_ADDR_TYPE;

#define TW_ATOMIC_INCR(glob)  glob++;
#define TW_ATOMIC_DCR(glob)  glob--;

#define SP_SIZE (1*128)              /* subpage */
#define PG_SIZE (16*1024)           /* page */
#define SP_ALIGNED(x) ((0x000003ffL & (long) (x)) == 0) /*subpage aligned?3fl*/
#define SP_IN_PAGE (SP_SIZE)/(PG_SIZE) /* number of subpages in page */

#ifndef _MAXFLOAT
#define _MAXFLOAT
#define MAXFLOAT        ((float)3.40282346638528860e+38)
#endif  

typedef unsigned long  TW_ULONG;
typedef int TW_LONG;

#endif /* SGI5 || SGI64 */

/**********************************************************************/
/**************************** SGIMP64 *********************************/
/**********************************************************************/
#if(defined(SGI5) || defined(SGI64) || defined(SGIMP64))
#include <ulocks.h>
#include <task.h>

#define TW_SHARED  
#define TW_PRIVATE 
#define TW_ALIGN 
#define TW_SYNCH 
#define TW_POSTSTORE(lck) 

#define TW_LOCK_TYPE 	ulock_t

#define TW_LOCK(lck)  	ussetlock(lck);			/* MHERE */
/* #define TW_LOCK(lck) uswsetlock(lck, 1); */

#define TW_UNLOCK(lck) 	usunsetlock(lck);

#define TW_INIT_LOCK(lck) lck = usnewlock(shm_handle); 	/* MHERE */
#if 0
#define TW_INIT_LOCK(lck)                             \
        lck = usnewlock(shm_handle);                  \
        if( lck == NULL )                             \
           {                                          \
           perror("usnewlock failed to alloc lock: ");\
           exit(1);                                   \
           }                                          \
        usinitlock(lck);
#endif

#define TW_BARRIER_TYPE barrier_t *

#define TW_BARRIER_SYNC_IO(bar,id)  	barrier((bar),TWnpe+1)
#define TW_RT_BARRIER_SYNC(bar,id)   	barrier(bar,TWnpe+1);
#define TW_BARRIER_SYNC(bar,id)  	barrier((bar),TWnpe)

#define barrier_pvt_t TW_BARRIER_TYPE			/* MHERE */

#define TW_INIT_BARRIER(bar,nproc) \
		bar = new_barrier(shm_handle);
#if 0
#define TW_INIT_BARRIER(bar,nproc)                                 \
        if( bar == NULL && (bar = new_barrier(shm_handle)) == NULL)\
           {                                                       \
           perror("new_barrier failed: ");                         \
           exit(1);                                                \
           }                                                       \
        init_barrier(bar);
#endif

#define TW_INIT_BARRIER_LOCAL(bar,id) 
/* typedef void *  TW_ADDR_TYPE; */
typedef char *  TW_ADDR_TYPE;

#define TW_ATOMIC_INCR(glob) glob++; 
#define TW_ATOMIC_DCR(glob)  glob--;

#define SP_SIZE (1*128)              /* subpage */
#define PG_SIZE (16*1024)           /* page */
#define SP_ALIGNED(x) ((0x000003ffL & (long) (x)) == 0) /*subpage aligned?3fl*/
#define SP_IN_PAGE (SP_SIZE)/(PG_SIZE) /* number of subpages in page */

#ifndef _MAXFLOAT
#define _MAXFLOAT
#define MAXFLOAT        ((float)3.40282346638528860e+38)
#endif  

typedef unsigned long  TW_ULONG;
typedef int TW_LONG;
#define MAXINT	INT_MAX	/* ULONG_MAX*/
/* MHERE end */

#endif /* SGIM64 || SGI5 || SGI64 */

#endif /* GTW_ARCH_H */
