#ifndef GTW_DEBUG_H
#define GTW_DEBUG_H

/******************************************************************************
*******************************************************************************
**
**
     FILE: gtw_debug.h

     DESCRIPTION:  

*******************************************************************************
******************************************************************************/

#ifdef IODEBUG
#define IODEBUGSIZE             1
#endif

/* debug flags */
#define DEBUG_MSG         0x0001  /* printout message sends and receives */
#define DEBUG_THREAD      0x0002  /* thread enter, exit */
#define DEBUG_ALL_GVT     0x0004  /* all GVT */
#define DEBUG_SOME_GVT    0x0008
#define DEBUG_FEEDBACK    0x0010  /* feedback parameters */
#define DEBUG_THROTTLE    0x0020  /* for throttling mechanism */
#define DEBUG_LBS         0x0040  /* Local Buffer Send: redistribution code */
#define DEBUG_SHOW_PROG   0x0080  /* display GVT every 5 seconds */
#define DEBUG_SILENT      0x0100  /* Do not display status messages */
#define DEBUG_INCR        0x0200
#define DEBUG_CALQ        0x1000
#define DEBUG_RESULTS_FILE   0x2000 /* write a results file at the end of
				     * simulation for post processing */
#ifdef IIODEBUG
#define  TWSetIIODebug(x) gtw_debug=x;
#else
/* #define  TWSetIIODebug(x)  */
#define  TWSetIIODebug(x) gtw_debug=x;
#endif /* IIODEBUG */


/* ***************************************************************************
** Check1 ********************************************************************
** ************************************************************************* */
#ifdef Check1
/* ---------------------------------------------------------------------------
## Terminate()
##
## DESCRIPTION:
##
##---------------------------------------------------------------------------*/
#define Terminate(b,a) printf("Exits: MyPE = %d GVT %g MsgQ %x CanQ %x" \
                " CalQ %4.2g  %d\n", b, GVT, MyState->MsgQ,       \
                 MyState->CanQ, Minim_Q(MyPE), a);

/* ---------------------------------------------------------------------------
** Check_ev_list()
**
** DESCRIPTION:  Checks the event list and make sure there are no broken
**      links.
**
** --------------------------------------------------------------------------*/
#define Check_ev_list(location)                                     \
                 {                                                  \
                 struct Event *sev;                                 \
                                                                    \
                 sev = statep->EList.Next;                          \
                 while(sev->Next != NULL) {                         \
                   if(sev->Next->Prev != sev){                      \
                     printf("%s Broken link list\n", (location));   \
                   }                                                \
                   sev = sev->Next;                                 \
                 }                                                  \
                 if(sev != statep->Last) {                          \
                   printf("%s: LP(%d): CL(%d) sev and Last don't match\n", \
                           (location), statep->EList.LP,            \
                           TWLP[statep->EList.LP].Cluster);         \
                   if( statep->EList.Next != NULL )                 \
                      {                                             \
                      printf("            EList.Next->Msg.RcvTS = %f\n",\
                              statep->EList.Next->Msg.RcvTS );      \
                      printf("            EList.Next->Next = %d\n", \
                              statep->EList.Next->Next );           \
                      }                                             \
                 }                                                  \
                 if( statep->Last->Next != NULL )                   \
                    printf("%s LP(%d)'s Last->Next != NULL \n",     \
                            (location), statep->ID );               \
                 sev = NULL;                                        \
                 }						    \


#define ck1_isNeg(loc,TS) \
	if ( TS <= 0.0 )	\
	error ("\n+** %s: Negative " #TS " TS %.2f \n", loc, TS );


/* do the below with MacroExansion */

/* make sure event is not NULL before testing */
#define ck1_inMsgQ(loc, event ) \
	if ( (event) && (event->in != 'm') ) \
	  {	\
	error ("+** %s: (%c) event not frm MsgQ on %d frm %d at" \
		" (%.4f) ex(%d) type(%d,%d)\n", loc,\
		event->in, event->LP, event->SrcLP, \
		event->Msg.RcvTS, event->Processed, \
		event->BirthType, event->InPDS ); \
	  fflush (stdout); \
	  }

#define ck1_inCalQ(event)\
   if ( (event) && (event->in != 'c') ) \
      error ( "+** About to processed an event to %d, not from (%c) CalQ?\n",\
		event->LP, event->in );

#define ck1_inEvQ(loc,event)\
	if ( (event) &&  (event->in != 'e')  ) \
	   error("+** %s: Error (%c) event not from EvQ\n", loc, event->in );

#define ck1_PrevNULL(loc,event) \
	if ( event->Prev == NULL ) \
	 	printf("Inconsistnet event list\n"); \

#define ck1_printNRet\
	printf("PE %d ret %d bufs to PE %d, NToRet=%d\n", \
		MyPE, j, i, MyState->NToReturn[i]); \
	fflush(stdout);




#else /* Check1 */

#define Terminate(b,a) ;
#define Check_ev_list(location) ;

#define ck1_isNeg(loc,TS) ;
#define ck1_inMsgQ(loc,event) ;
#define ck1_inCalQ(event) ;
#define ck1_inCalQ(event) ;
#define ck1_inEvQ(loc,event) ;
#define ck1_PrevNULL(loc,event) ;
#define ck1_printNRet ;


#endif /* Check1 */

/* ***************************************************************************
** Check *********************************************************************
** ************************************************************************* */
#ifdef Check

#define ck_sameGVT \
	if ( GVT == MyState->GVT )   \
		{   \
		NSameGVT++;   \
		if ( NSameGVT > MAX_SAME_GVT )   \
			{   \
			/* no increment, global synch */   \
			printf ("PE %d: compute same GVT (=%g)"   \
			  "  %d times in a row\n", MyPE, GVT, NSameGVT);   \
			fflush(stdout);   \
			}   \
		}	\
	else	\
		{	\
		NSameGVT = 0;	\
		}


#define ck_GVT(loc,RcvTS) \
	if ( RcvTS < MyState->GVT ) \
		error ("%s: LP %d: message TS (%g) < GVT (%g)\n", \
			loc, statep->ID, RcvTS, MyState->GVT);

#define ck_Event(loc,event)\
	if ( event == NULL ) \
		error ("%s: Unexpectedly found Last == NULL\n", loc ); \
	if ( event->Msg.RcvTS != statep->LVT )	\
		error ("%s: TS of Last (%g) != LVT (%g) e(%d) s(%d)\n", \
		loc,\
		event->Msg.RcvTS, statep->LVT, event->LP, statep->ID ); \
	if ( event->Msg.RcvTS < MyState->GVT ) \
		error ("incoming message has TS (%g) < GVT (%g)\n", \
                         event->Msg.RcvTS, MyState->GVT); 

#define ck_pPrev \
	if ( pPrev == NULL )	\
		{ \
		error("+** Pprev should never be null \n"); \
		}


/* MyState */
#define ck_FindPrevNULL(loc,event) \
	if ( event->Prev == NULL ) \
		{ \
		struct Event *p;   \
		struct LPState *statep=NULL;   \
		statep = MyState->CList[LPinPE(event->LP)];   \
		if ( statep->LVT > MyState->GVT)   \
			{   \
			for ( p=statep->Last; p->Prev != NULL; p=p->Prev )   \
			    {   \
			    if (p->Prev == NULL)   \
				{   \
				printf ("PE %d: Null Prev ptr found\n", MyPE); \
				break;   \
				}   \
			    printf ("PE %d: sendTS=%g, RcvTS=%g\n",   \
				MyPE, p->Msg.SendTS, p->Msg.RcvTS);   \
			    }   \
                        fflush(stdout);   \
		}   \
		error ("Cancel(%g): proc list wrong, GVT=%g\n",    \
			event->Msg.RcvTS, MyState->GVT);   \
                }

#define ck_EList(loc,event) \
	if ( event == &(statep->EList) )	\
		error ("%s: Rollback reached EList\n", loc);

#define ck_NotProc(loc,event) \
	if (event->Processed == FALSE) \
		error("rollback unprocessed event (%g)", event->Msg.RcvTS);

#define ck_GVTdecrease \
	if ( GVT < MyState->GVT ) \
		error("GVT decreases from %g to %g\n",  \
			MyState->GVT-OFFSET, GVT-OFFSET );

#define ck_ProcEvent(CurEvent) \
   if ( CurEvent->Processed ) \
        error("LP (%d) processing a processed event %d %.4f", \
	CurEvent->LP, CurEvent->Msg.RcvTS ); \
   if ( CurEvent->Msg.RcvTS < MyState->GVT )   \
    	error ("+** LP %d processing a PAST event w/ TS (%g) < GVT (%g)\n",\
        	CurEvent->LP, CurEvent->Msg.RcvTS, MyState->GVT);   \
   if ( CurEvent->Msg.RcvTS < statep->LVT )   \
        {   \
        struct Event *p; \
        /* if external should have rolled back LP */   \
	error ("+** LP (%d) not rolled back to (%.4f) at (%.4f) GVT:%.4f\n", \
		statep->ID, CurEvent->Msg.RcvTS, statep->LVT, MyState->GVT );\
        if ( statep->LVT > MyState->GVT )   \
                {   \
                for ( p=statep->Last; p->Prev != NULL; p=p->Prev )   \
                    {   \
                    if ( p->Prev == NULL )   \
                        {   \
                        printf ("PE %d: Null Prev pointer found\n", MyPE);   \
                        break;   \
                        }   \
                    printf ( "PE %d: sendTS=%g, RcvTS=%g\n", MyPE,   \
                        p->Msg.SendTS, p->Msg.RcvTS );   \
                    }   \
                fflush(stdout);   \
                }   \
        error("\t +*** LP %d: event TS too small, LVT TS=%g, Next TS=%g GVT=%g\n",   \
              CurEvent->LP, statep->LVT, CurEvent->Msg.RcvTS, MyState->GVT);   \
	}

#define   ck_nullRBSV(statep) \
	if (statep->RBSV == NULL) \
		error ("Error restoring state\n");


/* memory.c */
#define ck_NewBuf \
   if ( NewBuf == NULL ) \
      error("need to allocate a new message buffer, LP = %d", CurState->ID);\
   if ( NewBuf->SrcPE != MyPE ) \
      error ("PE %d: TWSend, bad source PE field (%d)\n",MyPE,NewBuf->SrcPE);

#define ck_PriorNewBuf \
   if ( NewBuf != NULL )\
      warn("TWGetMsg: Prior message buffer was allocated, but not sent \n");\
   if ( size > TWMsgSize )\
      error("TWGetMsg: Message size too large\n");\
   if ( -ts > EPSILON )\
      error("TWGetMsg: negative timestamp increment, incr. = %10.3f, LP = %d",\
            ts, CurState->ID);\
   if ( (dst<0) || (dst>TWvlp) )\
        error ("TWGetMsg: Illegal destination LP (%d) specified\n", dst);

#define ck_inPDS \
  if ( MyState->NewBuf->InPDS == TRUE ) \
        error ("GET_FREE allocated buffer in pending event set\n");


#else

#define ck_sameGVT ;
#define ck_GVT(loc,RcvTS) ;
#define	ck_Event(loc,event) ;
#define ck_pPrev ;
#define ck_EList(loc,event) ;
#define ck_FindPrevNULL(loc,event) ;
#define ck_NotProc(loc,event) ;
#define ck_GVTdecrease ;
#define ck_ProcEvent(CurEvent) ;
#define ck_nullRBSV(statep) ;
#define ck_NewBuf ;
#define ck_PriorNewBuf ;
#define ck_inPDS ;


#endif /* Check */


/* ***************************************************************************
** IO3 ***********************************************************************
** ************************************************************************* */
#ifdef IO3
#define printEXfirst(event) 	\
	if ( gtw_debug && ( event->EventType ==  EXTERNAL) ) 	\
		{ \
                double current_time;	\
		TWwallclock ( current_time );	\
		extern_debug[MyPE] = 1;	\
		printf( "\t*oo*:            ( MsgRcvTS,"	\
		       	"     CurTS,       GVT,   OptGVT)\n" ); \
      		printf("\t*oo*:  EX[%1d]: c1 (%9.4lf, %9.4lf, " \
		       "%9.4lf, %9.4lf)\n", \
			MyPE, \
			event->Msg.RcvTS-OFFSET, \
			current_time-OFFSET, \
			GVT-OFFSET, \
		        (OptGVT==MAXFLOAT) ? 9999.9999 : OptGVT-OFFSET \
		        ); \
		fflush (stdout); \
		}

#define printEX(loc) \
	if ( gtw_debug && extern_debug[MyPE] ) \
		{ \
		double current_time;\
		TWwallclock ( current_time );\
		printf("\t*oo*:  EX[%1d]: %s (%9.4lf, %9.4lf, %9.4lf"\
			", %9.4lf)\n",\
                MyPE, \
		loc, \
		0.0, \
		current_time-OFFSET, \
		GVT-OFFSET,\
                (OptGVT==MAXFLOAT) ? 9999.9999 : OptGVT-OFFSET);\
		}

#define printEXlast(loc) \
   if ( gtw_debug && extern_debug[MyPE] )   \
        {   \
        double current_time;   \
        TWwallclock ( current_time );   \
        printf("\t*oo*:  EX[%1d]: "   \
           "c4 (%9.4lf, %9.4lf, %9.4lf, %9.4lf)",   \
                MyPE,   \
                CurEvent->Msg.RcvTS-OFFSET,   \
                current_time-OFFSET,   \
                GVT-OFFSET,   \
                (OptGVT==MAXFLOAT) ? 9999.9999 : OptGVT-OFFSET   \
                );   \
        extern_debug[MyPE] = 0;   \
        if ( CurEvent->EventType != EXTERNAL )   \
                printf(" ** %d ** ", CurEvent->EventType );   \
        printf ("\n");   \
        }

#else

#define printEXfirst(event) ;
#define printEX(loc)  ;
#define printEXlast(loc) ;

#endif /* IO3 */


/* ***************************************************************************
** Check3 ********************************************************************
** ************************************************************************* */
#ifdef Check3

#define ck3_repTimer \
   if ( MyPE ==0 ) \
	{ /* TWtimer is currently a struct timeval Check1 */ \
        TWtimer tNow; \
        TWnow(tNow); \
        if ( GState[MyPE].repTimer == (TWtimer) 0 ) \
		GState[MyPE].repTimer = tNow; \
        else if ((tNow - GState[MyPE].repTimer) >= GState[MyPE].RepPeriod) \
                { \
                GState[MyPE].repTimer = tNow; \
                printf("*GVT=%g: NR[0][0] %d: NR[0][1] %d:" \
                        " NR[1][0] %d: NR[1][1] %d\n", \
                   MyState->GVT, \
                   GState[0].NToReturn[0], \
                   GState[0].NToReturn[1], \
                   GState[1].NToReturn[0], \
                   GState[1].NToReturn[1]); \
                   fflush(stdout); \
                } \
        }

#else

#define ck3_repTimer ;
  

#endif

#ifdef CheckIO4

#define ck4_probe(loc) \
	printf("\t\tProbe(%d): %s\n", MyPE, loc ); \
	fflush(stdout);

#else

#define ck4_probe(loc)

#endif /* CheckIO4 */


/* ************************************************************************* */
#endif /* GTW_DEBUG_H */




