#ifndef GTW_DS_H
#define GTW_DS_H

#include <limits.h>

#define DS_EXIT_FAIL	-1
#define DS_MAXSTRING	80
#define D_MAXPROC 	20
#define DS_MAXPROC 	20

#define MAXNAME		32 
#define MAXKERNELS 	64

#define MAX_PIGGYBACK  	8
#define SERVER 		0
#define DS_SERVER 	0
#define DS_CLIENT 	1
#define CLIENT 		1
#define DS_MAXMACHINES	16
#define MAXMACHINES	16
#define MAX_CLASS_NAME	32
#define RT_MAX_MESSAGE_SIZE 10000
#define MAX_KMSG_CLASSES 100
#define MAX_MAX_KMSG_CLASSES (MAX_KMSG_CLASSES + 5)
#define MAX_OB_GVT_MSG	10 /* max out of band GVT messages */

#define DS_MSG_CLASS_EVENT 		0
#define DS_MSG_CLASS_GVTMESSAGE 	1
#define DS_MSG_CLASS_TWSTATS 		2
#define DS_MSG_CLASS_TWAPPSTATS 	3
#define DS_MSG_CLASS_TWDS 		4
#define DS_MSG_CLASS_CLUSTER 		5
#define DS_MSG_CLASS_MAPPING		6


/* old */
#define KERNEL_MESSAGE_CLASS_EVENT 0
#define KERNEL_MESSAGE_CLASS_GVTMESSAGE 1
#define KERNEL_MESSAGE_CLASS_TWSTATS 2
#define KERNEL_MESSAGE_CLASS_TWAPPSTATS 3
#define KERNEL_MESSAGE_CLASS_KERNELCOMM_MACHINEINFOSTRUCT 4
#define KERNEL_MESSAGE_CLASS_CLUSTER 5
#define KERNEL_MESSAGE_CLASS_MAPPINGUPDATEMESSAGE 6
#define KERNEL_MESSAGE_CLASS_DLMHALTMESSAGE 7
#define KERNEL_MESSAGE_CLASS_DLMMOVEMESSAGE 8
#define KERNEL_MESSAGE_CLASS_DLMRESUMEMESSAGE 9
#define KERNEL_MESSAGE_CLASS_DLMSTATISTICSMESSAGE 10
#define KERNEL_MESSAGE_CLASS_DLMSTATUSMESSAGE 11
#define KERNEL_MESSAGE_CLASS_DLMACKMESSAGE 12


#define RT_MAX_BURST_EVENTS 		INT_MAX

#define GVT_WHITE 0
#define GVT_RED   1

#define ANY_MSG -1
typedef enum {NO_MSG, START_SIM_MSG, USERDEF_MSG, GVT_CONTROL_MSG,
             GVT_RESULT_MSG, MOVE_CLUSTER_MSG, DLM_CONFIG_MSG,
             DLM_UPDATE_REQUEST_MSG, DLM_STATS_MSG, DLM_STATUS_MSG,
             DLM_HALT_MSG, DLM_MOVE_MSG, DLM_RESUME_MSG, DLM_MAPPINGUPDATE_MSG,
             DLM_HALT_ACK_MSG, DLM_HALT_CONTINUE_MSG, DLM_MOVE_ACK_MSG,
             DLM_MAPPINGUPDATE_ACK_MSG, DLM_RESUME_ACK_MSG,
             ANTI_MSG, TW_STATS_MSG, TW_APPSTATS_MSG} rtmessage_t;

typedef enum{REFLECTORTHREAD, PROCESSGVTMESSAGE,
             PROCESSDLMHALTMESSAGE, PROCESSDLMRESUMEMESSAGE} rtwhere_t;

typedef enum{RT_BUSY, RT_IDLE, RT_REQUEST, RT_HALT_COMPLETE} rtstatus_t;




#ifndef MAXLP
#define MAXLP 100000
#endif 

#undef TRUE
#undef FALSE
typedef enum{FALSE,TRUE} Boolean_t;


/* probably can remove */
struct TWds_map_s
   {
   int LPToPE[MAXLP];
   int LPToCluster[MAXLP];
   };
typedef struct TWds_map_s TWds_map_t;

struct TWdsGVTMsg_s
   {
   int Color;
   double MinClock;
   double MinSend;
   int MsgCnts[MAXMACHINES];
   };
typedef struct TWdsGVTMsg_s TWdsGVTMsg_t;

/* one of these structures per machine */
struct TWds_s
	{
   	int 	Machines;                
   	int 	PEs;
   	int 	ServerMachine;
   	int 	MyMachine;
   	char 	HostName[MAXMACHINES][MAXNAME];
   	int 	PEsPerHost[MAXMACHINES];
   	int 	LocalPEToPE[MAXMACHINES][D_MAXPROC];
   	int 	PEToLocalPE[MAXMACHINES*D_MAXPROC];
   	int 	PEToMachine[MAXMACHINES*D_MAXPROC];
	double 	Speed[MAXMACHINES];
   	double 	TotalSpeed;
   	double 	EndTime;
   	};
typedef struct TWds_s TWds_t;


/* kernel message */
struct KMsg_s
   	{
   	int 		CID;
   	int  		KMT;
   	int  		Size;
   	void 		*Data;
   	void 		(*GetData)( void * );
   	void 		(*PutData)( void * );
   	struct KMsg_s 	*Next;
   	struct KMsg_s 	*Previous;
   };
typedef struct KMsg_s TWds_msg_t;

struct KMsgEntry_s
   	{
   	TWds_msg_t 	*(*NewPtr)( void );
   	char 		Name[MAX_CLASS_NAME+1];
   	};
/* KernelMessageDictionaryEntry */
typedef struct KMsgEntry_s KMsgEntry_t;

struct KMsgDic_s 
   {
   KMsgEntry_t 		Entry[MAX_MAX_KMSG_CLASSES];
   KMsgEntry_t 		FastEntry[MAX_MAX_KMSG_CLASSES];
   unsigned long 	FirstPrime;
   };
/* KernelMessageDictionary */
typedef struct KMsgDic_s TWds_msgDic_t;

/* one of these structures per machine */
struct TWds_comm_s
	{
	TWds_msg_t	*KMSendBuffer[MAXKERNELS][MAX_PIGGYBACK];
	int       	KMSendBufCount[MAXKERNELS];
	TWds_msg_t 	*KMRecvBuffer[MAX_PIGGYBACK];

	int		(*KMDeleteBuffer)(TWds_msg_t *);
	int		(*KMCopy)(TWds_msg_t *, TWds_msg_t *);

	int       	KMRecvBufHead;
   	int    		KMRecvBufTail;
   	int       	SendCount;
   	int       	RecvCount;
	int       	Tid[MAXKERNELS];
   	int       	MyTid;
   	int       	MyInstance;
   	int  		NumTids;
   	int       	KernelSendCount;
   	int       	Threadid;
	};
typedef struct TWds_comm_s TWds_comm_t;
/* KernelComm */

#endif /* GTW_DS_H */
