#ifndef GTW_RAND_H
#define GTW_RAND_H
/******************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.5                             ***
***         Copyright 1997 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/

/* --------------------------------------------------------------------------
** C implementation of Pierre L'ecuyer's "double congruential random number
** generator" as it appeared in EFFICIENT AND PORTABLE COMBINED RANDOM NUMBER
** GENERATORS paper in the Communications of the ACM, June 1988, pp. 742-774.  
**
** Based on the ADA implementation of William M. Graham. (Coded by William N.
** Graham, Applied Research Labs, 2-11-91.)
**
** Requirements: 32 bit (or larger) machine. 
** The period for this generator is reported to be ~ 2.3 * 10**18. 
**
** This implementation (version 0.1.sun4) coded by
** Ioanis Nikolaidis (nikolaid@cc.gatech.edu) Georgia Tech, 5th May 92
** Code tested only on sun4 (Sparcstation) machines.
**
** Added other distributions - Chris Carothers, Bellcore, 23rd June 94
** Tested only on sun4.  Tested the distribution against published tables. The
** accuracy is to 5 decimals places on 1 million iterations.
**
** Modified to GTW interface - Richard Fujimoto, Georgia Tech, 15th August 94
** ------------------------------------------------------------------------ */

#define M1 2147483563
#define A1 40014
#define Q1 53668
#define R1 12211
#define M2 2147483399
#define A2 40692
#define Q2 52774
#define R2 3791

#define M1_MINUS_1 (M1 - 1)
#define NORMALIZE 4.656613E-10

#define TWOPI 6.28318530718

#if(defined(SUNMP) || defined(SUN4SOL2))
typedef long TWEachSeed_t;
typedef struct {TWEachSeed_t S1;  TWEachSeed_t S2;} TWSeed;
#else
typedef long TWEachSeed_t;
typedef struct { long S1;  long S2;} TWSeed;
#endif

extern void TWRandInit(TWSeed*, int);
extern void TWRandTerminate(void);
extern void TWRandGetSeeds(TWSeed*, TWEachSeed_t*, TWEachSeed_t*);
extern void TWRandSetSeeds(TWSeed *, TWEachSeed_t, TWEachSeed_t);
extern double TWRandUnif(TWSeed*);
extern long TWRandInteger(TWSeed*, long,long);
extern long TWRandBinomial(TWSeed*, long, double);
extern double TWRandExponential(TWSeed*, double);
extern double TWRandGamma(TWSeed*, double, double);
extern long  TWRandGeometric(TWSeed*, double);
extern double TWRandNormal01(TWSeed*);
extern double TWRandNormalSD(TWSeed*, double, double);
extern long  TWRandPoisson(TWSeed*, double);


/* external variables */
extern TW_LOCK_TYPE RandFileLock;


#define MTWRandUnif(S,Result)                                         \
   {                                                                  \
   long z, k;                                                         \
   if (((S)->S1 == 0) && ((S)->S2 == 0))                              \
      printf("WARNING: Random number seeds are 0; not initialized?"); \
   k = (S)->S1 / Q1;                                                  \
   (S)->S1 = A1 * ((S)->S1 - k * Q1) - k * R1;                        \
   if ((S)->S1 < 0)                                                   \
      (S)->S1 = (S)->S1 + M1;                                         \
   k = (S)->S2 / Q2;                                                  \
   (S)->S2 = A2 * ((S)->S2 - k * Q2) - k * R2;                        \
   if ((S)->S2 < 0)                                                   \
      (S)->S2 = (S)->S2 + M2;                                         \
   z = (S)->S1 - (S)->S2;                                             \
   if (z < 1)                                                         \
      z = z + M1_MINUS_1;                                             \
   Result = (double) z * NORMALIZE;                                   \
   }

#define MTWRandExponential(S,Lambda,Result)                           \
   {                                                                  \
   MTWRandUnif(S,Result);                                             \
   Result =  -Lambda * log(Result);                                   \
   }                  

#endif
