#ifndef GTW_SIM_H
#define GTW_SIM_H

/******************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.5                             ***
***         Copyright 1997 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its
     documentation for any research purpose and without fee is hereby
     granted, provided that the above copyright notice appears in all
     copies and that both that copyright notice and this permission
     notice appear in supporting documentation.  The use or inclusion of
     this software or its documentation in any commercial product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE, INCLUDING
     ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/

/*****************************************************************************
**      TIME WARP kernel on Shared Memory Architectures.
******************************************************************************/

#include <sys/resource.h>

#include "gtw_ds.h"
#include "gtw_ds_glob.h"
#include "gtw_debug.h"
#include "gtw_glob.h"
#include "gtw_rand.h"
#include "gtw_iolist.h"
#include "gtw_io.h"
#include "gtw_incr.h"
#include "gtw_alloc.h"
#include "gtw_calQ.h"
#include "gtw_config.h"
#include "gtw_memory.h"
#include "gtw_utility.h"
#include "gtw_wrapup.h"


/* MARIA TODO */

TW_SHARED char **Argv;    

TW_SHARED int TWAppStatMsgSize=128;
int pvm_debug=0;
int RFthread=0;
TW_SHARED int RTPEEventMem=0;
TW_SHARED int RTGPEStateMem=0;
TW_SHARED int RTMsgBufferCheck=0;

TW_SHARED int 	gtw_debug = 0;
int extern_debug[MAXPROC] = {0};
int TWapp_flags=0;
TW_SHARED long MsgBufferCheck=0L;
int end_test=0;
int DO_PROCS=TRUE;
int elist_test=0;
int num_machines=1;

/* MARIA TODO */
/* local functions */
void SetDefaults ( void );
void set_core_limit ( void );
void InitTimeVars ( void );
void InitApplVars (void);
void CheckTWInit( void );
void ShowMapping(void);
void InitSynchVar(void);
void RunSim(void);
void RunSim(void);
void InitPE ( int );
void InitBufs(int );
void ProcessMsgQ( int );
void ProcessCanQ( int );
void LPScheduler ( void );

#if( defined(SUNMP) || defined(SUN4SOL2))
void * Scheduler ( void * );
#endif

#if( defined(SGI5) || defined(SGIMP64) || defined(SGI64) )
void Scheduler( void );
#endif

void LPScheduler( void );
void print_msg_q(struct Event *);

/* ---------------------------------------------------------------------------
**  System global variables and runtime constants
**-------------------------------------------------------------------------- */
#if( defined(SGI5) || defined(SGIMP64) || defined(SGI64) )
usptr_t * shm_handle;
char shared_area[80];
#endif /* SGI5 || SGIMP64 || SGI64 */

/* major modifiable data structures ---------------------------------------- */
TW_SHARED TW_ALIGN struct PEState GState[MAXPROC];  /* global PE state */

/* read-only variables ----------------------------------------------------- */
TW_SHARED int GlobEventMem=DEFAULT_GLOB_EV_MEM;/* bytes per PE global events */
TW_SHARED int PEEventMem=DEFAULT_EV_MEM;   /* bytes per PE for events */
TW_SHARED int TWnpe=0;			   /* number of PEs in system */
TW_SHARED int TWnlp=0;			   /* number of LPs in system */
TW_SHARED TWLPType TWLP[MAXLP];		   /* stuff related to LPs */
TW_SHARED int TWMsgSize=128;		   /* application supplied */
TW_SHARED int GPEStateMem=-1;		   /* Bytes memory per PE for state */
TW_SHARED int MaxStateSize;		   /* max of state sizes of LPs */
TW_SHARED int TWdebug = 0;		   /* turn on debuging (must be 32 bit)*/
TW_SHARED int TWstat = 0;		   /* statistics display level */

TW_SHARED int batch=1;			/* num. of iterations in batch loop */
TW_SHARED int LPNum[MAXLP];		/* tells which LP in PE */
TW_SHARED double ENDTIME=100;		/* time when simulation  ends */
TW_SHARED double OFFSET=0;		/* IO3 synchronize clients, and endsim */
TW_SHARED float RetThreshold;

int GVTPeriod=DEFAULT_GVT_PERIOD; 	/* MARIA */

/* writable variables ------------------------------------------------------ */
/* beware of cache overheads! */
TW_SHARED TW_SYNCH int NBusy;		/* # of PEs idling */

/* variables connected with memory management ------------------------------ */
TW_SHARED TW_ALIGN struct Event *FLast;	/* last event in global free list */
TW_SHARED TW_ALIGN struct Event *PFree;	/* first element in global free list */
TW_SHARED TW_ALIGN TW_LONG NumGlobalPool;
TW_SHARED TW_LOCK_TYPE Lck_NumGlobalPool;

/* variables associated with GVT computations ------------------------------ */
TW_SHARED TW_ALIGN TWTime LocalGVT;  	/* GVT on this Machines */
TW_SHARED TW_ALIGN TWTime GVT;		/* always ready approximation to GVT */
TW_SHARED int NSameGVT;			/* # consecutive times got same GVT  */
TW_SHARED TW_ALIGN TWTime Local_Min[MAXPROC];
TW_SHARED TW_SYNCH int GVTflag;		/* #PE doing GVT calc.     */
TW_SHARED TW_SYNCH int GVTProcessor;	/* GVT regulating processor */

/* barriers, go1 should be sufficient -------------------------------------- */
TW_SHARED TW_BARRIER_TYPE Go1;

/* reflector thread barrier */
TW_SHARED TW_BARRIER_TYPE RF1; 
TW_SHARED TW_BARRIER_TYPE RF2; 

/* global locks ------------------------------------------------------------ */
TW_SHARED TW_LOCK_TYPE Lck_PFree;
TW_SHARED TW_LOCK_TYPE Lck_NBusy;
TW_SHARED TW_LOCK_TYPE Lck_Prstate;
TW_SHARED TW_LOCK_TYPE Lck_GVTflag;

/* file name to write results to,  for easy parsing */
TW_SHARED char TW_results_file_name[80];

/* ---------------------------------------------------------------------------
**  Processor private globals
**---------------------------------------------------------------------------*/
TW_SHARED TW_LOCK_TYPE dumplck;	/* for dumping */
int randmixer=0;	/* used to mix random number generator */


#endif /* GTW_SIM_H */
