#ifndef GTW_TIMING_H
#define GTW_TIMING_H
/******************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.4                             ***
***         Copyright 1995 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/

/**********************************************************************
*   TWtimer is the timer unit
*   TWnow(reg) notes real time in some abstract timer unit in reg.
*   TWseconds() converts that unit to seconds and returns that value.
*   TWzero_timer will zero the timer
*   TWaccumulate_time will take the difference between end and start timers
*      and accumulate it in a timer.
***********************************************************************/


/* ************************************************** */
#ifdef sparc_sequential

#include <sys/time.h>
#include <sys/resource.h>

typedef struct timeval TWtimer;
struct timeval TWtime_accumulate();

static struct rusage usage;

#define TWInitClk() 

/* user and system time used so far */
#define TWnow(reg) \
 if(getrusage(RUSAGE_SELF, &usage) == -1){perror("getrusage");}\
 reg.tv_sec = usage.ru_stime.tv_sec + usage.ru_utime.tv_sec; \
 reg.tv_usec = usage.ru_stime.tv_usec + usage.ru_utime.tv_usec; 

/* convert to seconds */
#define TWseconds(x) ((x).tv_sec + 1.0e-6 * (double)(x).tv_usec)
#define TWzero_timer(reg) reg.tv_sec = 0; reg.tv_usec = 0;
struct timeval TW_add_time();
#define TWaccumulate_time(s,e,t) t = TW_add_time(s,e,t)
#endif

/***************************************************************************/
/*************************** SUN Timing Support ****************************/
/***************************************************************************/
#if(defined(SUNMP) || defined(SUN4SOL2))
#include <sys/time.h>

#ifdef NEWTIMER

#define TWnow(reg) reg = gethrtime();
#define TWLRnow(reg) reg = gethrtime();
#define TWInitClk()
typedef long long TWtimer;
#define TWseconds(x) (x * 1.0e-9)

#define TWzero_timer(reg) reg = 0;

typedef long GVTtimer_t;
typedef double ClockTicks;
typedef TWtimer GVTTimer_t;
#define TWTimer ClockTicks
#define TWwallclock(x)		\
	{			\
	TWtimer reg;		\
	TWnow (reg);		\
	x = TWseconds(reg);	\
	}

#else
#define TWnow(reg) reg = gethrvtime();
#define TWLRnow(reg) reg = gethrvtime();
#define TWInitClk()
#define TWzero_timer(reg) reg = 0;
typedef long long TWtimer;
#define TWseconds(x) (x * 1.0e-9)
typedef long GVTtimer_t;
typedef double ClockTicks;
typedef TWtimer GVTTimer_t;
#define TWTimer ClockTicks

#define TWwallclock(x)		\
	{			\
	TWtimer reg;		\
	TWnow (reg);		\
	x = TWseconds(reg);	\
	}

#define TWYield()
#endif


#endif


/****************************************************************************/
/*************************** SGI Timing Support *****************************/
/****************************************************************************/
#if(defined(SGI5) || defined(SGI64) || defined(SGIMP64))

typedef long GVTtimer_t;
typedef double ClockTicks;
#define TWTimer ClockTicks;

/* ------------------------------------------------------------------------ */
#ifdef NEWTIMER

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <fcntl.h>

#include <sys/time.h>
#include <sys/mman.h>
#include <sys/syssgi.h>
#include <sys/immu.h>
#include <sys/resource.h>


typedef double TWtimer;
typedef unsigned int fasttimer_t;

extern volatile fasttimer_t *fasttimer_addr;
extern int size_factor;
extern int pico_sec_per_cycle;
extern struct timeval tv;

#define TWnow(reg) (reg = (double)*(fasttimer_addr + size_factor))

#define TWLRnow(reg)                                              \
 if(gettimeofday(&tv, NULL) == -1){perror("gettimeofday");}       \
 reg = 1.0e6 * tv.tv_sec ;                                        \
 reg += (double)tv.tv_usec;                                       \
 reg = ((reg * 1.0e6) / (TWtimer)pico_sec_per_cycle)


#define TWseconds(reg) ((reg * (TWtimer)pico_sec_per_cycle) * 1.0e-12)

void TWInitTimer( void );

#define TWzero_timer(reg) reg=0;


#else
/* ------------------------------------------------------------------------ */
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#define TWnow(reg) gettimeofday(&reg, NULL);
#define TWInitClk()
#define TWzero_timer(reg) (reg).tv_sec = 0; (reg).tv_usec = 0;
typedef struct timeval TWtimer;
#define TWseconds(x) ((double)x.tv_sec + (double)x.tv_usec*1e-6)

#define TWwallclock(x)		\
	{			\
	TWtimer reg;		\
	TWnow (reg);		\
	x = TWseconds(reg);	\
	}

#endif 
/* ------------------------------------------------------------------------ */

/* RENTRANT timers: only time() returns value in whole seconds!  */ 
typedef time_t TWTIME;
#define TWWALLCLOCK(x) time (&x)

#define TWYield() sginap( 0 )


#endif /* SGI5 ||SGI64 || SGIMP64 */

#endif /* GTW_SIGNAL_H */


