/*****************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.4                             ***
***         Copyright 1997 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/

/*****************************************************************************
        TIME WARP kernel on Distributed & Shared Memory Architectures.
******************************************************************************/
#include <time.h> 
#include "gtw.h"
#include "pong.h"
#include "cs_api.h"     /* cloning api header file */
#include "hashcalculator.h"
#include "hashutils.h"

void IPong( struct MyLPState *, int );
void IPing( struct MyLPState *, int );

/* ---------------------------------------------------------------------------
** TWInitAppl ()
**
** DESCRIPTION: Initializes Application
**
**---------------------------------------------------------------------------*/
void TWInitAppl( int argc, char **argv )
{
   int my_lps=0, i=0;
   void IPing(), Ping(), FPing();
   noOfFound = 0 ;
   
   processingTime = 0 ;
   cacheOverhead = 0 ;

   TWnlp = my_lps;         /* specify number of LP */

   if ( argc>0 )
   {
	if( strcmp(argv[0],"-A")!=0 ) 
	{
		printf("\nUsage: ....-A -l <no. of Lp's> -r <range of state variables>.");
		exit(0) ;
	}
	else
	{
           for( i=1 ; i<argc-1 ; i++ )
	   {	
	     if( strcmp(argv[i],"-l")==0 )	
	     {
		i++ ;
		my_lps = atoi(argv[i]) ;	
		/* printf("\nLP--->%d.",my_lps); */
             }
	     else if( strcmp(argv[i],"-r")==0 )		
             {
		i++;
		RANGE = atoi(argv[i]) ;
		 /* printf("\nR-->%d.",RANGE) ;	*/
	     }
             else if( strcmp(argv[i],"-cache")==0 )
             {
                i++;
                MEMORY_POOL_SIZE = atoi(argv[i]) ;
                /* printf("\nCACHE-->%d.",MEMORY_POOL_SIZE) ; */
             }
	     else if( strcmp(argv[i],"-et")==0 )
             {
                i++;
                ET = atoi(argv[i]) ;
                /* printf("\nET-->%d.",ET) ; */
             }
             else if( strcmp(argv[i],"-factor")==0 )
             {
                i++;
                factor = atof(argv[i]) ;
                /* printf("\nET-->%d.",ET) ; */
             }	
	     else
	     {
		printf("\nUsage: ....-a/-A -l <no. of Lp's> -r <range of state variables>.");
                exit(0) ;
	     }
           }
	}
   }

   for( i=0 ; i < argc ; i++ )
   {
	memset( argv[i],NULL,strlen(argv[i]) ) ;
   }

   TWSetNumLPs(my_lps); 
   
   /* LP to PE mapping */
   for( i=0 ; i<my_lps ; i++ )
	TWLP[i].Map = i % TWGetNumGlobalPEs() ;	
   
   /* set initialization, event handler, and wrap-up procedures */
   for( i=0 ; i<my_lps ; i++ )
   {
   	TWLP[i].IProc = (FuncPtr2) 	IPing;
   	TWLP[i].Proc  = (FuncPtr) 	Ping;
   	TWLP[i].FProc = (FuncPtr2) 	FPing;
   	TWLP[i].IncrSave = TRUE;
   	TWLP[i].State = TWMalloc(sizeof(struct MyLPState));
   	TWLP[i].LPStateSize = sizeof(struct MyLPState);	
   	TWLP[i].CopySize = sizeof(struct MyAutoVars); 	
   }

   TWMsgSize = sizeof(struct MyMsgData);  
   TWAppStatMsgSize = sizeof( struct MyMsgData ); 

   CloneSim_InitAppl( argc, argv ); 
   fflush(stdout);

   if( createMemoryPool() != 1 )
   {
        printf("\nUnable to create Memory Pool.\nEXITING .....");
        exit(0);

   }
   if( createHashTable(my_lps) != 1 )
   {
        printf("\nUnable to create Hash.\nEXITING .....");
        exit(0);

   }


}

/* ---------------------------------------------------------------------------
** IPing ()
**
** DESCRIPTION: Initializes LP
**
**---------------------------------------------------------------------------*/
void IPing( struct MyLPState *SV, int MyPE)               /* Ping is LP 0 */
{   
   struct MyMsgData *TWMsg;
   int i ;
   TWTime ts ;
   TWSeed Seeds ;

   struct LPState * CurState;
   CurState = GState[MyPE].CurState; 

   /* SV is NULL since TWInit() did not allocate State */
   /* allocate memory for state vector, and set pointer to LP state */

   TWLP[TWMe()].State = TWMalloc(sizeof(struct MyLPState));
   SV = (struct MyLPState *) TWLP[TWMe()].State;

   /* specify variables to be automatically checkpointed */
   TWAutoCheck( (char *) &(SV->CVars), sizeof(struct MyAutoVars));

   if( TWMe() == 0 )
   {
   	/* initialize state variables */
   	SV->ROVars.NeighborID = rand() % TWnlp;   /* randomly getting next LP*/
	SV->CVars.var1 = rand() % RANGE ;
	SV->ISVars.var2 = rand() % RANGE ;

        if ( CurState->IncrSave )
        	TWCheck(&(SV->ISVars.var2));

   	TWRandInit (&(Seeds), 0);
 
   	/* send initial message */
   	ts = TWRandExponential(&(Seeds), 1.0);
   	TWGetMsg(ts, SV->ROVars.NeighborID, sizeof (struct MyMsgData));

        TWMsg->data1 = rand() % RANGE ;
        TWMsg->data2 = rand() % RANGE ;

   	TWSend();
   }

}

/* ---------------------------------------------------------------------------
** Ping  ()
**
** DESCRIPTION: Event Handler for Ping
**
**---------------------------------------------------------------------------*/
void Ping ( struct MyLPState *SV, struct MyMsgData *M, int MyPE )
{   
   struct MyMsgData *TWMsg= NULL;
   int hashindex=0 ;
   TWTime ts;
   TWSeed Seeds = {0,0} ;
   struct LPState *CurState=NULL;
   char HIT = 0 ;

   clock_t startAddingTocache=0, endAddingTocache=0, start=0, startProcessing=0, end=0 ;

   struct timespec *t1 = (struct timespec *)calloc( 1,sizeof(struct timespec) ) ;
   struct timespec *t2 = (struct timespec *)calloc( 1,sizeof(struct timespec) ) ;

   struct cacheStruct* temp = NULL ;
  
   t1->tv_sec = 0 ;
   t1->tv_nsec = 1000000000/ET ;
 
   CurState = GState[MyPE].CurState; 

   memcpy(&(globalMemStruct1->st),SV,sizeof(struct MyLPState));
   memcpy(&(globalMemStruct1->msg),M,sizeof(struct MyMsgData));

   /* update state */

   TWRandInit (&(Seeds), 0);
   ts = TWRandExponential(&(Seeds), 1.0);

   hashindex = hashIndexOf( (unsigned char *)globalMemStruct1,sizeof(struct MyStateData)) ;

   if ( CurState->IncrSave )
        TWCheck(&(SV->ISVars.var2));

   start = clock() ;
   /* printf("\nSTART->%ld",start) ; */

   /* printf("\nPrTime->%ld\n",processingTime) ;
   printf("\nCacheOv->%ld\n",cacheOverhead) ; */

   if( (processingTime - cacheOverhead ) >= 0 )
   {
     if(( temp = checkIfPresent(globalMemStruct1,hashindex,TWMe())) != NULL)
     {
     	HIT = 1 ;

	memcpy(SV,&(temp->result.st),sizeof(struct MyLPState));

	TWGetMsg(ts, SV->ROVars.NeighborID, sizeof(struct MyMsgData));

	memcpy(TWMsg,&(temp->result.msg),sizeof(struct MyMsgData));

     }
     else
     {
   	startProcessing = clock() ;
        /* printf("\nSTARTP->%ld",startProcessing) ; */

	SV->ROVars.NeighborID = rand() % TWnlp;   /* randomly getting next LP*/
	SV->CVars.var1 = rand() % RANGE ;
        SV->ISVars.var2 = rand() % RANGE ;

	TWGetMsg(ts, SV->ROVars.NeighborID, sizeof(struct MyMsgData));	

	TWMsg->data1 = rand() % RANGE ;
	TWMsg->data2 = rand() % RANGE ;

        startAddingTocache = clock() ;
        
	memcpy(&(globalMemStruct2->st),SV,sizeof(struct MyLPState));
	memcpy(&(globalMemStruct2->msg),TWMsg,sizeof(struct MyMsgData));

	addToHashTable(globalMemStruct1,globalMemStruct2,hashindex,TWMe()) ; 
 
        endAddingTocache = clock() ;

        timeTakenTocache = endAddingTocache - startAddingTocache ;
       
 	nanosleep(t1,t2) ;
     }
   
     end = clock() ;
     /* printf("\nToTime-->%ld\n",end -start) ; */
   
     if( HIT == 1)
	cacheOverhead = end - start ;
     else
     {
   	processingTime = (end - startProcessing) - timeTakenTocache + (t1->tv_nsec/1000) ;
   	cacheOverhead = 0  ;
     } 

     if( cacheOverhead > processingTime)
	printf("\nTIME SENSTIVITY ACHIEVED\n");

   }
   else
   {
   	SV->ROVars.NeighborID = rand() % TWnlp;   /* randomly getting next LP*/
        SV->CVars.var1 = rand() % RANGE ;
        SV->ISVars.var2 = rand() % RANGE ;

        TWGetMsg(ts, SV->ROVars.NeighborID, sizeof(struct MyMsgData));

        TWMsg->data1 = rand() % RANGE ;
        TWMsg->data2 = rand() % RANGE ; 
        
   	nanosleep(t1,t2) ;
   }
 
   HIT = 0 ;
   timeTakenTocache = 0 ; startAddingTocache=0; endAddingTocache=0; start=0; startProcessing=0; end=0 ;
   TWSend();

}

/* ---------------------------------------------------------------------------
** FPing  ()
**
** DESCRIPTION: Wrap up for Ping
**
**---------------------------------------------------------------------------*/
/* void FPing (int LP, struct MyLPState *SV, TW_ADDR_TYPE Data ) */
void FPing ( struct MyLPState *SV, int MyPE )

{  
   printf("FPing is done with application\n");

}

/* ---------------------------------------------------------------------------
** TWWrapAppl  ()
**
** DESCRIPTION: WrapUp for Application
**
**---------------------------------------------------------------------------*/
void TWWrapAppl( void )
{
   int i = 0 ;

   free(memoryPool) ;
   free(globalMemStruct1);
   free(globalMemStruct2) ;

   for(i=0 ; i < TWnlp ; i++ )
   {
   	free( LPCache[i] );
        free( LPFlags[i] ) ;
   }

   printf("Application done \n");
   printf("\nNumber of FOUND->%d.",noOfFound) ;
   CloneSim_CloseAppl();
}
