/******************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.4                             ***
***         Copyright 1997 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/

/*****************************************************************************
        TIME WARP kernel on Distributed & Shared Memory Architectures.
******************************************************************************/
#include "gtw.h"
#include "pong.h"
#include "cs_api.h"     /* cloning api header file */

void IPong( struct MyLPState *, int );
void IPongInit( struct MyLPState *, int );

int noInteractions = 0 ;

/* ---------------------------------------------------------------------------
** TWInitAppl ()
**
** DESCRIPTION: Initializes Application
**
**---------------------------------------------------------------------------*/
void TWInitAppl( int argc, char **argv )
{
   int my_lps = 0, no_PEs = 0, i = 0 ;

   void IPong(), Pong(), FPong(), IPongInit();

   printf("\nEnter the value of TWnlp :  ") ;
   scanf("%d",&my_lps) ;

   TWnlp = my_lps ;         /* specify number of LP */

   TWSetNumLPs(my_lps); 

   /* LP to PE mapping */

   no_PEs = TWGetNumGlobalPEs() ;
   
   for(i=0 ; i<TWnlp ; i++ )
	TWLP[i].Map = i % no_PEs ; 
   
/* set initialization, event handler, and wrap-up procedures */
   for( i=0 ; i < TWnlp ; i++ )
   { 
	if( i == 0)
		TWLP[i].IProc = IPongInit ;
	else
		TWLP[i].IProc = IPong ;
	TWLP[i].Proc  = Pong;
   	TWLP[i].FProc = FPong; 
   	TWLP[i].IncrSave = TRUE;
   	TWLP[i].State = TWMalloc(sizeof(struct MyLPState));
   	TWLP[i].LPStateSize = sizeof(struct MyLPState);
   	TWLP[i].CopySize = sizeof(struct MyAutoVars); 
   }

   TWMsgSize = sizeof(struct MyMsgData);  
   TWAppStatMsgSize = sizeof( struct MyMsgData ); 

   CloneSim_InitAppl( argc, argv ); 
   fflush(stdout);
}

/* ---------------------------------------------------------------------------
** IPong ()
**
** DESCRIPTION: Initializes LP
**
**---------------------------------------------------------------------------*/
void IPongInit( struct MyLPState *SV, int MyPE)               /* Used for 1st LP */
{
   struct MyMsgData *TWMsg;
   int i;
   TWTime ts;

   struct LPState * CurState;
   CurState = GState[MyPE].CurState;

   /* SV is NULL since TWInit() did not allocate State */
   /* allocate memory for state vector, and set pointer to LP state */

   TWLP[TWMe()].State = TWMalloc(sizeof(struct MyLPState));
   SV = (struct MyLPState *) TWLP[TWMe()].State;

   /* specify variables to be automatically checkpointed */
   TWAutoCheck( (char *) &(SV->CVars), sizeof(struct MyAutoVars));

   /* initialize state variables */
   SV->ROVars.NeighborID = ( TWMe() + 1 )% TWnlp ;   /* The next LP in the ring */
   SV->CVars.Count = 0;
   for (i=0; i<1000; i++) SV->ISVars.LastTS[i] = 0.0;
   TWRandInit (&(SV->CVars.Seeds), 0);

   /* update state */
   TWCheckTWTime(&(SV->ISVars.LastTS[SV->CVars.Count % 1000]));
   SV->ISVars.LastTS[SV->CVars.Count % 1000] = TWNow();


   /* send initial message */
   ts = TWRandExponential(&(SV->CVars.Seeds), 1.0);
   TWGetMsg(ts, SV->ROVars.NeighborID, sizeof (struct MyMsgData));
   TWMsg->count = 0;
   printf("ipongInit\n");
   fflush(stdout);

   printf("\nSending to LP%d TIME : %f", SV->ROVars.NeighborID,TWNow()) ;
   noInteractions++ ;
   TWSend();

}


/* ---------------------------------------------------------------------------
** IPong ()
**
** DESCRIPTION: Initializes LP
**
**---------------------------------------------------------------------------*/
void IPong( struct MyLPState *SV, int MyPE)               
{
   struct MyMsgData *TWMsg;
   int i;

   struct LPState * CurState;
   CurState = GState[MyPE].CurState;

   /* allocate memory for state vector, and set pointer to LP state */
   SV = (struct MyLPState *) TWLP[TWMe()].State;

   /* specify variables to be automatically checkpointed */
   TWAutoCheck( (char *) &(SV->CVars), sizeof(struct MyAutoVars));

   /* initialize state variables */
   SV->ROVars.NeighborID = ( TWMe() + 1 )% TWnlp ;   /* The next LP in the ring */
   SV->CVars.Count = 0;
   for (i=0; i<1000; i++) SV->ISVars.LastTS[i] = 0.0;
   TWRandInit (&(SV->CVars.Seeds), 0);

   printf("ipong\n");
   fflush(stdout);
}

/* ---------------------------------------------------------------------------
** Pong  ()
**
** DESCRIPTION: Event handler for Pong! 
**
**---------------------------------------------------------------------------*/
void Pong ( struct MyLPState *SV, struct MyMsgData *M, int MyPE )
{  
   struct MyMsgData *TWMsg;
   TWTime ts;
   float roundTripTime = 0.0 ;

   struct LPState * CurState;
   CurState = GState[MyPE].CurState; 
   
   fflush(stdout);

   /* update state */
   TWCheckTWTime(&(SV->ISVars.LastTS[SV->CVars.Count % 1000])); 
   SV->ISVars.LastTS[SV->CVars.Count % 1000] = TWNow();

   /* Calculating Round Trip time */
   if( noInteractions%TWnlp == 0 )
        printf("\n\t\t ROUND TRIP TIME : %f", SV->ISVars.LastTS[(SV->CVars.Count) % 1000]-SV->ISVars.LastTS[(SV->
CVars.Count-1) % 1000]) ;  
   ts = TWRandExponential(&(SV->CVars.Seeds), 1.0);

   /* Forwarding message to the next LP */
   TWGetMsg(ts, SV->ROVars.NeighborID, sizeof (struct MyMsgData));
   TWMsg->count = M->count + 1;

   printf("\nSending to LP-%d TIME : %f", SV->ROVars.NeighborID, TWNow()) ;

   noInteractions++ ;
   SV->CVars.Count++;

   TWSend();
}

/* ---------------------------------------------------------------------------
** FPong  ()
**
** DESCRIPTION: Wrap up for Pong
**
**---------------------------------------------------------------------------*/
/* void FPong (int LP, struct MyLPState *SV, TW_ADDR_TYPE Data ) */
void FPong ( struct MyLPState *SV, int MyPE )
{  int i;

   if (SV->CVars.Count < 1000) /* less than 1000 messages received */
      for (i=SV->CVars.Count; i>=0; i--)
	/* printf ("%f\n", SV->ISVars.LastTS[i])*/ ;
   else /* 1000 or more messages received */
      for (i=0; i<1000; i++)
	/* printf ("%f\n",SV->ISVars.LastTS[(SV->CVars.Count-i+1000)%1000])*/ ;

   printf("FPong is done with application\n");
}

/* ---------------------------------------------------------------------------
** TWWrapAppl  ()
**
** DESCRIPTION: WrapUp for Application
**
**---------------------------------------------------------------------------*/
void TWWrapAppl( void )
{
   printf("Application done \n");
   printf("\nNumber of Interactions 	:	%d\n",noInteractions) ;
   CloneSim_CloseAppl();
}
