/******************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.5                             ***
***         Copyright 1997 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/


/*****************************************************************************
**  	TIME WARP kernel on Distributed & Shared Memory Architectures.
******************************************************************************/

#include "gtw_glob.h"

#if( defined(SUNMP) || defined(SUN4SOL2) )

void barrier_init_local(barrier_pvt_t * B)
{
B->Local_sense = False;
}


void barrier_init(barrier_t * b, int n)
{
  TW_INIT_LOCK(b->lock);
  b->Count = b->Threads = n;
  b->Sense = False;
}

void barrier_wait(barrier_t * b, barrier_pvt_t * p, int id)
{

  (p->Local_sense) ^= True;

  /* printf("Local sense %d global is %d, id is %d, count %d\n",
	  p->Local_sense, b->Sense, id, b->Count); 
  print_barrier(b); */

  TW_LOCK(b->lock);
  b->Count --;
  TW_UNLOCK(b->lock);

  /* printf("Count %d \n", b->Count);  */

  if(b->Count != 0){
    while(b->Sense != (p->Local_sense)){
      /* printf("id %d Local sense %d global is %d \n", id,
	     p->Local_sense, b->Sense);  
      fflush(stdout);*/


    }
    return;
  }
  
  b->Count = b->Threads;
  b->Sense = (p->Local_sense);
  return;
}

void print_barrier(barrier_t * b)
{
  printf("count %d, Sense %d, Threads %d\n", b->Count, b->Sense, b->Threads);
}

#endif /* SUNMP || SUN4SOL2 */

