/******************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.4                             ***
***         Copyright 1995 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#undef MAXFLOAT

#include "gtw_glob.h"
#include "gtw_incr.h"
#include "gtw_insim.h"
#include "gtw_utility.h"
#include "gtw_debug.h"

#include "gtw_ds.h"
#include "gtw_ds_glob.h"

/* forward refererences */
void 	to_uppercase ( char * );
void 	TWFeedBack( void );
int 	build_file_name ( char **file_name, int top_level );

/* GLOBALS */
char TWConfigFile[GTW_MAX_FILE_NAME_LENGTH]="" ;
char TWMachineFile[GTW_MAX_FILE_NAME_LENGTH]="";

int machine_file=0;
int my_machine=-1;

/* IO4 */
extern int 	show_gvt;
int           show_noprint=10000;
/*int             show_noprint=0; */     /* production, and releases */
int 		show_gvt_dump=0; 
int 		show_gvt_all=0; 
int 		spread_clone=0;
int 		limit_output=0;
int 		batch_sampling=0; 
int 		batch_processing=0; 
double 		newMIT=0.0;

int 		clone_multiple =1;
int 		minimize_copy =1;
int 		do_mh_memory=2;
int 		do_headers=1;
int 		print_tables[10] = {0};
int		CloneSim_scale_size=1;
int		CloneSim_map_msgs=0;
int		CloneSim_read_files=0;


char 		*D_MITfix=NULL;
int 		do_restrictions=0;

/* generic application */
int 		do_z=0;		
int 		do_Z=0;		


/*-----------------------------------------------------------------------------
**  to_uppercase()
**
**  DESCRIPTION:
**	converts string s to all uppercase 
**
**---------------------------------------------------------------------------*/
void to_uppercase( char *s )
{
   while( *s != '\0' ) 
	{
      	if( (*s >= 'a') && (*s <= 'z') )
	 	*s = *s - 'a' + 'A';
      	s++;
   	}
}

/*-----------------------------------------------------------------------------
**  build_file_name()
**
**  DESCRIPTION:  top_level = 0 gives a .tw-config file in an arch/dependent
**	data directory top_level = 1 gives a machine file in  and
**	arch.independent data directory
**	
**---------------------------------------------------------------------------*/
int build_file_name( char **file_name, int top_level )
{
   char thishost[GTW_MAX_STRING];
   char thisfile[GTW_MAX_FILE_NAME_LENGTH]="" ;
   struct stat status_struct;

   char *gtw_root;
   char *arch;

   char *return_string;

   if( gethostname ( thishost, GTW_MAX_STRING )  == -1 )
	{
	perror( "build_file_name: gethostname" );
	exit ( EXIT_FAIL );
	}
 
   /* get environmental variables */
   if( ( gtw_root = getenv( "GTW_ROOT" ) ) == NULL )
	{
	perror( "Configure: GTW_ROOT environment variable not set" );
	exit( -1);
	}

   if( ( arch = getenv( "ARCH" ) ) == NULL )
	{
	perror( "Configure: ARCH environment variable not set" );
	exit( -1);
	}


   if( top_level == 0 )
	{
	sprintf( thisfile, "%s/config/%s/.tw-config", gtw_root, arch );
	if( stat (thisfile, &status_struct) == -1 ) return FALSE;
	}

   else
	{
	sprintf( thisfile, "%s/config/machines", gtw_root );
	if( stat (thisfile, &status_struct) == -1 ) return FALSE;
	machine_file=1;
	}

   return_string = strdup (thisfile);
   *file_name = return_string;

   return TRUE;
}


/*-----------------------------------------------------------------------------
**  Configure()
**
**  DESCRIPTION:
**	 Read and parse .tw-config file. Borrowed from the PAT code gotten
**	 from Kevin Smith -- SRD 8/12/94  
**
**---------------------------------------------------------------------------*/
void Configure( char *thisfile )
{
   FILE *fp;
   int line = 0;
   char val[GTW_MAX_STRING]="";
   char token[GTW_MAX_STRING]="";
   char thishost[GTW_MAX_STRING]="";
   int c;
   int i;
   int current_machine=-1;

   if( gethostname ( thishost, GTW_MAX_STRING )  == -1 )
	{
	perror( "Configure: gethostname" );
	exit ( EXIT_FAIL );
	}
   else
	{
	++current_machine;
	strcpy ( TWds.HostName[current_machine], thishost );
	}

   if( ( fp = fopen( thisfile, "r") ) == NULL )
	{
	fprintf ( stderr, "file: %s\n", thisfile );
	fflush (stderr);
	perror( "Configure: missing or unreadable");
	exit(-1);
	}

   /* scan file and get parameters */
   while( ((int)(c = getc(fp))) != EOF ) 
	{
      	line++;
      	if( c == '#' ) 
		{		
		/* skip the comment line */
	 	for( ; (c != '\n') && ( (int)c != EOF ); c = getc(fp) );
	 	if( (int)c == EOF )
	    		break;
	 	continue;
      		}
	/* if line only contains a newline skip it */
      	if( c == '\n' )
	 	continue;
      	else if( (int)c == EOF )
	 	break;

	/* skip white space */
      	else if( (c == ' ') || (c == '\t') ) 
		{
	 	if( fscanf( fp, "%s", token ) == EOF )
	    		break;		/* end of file */
      		} 
	else 
		{
		/* get first token: format of file is TOKEN = VALUE */
	 	token[0] = c;
	 	if( fscanf( fp, "%s", &token[1] ) == EOF )
	    		break;		/* end file */
      		}

	/* find = sign after token, while skipping white spaces */
      	while( ((c = getc(fp)) == ' ') || (c == '\t') );
      	if( c != '=' ) 
		{
	 	warn( "Bad command on line %d of .tw-config\n", line );
	 	if( (int)c == EOF )
	    		break;
	 	for( ; (c != '\n') && ((int)c != EOF); c = getc(fp) );
	 	if( (int)c == EOF )
	    		break;
	 	continue;
      		}
  	/* get assigned value of token */
      	fscanf( fp, "%s", val );	/* value of initializer */

      	/* parse token */
      	to_uppercase( token );

      	/* leave val is lowercase */
      	/* to_uppercase( val ) */

      	if( strcmp( token, "PE_EVENT_MEMORY" ) == 0 )
	 	PEEventMem = atoi( val );
      	else if( strcmp( token, "APP_FLAGS" ) == 0 )
	 	TWapp_flags = atoi( val );
      	else if( strcmp( token, "GVT_PERIOD" ) == 0 )
	 	GVTPeriod = (long) atoi( val );
      	else if( strcmp( token, "MSG_BUFFER_CHECK" ) == 0 )
	 	MsgBufferCheck = (long) atoi( val );
      	else if( strcmp( token, "GLOB_EVENT_MEMORY" ) == 0 )
	 	GlobEventMem = atoi( val );
      	else if( strcmp( token, "PE_STATE_MEMORY" ) == 0 )
	 	GPEStateMem = atoi( val );
      	else if( strcmp( token, "RAND_MIXER" ) == 0 )
	 	randmixer = atoi( val );
      	else if( strcmp( token, "INCR_N_ARRAYS" ) == 0 )
	 	NLogArrays = atoi( val );
      	else if( strcmp( token, "INCR_ARRAY_SIZE" ) == 0 )
	 	TW_Incr_Array_Size = atoi( val );
      	else if( strcmp( token, "BATCH" ) == 0 )
	 	batch = atoi( val );
      	else if( strcmp( token, "TWMSGSIZE" ) == 0 )
	 	TWMsgSize = atoi( val );

      	else if( strcmp( token, "DEBUG" ) == 0 ) 
		{
	 	if( strcmp( val, "MESSAGE" ) == 0 )
	    		TWdebug += DEBUG_MSG;
	 	else if( strcmp( val, "THREAD" ) == 0 )
	    		TWdebug += DEBUG_THREAD;
	 	else if( strcmp( val, "ALL_GVT" ) == 0 )
	    		TWdebug += DEBUG_ALL_GVT;
	 	else if( strcmp( val, "SOME_GVT" ) == 0 )
	    		TWdebug += DEBUG_SOME_GVT;
	 	else if( strcmp( val, "FEEDBACK" ) == 0 )
	    		TWdebug += DEBUG_FEEDBACK;
	 	else if( strcmp( val, "SILENT" ) == 0 )
	    		TWdebug += DEBUG_SILENT;
	 	else if( strcmp( val, "LBS" ) == 0 )
	    		TWdebug += DEBUG_LBS;
	 	else if( strcmp( val, "PROGRESS" ) == 0 )
	    		TWdebug += DEBUG_SHOW_PROG;
	 	else 
			{
	   		warn( "Bad TWdebug value (%s) on line %d"
				" of .tw-config", token, line );
	   		continue;
	 		}
       		}

      	else if( strcmp( token, "RETTHRESHOLD" ) == 0 )
		RetThreshold = atof( val);
      	else if( strcmp( token, "STATISTICS" ) == 0 ) 
		{
		if( strcmp( val, "MAPPING" ) == 0 )
	  		TWstat += STAT_MAPPING;
		else if( strcmp( val, "LPS" ) == 0 )
	  		TWstat += STAT_LPS;
		else 
			{
	  		warn( "Bad initializer on line %d"
				" of .tw-config", line );
	  		continue;
			}
      		} 

/* BEGIN TWds */

        else if( strcmp( token, "END_TIME" ) == 0 )
		{
                TWds.EndTime = ENDTIME = (double) atof( val );
		}
        else if( strcmp( token, "NUM_MACHINES" ) == 0 )
                {
                TWds.Machines = atoi ( val );
		num_machines = TWds.Machines;
		if( TWds.Machines == 1 )
			RFthread=0;
		else
			RFthread=1;

                }
        else if( strcmp ( token, "MACHINE_HOSTNAME" ) == 0 )
                { /* first hostname then NUMPES must be give */
		if( strcmp ( TWds.HostName[current_machine], thishost ) != 0 )
                	++current_machine;
                /* strcpy ( TWds.HostName[current_machine], val ); */
                strcpy ( TWds.HostName[current_machine], thishost ); 
                }

        else if( strcmp ( token, "MACHINE_NUMPES" ) == 0 )
                { /* first hostname then NUMPES must be give */
		TWds.PEsPerHost[current_machine] = atoi( val );
		if( strcmp ( TWds.HostName[current_machine], thishost ) == 0 )
			{
			my_machine=current_machine;
			TWnpe = TWds.PEsPerHost[current_machine];
			}
                }
        else if( strcmp ( token, "MACHINE_SPEED" ) == 0 )
                { /* first hostname then NUMPES must be give */ 
		TWds.Speed[current_machine] = atof ( val );
                }
/* END TWds */

/* BEGIN IO3 */
      	else if( strcmp( token, "RESULTS_FILE_NAME" ) == 0 )
        	strcpy( TW_results_file_name, val );
      	else if( strcmp( token, "END_TEST" ) == 0 )
		{
		/* if a letter convert to uppercase */
      	 	to_uppercase( val );
        	if( strcmp( val, "ON" ) == 0 )
          		{
          		end_test = 1; 
			}
        	if( strcmp( val, "OFF" ) == 0 )
          		{
          		end_test = 0; 
			}
		}

      	else 
		{
		warn ( "Bad keyword token (%s) on line %d of .tw-config",
				token, line );
		continue;
      		}

      	for( ; (c != '\n') && ((int)c != EOF); c = getc(fp) );
      	if( (int)c == EOF )
		break;
      	} /* end scanning .tw-config */

   /* check that results file name is specified if results_file is set */
   if( TWdebug & DEBUG_RESULTS_FILE )
	if( strcmp(TW_results_file_name,"")== 0)
		error("\tSpecify RESULTS_FILE_NAME in .tw-config\n");
}


/*-----------------------------------------------------------------------------
**  TWFeedBack()
**
**  DESCRIPTION:
**	 print out the parameter values
**
**---------------------------------------------------------------------------*/
void 
TWFeedBack(void)
{
   char thishost[GTW_MAX_STRING];
   int i;

   if( gethostname( thishost, GTW_MAX_STRING )  == -1 )
      {
      perror ("TWFeedBack: gethostname");
      exit ( EXIT_FAIL );
      }
   printf( "\t----------------------------------------------------------- \n" );

   printf("\t  On Host(%s), Will Run GTW with %d Threads\n", thishost, TWnpe );

   printf("\t   EndTime = %lf PEEvent = %d\n", ENDTIME, PEEventMem );

   printf("\t   Allocating %4.2f M for events, %4.2f M for states \n",
          (double) PEEventMem/1000000.0,  (double)GPEStateMem/1000000.0);

   printf("\t   Mixing random by %d\n", randmixer);
   printf("\t   No. of events in a batch is %d \n", batch);

   printf("\t   Return Threshold is  %.2f \n", RetThreshold);
   printf("\t   Debugging bit vector is 0x%0x \n", TWdebug);
   printf("\t   Statistics display level is %d \n", TWstat);


   if( machine_file == 1 )
        printf("\t machine file:\n\t    %s\n", TWMachineFile );

   printf("\t configuration file:\n\t    %s\n", TWConfigFile );
   printf( "\t----------------------------------------------------------- \n" );
}


/* ---------------------------------------------------------------------------
**  TWGetCommandLine ()
**
**  DESCRIPTION:
**      gets commandline arguments.
**      dont use getopt b/c command line may for brake for application.
**---------------------------------------------------------------------------*/
int TWGetCommandLine(int argc, char **argv, int *argc_appl, char ***argv_appl )
{
char **tmp_argv=NULL;
int tmp_argc=0;
int ret_val=0;

double EndTime;
int 	not_done=1;
int 	i;

/* command line input */
int 	x_int;
double 	x_doub;
char 	x_str[10];

char *option;

print_tables[3] = 1; /* default, only bit that should be turned on */

for( i = 1; ( i < argc ) && ( not_done ); i++ )
   {
   option = argv[i];
   if( ( (int) strlen (option) > 1 ) && ((*option == '-') || (*option=='+')))
	{
      	switch ( *(option+1) )
       	    {
            case 'A':
		{
		/* we point to A */
                tmp_argv = &argv[i]; /* points to the -A */ 
                tmp_argc = argc - i;
		ret_val=1;
		not_done=0;
		break;
		}
            case 'a':
		{
            	if( *option == '+' ) 
		  	{
			DO_PROCS=FALSE;
			}
		break;
		}

            case 'b':
		{
            	if( *option == '+' ) 
		  	{
			batch_sampling = 1; 
			}
            	else if( *option == '-' ) 
		  	{
			batch_sampling = 0; 
			}
		break;
		}

            case 'B':
		{
            	if( *option == '+' ) 
		  	{
		    	i++;
			batch_processing = 1; 
          	    	if( (i == argc) || ! sscanf(argv[i],"%lf",&x_doub ) )
               			{
                        	fprintf ( stderr, "missing argument +B\n");
                        	exit(1);
                        	}
			else
             			{
      				newMIT = (double) x_doub;
				if( newMIT < 4 )
					newMIT=4.0;
				}
			}
            	else if( *option == '-' ) 
		  	{
		    	i++;
			batch_processing = 0; 
			}
		break;
		}

       	    case 'd':
       	    	{
		i++; /*  get to argument */
            	if( *option == '-' ) 
		  	{
          	  	if( (i == argc) || !sscanf(argv[i],"%d",&x_int ) )
               			{
                       	 	fprintf ( stderr, "missing argument of -d\n");
                        	exit(1);
                        	}
      			else 
             			{
      				show_gvt_dump = x_int;
      				}
			}
		break;
		}
       	    case 'D':
       	    	{
		i++; /*  get to argument */
            	if( *option == '-' ) 
		  	{
          	  	if( (i == argc) || !sscanf(argv[i],"%s", x_str ) )
               			{
                       	 	fprintf ( stderr, "missing argument of -D\n");
                        	exit(1);
                        	}
      			else 
             			{
      				D_MITfix = strdup(x_str);
		    		do_restrictions=1;
      				}
			}
		break;
		}
            case 'e':
		{
            	if( *option == '+' ) 
		  	{
			elist_test=1;
			}
		break;
		}

            case 'f':
		{
            	if( *option == '-' ) 
		  	{
			show_gvt_all=1;
			}
		break;
		}

            case 'F':
		{
            	if( *option == '-' ) 
		  	{
			CloneSim_read_files=1;
			}
		break;
		}

       	    case 'g':
       	    	{
		i++; /*  get to argument */
             	if( *option == '-' )
                   	{
          	  	if( (i == argc) || !sscanf(argv[i],"%d", &x_int ) )
               			{
                       	 	fprintf ( stderr, "missing argument of -g\n");
                        	exit(1);
                        	}
      			else 
             			{
      				show_gvt = x_int;
      				}

                    	}
                break;
		}

       	    case 'i':
       	    	{
             	if( *option == '-' )
                    {
                    i++;
                    if( (i == argc) || !sscanf(argv[i],"%d",&x_int ) )
                        {
		        fprintf ( stderr,  "sim: missing argument of -i\n");
		    	exit(1);
                        }
                   else
                       	{
		    	show_noprint=x_int;
                        }
                    }
                break;
		}

       	    case 'l':
       	    	{
             	if( *option == '-' )
                   	{
		    	limit_output=1;
                    	}
                break;
		}

       	    case 'M':
       	    	{
             	if( *option == '-' )
                    {
                    i++;
                    if( (i == argc) || !sscanf(argv[i],"%d",&x_int ) )
                        {
		        fprintf ( stderr,  "sim: missing argument of -M\n");
		    	exit(1);
                        }
                   else
                       	{
		    	clone_multiple=x_int;
                        }
                    }
                break;
		}

            case 'm':
		{
		i++;
            	if( *option == '-' ) 
		  	{
          	  	if( (i == argc) || !sscanf(argv[i],"%d",&x_int ) )
               			{
                       	 	fprintf ( stderr, "missing argument of -mem\n");
                        	exit(1);
                        	}
      			else 
             			{
   				if( num_machines == 1 )
				    {
      				    PEEventMem = x_int;
      				    fflush ( stdout );
				    }
      				}
			}
		break;
		}

       	    case 'P':
       	    	{
             	if( *option == '-' )
                    {
                    i++;
                    if( (i == argc) || !sscanf(argv[i],"%s", x_str ) )
                        {
		        fprintf ( stderr,  "sim: missing argument of -P\n");
		    	exit(1);
                        }
                   else
                       	{
			int k;
			for( k = 0; k < strlen(x_str) ; k++ )
				{
				if( x_str[k] == '1' )
					{
					print_tables[k] = 1;
					}
				else if( x_str[k] == '0' )
					{
					print_tables[k] = 0;
					}
				else
					{
					printf("** print_tables[%d] = %c\n", k, x_str[k] );
					exit(1);
					}
				}
                        }
                    }
                break;
		}

       	    case 'p':
       	    	{
             	if( *option == '-' )
                    {
                    i++;
                    if( (i == argc) || !sscanf(argv[i],"%d",&x_int ) )
                        {
		        fprintf ( stderr,  "sim: missing argument of -p\n");
		    	exit(1);
                        }
                   else
                       	{
			if( num_machines == 1 )
			    {
                            TWnpe = x_int;
                            TWds.PEs = x_int;
			    TWds.PEsPerHost[my_machine] = x_int;
			    }
                        }
                    }
                break;
		}

       	    case 'r':
       	    	{
             	if( *option == '-' )
                   	{
		    	do_restrictions=1;
                    	}
                break;
		}

            case 'R':
		{
		i++;
            	if( *option == '-' ) 
		  	{
          	  	if( (i == argc) || !sscanf(argv[i],"%d",&x_int ) )
               			{
                       	 	fprintf ( stderr, "missing argument of -R re-map messages\n");
                        	exit(1);
                        	}
      			else 
             			{
      				CloneSim_map_msgs = x_int;
      				}
			}
		break;
		}

            case 'S':
		{
		i++;
            	if( *option == '-' ) 
		  	{
          	  	if( (i == argc) || !sscanf(argv[i],"%d",&x_int ) )
               			{
                       	 	fprintf ( stderr, "missing argument of -S scale\n");
                        	exit(1);
                        	}
      			else 
             			{
      				CloneSim_scale_size = x_int;
      				}
			}
		break;
		}

       	    case 's':
       	    	{
             	if( *option == '-' )
                   	{
		    	spread_clone=1;
                    	}
                break;
		}

            case 't':
		{
            	if( *option == '-' ) 
	     	    	{
		    	i++;
          	    	if( (i == argc) || !sscanf(argv[i],"%lf",&x_doub ) )
               			{
                        	fprintf ( stderr, "missing argument -t\n");
                        	exit(1);
                        	}
      		    	else
             			{
   				if( num_machines == 1 )
				    {
      				    EndTime = (double) x_doub;
				    ENDTIME = EndTime;
      				    fflush ( stdout );
				    }
      				}
		    	}
		break;
		}
       	    case 'T':
       	    	{
             	if( *option == '+' )
                   	{
		    	do_headers=1;
                    	}
            	else if( *option == '-' ) 
	     	    	{
		    	do_headers=0;
                    	}
                break;
		}
       	    case 'x':
       	    	{
             	if( *option == '+' )
                   	{
		    	minimize_copy=1;
                    	}
            	else if( *option == '-' ) 
	     	    	{
		    	minimize_copy=0;
                    	}
                break;
		}


       	    case 'X':
       	    	{
             	if( *option == '-' )
                    {
                    i++;
                    if( (i == argc) || !sscanf(argv[i],"%d",&x_int ) )
                        {
		        fprintf ( stderr,  "sim: missing argument of -X\n");
		    	exit(1);
                        }
                   else
                       	{
		    	do_mh_memory=x_int;
                        }
                    }
                break;
		}


       	    case 'Z':
       	    	{
             	if( *option == '-' )
                    {
                    i++;
                    if( (i == argc) || !sscanf(argv[i],"%d",&x_int ) )
                        {
		        fprintf ( stderr,  "sim: missing argument of -Z\n");
		    	exit(1);
                        }
                   else
                       	{
		    	do_Z=x_int;
                        }
                    }
                break;
		}

       	    case 'z':
       	    	{
             	if( *option == '-' )
                    {
                    i++;
                    if( (i == argc) || !sscanf(argv[i],"%d",&x_int ) )
                        {
		        fprintf ( stderr,  "sim: missing argument of -z\n");
		    	exit(1);
                        }
                   else
                       	{
		    	do_z=x_int;
                        }
                    }
                break;
		}


            case '?':
        	{
          	fprintf ( stderr, "sim: unrecognized option: %s",option);
          	break;
        	}
            default:
       		{
       		fprintf ( stderr, "sim: unrecognized option %s",option);
         	break;
        	}
	    } /* end switch */
	 } /* end if */
      else
	{
	fprintf( stderr, "sim: unrecognized option: %s\n",  argv[i] );
	}

   } /* for loop */

   *argc_appl = tmp_argc;
   *argv_appl = tmp_argv;
   return ret_val;
}


/*-----------------------------------------------------------------------------
**  TWConfigure()
**
**  DESCRIPTION: Gets .tw-config file, sets kernel info structure
**
**---------------------------------------------------------------------------*/

#define TWCONFIG_FILE 0
#define MACHINE_FILE 1

int TWConfigure( int argc, char **argv, int *argc_appl, char ***argv_appl )
{
char **tmp_argv=NULL;
int tmp_argc=0;
int top_level=0;
char *config_file=NULL;

   /* initialize TWds structure */
   ds_InitTWdsInfo ( &TWds );

   /* .tw-config must be called first and must exist */
   if( build_file_name ( &config_file, TWCONFIG_FILE ) == FALSE )
        {
        perror( ".tw-config: missing or unreadable");
        exit(EXIT_FAIL);
        }
   Configure ( config_file ); /* arch. specific file must be called first */

   strcpy ( TWConfigFile, config_file );
   free ( config_file );
   config_file=NULL;

   /* .tw-config must be called first and must exist */
   if( build_file_name ( &config_file, MACHINE_FILE ) == FALSE )
        {
        /* fprintf ( stderr,  "no machine file: OK to skip\n") ; */
        /* no space for it allocated if function fails */
        }
   else
        {
        Configure( config_file );
        strcpy( TWMachineFile, config_file );
        free( config_file );
        }
   TWGetCommandLine ( argc, argv, &tmp_argc, &tmp_argv );
   ds_CheckTWdsInfo ( &TWds );

   ds_BuildMappings ( &TWds );
   ds_TotalSpeed ( &TWds );

   if( show_noprint == 0 )
        {
        printf( "\n%s\n\n", GTW_VERSION );

        TWFeedBack();
        }

   /* return to caller */
   *argc_appl = tmp_argc;
   *argv_appl = tmp_argv;

   /* if there are no return arguments returns 0 */
   return tmp_argc;
}

