#include <stdio.h>
#include <stdlib.h>
#include <unistd.h> /* gethostname */

#include "gtw_ds.h"

/* ALL globals here */
TWds_t 		TWds;

TWds_comm_t 	TWds_comm;
TWds_comm_t 	*KC; 		/* lives in pg-data in chris' */
TWds_msgDic_t 	*KMD;

TWds_msg_t  *OutofBandGVTMessage[MAX_OB_GVT_MSG] =
       {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};

/* forward reference */
void ds_InitTWdsInfo ( TWds_t * );
void ds_CheckTWdsInfo ( TWds_t * );
void ds_BuildMappings ( TWds_t * );
void ds_TotalSpeed ( TWds_t * );
int ds_GetNumPEs ( TWds_t, char * );

/*-----------------------------------------------------------------------------
**  ds_GetNumPEs ( char *name )
**
**  DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
int ds_GetNumPEs ( TWds_t TWds, char *name )
{
   int i;
   int r_val=0;

   for( i = 0; i < TWds.Machines; i++ )
        {
        if( strncmp( TWds.HostName[i], name, strlen(name)) == 0  )
                {
                r_val = TWds.PEsPerHost[i];
                break;
                }
        }
   return r_val;
}

/*-----------------------------------------------------------------------------
**  ds_TotalSpeed ( char *name )
**
**  DESCRIPTION: this is related to DLM 
**
**---------------------------------------------------------------------------*/
void ds_TotalSpeed ( TWds_t *TWds )
{
   int i;

   for( i = 0; i < TWds->Machines; i++ )
	{
	TWds->TotalSpeed += TWds->Speed[i] * ( (double) TWds->PEsPerHost[i] );
        }
}

/*-----------------------------------------------------------------------------
**  ds_InitTWdsInfo ()
**
**  DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void ds_InitTWdsInfo ( TWds_t *TWds )
{
   int i, j;
   TWds->Machines       = 0;
   TWds->PEs            = 0;
   TWds->ServerMachine  = 0;
   TWds->MyMachine      = 0;
  
   for( i = 0; i < MAXMACHINES; i++ )
	{
	TWds->HostName[i][0] = (char) 0;
	TWds->Speed[i]       = 1.0;
	
	TWds->PEsPerHost[i]  = 0;
	for( j = 0; j < D_MAXPROC; j++ )
         	{
         	TWds->LocalPEToPE[i][j] = 0;
         	TWds->PEToLocalPE[(i * D_MAXPROC) + j] = 0;
         	TWds->PEToMachine[(i * D_MAXPROC) + j] = 0;
         	}
      	}
   TWds->EndTime = 0.0;
}


/*-----------------------------------------------------------------------------
**  ds_CheckTWdsInfo ()
**
**  DESCRIPTION:  checks name(s)
**
**---------------------------------------------------------------------------*/
void ds_CheckTWdsInfo ( TWds_t *TWds )
{
   int i;
   char thishost[DS_MAXSTRING];

   if( gethostname ( thishost, DS_MAXSTRING )  == -1 )
	{
      	perror ("CheckMachineInfo: gethostname");
      	exit (DS_EXIT_FAIL);
      	}

   for( i = 0; i < TWds->Machines; i++ )
      	{
      	if( TWds->HostName[i][0]  == NULL )
         	{
         	fprintf ( stderr,"ds_Check: NULL Hostname Machine %d\n", i );
         	exit (DS_EXIT_FAIL);
         	}
	if( TWds->Speed[i] < 0.0 )
         	{
         	fprintf ( stderr,"ds_Check: Speed[%d](%f) < 0\n",
			i, TWds->Speed[i] );
         	exit (DS_EXIT_FAIL);
         	}
      	}
}


/*-----------------------------------------------------------------------------
**  ds_BuildMappings ()
**
**  DESCRIPTION:  Builds the Machine Mapping Structure contained in
**	MachineInfo structure, gives each PE a global unique number
**	that can then be referenced to find machine it is on and which
**	PE on that machine.
**
**---------------------------------------------------------------------------*/
void ds_BuildMappings ( TWds_t *TWds )
{
   char thishost[DS_MAXSTRING];
   int i, j, count;

   if( gethostname ( thishost, DS_MAXSTRING )  == -1 )
	{
      	perror ("CheckMachineInfo: gethostname");
      	exit (DS_EXIT_FAIL);
      	}

   /* get number for MyMachine */
   TWds->MyMachine = -1;
   for( i = 0; i < TWds->Machines; i++ )
	if( !strncmp( TWds->HostName[i], thishost, strlen(thishost) ) )
		TWds->MyMachine = i;

   /* if not yet defined something is wrong */
   if( TWds->MyMachine == -1 )
	{
	if( TWds->Machines > 1 ) 
		{
		printf("ds_BuildMappings:thishost(%s) not found:\n", thishost );
		for( i = 0; i < TWds->Machines; i++)
		    printf("TWds->HostName[%d] = %s\n", i, TWds->HostName[i] );
		exit(DS_EXIT_FAIL);
		}
	else
		{
		printf("ds_BuildMappings:thishost(%s) not found: ignoring\n",
			thishost ) ; 
		}
	}

   /* count total number of PEs in whole system */
   TWds->PEs = 0;
   for( i = 0; i < TWds->Machines; i++ )
	{
	TWds->PEs += TWds->PEsPerHost[i];
	}
  
   count = 0;
   for( i = 0; i < TWds->Machines; i++ )
	{
	for( j = 0; j < TWds->PEsPerHost[i]; j++ ) 
		{
		/* each PE gets a uniqueue number */
		TWds->LocalPEToPE[i][j] = count; 

		/* reference the PE # with these arrays */
		TWds->PEToLocalPE[count] = j;
		TWds->PEToMachine[count] = i;
		count++;
        	}
	}
}

