/******************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.5                             ***
***         Copyright 1997 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission  to use, copy   and   modify   this   software and its
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/


/*****************************************************************************
  	TIME WARP kernel on Distributed & Shared Memory Architectures.
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "gtw_glob.h"
#include "gtw_utility.h"

int mh_incr[D_MAXPROC] = {0};
int mh_incr2[D_MAXPROC] = {0};
extern TW_SHARED TW_ALIGN TWTime GVT;

/* API's */
void check(int *, struct LPState *, int);
void check_long(long *, struct LPState *, int);
void check_time(TWTime *, struct LPState *, int);
void check_float(float *, struct LPState *, int);
void check_double(double *, struct LPState *, int);
void check_short(short *, struct LPState *, int);
void check_char(char *, struct LPState *, int);
void Restore(struct log_entry *, struct LPState *);
void AllocateLPLog( struct LPState *, int);

/* local forward declarations */
void GarbageCollectLog(struct LPState *, struct log_entry *);
void PrintAll( struct LPState *);
void PrintArray(struct array_ptr *, struct LPState *);
void PrintLog( struct LPState *);
void RestoreArray(struct log_entry *, struct log_entry *);
void RestoreState(struct log_entry *, struct log_entry *);
void RestoreStraight(struct log_entry *, struct log_entry *);
void save_state(int *, struct LPState *);

static int firstTime;


#define alloc_1 \
 {\
 array=(struct log_entry *)calloc(sizeof(struct log_entry),TW_Incr_Array_Size);\
 /* MH_HERE */ mh_incr[MyPE] += sizeof(struct log_entry)*TW_Incr_Array_Size;\
 p = (struct array_ptr *) calloc(1,sizeof(struct array_ptr)); \
 /* MH_HERE */ mh_incr[MyPE] += sizeof(struct array_ptr);\
 p->first = array;  \
 p->last = & array[TW_Incr_Array_Size-1];\
 }

#define alloc_2 \
 {\
 array=(struct log_entry *)calloc(sizeof(struct log_entry),TW_Incr_Array_Size);\
 /* MH_HERE */ mh_incr2[MyPE] += sizeof(struct log_entry)*TW_Incr_Array_Size;\
 p = (struct array_ptr *) calloc(1,sizeof(struct array_ptr)); \
 /* MH_HERE */ mh_incr2[MyPE] += sizeof(struct array_ptr);\
 p->first = array;  \
 p->last = & array[TW_Incr_Array_Size-1];\
 }


#define alloc_NEW \
 {\
 array=(struct log_entry *) m_calloc(sizeof(struct log_entry),TW_Incr_Array_Size, MyPE);\
 p = (struct array_ptr *) m_calloc(1,sizeof(struct array_ptr), MyPE); \
 p->first = array; \
 p->last = & array[TW_Incr_Array_Size-1];\
 }

/* GLOBALS */
/* used to be 16 each */
/* int NLogArrays = 128, TW_Incr_Array_Size = 128; before fred kernel */

int NLogArrays = 10, TW_Incr_Array_Size = 1000; /* IO5 */

extern void * m_calloc (size_t nelem, size_t elsize, int MyPE);

/* ---------------------------------------------------------------------------
**  incr_main()
**
**  DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void incr_main( void )
{
   int state, i, j;
   struct LPState lpstate, * statep;
   int MyPE = 0;
   state = 0;
   statep = &lpstate;
   AllocateLPLog(statep, MyPE);

   for(i = 0; i < 25; i++)
	{
    	check(&state, statep, MyPE);
    	state ++;
  	}

   GarbageCollectLog(statep, &statep->Log->next->first[5]);

   PrintLog(statep);
   printf("New value of state is %d\n", state); 
   PrintAll(statep);

   for(j = 0 ; j < 3 ; j++)
	{
    	for(i = 0; i < 5; i++)
		{
      		check(&state, statep, MyPE);
      		state ++;
    		}

 	PrintLog(statep);
    	Restore(& statep->Log->first[TW_Incr_Array_Size-1], statep);
    	PrintLog(statep);
    	printf("New value of state is %d\n", state); 
  	}
}


/* ---------------------------------------------------------------------------
**  AllocateLPLog()
**
**  DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void AllocateLPLogOLD ( struct LPState *statep, int MyPE )
{
   int n = NLogArrays, i;
   struct log_entry * array;
   struct array_ptr *p, *prev, *last;

   alloc_1;
   prev = p;                                                            
   last = p;
   p->TS = NEGINFINITY;		/* signifying unused */

   for ( i = 0; i < n-1; i++ )
	{
    	alloc_1;
    	p->next = prev;
    	prev->prev = p;
    	prev = p;                                                            
    	p->TS = NEGINFINITY;	/* unused slot */ 
  	}

   /* link in as a circular link list */
   p->prev = last;
   last->next = p;

   statep->CurArray = statep->Log = p;
   statep->LogHead = p->first - 1;
   statep->Log = p;

   /* dummy event delimits log, in case all events are rolled back */
   statep->EList.LogPtr = statep-> LogHead; 
}


void AllocateLPLogNEW ( struct LPState *statep, int MyPE )
{
   int n = NLogArrays, i;
   struct log_entry * array;
   struct array_ptr *p, *prev, *last;

   alloc_NEW;
   prev = p;                                                            
   last = p;
   p->TS = NEGINFINITY;		/* signifying unused */

   for ( i = 0; i < n-1; i++ )
	{
    	alloc_NEW;
    	p->next = prev;
    	prev->prev = p;
    	prev = p;                                                            
    	p->TS = NEGINFINITY;	/* unused slot */ 
  	}

  if(firstTime)
    {
    firstTime = 0;
  fprintf(stderr,"AllocateLPLog: allocated %d bytes per LP (%d total)\n",   /* FPW */
     n * sizeof(struct log_entry) * TW_Incr_Array_Size,
     n * sizeof(struct log_entry) * TW_Incr_Array_Size * 1221);

    }

   /* link in as a circular link list */
   p->prev = last;
   last->next = p;

   statep->CurArray = statep->Log = p;
   statep->LogHead = p->first - 1;
   statep->Log = p;

   /* dummy event delimits log, in case all events are rolled back */
   statep->EList.LogPtr = statep-> LogHead; 
}


void AllocateLPLog( struct LPState *statep, int MyPE )
{
  extern int do_mh_memory;

  if(do_mh_memory > 0 )
	AllocateLPLogNEW( statep, MyPE );
  else
	AllocateLPLogOLD( statep, MyPE );
}


/* ---------------------------------------------------------------------------
**  check()
**
**  DESCRIPTION: 
**	check_macro in gtw_glob.h 
**
**---------------------------------------------------------------------------*/

extern TW_SHARED struct PEState GState[MAXPROC];

void check(int *ip, struct LPState *statep, int MyPE)
{
   TW_PRIVATE struct Event *CurEvent;	/* current event in process */
   CurEvent = GState[MyPE].CurEvent;

   /* FPW: modified code */
   /* End of one array */
   if( statep -> LogHead == statep->CurArray->last )
  	{
    	if( statep->CurArray->next->TS >= GVT )
    		{
      		struct log_entry *array;
      		struct array_ptr *p;

      		alloc_2;
      		p->prev = statep->CurArray;
      		p->next = statep->CurArray->next;
      		statep->CurArray->next->prev = p;
      		statep->CurArray->next = p;
      		p->TS = NEGINFINITY;
#if 0
      		fprintf(stderr, "LP %d, PE %d, time %f; alloced %d bytes"
			" for incremental state\n",
              		statep->ID, 
			MyPE,
              		statep->Last->Msg.RcvTS,
              		(sizeof(struct log_entry) * TW_Incr_Array_Size) 
		      + sizeof(struct array_ptr));
#endif
    		}

	/* TS of an array is the TS of last message to write into it */
      	/* Reuse array only if TS of last log in array < GVT */

   	statep->CurArray->TS = CurEvent->Msg.RcvTS;
      	statep->CurArray = statep->CurArray->next;
      	statep->LogHead = statep->CurArray->first;
  	}
   else 
	{
	statep->LogHead++;
	}

   /* content of what it points to is an int save it and its address which is
      long */
   statep->LogHead->value = ( int) *ip;
   statep->LogHead->addr  = ( long) ip;

   if( statep->LogHead->addr == 0 )
	{
	printf("******************** incr check: addr is 0 \n");
	error("check: incremental problem\n"); 
	}

#ifdef ConsistencyChecks
   if( ip == 0 ) 
	{
    	printf("NULL address checkpointed, should not happen");
  	}

   if(statep->Log == NULL || statep->CurArray == NULL) 
	{
    	puts("CurArray or Log is null, should not happen");
    	fflush(stdout);
  	}
#endif 
}

/* ---------------------------------------------------------------------------
**  check_long()
**
**  DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void check_long(long *ip, struct LPState *statep, int MyPE)
{
   int * iip;

   iip = ( int *)ip;
   check(iip, statep, MyPE);
   check(iip + 1, statep, MyPE);
}


/* ---------------------------------------------------------------------------
**  check_time()
**
**  DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void check_time(TWTime *ip, struct LPState *statep, int MyPE)
{
   int * iip;

   iip = ( int *)ip;
   check(iip, statep, MyPE);
   check(iip + 1, statep, MyPE);
}

/* ---------------------------------------------------------------------------
**  check_float()
**
**  DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void check_float(float *ip, struct LPState *statep, int MyPE)
{
  int * iip;

  iip = ( int *)ip;
  check(iip, statep, MyPE);
  check(iip + 1, statep, MyPE);

}

/* ---------------------------------------------------------------------------
**  check_double()
**
**  DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void check_double(double *ip, struct LPState *statep, int MyPE)
{
   int * iip;

   iip = ( int *)ip;
   check(iip, statep, MyPE);
   check(iip + 1, statep, MyPE);
}

/* ---------------------------------------------------------------------------
**  check_short()
**
**  DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void check_short(short *ip, struct LPState *statep, int MyPE)
{
   int * iip;

   iip = (int *)((unsigned long)ip & ~3);
   check(iip, statep, MyPE);
}

/* ---------------------------------------------------------------------------
**  check_char()
**
**  DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void check_char(char *ip, struct LPState *statep, int MyPE)
{
   int * iip;

   iip = (int *)((unsigned long)ip & ~3);
   check(iip, statep, MyPE);
}

/* ---------------------------------------------------------------------------
**  RestoreStraight()
**
**  DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
#define in(a,b) (b >= (a->first-1) && (b <= a->last))
/* if b is anywhere from the start of the array to */
void RestoreStraight(struct log_entry *FromLog, struct log_entry *ToLog)
{
  /* works even if no variables are written */
  while ( ToLog != FromLog )
	{
	int me=0;
	if( FromLog->addr == 0 )
		{
		error("*************** incr:RestoreStraight:addr is 0 \n");
		me=1;
		error("check: incremental RestoreStraight problem\n");
		}
    	*(( int *)FromLog->addr) = FromLog->value;
    	FromLog--;

	if( me == 1 )
		error("*** ------- *** incr:RestoreStraight:addr is 0 \n");
  	}
}

/* ---------------------------------------------------------------------------
**  RestoreArray()
**
**  DESCRIPTION:  Same as restore state, except that the last log is also 
**	 restored.
**
**---------------------------------------------------------------------------*/
void RestoreArray(struct log_entry *FromLog, struct log_entry *ToLog)
{
int me=0;
   /* works even if no variables are written */
   while ( ToLog != FromLog )
	{
	if( FromLog->addr == 0 )
		{
		me=1;
		printf("******************** incr:RestoreArray: addr is 0 \n");
		error("check: incremental RestoreArray problem\n");
		}
    	*(( int *)FromLog->addr) = FromLog->value;
    	FromLog--;
  	}
   *(( int *)FromLog->addr) = FromLog->value;

   if( me == 1 )
	error("*** ------- *** incr:RestoreArray:addr is 0 \n");
}

/* ---------------------------------------------------------------------------
**  ResetLog()
**
**  DESCRIPTION: 
**	resets array ptrs without restoring values
**
**---------------------------------------------------------------------------*/
void ResetLog(struct log_entry *ToLog, struct LPState *statep)
{
   BOOLEAN done;
   struct log_entry * FromLog;

   done = FALSE;
   FromLog = statep->LogHead;

   if( FromLog == ToLog ) 
	{
	/* nothing to restore (nothing was checkpointed) */
	return;	
	}

   /* if need be the else is executed first */
   while ( !done )
	{
    	if( in(statep->CurArray, ToLog) )
      		done = TRUE;
	else
		{ 
		if( FromLog->addr == 0 )
			error("*** ------- *** incr:Restore:addr is 0 \n");
      		statep->CurArray->TS = NEGINFINITY;
      		statep->CurArray = statep->CurArray->prev;
      		FromLog = statep->CurArray->last;
    		}
  	}

   statep->LogHead = ToLog;
   if( statep->Log == NULL || statep->CurArray == NULL )
	{
    	puts("CurArray or Log is null, should not happen");
    	fflush(stdout);
  	}
}


/* ---------------------------------------------------------------------------
**  Restore()
**
**  DESCRIPTION: 
** 	restore state from the log 
** 	first restore whole arrays 
** 	Then restore partial array 
** 	Turn back LogHead pointer 
**
**---------------------------------------------------------------------------*/
void Restore(struct log_entry *ToLog, struct LPState *statep)
{
   BOOLEAN done;
   struct log_entry * FromLog;

   done = FALSE;
   FromLog = statep->LogHead;

   if( FromLog == ToLog ) 
	{
	/* nothing to restore (nothing was checkpointed) */
	return;	
	}

   /* if need be the else is executed first */
   while ( !done )
	{
    	if( in(statep->CurArray, ToLog) )
		{ /* partial restore */
      		RestoreStraight(FromLog, ToLog);
      		done = TRUE;
    		} 
	else
		{ /* restore whole array */
#ifdef MH_PRINT
		int err=0;
		int cnt=0;

   		struct log_entry *l;
		l = FromLog;

		if( l == statep->CurArray->first )
			printf("*** ------- *** incr:From log is first\n");
	
  		for ( l = FromLog;
			l != statep->CurArray->first; l-- )
			{
			if( l-> addr == 0 ) 
				{
				err=1;
				error ( "(%d) **addr=0 cnt=%d\n", 
					statep->ID, cnt);
				}
			cnt++;
			}
		if( err == 1 )
			{
  			for ( l = FromLog;
				l != statep->CurArray->first; l-- )
				{
				printf ("** err== 1: (%d, %ld)\n",
					l->value, l->addr);
				}
    			error("HERE ** addr is 0\n" );
			}

		if( FromLog->addr == 0 )
			{
			error("*** ------- *** incr:Restore:addr is 0 \n");
			}
#endif
      		RestoreArray(FromLog, statep->CurArray->first); 
      		statep->CurArray->TS = NEGINFINITY;
      		statep->CurArray = statep->CurArray->prev;
      		FromLog = statep->CurArray->last;
    		}
  	}

   statep->LogHead = ToLog;

   if( statep->Log == NULL || statep->CurArray == NULL )
	{
    	puts("CurArray or Log is null, should not happen");
    	fflush(stdout);
  	}
}


/* ---------------------------------------------------------------------------
**  PrintAll()
**
**  DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void PrintAll( struct LPState *statep)
{
   int i;
   struct array_ptr * array;

   for ( array = statep->Log; array != NULL; array = array->next )
	{
  	for ( i = 0; i < TW_Incr_Array_Size; i++ )
    		printf("%d ", array->first[i].value);
  	printf("\n");
	}
}


/* ---------------------------------------------------------------------------
**  PrintLog()
**
**  DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void PrintLog( struct LPState *statep)
{
   struct array_ptr * array;

   for ( array = statep->Log; array != statep->CurArray; array = array->next )
  	PrintArray(array, statep);

   PrintArray(array, statep);
}


/* ---------------------------------------------------------------------------
**  PrintArray()
**
**  DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void PrintArray(struct array_ptr *array, struct LPState *statep)
{
  struct log_entry * l;
  
  for ( l = array->first; l != statep->LogHead && l != array->last; l++ )
    	printf("%d ", l->value);
  printf("%d \n", l->value);
}


/* ---------------------------------------------------------------------------
**  GarbageCollectLog()
**
**  DESCRIPTION:  garbage collects log entries in statep, starting from 
**	GarbageFrom going back to the first event 
**
**---------------------------------------------------------------------------*/
void GarbageCollectLog( struct LPState *statep, struct log_entry *GarbageFrom)
{
   struct log_entry * GarbageTo;
   struct array_ptr * OldLogPtr, * end;

   GarbageTo = statep->Log->first;
   OldLogPtr = statep->Log;

   if( statep->CurArray == statep->Log ) return;

   /* cannot garbage collect because space is collected in chunks of arrays */
   if( in(statep->Log, GarbageFrom) ) return; 

   /* find arrays to garbage collect */
   while ( !(in(statep->Log, GarbageFrom)) ) statep->Log = statep->Log->next;

   /* splice reclaimed arrays to the end of link list */
   statep->Log->prev->next = NULL;
   statep->Log->prev = NULL;
  
   end = statep->Log;
   while ( end->next != NULL ) end = end->next;

   end->next = OldLogPtr;
   OldLogPtr->prev = end;

   /* reset lpstate->dummy event's log pointer to point to the new start
    * of log, in case all events are rolled back */
   statep->EList.LogPtr = statep->Log->first - 1;

   if( statep->Log == NULL || statep->CurArray == NULL ) 
	{
    	puts("CurArray or Log is null, should not happen");
    	fflush(stdout);
  	}
}

#ifdef SimpleIncrStateSaving

#include <stdio.h>
#include "gtw_glob.h"

/* ---------------------------------------------------------------------------
**  save_state()
**
**  DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void save_state(int *p, struct LPState *lp)
{
   lp->LogHead++;
   lp->LogHead->value = (int) (* p);

   lp->LogHead->addr = (long) p;

   if( lp->LogHead->addr == 0 )
	error("******************** incr:save_state:addr is 0 \n");
}

/* ---------------------------------------------------------------------------
**  RestoreState()
**
**  DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void RestoreState(struct log_entry *FromLog, struct log_entry *ToLog)
{
   /* works even if no variables are written */
   while(ToLog != FromLog)
	{
   	if( FromLog->addr == 0 )
		error("******************** incr:RestoreState:addr is 0 \n");
    	*(( int *)FromLog->addr) = FromLog->value;
    	FromLog--;
  	}
}
#endif /* SimpleIncrStateSaving */

