/******************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.5                             ***
***         Copyright 1997 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/


/******************************************************************************
*******************************************************************************
**
**
     FILE: gtw_io.c

     DESCRIPTION:

     HISTORY:

      Initially implemented by Maria Hybinette June 1995.

     TODO:
      Clean-up

*******************************************************************************
******************************************************************************/

#include <stdio.h>

#include "gtw_glob.h"
#include "gtw_debug.h"

#include "gtw_iolist.h"
#include "gtw_utility.h"
#include "gtw_calQ.h"
#include "gtw_insim.h"
#include "gtw_incr.h"
#include "gtw_memory.h"

/* LOCAL */
static double current_time;

/* IO2 BEGIN */
/*-----------------------------------------------------------------------------
**  _TWOptimisticIOSend(int MyPE)
**
**  DESCRIPTION:  
**	Sends IO message to destination LP. This is almost exactly the same
**	As a regular message.
**
**---------------------------------------------------------------------------*/
int
_TWOptimisticIOSend (	IOFuncPtr io_func_ptr, 
			IOFuncPtr rb_func_ptr,
			int MyPE
			)
{
  struct PEState *MyState;
  struct Event   *NewBuf;

  MyState =  &GState[MyPE];
  NewBuf = MyState->NewBuf;

  NewBuf->EventType = OPTIMISTIC;
  NewBuf->EvType = 'O';        
  NewBuf->IOFunc = io_func_ptr;
  NewBuf->RBFunc = rb_func_ptr;
  NewBuf->Processed = 0;
  
  _TWSend( MyPE );
  return 1;
}
/* IO2 END */

/*-----------------------------------------------------------------------------
**  _TWBlockingIO ( MyPE )
**
**  DESCRIPTION:
**
**---------------------------------------------------------------------------*/
int
_TWBlockingIOSend(IOFuncPtr io_func_ptr, int MyPE)		
{							

  struct PEState *MyState; 
  struct Event 	 *NewBuf;

  MyState =  &GState[MyPE];
  NewBuf = MyState->NewBuf;

  NewBuf->EventType = BLOCKING;
  NewBuf->EvType = 'B';		
  NewBuf->IOFunc = io_func_ptr;                
  NewBuf->Processed = 0;		

  _TWSend( MyPE ); 
  return 1;	
}					

/*-----------------------------------------------------------------------------
**  _TWNonBlockingIOSend(int MyPE)
**
**  DESCRIPTION:
**
**---------------------------------------------------------------------------*/
int
_TWNonBlockingIOSend(IOFuncPtr io_func_ptr, int MyPE)		
{							
  struct PEState *MyState; 
  struct Event 	 *NewBuf;

  MyState =  &GState[MyPE];
  NewBuf = MyState->NewBuf;

  NewBuf->EventType = NONBLOCKING;
  NewBuf->EvType = 'N';			
  NewBuf->IOFunc = io_func_ptr;       
  NewBuf->Processed = 0;
   
  _TWSend( MyPE ); 	
  return 1;	
}					

/*-----------------------------------------------------------------------------
**  _TWBlockingProcessing()
**
**  DESCRIPTION: 
**
**		- Blocking messages should have State
**		- The first blocking message should be processed first.
**		- mutual excludsion is guaranteed becuase it is per procesor
**
**---------------------------------------------------------------------------*/
int
_TWBlockingProcessing(int MyPE, TWTime compTS )		
{							

  struct LPState *lp_statep;
  struct LPState *CurLPState;
  struct PEState *CurPEState;
  struct PEState *MyState;
  int cpe;
  struct Event *c, *NCause;
  struct Event * CurIOEvent;
  struct Event * NextIOEvent;

  CurPEState = MyState = &GState[MyPE];

  if (0)
  	printf("_TWBlockingProcessing: BEGIN: Processing IO Event\n");

  if ( CurPEState->FirstBlockedLP )
     {
	CurIOEvent = CurPEState->FirstBlockedLP->IoBList.IONext;
        GState[MyPE].CurEvent = CurIOEvent;

	while (CurIOEvent != NULL)
   		{	
   		if ( compTS >= CurIOEvent->Msg.RcvTS)
			{
		        /*  
			 *  Dequeue: returns smallest timestamped
		         *  message, returns TAIL of lp list to CalQ
			 */

/* PHERE begin */
			if ( 
				(CurIOEvent->IONext != NULL) 
				&& (CurIOEvent->IONext->EventType == BLOCKING) 
				&& ((CurIOEvent->IONext->Msg.RcvTS) == (CurIOEvent->Msg.RcvTS))
			   )
				{
  				struct Event * SameTS_IOEvent;
				struct LPState *SameLPState;

printf("IO blocking processing same TS here uid=%d next uid=%d at=%lf\n", 
	CurIOEvent->LP, CurIOEvent->IONext->LP, CurIOEvent->IONext->Msg.RcvTS );

				SameLPState    = CurPEState->FirstBlockedLP;
				SameTS_IOEvent = CurIOEvent->IONext;

				SameLPState->IoBList.IONext = SameTS_IOEvent;
				SameTS_IOEvent->IOPrev = CurIOEvent->IOPrev; 

				CurIOEvent->IONext = NULL;
				CurIOEvent->IOPrev = NULL;
				CurIOEvent->InList = INNOLIST;
				}
			else
/* PHERE end */
				{
				CurIOEvent = IOBlockedDeQueue(CurIOEvent, MyPE);
				}

			lp_statep = LPST(CurIOEvent->LP);
			CurLPState = GState[MyPE].CurState = lp_statep;
			CurIOEvent->Processed = IN_PROCESS;

   			/* REC: save state in previous event 
			** REC: or restore state */
   			if ( lp_statep->StateValid )
				{
				/* save state */
				/* allocate state vector if necessary */
        			if ( CurIOEvent->SV == NULL )
                			{
                			/* try to get more state vectors
                			** if pool empty */
                			if ( CurPEState->PState == NULL )
                        			{
                        			error ("PE %d ran out of"
						   " state vectors\n", MyPE);
                        			}
                			else
                        			{
                        			CurIOEvent->SV = 
						   CurPEState->PState;

                        			CurPEState->PState =
                                  		   CurPEState->PState->next;
                        			}

                			} /* if SV == NULL */

        			MCpSt ( lp_statep->auto_chkpt_state, 
						CurIOEvent->SV->state,
                   				lp_statep->CopySize );

       	 			} /* if valid state */
   			else
        			{

        			/* -----------------------------------------
        			** restore state if not valid
        			** ----------------------------------------*/
#ifdef Check
                                if (lp_statep->RBSV == NULL)
                                        error ("Error in restoring state\n");
#endif

        			MCpSt(lp_statep->RBSV->state,
                		   	lp_statep->auto_chkpt_state, 
					lp_statep->CopySize);

        			lp_statep->StateValid = TRUE;

        			/* if CurEvent already has a SV, save it */
        			if (CurIOEvent->SV != NULL)
                			{
                			CurIOEvent->SV->next = 
						CurPEState->PState;
                			CurPEState->PState = CurIOEvent->SV;
                			}
        			CurIOEvent->SV = lp_statep->RBSV;
        			lp_statep->RBSV = NULL;
        			}

   			if (TWdebug & DEBUG_MSG)
   				printf("RCV EVENT: LP %d, time %f \n",
                      			lp_statep->ID, CurIOEvent->Msg.RcvTS);

   			/* Save pointer into Log, before processing
   			** an event -- probably dont need since
			** BLOCKING events do not roll back
			** but it could get ABORTED 
			*/
   			if (lp_statep->IncrSave)
        			CurIOEvent->LogPtr = lp_statep->LogHead;

			/* -------------------------------------------------
			** actually execute  event
			** -----------------------------------------------*/

  /* =============== Begin I/O MChanges ==Sched()=================== REC === */
		   
#if IIODEBUG
		    	if (MyPE==0)
				{
				printf("BP %6.4f (%d, %d)\n", 
				   CurIOEvent->Msg.RcvTS, MyPE, getpid() );
				}
#endif

			CurIOEvent->IOFunc(lp_statep->state, 
					CurIOEvent->Msg.Data, MyPE);

   			/* number of Blocking events processed */
  	 		(lp_statep->NBLEvent)++;
  	 		(lp_statep->NEvent)++;

  /* ================= End I/O MChanges ============================ REC === */


			/* if aborted, cancel msgs and put back into queue */
			if (CurIOEvent->Processed == ABORT)
        			{
        			struct EvState *tmpState;

				{
				extern int show_gvt;
                		if ( show_gvt == 2 ) /* PHERE */
					{
					printf("[ABRT]: * aborted blocking event "
						"uid=%d rcv_time=%lf\n",
					CurIOEvent->LP, CurIOEvent->Msg.RcvTS );
					}
				}

        			/* number of events aborted */
        			(lp_statep->NAbort)++;
        			(lp_statep->NBLAbort)++;
        			CurIOEvent->Processed = FALSE;

        			for (c=CurIOEvent->CauseList;c!=NULL;c=NCause)
					{
                			CANCEL;
					}
        			CurIOEvent->CauseList = NULL;

        			/* set up for state restore */
        			/* if (lp_statep->RBSV != NULL) */
        			tmpState = lp_statep->RBSV;
        			lp_statep->RBSV = CurIOEvent->SV;
        			CurIOEvent->SV = tmpState;
        			lp_statep->StateValid = FALSE;

        			/* roll back log and incremental state */
        			if (lp_statep->IncrSave)
                			{
                			Restore(CurIOEvent->LogPtr, lp_statep);
                			}

        			EnQueue(CurIOEvent,MyPE);
        			CurIOEvent->in = 'c';

        			CurPEState->TWMsg = NULL;
        			CurPEState->NewBuf = NULL;

        			CurIOEvent 
					= GState[MyPE].CurEvent 
					= NULL;

        			lp_statep 
					= CurLPState 
					= GState[MyPE].CurState 
					= NULL;

        			} /* ABORTED EVENT */
   			else
				{
   				/* successfully processed event */
   				CurIOEvent->Processed = TRUE;

   				/* thread into the event list */
   				CurIOEvent->in = 'e';
   				lp_statep->Last->Next = CurIOEvent;
   				CurIOEvent->Prev = lp_statep->Last;

   				CurIOEvent->Next = NULL;
   				lp_statep->Last = CurIOEvent;
   				/* CurIOEvent->PrevTS = lp_statep->LVT; */
   				lp_statep->LVT = CurIOEvent->Msg.RcvTS;

   				/* also thread into the free list */
   				PUT_FREE_E(CurIOEvent,CurIOEvent->SrcPE);
				/* successfully processed event */
				}

			/* are there more Current Events? */
			if (CurPEState->FirstBlockedLP != NULL)
				{
        			CurIOEvent = 
				    CurPEState->FirstBlockedLP->IoBList.IONext;
        			GState[MyPE].CurEvent = CurIOEvent;
				}
			else CurIOEvent = NULL;
	
			}
   		else
      			break;
   		}
	}
   return 1;
}


/*-----------------------------------------------------------------------------
**  _TWNonBlockingProcessing()
**
**  DESCRIPTION: 
**
**
**---------------------------------------------------------------------------*/
int
_TWNonBlockingProcessing( char c_from, int MyPE, TWTime compTS )		
{							
  struct LPState *lp_statep;
  struct PEState *CurPEState;
  struct PEState *MyState;
  int cpe;
  struct Event *c, *NCause;
  struct Event * CurIOEvent;
  struct Event * NextIOEvent;
  static int ncnt=0;

  int i;

  CurPEState = MyState = &GState[MyPE];

  CurIOEvent = IO_NBListDeQueue ( MyPE, compTS ); /*  CurPEState->GVT */

  ncnt=0;
  while (CurIOEvent != NULL)
   	{	
	GState[MyPE].CurEvent = CurIOEvent;
	lp_statep = LPST(CurIOEvent->LP);

	/*
	**  A nonblocking event should not use state!
	**  lp_statep = LPST(CurIOEvent->LP);
	**  CurLPState = GState[MyPE].CurState = lp_statep;
    	*/

	CurIOEvent->Processed = IN_PROCESS;

	/* -------------------------------------------------
	** actually execute  event
	** -----------------------------------------------*/

	CurIOEvent->IOFunc(lp_statep->state, CurIOEvent->Msg.Data, MyPE);

   	/* number of NonBlocking events processed */
	(lp_statep->NEvent)++;
   	(lp_statep->NNBEvent)++;

	/* if aborted, cancel msgs and put back into queue */
	/* since NB does not schedule events this should not happen */
	if (CurIOEvent->Processed == ABORT)
        	{
		printf("**** Aborted NB event in PE %d\n", MyPE);
		fflush(stdout);
       		(lp_statep->NAbort)++;
     		(lp_statep->NNBAbort)++;
      		CurIOEvent->Processed = FALSE;

     		for (c=CurIOEvent->CauseList;c!=NULL;c=NCause)
               			CANCEL;
        	CurIOEvent->CauseList = NULL;

        	EnQueue(CurIOEvent,MyPE);

       		CurIOEvent = GState[MyPE].CurEvent = NULL;
        	CurIOEvent->in = 'c';
       		CurPEState->TWMsg = NULL;
       		CurPEState->NewBuf = NULL;
		printf("x");
        	}
   	else
		{
   		/* successfully processed event */
   		CurIOEvent->Processed = TRUE;

 		/* thread into front of freelist 
	 	* (regular are inserted @ end) */
		CurIOEvent->in='f';
		CurIOEvent->EvType='P';
   		PUT_FREE_E( CurIOEvent, CurIOEvent->SrcPE);
               	fflush(stdout);
		} /* else successful processing */
	/* get next event */
	CurIOEvent = IO_NBListDeQueue(MyPE, compTS );
   	} /* while */
   return 1;
}

/*-----------------------------------------------------------------------------
**  _TWProcessing()
**
**  DESCRIPTION:  Need to go through a list of events
**
**---------------------------------------------------------------------------*/
int
_TWProcessing(int MyPE)		
{							
  printf("Normal Processing\n");
  return 1;
}
