 /*****************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.4                             ***
***         Copyright 1995 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its 
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/

/******************************************************************************
*******************************************************************************
**
**
     FILE: io_list.c

     DESCRIPTION:

     HISTORY:

      	Initially implemented by Maria Hybinette 	June		1995.
	Added Optimistic Routines		 	September	1996
	Removed Optimistic Routines			July		1997

     TODO: 
	- simplify
	- modularize more

*******************************************************************************
******************************************************************************/
#include <stdio.h>

#include "gtw_glob.h"
#include "gtw_insim.h"
#include "gtw_calQ.h"

/* forward reference */
int IO_NBListDeThread(struct Event * DeThread_Event, int Pe);
int IO_BlockingListDeThread(struct Event * DeThread_Event, int MyPE);


/* ---------------------------------------------------------------------------
** print_iolist()
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
void print_iolist(struct Event * ev, char * s)
{
}


/* ---------------------------------------------------------------------------
** IO_EvListEnQueue()
**
** DESCRIPTION:  Enqueues Specified IO Event into Event List.
**	This is the same as enqueing of a regular event accept that the
**	nonblocking event is inserted somewhere in the list.
**
**	Need to make sure that the TS of the previous event is
**	smaller than the IO message.  (It is important that
**	the inclusion of nonblocking events are interlieved in order
**	into the event list. It will not cause a rollback however.
**
**	Need to thread into free list separately.
**
**	Should call this IO_EventListInsert() instead.
**	
** CALLEES: all from this FILE
**
**---------------------------------------------------------------------------*/
int 
IO_EvListEnQueue(struct Event * NewEvent, struct LPState *lpstatep)
{

struct Event *Current;
Current = lpstatep->Last;

/*
** Find the position in list where it should be inserted.
** Look for Event that has a smaller time stamp than the
** new event ( the NewEvent should be inserted after that event) */

while ((Current != NULL) && (Current->Msg.RcvTS > NewEvent->Msg.RcvTS))
        Current = Current->Prev;

/* Insert Event behind this event */
/* It will never become the first event */
if (Current->Next == NULL)
        {
	/* Event becomes last Event in event list */
        Current->Next = NewEvent;
        NewEvent->Prev = Current;
        NewEvent->Next = NULL;
        lpstatep->Last = NewEvent;
        }
else
        {
        NewEvent->Prev = Current;
        NewEvent->Next = Current->Next;

        Current->Next->Prev = NewEvent;
        Current->Next = NewEvent;

        }

NewEvent->in = 'e';
NewEvent->Processed = TRUE;
	
/*
** Check_ev_list("nonblock_in_eventlist");
*/
return 1;

}


/* ---------------------------------------------------------------------------
** IO_EvListDeThread()
**
** DESCRIPTION:  Dequeues the specified Event from Event List.
**	This is the same as dequeuing a regular event.
**
** 	Does not REMOVE from FreeList.
**
**	CALLES: all from this file io-list.c
**	
**---------------------------------------------------------------------------*/
int 
IO_EvListDeThread(struct Event * DeThread_Event)
{
struct Event * NextEvent;
struct Event * PrevEvent;

struct PEState * pestate_pt;
struct LPState * lpstate_pt;

int DestPE;

printf ( "DeThread event from Event List\n" );
   DestPE = LP2PE(DeThread_Event->LP);	
   pestate_pt = &GState[DestPE];
   lpstate_pt = pestate_pt->CList[LPinPE(DeThread_Event->LP)];

   NextEvent = DeThread_Event->Next;
   PrevEvent = DeThread_Event->Prev;

   if (PrevEvent == NULL) 
	/* head of list */
	/* this will not happen */
	{ 
	lpstate_pt->EList.Next = NextEvent;
   	}
   else 
	{
      	PrevEvent->Next = NextEvent;
   	}

   if (NextEvent == NULL) 
	/* DeThread_Event is last */
	/* then NextEvent == NULL */
	{ 
      	lpstate_pt->Last = PrevEvent;
   	}
   else 
	{
      	NextEvent->Prev = PrevEvent;
   	}
DeThread_Event->in = 'X';
return 1;
}



/* ---------------------------------------------------------------------------
** IO_NBListEnQueue()
**
** DESCRIPTION:  
**
**	(1).	Enqueues an event into the IO_NBList queue.  This list
**		is sorted.  
**
**	(2). 	Need to Threads the event onto it's handling LP's event list. 
**		If the IOEvent list was involved in a roll back may
**		already be in IO_NBList. Need to check this before 
**		calling that procedure. THIS IS NOT DONE IN THIS
**		PROCEDURE.
**
**	The Event may already be in the NBList (since it can reappear
**	in the calque if there is a rollback. Check pointers first.
**	Probably should check this BEFORE calling this procedure.
**
** 	Not to worry about if NB events are not threaded into event queu.
**	
**---------------------------------------------------------------------------*/
int 
IO_NBListEnQueue(struct Event * NewEvent, int pe)
{
   struct Event *Current;
   struct PEState *pestate_pt;

   pestate_pt = &GState[pe];

   /* 
   ** Find place of new IO event in non-blocking event list 
   ** Starting at event with largest time stamp */

   if (NewEvent->InList != INIOPELIST)
	{
	Current = pestate_pt->LastIoNBList;
	/* if the empty is empty found place */
	if (Current == NULL)
		{
		/* enqueue into IO NB List */
		pestate_pt->FirstIoNBList = NewEvent;
		pestate_pt->LastIoNBList = NewEvent;
		NewEvent->IOPrev = NULL;
		NewEvent->IONext = NULL;
		}
	else 
		{
		/*
		** Find the position in NB list where it should be inserted.
		** Look for Event that has a smaller time stamp than the
		** new event ( the NewEvent should be inserted after 
		** that event) */
		while ((Current != NULL) && 
				(Current->Msg.RcvTS > NewEvent->Msg.RcvTS))
			{
			Current = Current->IOPrev;
			}

		/* Insert Event behind this event */
		if (Current == NULL)
			{
			NewEvent->IONext = pestate_pt->FirstIoNBList;
			NewEvent->IOPrev = NULL;
			pestate_pt->FirstIoNBList->IOPrev = NewEvent;
			pestate_pt->FirstIoNBList = NewEvent;
			}
		else if (Current->IONext == NULL)
			{
			Current->IONext = NewEvent;
			NewEvent->IOPrev = Current;
			NewEvent->IONext = NULL;
			pestate_pt->LastIoNBList = NewEvent;
			}
		else 
			{
			NewEvent->IOPrev = Current;
			NewEvent->IONext = Current->IONext;

			Current->IONext->IOPrev = NewEvent;
			Current->IONext = NewEvent;
			}
   		}
	NewEvent->InList = INIOPELIST;
	NewEvent->in = 'n';
	}
   else
	printf ("************ already enqued into INIOPELIST\n");

 return 1;
}


/* ---------------------------------------------------------------------------
** NB_IO_Delete()
**
** DESCRIPTION:  
**
**	(1). 	DeThread a nonblocking event from IO_NBList.
**
**	(2).	Unthreads event from the lp's event list as well if InEvList.
**
**	(3).	Finally puts event into free list if DoInTW is set to FALSE.
**
**	Clears out pointer to NULL.
**	Assumes that Pe is known.
** 	 	 __       __       __ 
**	    <-- |  | <-- |  | <-- |  | 
**	  	|__| --> |__| --> |__| --->
**
**---------------------------------------------------------------------------*/
int 
NB_IO_Delete(struct Event * Delete_Event, int Pe, int DoInTW, int InEvList)
{

   struct LPState *lpstate_pt;
   struct PEState *pestate_pt;
   struct PEState *MyState;
   char fn[]= "[NB_IO_Delete]";

   pestate_pt = &GState[Pe];
   lpstate_pt = pestate_pt->CList[LPinPE(Delete_Event->LP)];

   IO_NBListDeThread(Delete_Event,  Pe);

   /* clear io pointers */
   Delete_Event->IONext = NULL;
   Delete_Event->IOPrev = NULL;

   /* Second Remove Event from evlist list and put event in free list */
   /* This does not involve rollback since non blocking */
   /* events do not effect the state */
   /* Put this in a separate procedure */

   if (InEvList)	
	{
	printf ("%s: SHOULD NOT HAPPEN*****************\n", fn );
   	IO_EvListDeThread(Delete_Event);
	}

   Delete_Event->Next = NULL;
   Delete_Event->Prev = NULL;

   Delete_Event->IOFunc = NULL;
   Delete_Event->EvType = 'D';

   /* clear in I/O list bit */
   Delete_Event->InList = INNOLIST;

   if (!(DoInTW))
	{
   	/* in free list */
   	/* put in front of free list */
	printf ("%s: SHOULD NOT HAPPEN*****************\n", fn);
   	MyState = pestate_pt;
   	PUT_FREE_F(Delete_Event, Delete_Event->SrcPE);
	printf("%s: WOOPS\n", fn);
   	Delete_Event->in = 'f';
   	Delete_Event->EventType = UNDEFINED;
   	Delete_Event->EvType = 'D';
	}
  return 1;
}


/* ---------------------------------------------------------------------------
** IO_NBListDeThread()
**
** DESCRIPTION:  DeThread a nonblocking event from IO_NBList.
**	Need to unthread this event from the lp's event list as well
**	This assumes that we now NB. It does not dequeu it from the
**	event list.
**
**  	 __       __       __ 
**  <-- |  | <-- |  | <-- |  | 
**  	|__| --> |__| --> |__| --->
**
**---------------------------------------------------------------------------*/
int IO_NBListDeThread(struct Event * DeThread_Event, int Pe)
{
   struct Event *NextEvent;
   struct Event *PrevEvent;
   struct PEState *pestate_pt;

   pestate_pt = &GState[Pe];

   /* First remove the Event from the IO_NBList */
   NextEvent = DeThread_Event->IONext;
   PrevEvent = DeThread_Event->IOPrev;

   if (PrevEvent == NULL) 
	/* head of list */
	{ 
	pestate_pt->FirstIoNBList = NextEvent;
   	}
   else 
	{
      	PrevEvent->IONext = NextEvent;
   	}

   if (NextEvent == NULL) 
	/* tail of list */
	{ 
      	pestate_pt->LastIoNBList = PrevEvent;
   	}
   else 
	{
      	NextEvent->IOPrev = PrevEvent;
   	}
  return 1;
}



/* ---------------------------------------------------------------------------
** IO_NBListDeQueue()
**
** DESCRIPTION: This seams easy just take first event. Smallest Timestamp.
**	Dequeues events if they are larger than the given TS.
**	agressive dequeuing requires a ts of MAXFLOAT.
**
**
**---------------------------------------------------------------------------*/
struct Event * 
IO_NBListDeQueue(int pe, TWTime comp_TS)
{
struct PEState *pestate_pt;
struct Event *DequeueEvent;
struct Event *NextIOEvent;

pestate_pt = &GState[pe];

if ( (pestate_pt->FirstIoNBList) == NULL)
	return NULL;
else 
	{
	DequeueEvent = pestate_pt->FirstIoNBList;

#ifdef IO_DEBUG
	printf("compTS = %f DequeueEvent->Msg.RcvTS = %f\n", 
		comp_TS, DequeueEvent->Msg.RcvTS);
#endif
	if (comp_TS >= DequeueEvent->Msg.RcvTS)
		{
		NextIOEvent =  DequeueEvent->IONext;

		pestate_pt->FirstIoNBList = NextIOEvent;

		if (NextIOEvent == NULL) 
			{
			pestate_pt->LastIoNBList = NULL;
			}
		else 	{
			NextIOEvent->IOPrev = NULL;
			}

		DequeueEvent->IONext = NULL;
		DequeueEvent->IOPrev = NULL;

		DequeueEvent->InList = INNOLIST;

		return DequeueEvent;
		}
	else 
		return NULL;
	}
}

/* ========================= BLOCKING STUFF BELOW ========================= */


/* ---------------------------------------------------------------------------
** IO_LpListAppend()
**
** DESCRIPTION: Add an event to the LP end of the LP list when
**	The processor is BLOCKED.  Blocking Events and Regular event
**	are held here while LP is blocked. 
**	statep needs to be defined before calling this procedure.
**	lpstate_pt->Last should remain where it is.
**
** TODO
**	Make a special case of blocking events to be inserted for
**	blocked LPs -- if its TS is smaller than the blocking Event.
**	All events in the attached list are guarnateed to be larger.
**
**---------------------------------------------------------------------------*/
int 
IO_LpListAppend(struct Event * CurEvent, struct LPState *lpstate_pt )
{

lpstate_pt->IoBList.LastIoBList->IONext = CurEvent;
CurEvent->IOPrev = lpstate_pt->IoBList.LastIoBList;

CurEvent->IONext = NULL;

lpstate_pt->IoBList.LastIoBList = CurEvent;

CurEvent->in = 'b';
CurEvent->InList = INBLOCKINGEVENTLIST;

CurEvent->Next = NULL;
CurEvent->Prev = NULL;

/* lpstate_pt->Last = CurEvent; 		*/
/* should this be the case?  Not in this case 	*/
/* lpstate_pt->LVT = CurEvent->Msg.RcvTS; 	*/

/*
** Check_ev_list("blocking_in_eventlist");
*/

  return 1;
}


/* ---------------------------------------------------------------------------
** IO_LpListDetach()
**
** DESCRIPTION: reenter a list of events into cal queue.
**
**---------------------------------------------------------------------------*/
int 
IO_LpListDetach(struct Event * first_event, int MyPE)
{

/* first_event->lpstate_pt->IoBList.IONext */

struct Event *Current;
struct Event *Next;

Current = first_event;
while (Current != NULL) 
	{
	Next = Current->IONext;
	Current->IONext = NULL;
	Current->IOPrev = NULL;
	EnQueue(Current, MyPE);
	Current->in = 'c';
	Current->InList = INNOLIST;
	Current = Next;
	}

  return 1;
}


/* ---------------------------------------------------------------------------
** IO_BlockedEnQueue()
**
** DESCRIPTION: Insert an event into IO_Blocked Queue this list is
**	Sorted.  If it is the event to be inserted has a
**	smaller TS -- insert in front (not events appended all
**	have larger TS -- change flag of previous blocking event
**	to not be INBLOCKINGLIST.  
**
** TODO
**	If the event has a larger TS then append to event list.
**	This should not happen here because this assumes the LP is
**	not blocked already. blocking events may come through this route.
**	event if processor is already blocked.
**
**	- THIS routine needs to be simplified!
**	- bad practice, to have a return in middle
**---------------------------------------------------------------------------*/
int 
IOBlockedEnQueue(struct Event *BlockingEvent, struct LPState * lpstate_pt)
{
   struct Event *Current=NULL;
   struct LPState *CurrentLPState=NULL;
   struct PEState *pestate_pt=NULL;
   int MyPE;

   MyPE = LP2PE(lpstate_pt->ID);	

/* if NOT blocking already */
if ( lpstate_pt->IoBList.IONext == NULL ) 
   {
   /* 
   ** Find place of new IO event in blocking event list 
   ** Starting at event with largest time stamp */

        pestate_pt = &GState[MyPE];
	CurrentLPState = pestate_pt->LastBlockedLP;

	/* if no blocking events for this PE then found place */
	if (CurrentLPState == NULL)
		/* if a bogus event this will go away */
		{
		/* enqueue into IOBlockingList */
		pestate_pt->FirstBlockedLP = lpstate_pt;
		pestate_pt->LastBlockedLP = lpstate_pt;

		lpstate_pt->IoBList.IONext = BlockingEvent;
		lpstate_pt->IoBList.IOPrev = NULL;

		lpstate_pt->IoBList.Up = NULL;
		lpstate_pt->IoBList.Down = NULL;
		lpstate_pt->IoBList.LastIoBList = BlockingEvent;


                /* points to last regular event in event list */
		BlockingEvent->IOPrev = lpstate_pt->Last;
		BlockingEvent->IONext = NULL;
	        /* BlockingEvent->IoLP = lpstate_pt; */
		BlockingEvent->InList = INBLOCKINGLIST;
		BlockingEvent->in = 'b';

		return 1;
		}
	else
		{
		/* We know there is one event at CurrentLPState */
		/* Since Last pointed here */
		Current = CurrentLPState->IoBList.IONext;
		}

	/*
	** Find the position in list where it should be inserted.
	** Look for Event that has a smaller time stamp than the
	** new event ( the NewEvent should be inserted after that event) */

	/* We can assume that this event is not already in IoBlocking list */
	while ( (Current != NULL) && 
	       	(Current->Msg.RcvTS > BlockingEvent->Msg.RcvTS)) 
		{
		CurrentLPState = CurrentLPState->IoBList.Up;
		if (CurrentLPState != NULL)
			{
			Current = CurrentLPState->IoBList.IONext;
			}
		else
			{
			/* it is the first blocked event now */
			Current = NULL;
			}
		}

	/* Insert Event after current if there is one this event */
	if (Current == NULL)
		{
		lpstate_pt->IoBList.Down = pestate_pt->FirstBlockedLP;
		lpstate_pt->IoBList.Up = NULL;

		pestate_pt->FirstBlockedLP->IoBList.Up = lpstate_pt;
		pestate_pt->FirstBlockedLP = lpstate_pt;
		}
	else if (CurrentLPState->IoBList.Down == NULL)
		{
		CurrentLPState->IoBList.Down = lpstate_pt;
                lpstate_pt->IoBList.Up = CurrentLPState;
                lpstate_pt->IoBList.Down = NULL;
		pestate_pt->LastBlockedLP = lpstate_pt;
		}
	else 
		{
		lpstate_pt->IoBList.Up = CurrentLPState;
		lpstate_pt->IoBList.Down = CurrentLPState->IoBList.Down;

		CurrentLPState->IoBList.Down->IoBList.Up = lpstate_pt;
		CurrentLPState->IoBList.Down = lpstate_pt;
		}

	BlockingEvent->IOPrev = lpstate_pt->Last;
	BlockingEvent->IONext = NULL;
	/* BlockingEvent->IoLP = lpstate_pt; */
	BlockingEvent->InList = INBLOCKINGLIST;
	BlockingEvent->in = 'b';

	/* since it is the first to be inserted it is also the last event */
	lpstate_pt->IoBList.IONext = BlockingEvent;
	lpstate_pt->IoBList.IOPrev = NULL;
	lpstate_pt->IoBList.LastIoBList = BlockingEvent;

        return 1;
  }
  else
	{
	/* it is blocked already */
        /* append to the list this could happen */

	if ( lpstate_pt->IoBList.IONext->Msg.RcvTS == BlockingEvent->Msg.RcvTS)
		{  /* find place in fron in LpListAppend */
		/* IO_LpListAppend(BlockingEvent, lpstate_pt); */
		struct Event *NotSameTSEvent;

		NotSameTSEvent = lpstate_pt->IoBList.IONext;
		while ( (NotSameTSEvent != NULL) 
		  && (BlockingEvent->Msg.RcvTS == NotSameTSEvent->Msg.RcvTS)
		  )
			{
			NotSameTSEvent = NotSameTSEvent->IONext;
			}

		if ( NotSameTSEvent == NULL )
			{
			IO_LpListAppend(BlockingEvent, lpstate_pt); 
			}
		else
			{
			BlockingEvent->IOPrev = NotSameTSEvent->IOPrev;
			BlockingEvent->IONext = NotSameTSEvent;

			BlockingEvent->IOPrev->IONext = BlockingEvent;
			NotSameTSEvent->IOPrev = BlockingEvent; 
			}
		}
	else if (lpstate_pt->IoBList.IONext->Msg.RcvTS 
						<= BlockingEvent->Msg.RcvTS) 
		{
		IO_LpListAppend(BlockingEvent, lpstate_pt);
		}
	else    
		{
		/* TODO very inefficient  */
		struct Event *OldBlockingEvent;

		OldBlockingEvent = lpstate_pt->IoBList.IONext; 

fprintf( stdout, "[IOBlockedEnQueue]: BL_RcvTS=%lf O.BL_RcvTS=%lf\n", BlockingEvent->Msg.RcvTS, OldBlockingEvent->Msg.RcvTS );
fprintf( stderr, "[IOBlockedEnQueue]: BL_RcvTS=%lf O.BL_RcvTS=%lf\n", BlockingEvent->Msg.RcvTS, OldBlockingEvent->Msg.RcvTS );
fflush(stderr);
fflush(stdout);

 		IO_BlockingListDeThread(lpstate_pt->IoBList.IONext, MyPE);

fprintf( stdout, "[IOBlockedEnQueue]: here 2 \n");
fprintf( stderr, "[IOBlockedEnQueue]: here 2 \n");

		IOBlockedEnQueue(BlockingEvent, lpstate_pt);

fprintf( stdout, "[IOBlockedEnQueue]: here 3 \n");
fprintf( stderr, "[IOBlockedEnQueue]: here 3 \n");

		IOBlockedEnQueue(OldBlockingEvent, lpstate_pt);

fprintf( stdout, "[IOBlockedEnQueue]: here 4 \n");
fprintf( stderr, "[IOBlockedEnQueue]: here 4 \n");
		fflush(stdout);
       		return 1;
		}
	}

  return 1;
}


/* ---------------------------------------------------------------------------
** IO_BlockedMinim_Q()
**
** DESCRIPTION: 
** 	Finds the smallest time stamped blocking message 
**	Returns the timestamp.	
**---------------------------------------------------------------------------*/
double 
IOBlockedMinim_Q(int Pe)
{

struct PEState *pestate_pt;
struct LPState *lpstate_pt;

struct Event *MinimEvent;
struct Event *UpIOEvent;
struct LPState *DownIoB;

struct Event *TailLpList;

pestate_pt = &GState[Pe];
lpstate_pt = pestate_pt->FirstBlockedLP;

if (lpstate_pt != NULL)
	{
	MinimEvent = lpstate_pt->IoBList.IONext;

	if ( MinimEvent == NULL )
		return (MAXFLOAT);
	else 
		{
		return MinimEvent->Msg.RcvTS;
		}
	}
else
	return (MAXFLOAT);
}

/* ---------------------------------------------------------------------------
** IOBlockedDeQueue()
**
** DESCRIPTION: 
**	(1) Dequeues the smallest time stamped blocking message 
**	    If it is dequed needs to unblock the LP and return the
**	    tail of the lp list. Recall that blocking is implicit when
**	    there is an event in IoBList.IONext Recall that blocked
**	    events schedules events.
**	
**	(2) Enqueus tail of list into CalQ
**
**    pe is the same as MyPE
**
**---------------------------------------------------------------------------*/
struct Event * 
IOBlockedDeQueue(struct Event * ev, int MyPE)
{
/* detach IO event from List then detach tail of list */
/* TODO should NOT send in an event since this is a dequeue operation  */
/* Maybe send in a comp_TS so that we only dequeue if comp_TS > TS of
*  dequeued message */

struct PEState *pestate_pt;
struct LPState *lpstate_pt;

struct Event *DequeueEvent;
struct Event *UpIOEvent;
struct LPState *DownIoB;

struct Event *TailLpList;

pestate_pt = &GState[MyPE];
lpstate_pt = pestate_pt->FirstBlockedLP;

if (lpstate_pt == NULL)
	return NULL;
else
	{
	DequeueEvent = lpstate_pt->IoBList.IONext;

	if ( DequeueEvent == NULL)
		return NULL;
	else 
		{
		TailLpList = DequeueEvent->IONext;

		DownIoB = lpstate_pt->IoBList.Down;

		pestate_pt->FirstBlockedLP = DownIoB;

		if (DownIoB == NULL) 
			{
			pestate_pt->LastBlockedLP = NULL;
			}
		else 	{
			DownIoB->IoBList.Up = NULL;
			}

		DequeueEvent->IONext = NULL;
		DequeueEvent->IOPrev = NULL;

		DequeueEvent->InList = INNOLIST;

		/* detach tail */
		if (TailLpList != NULL)
			{
			IO_LpListDetach(TailLpList, MyPE);
			}

		lpstate_pt->IoBList.IONext = NULL;
		lpstate_pt->IoBList.IOPrev = NULL;
		lpstate_pt->IoBList.LastIoBList = NULL;

		if ( lpstate_pt->IoBList.Down != NULL )
		     lpstate_pt->IoBList.Down->IoBList.Up = NULL;  /* PHERE */

		return DequeueEvent;
		}
	}

}


/* ---------------------------------------------------------------------------
** Blocking_EvListDeThread()
**
** DESCRIPTION:  Denqueus blocking IO Event or regular from Event List
**	That is blocked this is NOT the blocking event.
**	Need to check if it is in INBLOCKINGEVENTLIST and NOT in
**	INBLOCKINGLIST
**	
**---------------------------------------------------------------------------*/
int 
Blocking_EvListDeThread(struct Event * DeThread_Event)
{
struct Event * NextEvent;
struct Event * PrevEvent;
struct PEState * pestate_pt;
struct LPState * lpstate_pt;

pestate_pt = &GState[LP2PE(DeThread_Event->LP)];
lpstate_pt = pestate_pt->CList[LPinPE(DeThread_Event->LP)];

NextEvent = DeThread_Event->IONext;
PrevEvent = DeThread_Event->IOPrev;

   PrevEvent->IONext = NextEvent;

   if (NextEvent == NULL) 
	/* DeThread_Event is last */
	/* then NextEvent == NULL */
	{ 
      	lpstate_pt->IoBList.LastIoBList = PrevEvent;
   	}
   else 
	{
      	NextEvent->IOPrev = PrevEvent;
   	}

DeThread_Event->InList = INNOLIST;
 
return 1;

}



/* ---------------------------------------------------------------------------
** IO_BlockingListDeThread()
**
** DESCRIPTION:  DeThread a blocking event from  Blocking list.
**
**       __       __       __
**  <-- |  | <-- |  | <-- |  |
**      |__| --> |__| --> |__| --->
**
**---------------------------------------------------------------------------*/
int 
IO_BlockingListDeThread(struct Event * DeThread_Event, int MyPE)
{
   struct Event *NextEvent;
   struct Event *PrevEvent;

   struct LPState *DownIoB;
   struct LPState *UpIoB;
   struct PEState *pestate_pt;
   struct LPState *lpstate_pt;

   struct Event *TailLpList;
   
   pestate_pt = &GState[LP2PE(DeThread_Event->LP)];
   lpstate_pt = pestate_pt->CList[LPinPE(DeThread_Event->LP)];

/*
   printf("[IOBDethread]: %d, %d\n", DeThread_Event->LP, 
		LPinPE(DeThread_Event->LP)
	);
*/

   /* First remove the Event from the BlockingList */
   DownIoB = lpstate_pt->IoBList.Down;
   UpIoB = lpstate_pt->IoBList.Up;

   TailLpList = DeThread_Event->IONext;

   if (UpIoB == NULL)
        /* dethread event head of list */
        {
        pestate_pt->FirstBlockedLP = DownIoB;
        }
   else
        {
        UpIoB->IoBList.Down = DownIoB;
        }

   if (DownIoB == NULL)
        /* tail of list */
        {
        pestate_pt->LastBlockedLP = UpIoB;
        }
   else
        {
        DownIoB->IoBList.Up = UpIoB;
        }

   /* detach tail */
   if (TailLpList != NULL)
	{
	IO_LpListDetach(TailLpList, MyPE);
	}

   /* Clear IoBList */
   lpstate_pt->IoBList.LastIoBList = NULL;
   lpstate_pt->IoBList.IONext = NULL;
   lpstate_pt->IoBList.IOPrev = NULL;

   /* Clear Event itself */
   DeThread_Event->IONext = NULL;
   DeThread_Event->IOPrev = NULL;

   DeThread_Event->InList = INNOLIST;

  return 1;
}


/* ---------------------------------------------------------------------------
** IOBlockedDelete()
**
** DESCRIPTION: Removes a blocked event (blocked events schedule 
**	events, but if blocked event are not processed then it
**	cannot have schedule events).
**
**	If event is in front of list then it is event that is blocking the 
**	LP if it is in INBLOCKINGLIST.
**	
**	Event deleted is the blocking event, otherwise need to
**	delete from event list. Does NOT put in Free LIST
**
**---------------------------------------------------------------------------*/
int 
IOBlockedDelete(struct Event * Delete_Event, int MyPE)
{
struct PEState * pestate_pt;
struct LPState * lpstate_pt;

/* if blocking dethread  from blocking list */
if (Delete_Event->InList == INBLOCKINGLIST)
	{
 	IO_BlockingListDeThread(Delete_Event, MyPE);
	}
else
	{
	Blocking_EvListDeThread(Delete_Event); 
	}
  return 1;
}
