/******************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.7                             ***
***         Copyright 1997 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/

/******************************************************************************
*      Time Warp memory allocate/deallocate routines.
*      Send() for allocation, and FossilCollect() for deallocation
******************************************************************************/
#include <stdio.h>

#include "gtw_ds.h"
#include "gtw_ds_glob.h"

#include "gtw_glob.h"
#include "gtw_debug.h"
#include "gtw_io.h"
#include "gtw_iolist.h"
#include "gtw_insim.h"
#include "gtw_utility.h"
#include "gtw_calQ.h"
#include "op_gtw.h" /* IO4 */

/* forward declarations */
void _TWSend( int );
void _TWGetMsg ( TWTime, int, int , int );
void FossilCollect(int );
struct Event * get_free_partitioned(int, int );
struct Event * get_free_external ( int, int );
extern void cntFree( int pe, int MyPE, char *fn ); /* IO5 */
extern int	show_gvt; /* IO5 */



/*-----------------------------------------------------------------------------
**  _TWSend(int MyPE)
**
**  DESCRIPTION:  Send message into destination LP
**
**  NOTES: in gtw.h: #define TWSend() _TWSend(MyPE);
**
**---------------------------------------------------------------------------*/
void _TWSend(int MyPE)
{
   int DestPE;

   struct Event   *CurEvent;    /* current event that */
   struct LPState *CurState;    /* current LP state */
   struct PEState *MyState;     /* state for this PE */
   struct Event   *NewBuf;

   MyState =  &GState[MyPE];
   CurEvent = MyState->CurEvent;
   CurState = MyState->CurState;

   NewBuf   = MyState->NewBuf;

   /* don't send abort buffer! */
   if ( NewBuf == GState[MyPE].AbortBuf )
	{
	MyState->TWMsg = NULL;
     	MyState->NewBuf = NULL;
     	return;
   	}

   DestPE = LP2PE(NewBuf->LP);

   if ( TWdebug & DEBUG_MSG ) 
	{
	TW_LOCK(dumplck);
      	printf("Send from LP %d to %d, time %f \n", CurState->ID,
	NewBuf->LP, NewBuf->Msg.RcvTS);
      	TW_UNLOCK(dumplck);
   	}

   ck_NewBuf

   /* update causality pointers; assume message already filled in */
   NewBuf->NextCause = CurEvent->CauseList;
   NewBuf->CauseList = NULL;
   CurEvent->CauseList = NewBuf;

   /* fast action for local sends that are not stragglers  */
   /* guaranteed not to rollback itself or its parent */
   if ( (DestPE == TWds.LocalPEToPE[ TWds.MyMachine ][MyPE])
	&& (LPST(NewBuf->LP)->LVT <= NewBuf->Msg.RcvTS)
	&& (OP_DoEnQueueOnSend( NewBuf->LP, NewBuf->Msg.RcvTS)) /*IO4*/
      ) 
	{
        EnQueue(NewBuf, MyPE);
	NewBuf->in = 'c';	/* cal q */
   	} 

   /* remote (on local machine) all variations or straggler */
   else if ( TWds.MyMachine == TWds.PEToMachine[DestPE] )
	{ 
	/* remote send */
      	if ( DestPE != TWds.LocalPEToPE[ TWds.MyMachine ][MyPE] )  
		{ /* detach state */
         	if (NewBuf->SV != NULL) 
			{
            		if (MyState->PState == NULL) 
				{
               			MyState->PState = NewBuf->SV;
	       			MyState->PState->next = NULL;
            			} 
			else 
				{
	       			NewBuf->SV->next = MyState->PState;
	       			MyState->PState = NewBuf->SV;
            			}
            		NewBuf->SV = NULL;
	 		}

         	MyState->NToReturn[ TWds.PEToLocalPE[DestPE] ]--;
       		}

      	/* put the event in the remote MsgQ */
      	NewBuf->Prev = NULL;

      	TW_LOCK(GState[ TWds.PEToLocalPE[DestPE] ].Lck_MsgQ);
      	    NewBuf->in = 'm';
      	    NewBuf->Next = GState[ TWds.PEToLocalPE[DestPE] ].MsgQ;
      	    if (NewBuf->Next != NULL)
	 	NewBuf->Next->Prev = NewBuf;
      	    GState[ TWds.PEToLocalPE[DestPE] ].MsgQ = NewBuf;
      	TW_UNLOCK(GState[ TWds.PEToLocalPE[DestPE] ].Lck_MsgQ);

 	if ( NewBuf->EventType != NONBLOCKING  )
		/* update TransMsgTS if in GVT computation */
      		if (GVTflag > 0) 
			{
	 		(MyState->TransMsgTS) 
			    = min((MyState->TransMsgTS),(NewBuf->Msg.RcvTS));
      			}

   	} /* remote (local machine) send or straggler */
   else
	{ /* between different machine, send to reflector thread */

	/* detatch state */
         if ( NewBuf->SV != NULL ) 
		{
            	if (MyState->PState == NULL) 
			{
               		MyState->PState = NewBuf->SV;
	       		MyState->PState->next = NULL;
            		} 
		else 
			{
	       		NewBuf->SV->next = MyState->PState;
	       		MyState->PState = NewBuf->SV;
            		}
            	NewBuf->SV = NULL;
	 	}
	       MyState->NToReturn[ TWnpe ]--;

        /* put the event in the ReflectorThread MsgQ */
        NewBuf->Prev = NULL;
        TW_LOCK(GState[TWnpe].Lck_MsgQ);
            NewBuf->in = 'm';
            NewBuf->Next = GState[TWnpe].MsgQ;
            if (NewBuf->Next != NULL)
                NewBuf->Next->Prev = NewBuf;
            GState[TWnpe].MsgQ = NewBuf;
        TW_UNLOCK(GState[TWnpe].Lck_MsgQ);

	/* nonblocking/external events not sent b/w machines */
        if ( (GVTflag > 0) && (NewBuf->EventType != NONBLOCKING) )
            MyState->TransMsgTS=min(MyState->TransMsgTS,NewBuf->Msg.RcvTS);
	} /* send between different machines */

   /* reset the pointers */
   MyState->TWMsg = NULL;
   MyState->NewBuf = NULL;

   return;
} /* end TWSend() */

/*-----------------------------------------------------------------------------
**  _TWGetMsg(ts, dst, size, MyPE)
**
**  DESCRIPTION:  Get a new message buffer. The buffer is NewBuf. The data
**      part is in TWMsg, which is seen by the application.
**---------------------------------------------------------------------------*/
void _TWGetMsg 
		( 
		TWTime ts, 	/* increment from current event 	*/
		int dst,	/* receiver LP 				*/ 
		int size,	/* of message 				*/ 
		int MyPE	/* if myPE == TWnpe => external PE 	*/
		)
{
   struct Event   *CurEvent;    /* current event that */
   struct LPState *CurState;    /* current LP state */
   struct PEState *MyState;     /* state for this PE */
   struct Event   *NewBuf;
   int		  glob_dst;

   MyState  = &GState[MyPE];
   CurEvent = MyState->CurEvent;
   CurState = MyState->CurState;
   NewBuf   = MyState->NewBuf;

   ck_PriorNewBuf

   if ( CurEvent->Msg.RcvTS + ts > ENDTIME ) 
	{
	MyState->NewBuf = GState[MyPE].AbortBuf;
   	} 
   else 
	{
      	/* get a buffer at the beginning of the local pool */
     	/* sets NewBuf */

	/* of on local machine */
	if ( TWds.PEToMachine[LP2PE(dst)] == TWds.MyMachine )
       		{
     		MyState->NewBuf 
		 = get_free_partitioned( TWds.PEToLocalPE[ LP2PE(dst) ], MyPE );
		}
	else
		{
     		MyState->NewBuf = get_free_partitioned( TWnpe, MyPE );
		}
	} /* get buffer and before end time */

   ck_inPDS

   /* have a buffer (could be dummy buffer), fill it in */

#if 0
   MyState->NewBuf->CloneID = CurState->CloneID;  	/* IO4 GetMsg */
#endif
   MyState->NewBuf->BirthType = -1; 			/* IO4 GetMsg */

   MyState->NewBuf->Processed = FALSE;
   MyState->NewBuf->InPDS = FALSE;
   MyState->NewBuf->LP = dst;
   MyState->NewBuf->SrcLP = CurState->ID;
   MyState->NewBuf->SrcPE = MyPE;

   MyState->NewBuf->Msg.SendTS = CurEvent->Msg.RcvTS;
   MyState->NewBuf->Msg.RcvTS = CurEvent->Msg.RcvTS + ts;

   ck1_isNeg("TWGetMsg",MyState->NewBuf->Msg.RcvTS)

   MyState->NewBuf->Next = NULL;
   MyState->NewBuf->Prev = NULL;

   MyState->NewBuf->NextCause = NULL;
   MyState->NewBuf->CauseList = NULL;

   MyState->NewBuf->FNext = NULL;
   MyState->NewBuf->FPrev = NULL;

   MyState->NewBuf->InList = INNOLIST;             
   MyState->NewBuf->IONext = NULL;                
   MyState->NewBuf->IOPrev = NULL;               

   /* default message is a regular message */
   MyState->NewBuf->EventType = REGULAR;        
   MyState->NewBuf->EvType = 'R';              
   MyState->NewBuf->IOFunc = NULL;            
   MyState->NewBuf->RBFunc = NULL;           

   MyState->TWMsg = MyState->NewBuf->Msg.Data;	/* goes to application */
}


/*----------------------------------------------------------------------------
**  FossilCollect(int MyPE)
**
**  DESCRIPTION:   Fossil collection.  Assumes that GVT is already computed.
**      redistributes buffers amongst free pools after fossil collection.
**
**---------------------------------------------------------------------------*/
void 
FossilCollect( int MyPE )
{
   struct Event   *CurEvent;      /* current event that */
   struct LPState *CurState;    /* current LP state */
   struct PEState *MyState;     /* state for this PE */
   struct Event *NewBuf;
   int i,j;
   struct Event  *first, *last, *p;
   extern TW_PRIVATE int Qsize;

   MyState = &GState[MyPE];
   CurEvent = MyState->CurEvent;
   CurState = MyState->CurState;
   NewBuf   = MyState->NewBuf;

   MyState->NumFC++;

   ck3_repTimer


   _TWBlockingProcessing ( MyPE, MyState->GVT ) ;
   _TWNonBlockingProcessing ( 'm', MyPE, MyState->GVT );

#if 0
   if ( show_gvt_dump > 0 )
	{
	ipdes_stat_gvt ( MyPE ); /*IO4 */
	}
#endif



   if( show_gvt > 0 )
        {
        int cnt=0;
        for (cnt=0; cnt<TWnpe; cnt++)
                {
                cntFree( cnt, MyPE, "FossilCollect begin" );
                }
        }

   for ( i = 0 ; i < TWnpe ; i++ ) 
	{
      	if (i==MyPE) continue;

      	if (MyState->RetTrig[i] == 0) continue;

      	if (MyState->NToReturn[i] < MyState->RetTrig[i]) continue;

      	p = MyState->PFree[i];
      	if (p == NULL) 
		continue;

      	if (p->Msg.RcvTS >= MyState->GVT) 
		continue;

      	first = p;		/* at least one item to return */

#if 0
printf( "* MyPE=%d NToReturn[%d]=%d\n", MyPE, i, MyState->NToReturn[i] );
#endif
      	for(j=0; j < (MyState->NToReturn[i]); j++) 
		{
	 	if (p->Msg.RcvTS < MyState->GVT) 
			{
	   		/* if p is in the event list, fix event list */
	   		if(p->in =='e')
				{
				ck1_inEvQ("FC",p) 
				ck1_PrevNULL("FC",p)

	     			if ( p->Next ) 
					p->Next->Prev = p->Prev;
	     			else
					{ /* last element on list */
	       				struct LPState * statep;
	       				statep = MyState->CList[LPinPE(p->LP)];
	       				statep->Last = p->Prev;
	     				}
				/* p->Prev is always non null */
	    			p->Prev->Next = p->Next; 
	   			}
			p->Next = NULL;
	   		p->Prev = NULL;
	   		p->in = 'f';		/* p is now in free list */
                        p->EvType = 'M';	/* IO1 */

	   		/* detach state */
	   		if (p->SV != NULL) 
				{
	     			/* and keep with my pool of state vectors */
	     			if (MyState->PState == NULL) 
					{
               				MyState->PState = p->SV;
	       				MyState->PState->next = NULL;
	     				} 
				else 
					{
	       				p->SV->next = MyState->PState;
	       				MyState->PState = p->SV;
	     				}
	     			p->SV = NULL;
	   			}

	    		last = p;
	 		} 
		else
	    		break;

 		if (p->FNext == NULL) 
			break;
	 	p = p->FNext;
      		}	/* for a buffer to return to  pe = i */

      	MyState->NToReturn[i] -= j;
      	MyState->PFree[i] = last->FNext;
      	if (MyState->PFree[i] != NULL) 
		{
         	MyState->PFree[i]->FPrev = NULL;
      		}
      	else 
		{
         	MyState->FLast[i] = NULL;
      		}

      	MyState->Nret += 1;
      	/* partitioned: return buffers */
      	TW_LOCK(MyState->Ret[i].Lck_RetQ);
      	first->FPrev = MyState->Ret[i].RLast;
      	last->FNext = NULL;
      	if (MyState->Ret[i].RFree == NULL) 
		{
         	MyState->Ret[i].RFree = first;
      		}
      	else 	
		{
         	MyState->Ret[i].RLast->FNext = first;
      		}

      	MyState->Ret[i].RLast = last;
      	MyState->Ret[i].NBufs += j;

      	TW_UNLOCK(MyState->Ret[i].Lck_RetQ);

   } /* for each pe */

 if( show_gvt > 0 )
        {
        int cnt=0;
        for (cnt=0; cnt<TWnpe; cnt++)
                {
                cntFree( cnt, MyPE, "FossilCollect end" );
                }
	}
}


/*----------------------------------------------------------------------------
**  get_free_partitioned (pe, myPE)
**
**  DESCRIPTION:  get an event ev from front of local free list
##      indexed by destination PE pe, on MyPE. Each PE has a dedicated
##      list for all the destinations that it sends to.
**
**---------------------------------------------------------------------------*/
struct Event * 
get_free_partitioned(int pe, int MyPE)
{
   struct Event *ev;
   struct Event **_first;
   struct RetPool *r;

   struct Event   *CurEvent;      	/* current event that */
   struct LPState *CurState;    	/* current LP state */
   struct PEState *MyState;     	/* state for this PE */

   MyState =  &GState[MyPE];
   CurEvent = MyState->CurEvent;
   CurState = MyState->CurState;

   cntFree( pe, MyPE, "[get_free] begin" );

  _first = &(MyState->PFree[(pe)]);

  (ev) = *_first;

  /* while( ev && ev->Msg.RcvTS > MyState->GVT ) ev = ev->FNext;*/ 
  /* add 10/07/98 */

#if 0
  { /* ADDING 02/27/99 */
  int count=0;
  while(((ev) != NULL) && ((ev)->Msg.RcvTS > MyState->GVT) && count <128)
     {
     ev = ev->FNext;
     count++;
     }
  }
#endif

  if ( ((ev) == NULL) || ((ev)->Msg.RcvTS > MyState->GVT) ) 
	{ 
     	/* get buffers from destianation pe's return pool if there are any */
     	if ( GState[pe].Ret[MyPE].RFree != NULL ) 
		{
		r = &(GState[pe].Ret[MyPE]);
        	(MyState->Ntaken)++;
		TW_LOCK(r->Lck_RetQ);
		r->RLast->FNext = *_first;
		if (*_first == NULL) 
			MyState->FLast[(pe)] = r->RLast;
		else 
			(*_first)->FPrev = r->RLast;

		*_first = r->RFree;
		(*_first)->FPrev = NULL; /* shouldn't be needed? */
		r->RFree = NULL;
		r->RLast = NULL;
		MyState->NToReturn[(pe)] += (r->NBufs);
		r->NBufs = 0;
		TW_UNLOCK(r->Lck_RetQ);
        	(ev) = *_first;
        	*_first = (ev)->FNext;

        	if (*_first == NULL)
	   		MyState->FLast[(pe)] = NULL;
        	else
           		(*_first)->FPrev = NULL;
     		}
     	else 	
		{
		if ( (ev) != NULL ) 
			{
			extern int show_noprint;
        		if ( show_noprint == 0 )
			  {
#if 0
			  printf (" cant reclaim %.4f (%.4f) from (%c) LP(%d)"
			    " frm (%d) proc'd (%d)\n",
				(ev)->Msg.RcvTS, MyState->GVT,
				(ev)->in, (ev)->LP, (ev)->SrcLP,
				(ev)->Processed );
#endif
			   }
#if 0
			printf ("reclaim event is %lf and GVT = %lf\n", 
				(ev)->Msg.RcvTS, MyState->GVT );
			cntFree( pe, MyPE, "getFree" );
			exit(1); /* IO5 */
#endif
			}
		else
			{
#if 0
			printf (". reclaim is NULL\n");
			exit(1); /* IO5 */
#endif
			}
		(ev) = GState[MyPE].AbortBuf;
		CurEvent->Processed = ABORT; /* get_free_partioned */
     		}
  	}

   else 
	{
     	*_first = (ev)->FNext;
     	if ( *_first == NULL ) 
		{
		MyState->FLast[(pe)] = NULL;
    		}
     	else
        	(*_first)->FPrev = NULL;

    	/* remove from event list */
    	if ( (ev)->in =='e' )
		{
      		if ( (ev)->Next ) 
			(ev)->Next->Prev = (ev)->Prev;
      		else
			{ 
			/* last element on list */
			struct LPState * statep;

			if ( LP2PE( (ev)->LP) 
			   == TWds.LocalPEToPE[TWds.MyMachine][MyPE] )
				{
				statep = MyState->CList[LPinPE((ev)->LP)];
				statep->Last = (ev)->Prev;
				}
      			}
		/* ev->Prev is always non null */
		if( (ev)->Prev != NULL )
      			(ev)->Prev->Next = (ev)->Next; 
		else 	
			{
			printf ("FC ** Error\n"); exit(-1);
			}
    		}
  	(ev)->Next = (ev)->Prev = NULL;

  	if ( (ev)  == (MyState->PFree[(pe)])  )
		error ("in get free\n");
  	}

  cntFree( pe, MyPE, "end get free" );
  return (ev);
}

