/******************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.7                             ***
***         Copyright 1997 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its
     documentation for any research purpose and without fee is hereby
     granted, provided that the above copyright notice appears in all
     copies and that both that copyright notice and this permission
     notice appear in supporting documentation.  The use or inclusion of
     this software or its documentation in any commercial product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE, INCLUDING
     ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/

/*****************************************************************************
**      TIME WARP kernel on Shared Memory Architectures.
******************************************************************************/

#include "gtw_sim.h"
#include "gtw_timer.h"
#include "op_gtw.h" /*IO4*/

#define 	MH_INDEX 40

TWtimer		mh_init[MH_INDEX];
ClockTicks 	mh_secs[MH_INDEX];

int 		same_gvt=0;
int 		show_gvt=0;

/* DEBUG from config.c */
extern int 	clone_multiple;
extern int	print_tables[10];
extern int 	do_mh_memory;
extern int 	do_headers;
extern int	show_noprint;


/* USER defined */
extern void TWInitAppl(int argc, char ** argv);
extern void TWWrapAppl(void);

#ifdef MITRE_CUSTOM
 extern void TWStartAppl( int MyPE );
#endif


/* ---------------------------------------------------------------------------
**  GetLPsLVT() 
**
**  DESCRIPTION:   This is also set in declaration - maybe remove this
**
**---------------------------------------------------------------------------*/
double GetLPsLVT( int LP ) /*IO4*/
{ 
   struct LPState       *ParentState=NULL;
   struct PEState       *MyState=NULL;
   int                  MyPE;

   MyPE = LP2PE(LP);
   MyState = &(GState[MyPE]);
   ParentState = MyState->CList[LPinPE(LP)];
   return ParentState->LVT;
}


/* ---------------------------------------------------------------------------
**  SetDefaults()
**
**  DESCRIPTION:   This is also set in declaration - maybe remove this
**
**---------------------------------------------------------------------------*/
void
SetDefaults ( void )
{
   /* set default parameter values */
   GPEStateMem = -1;
   PEEventMem = DEFAULT_EV_MEM;

   MsgBufferCheck = DEFAULT_MSG_BUFFER_CHECK;
   GlobEventMem = DEFAULT_GLOB_EV_MEM;
   TWdebug = TWstat = 0;
   randmixer = 0;
   batch = 1;
   RetThreshold = 4.0; /* v3_0 -- is 8.0 */

   /* distributed computing */
   RTGPEStateMem = -1;
   RTPEEventMem = DEFAULT_EV_MEM;
   RTMsgBufferCheck = DEFAULT_RT_MSG_BUFFER_CHECK;
}

/* ---------------------------------------------------------------------------
** set_core_limit(void)
**
** DESCRIPTION:  calls initializing routines, then runs simulation,
**      and finally collects and display statistics
**
**---------------------------------------------------------------------------*/
void
set_core_limit ( void )
{
   struct rlimit rl;

   rl.rlim_cur = 1;
   rl.rlim_max = 1;
   if( setrlimit( RLIMIT_CORE, &rl ) < 0 )
      	perror("setrlimit failed" );
}


/* ---------------------------------------------------------------------------
**  InitTimeVar()
**
**  DESCRIPTION:  initializes GVT, clock and other time variables
**      initializes real time clock and figure out clocks per millisecond.
**
**---------------------------------------------------------------------------*/
void
InitTimeVars ( void )
{

   /* initialize real time clock and figure out clocks per millisecond */

#ifndef NEWTIMER
   TWInitClk();
#else
   TWInitTimer();
#endif

   /* set frequency of GVT invocations */
   GVT = 0.0;
   LocalGVT = 0.0;

   if( show_noprint == 0 )
   	printf ("\n GVT Computed every %ld ticks\n\n", GVTPeriod ) ;
   GVTflag = -(TWnpe+RFthread);
   GVTProcessor = 0;
}

/* ---------------------------------------------------------------------------
**  InitApplVars()
**
**  DESCRIPTION:   initialize application handlers
**
**---------------------------------------------------------------------------*/
void InitApplVars ( void )
{
   int i;

   for (i = 0; i < MAXLP; i++) 
	{
	TWLP[i].Map = -1;
	TWLP[i].IProc = NULL;
	TWLP[i].Proc = NULL;
	TWLP[i].FProc = NULL;

	TWLP[i].CProc = NULL; /*IO4*/

	TWLP[i].IncrSave = FALSE;
	TWLP[i].State = NULL;
	TWLP[i].CopySize = 0;

	LPNum[i] = MAXINT;
   	}
}


/* ---------------------------------------------------------------------------
** main()
**
** DESCRIPTION:  calls initializing routines, then runs simulation,
**      and finally collects and display statistics
**
**---------------------------------------------------------------------------*/
int main( int argc, char **argv )
{
   int i;
   int argc_appl=0;
   char **argv_appl=NULL;

   Argv = argv;

   set_core_limit();
   SetDefaults();
   TWConfigure( argc, argv, &argc_appl, &argv_appl );

   InitTimeVars();

   for (i=0; i<MH_INDEX;i++)
	{
	TWzero_timer(mh_init[i]);
	mh_secs[i] = 0;
	}

   TWLRnow(mh_init[0]);

   InitSynchVar();
   InitApplVars ();			/* also initializes LPNum[] */
   InitMemMap();   			/* mapping of buffers:alloc.c */

   TWLRnow(mh_init[1]);

   TWInitAppl(argc_appl, argv_appl);

   TWLRnow(mh_init[2]);
   TWLRnow(mh_init[3]);

   CheckTWInit();
   TWLRnow(mh_init[4]);
   RunSim(); 				/* run simulation */

   TWLRnow(mh_init[12]);

   WrapUpAppl(); 			/* wrap up and display statistics */

   TWLRnow(mh_init[13]);

   TWWrapAppl();			/* user defined */
   WrapUpSim(); 


   fflush( stdout ) ;
   fflush( stderr ) ;
   return(0);
}

/*-----------------------------------------------------------------------------
**  CheckTWInit()
**
**  DESCRIPTION:  Check parameters set by TWInit
** H
**---------------------------------------------------------------------------*/
void CheckTWInit( void )
{
   int i, j, TotalMemMap;
   void ShowMapping(void);

   if (TWnlp <= 0) error ("TWInitAppl did not initialize TWnlp\n");
   if (TWstat > 1) ShowMapping();

   for (i = 0; i < TWnlp; i++) 
	{
      	if ((TWLP[i].Map < 0) || (TWLP[i].Map >= TWds.PEs )) 
		{
	 	fprintf(stderr, "Illegal Mapping\n");
		printf ("TWLP[%d].Map = %d and TWnpe = %d\n", 
			i, TWLP[i].Map, TWnpe );
	 	exit(1);
      		}
   	}

   /* Total of weights in MemMap array */
   TotalMemMap = 0;
   for (i=0; i<TWnpe+RFthread; i++)
      	for (j=0; j<TWnpe+RFthread; j++)
	 	TotalMemMap += TWMemMap[i][j];

   /* set up default mapping of buffers if not set (equal weight) */
   if ( TotalMemMap == 0 )
	{
      	for (i=0; i<TWnpe+RFthread; i++)
         	for (j=0; j<TWnpe+RFthread; j++)
			{
	    		TWMemMap[i][j] = (int) 1;
	 		TotalMemMap += TWMemMap[i][j];
			}
   	}

}


/*-----------------------------------------------------------------------------
**  ShowMapping()
**
**  DESCRIPTION:  Print out the mapping of LP to PE
**
**---------------------------------------------------------------------------*/
void
ShowMapping(void)
{
   int i, j;

   printf("\n *** Mapping of LP to PE *** \n");
   for (i = 0; i < TWnpe; i++) 
	{
      	printf(" Proc %d has the LPs ", i);
      	for (j = 0; j < TWnlp; j++)
	 	if (TWLP[j].Map == i)
	    		printf("%d,", j);
      		printf("\n");
   	}
   printf("\n");
}


/*-----------------------------------------------------------------------------
**  InitSynchVar()
**
**  DESCRIPTION:  Initialize synchronization variables: Locks and Barriers
** 	must be called after TWnpe had been determined
**
**---------------------------------------------------------------------------*/
void InitSynchVar(void)
{
   int i, j;

   /* barrier initializations */

#if( defined(SGI5) || defined(SGIMP64) || defined(SGI64) )
   /* initialize shared area */
   strcpy(shared_area, mktemp("/tmp/TWXXXXXX"));
   if( (shm_handle = usinit(shared_area)) == NULL )
	{
    	perror("usinit");
    	exit(-1);
  	}
#endif /* SGI5 || SGIMP64 || SGI64 */

   TW_INIT_LOCK(RandFileLock);
   TW_INIT_BARRIER(Go1, TWnpe);

   TW_INIT_BARRIER(RF1,TWnpe+1);
   TW_INIT_BARRIER(RF2,TWnpe+1);

   /* init locks */
   /* insert lock  initializations here, if not dist then we just need to
    * allocate one lock  **/
   for (i = 0; i < TWnpe; i++) 
	{
  	TW_INIT_LOCK(GState[i].Lck_MsgQ);
      	TW_INIT_LOCK(GState[i].Lck_CanQ);
      	TW_INIT_LOCK(GState[i].Free_Lock);
      	for (j = 0; j < TWnpe+RFthread; j++)
		{
         	TW_INIT_LOCK(GState[i].Ret[j].Lck_RetQ);
      		}
   	}
   TW_INIT_LOCK(Lck_NumGlobalPool);
   TW_INIT_LOCK(Lck_GVTflag);
   TW_INIT_LOCK(dumplck);

   NBusy = TWnpe;

   /* intitialize random variable generator */
   /* call after TWConfigure and after above InitSynch Code o/w SEG FAULT */
   if ( randmixer > 0 )
	{
      	TWSeed dummy;
      	TWRandInit (&dummy, randmixer-1);
      	}
}


#if( defined(SGI5) || defined(SGIMP64) || defined(SGI64) )
/*-----------------------------------------------------------------------------
**  RunsSimSGI()
**
**  DESCRIPTION:  Runs Simulation
**
**----------------------------------------------------------------------------*/
void RunSim ( void )
{
   void Scheduler ( void );
   int error;

   m_set_procs( TWnpe+RFthread ); 

   if ( (error = m_fork (Scheduler) ) < 0 )
	{
         perror ( "m_fork failed: " );
         exit( EXIT_FAIL );
         }
  
   /* after all threads done */
   usdetach ( shm_handle );
   if ( unlink ( shared_area ) == -1 )
	{
     	perror ( "unlink" );
     	exit ( EXIT_FAIL );
   	}
}
#endif /* SGI5 || SGIMP64 || SGI64 */


#if( defined(SUNMP) || defined(SUN4SOL2) )
/*-----------------------------------------------------------------------------
**  RunsSim()
**
**  DESCRIPTION:  Runs Simulation
**
**----------------------------------------------------------------------------*/
void RunSim(void)
{
   void * Scheduler (void *);
   thread_t tid[MAXPROC];
   struct args a1[MAXPROC];

   int i;
   int num_threads=TWnpe+RFthread;

   num_threads = ( TWnpe == 1 ) ? 1 : (TWnpe+RFthread);

   /* a uniprocessor should not create an extra thread */

   /* creating extra threads, in addition to the original one */
   for ( i = 0; i < num_threads - 1; i++ )
	{
	int error;    
      	a1[i].id = i;
     	error = thr_create(NULL, 0, Scheduler, (void *)&a1[i],
			THR_NEW_LWP, &(tid[i]));
	printf("Created %d %d %d\n", tid[i], error, thr_getconcurrency());
    	}

   a1[num_threads-1].id = num_threads-1;
   Scheduler((void *)&a1[num_threads-1]);

   for(i = 0; i < num_threads - 1; i++)
    	{
      	thr_join(tid[i], 0 , 0);
    	}

}
#endif /* SUNMP || SUN4SOL2 */


/*-----------------------------------------------------------------------------
**  InitLP()
**
**  DESCRIPTION:  Initialize processor LP data structures for PE
**	not called by reflector thread
**
**  CALLEES: InitPE
**
**----------------------------------------------------------------------------*/
void InitLPs ( int PE )
{
   struct LPState *statep;
   int i;
   int GlobalPE=-1;
   TWtimer  mh_1, mh_2;
   TWtimer  mh_3, mh_4;

   GlobalPE = TWds.LocalPEToPE[TWds.MyMachine][PE];

   AllocateLPState(PE);

   if ( PE == 0 )
        {
        TWLRnow(mh_init[19]);
        }

   for (i = 0; i < TWnlp; i++) 
	{
      	int Incr_Print_Flag=0;

      	if  ( TWLP[i].Map != GlobalPE ) 
		{
		continue;
		}
	else
		{
        	if ( PE == 0 )
        		{
        		TWLRnow(mh_3);
			}
		}

      	statep = LPST(i);		/* pointer to state of LP */
      	statep->ID = i;
      	statep->EList.LP = i;

   	/* These arrays are set by the applications */
    	statep->ProcMsg = TWLP[i].Proc;
      	statep->IProcMsg = TWLP[i].IProc;
      	statep->FProcMsg = TWLP[i].FProc;

      	statep->EList.Processed = FALSE;	/* hdr evnt is end */
      	statep->EList.InPDS = FALSE;	/* hdr evnt is end */
      	statep->EList.Msg.SendTS = NEGINFINITY;
      	statep->EList.Msg.RcvTS = -1.0;
      	statep->EList.Next = statep->EList.Prev = NULL;
      	statep->EList.FNext = statep->EList.FPrev = NULL;
      	statep->EList.SrcLP = statep->EList.SrcPE = MAXINT;
      	statep->EList.SV = NULL;
      	statep->EList.CauseList = NULL;
      	statep->EList.NextCause = NULL;

      	statep->LVT = 0.0;

        statep->IoBList.IONext = NULL;    
        statep->IoBList.IOPrev  = NULL;  
        statep->IoBList.Up = NULL;     
        statep->IoBList.Down = NULL;  
        statep->IoBList.LastIoBList = NULL;   

      	/* no state until allocated */
      	statep->state = NULL;
      	statep->auto_chkpt_state = NULL;
      	statep->StateValid = TRUE;
      	statep->RBSV = NULL;

      	/* set up statistics for LP behavior */
      	statep->NEvent = 0;
      	statep->NAbort = 0;
      	statep->NCancel = 0;
      	statep->NStraggler = 0;
      	statep->NRBStraggler = 0;
      	statep->NSecondary = 0;
      	statep->NRBSecondary = 0;
      	statep->NArt = 0;
      	statep->NRBArt = 0;
      	statep->LongestRB = 0;

      	TWzero_timer(statep->EvExecTime);

      	statep->NOPEvent = 0;                   
      	statep->NEXEvent = 0;                  
      	statep->NNBEvent = 0;                 
      	statep->NBLEvent = 0;                

      	statep->NOPAbort = 0;                            
      	statep->NEXAbort = 0;                           
      	statep->NNBAbort = 0;                          
      	statep->NBLAbort = 0;                         

      	statep->NOPStraggler = 0;                    
      	statep->NEXStraggler = 0;                   
      	statep->NNBStraggler = 0;                  
      	statep->NBLStraggler = 0;                 
      	statep->NRBOPStraggler = 0;              

      	statep->NOPCancel = 0;                  
      	statep->NRBOPSecondary = 0;            

      	statep->NRBDelays = 0;                
      	statep->NRBDelaysR = 0;              

      	statep->NEventsBlocked = 0;         
      	statep->NBlockings = 0;            

      	statep->NCanOPCQ = 0;             
      	statep->NCanEXCQ = 0;            
      	statep->NCanNBCQ = 0;           
      	statep->NCanBLCQ = 0;          
      	statep->NCanNBQ = 0;          
      	statep->NCanOQ = 0;          
      	statep->NRCanBQ = 0;        
      	statep->NBCanBQ = 0;       

      	statep->NBBLEnQ = 0;                      
      	statep->NBLEnQ = 0;                      
      	statep->NOPEnQ = 0;                     
      	statep->NNBEnQ = 0;                    

      	/* incremental state save */
      	statep->IncrSave = TWLP[i].IncrSave;

        if ( PE == 0 )
        	{
       		TWLRnow(mh_4);
		mh_secs[22] += (TWseconds(mh_4) - TWseconds(mh_3));

        	TWLRnow(mh_1);
		}

      	if (TWLP[i].IncrSave)
		{
        	if ((PE == 0)&&(Incr_Print_Flag==0) && (TWdebug & DEBUG_INCR)) 
			{
	  		printf("Allocate IncrLog (%d arrays of size %d) ...\n",
		 		NLogArrays, TW_Incr_Array_Size);
	  		fflush(stdout);
			}
		Incr_Print_Flag = 1;
		AllocateLPLog ( statep, PE );
      		}
        if ( PE == 0 )
        	{
		TWLRnow (mh_2);
 		mh_secs[21] += (TWseconds(mh_2) - TWseconds(mh_1));
		}

   	} /*for all lps */

   if ( PE == 0 )
        {
        TWLRnow(mh_init[20]);
        }
}


/*-----------------------------------------------------------------------------
**  InitPE()
**
**  DESCRIPTION:  Initialize processor PE data structures
**
**  CALLEES:  scheduler procesees
**
**----------------------------------------------------------------------------*/
void InitPE ( int PE )
{  int i, j, k;
   struct LPState *statep;
   int GlobalPE=-1;
   int MyPE = PE;

   /* get my Global PE number */
   if ( PE == TWnpe || (PE == TWnpe + 1 && TWnpe == 1) )
	{ /* reflector thread */
   	AllocateEventMemory(RTPEEventMem, GlobEventMem, PE);
	GlobalPE = MAXINT;
	}
   else
	GlobalPE = TWds.LocalPEToPE[TWds.MyMachine][PE];

   if ( TWnpe != PE ) 
   	AllocateEventMemory(PEEventMem, GlobEventMem, PE);

   if ( MyPE == 0 )
        {
        TWLRnow(mh_init[14]);
        }

   for (j=0; j<TWnpe+RFthread; j++) 
	{
      	GState[PE].Ret[j].RFree = NULL;
      	GState[PE].Ret[j].RLast = NULL;
      	GState[PE].Ret[j].NBufs = 0;
   	}

   GState[PE].Nret = 0;
   GState[PE].Ntaken = 0;
   Local_Min[PE] = GVT;
   GState[PE].MsgQ = NULL;
   GState[PE].CanQ = NULL;

   GState[PE].FirstBlockedLP = NULL;                             
   GState[PE].LastBlockedLP = NULL;                             

   GState[PE].FirstIoNBList = NULL;                            
   GState[PE].LastIoNBList = NULL;                            

   GState[PE].FirstIoOptList = NULL;                         
   GState[PE].LastIoOptList = NULL;                         

   /* Registers for memory management */
   GState[PE].PELVT = 0.0;	/* min LVT for PE */
   GState[PE].GVT = 0.0;
   GState[PE].GVT_Progress = NORMAL;
   GState[PE].TransMsgTS = MAXFLOAT;

   GState[PE].NumFC = 0;
   GState[PE].NumSR = 0;
   GState[PE].NumTerminate = 0;

   /* * initialize LP states */
   GState[PE].LPsOnPE = 0;
   for (i = 0; i < TWnlp; i++) 
	{
      	if ( TWLP[i].Map != GlobalPE ) continue;
      	LPNum[i] = GState[PE].LPsOnPE;
      	(GState[PE].LPsOnPE)++;		/* number of LPs in PE */
   	}

   if (GState[PE].LPsOnPE > MAXLPONPE) 
	error("Too many (%d) LPs on Proc %d MAX (%d)\n",
		GState[PE].LPsOnPE, PE, MAXLPONPE );

   if ( MyPE == 0 )
        {
        TWLRnow(mh_init[15]);
        }

   if ( PE != TWnpe ) 
	{
	extern void mh_init_calloc(int);
	if ( do_mh_memory > 0 )
		{
		mh_init_calloc( PE );
		}
	InitLPs ( PE );
	}

   if ( MyPE == 0 )
        {
        TWLRnow(mh_init[16]);
        }

   InitQ(PE); 		/*Initialize priority queue data structure */
   if ( MyPE == 0 )
        {
        TWLRnow(mh_init[17]);
        }
   InitBufs(PE); 	/* Initialize memory buffers */
   if ( MyPE == 0 )
        {
        TWLRnow(mh_init[18]);
        }
}

/*-----------------------------------------------------------------------------
**  InitBufs()
**
**  DESCRIPTION:  Initialize message buffer for each LP on this PE.
**      Called from  Scheduler()
**
**---------------------------------------------------------------------------*/
void InitBufs(int id)
{
   int i, bp, obp;
   struct LPState *statep;
   struct Event *p;
   struct PEState * MyState;     /* state for this PE */
   int MyPE;

   MyPE = id;
   MyState = &GState[id];

   /* write into free buffers to map them into local memory */
   for ( i = 0; i < TWnpe+RFthread ; i++ ) 
	{
	/* for all my free buffers */
      	for ( p = MyState->PFree[i] ; p != NULL ; p = p->FNext ) 
		{
                p->IONext = NULL;                             
                p->IOPrev = NULL;                            
                p->EventType = UNDEFINED;                   
                p->EvType = 'I';                           
                p->IOFunc = NULL;                         
                p->InList = INNOLIST;                    
		p->RBFunc = NULL;  
		p->Next = NULL;
		p->Prev = NULL;
         	p->Processed = FALSE;
         	p->InPDS = FALSE;
         	p->LP = -1;

                p->PCloneID = 0;	/* InitBufs: IO4 */
                p->CloneID = -1;	/* InitBufs: IO4 */

         	p->Msg.RcvTS = NEGINFINITY;
         	p->CauseList = NULL;
         	p->NextCause = NULL;
	 	p->in = 'f';		/* initial free list */
      		}
   	} /* for all pe's */

   /* LPs add a dummy buffer to each event list */
   /* get buffer from bufferpool (bp) round robin */
   bp = 0;		/* buffer pool to get message buffer from */
   for ( i = 0 ; i < MyState->LPsOnPE ; i++ ) /* RF has no LPs */  
	{
      	obp = bp;
      	while ( MyState->PFree[bp] == NULL ) 
		{ /* get from next buffer pool */
	 	bp = (bp+1)%TWnpe;
	 	if (bp == obp) error ("Need more event memory\n");
      		}

      	statep = MyState->CList[i];

      	p = MyState->PFree[bp];		/* allocate buffer */
      	MyState->PFree[bp] = p->FNext;
      	if (MyState->PFree[bp] == NULL)
	 	MyState->FLast[bp] = NULL;
      	else
         	MyState->PFree[bp]->FPrev = NULL;

      	/* statep->Last = &(statep->EList); */

      	statep->EList.Next = p;
      	statep->EList.Prev = NULL;
      	p->LP = statep->ID;

	p->PCloneID = -1; 		/* InitBufs: IO4 */

      	p->Prev = &(statep->EList);

      	p->Next = NULL;
      	p->Processed = FALSE;
      	p->InPDS = FALSE;
      	p->SrcPE = bp;
      	p->FPrev = p->FNext = NULL;
	
      	p->Msg.SendTS = NEGINFINITY;
      	p->Msg.RcvTS = 0.0;	/* initial message */
      	p->CauseList = NULL;
      	p->NextCause = NULL;

      	p->in = 'e';
      	statep->Last = p;
      	bp = (bp+1) % TWnpe;
   	}

   /* add dummy buffer for aborted events */
   obp = bp;
   while ( MyState->PFree[bp] == NULL ) 
	{
      	bp = (bp+1)%TWnpe;
      	if (bp == obp)
		error ("Insufficient event memory, processor %d\n", MyPE);
   	}
   p = MyState->PFree[bp];		/* allocate buffer */
   MyState->PFree[bp] = p->FNext;
   if (MyState->PFree[bp] == NULL)
      	MyState->FLast[bp] = NULL;
   else
      	MyState->PFree[bp]->FPrev = NULL;

   GState[MyPE].AbortBuf = p;
   GState[MyPE].AbortBuf->Next = NULL;
   GState[MyPE].AbortBuf->Prev = NULL;
   GState[MyPE].AbortBuf->Processed = FALSE;
   GState[MyPE].AbortBuf->InPDS = FALSE;
   GState[MyPE].AbortBuf->CauseList = NULL;
   GState[MyPE].AbortBuf->NextCause = NULL;
   GState[MyPE].AbortBuf->FNext = NULL;
   GState[MyPE].AbortBuf->FPrev = NULL;
   GState[MyPE].AbortBuf->Msg.RcvTS = -3.0;
   GState[MyPE].AbortBuf->SV = NULL;
}

/*-----------------------------------------------------------------------------
**  ProcessMsgQ()
**
**  DESCRIPTION: Places Message in the Message in the Proper LP eventlist
**
**----------------------------------------------------------------------------*/
void ProcessMsgQ( int MyPE )
{
   int 		cpe;
   short 	RBDist;
   struct Event	*p=NULL, *q=NULL, *c=NULL, *nextq=NULL;
   struct Event	*NCause=NULL, *pPrev=NULL;
   struct Event	*mp=NULL;
   struct PEState *MyState=NULL;   
   struct LPState *statep=NULL;
   struct LPState *CurState=NULL;  
   TWTime RcvTS;

   MyState = &(GState[MyPE]);

   ck4_probe("MsgQ begin")

   if ( MyState->MsgQ == NULL )
	{
	ck4_probe("MsgQ end")
	return;
	}

   TW_LOCK(MyState->Lck_MsgQ);
     mp = MyState->MsgQ;
     MyState->MsgQ = NULL;	/* Reset MsgQ */
   TW_UNLOCK(MyState->Lck_MsgQ);

   /* insert events into event lists or calq */
   for ( q = mp; q != NULL; q = nextq ) 
	{
	ck1_inMsgQ("MsgQ",q)

	nextq = q->Next;

	(MyState->NToReturn[q->SrcPE])++;

	printEXfirst(loc)

    	RcvTS = q->Msg.RcvTS;

 	ck1_isNeg("MsgQ",RcvTS)

	CurState 
		= GState[MyPE].CurState 
		= statep 
		= MyState->CList[LPinPE( OP_GetRBLP(q->LP, RcvTS) ) ]; /*IO4*/

	p = statep->Last;	/* last processed event */

	ck_GVT("MsgQ",RcvTS)	

	if ( RcvTS < p->Msg.RcvTS )
		{ 
		/* case I: rollback dst LP */
	       	struct Event *lastp=NULL;

		ts_set(t5)
		ck_Event("MsgQ",p) 

	       	RBDist = 0;

		if ( show_gvt >= 2 ) /* PHERE */
			{
			printf ("[MSQ RB]: straggler"
				" rcvLP=%d srcLP=%d"
				" RcvTS=%lf\n",
				q->LP, q->SrcLP,
				q->Msg.RcvTS );
			}

		/* roll back the event queue starting with last */
	       	for ( ; RcvTS < p->Msg.RcvTS; p = pPrev ) 
			{
			if ( show_gvt >= 2 ) /* PHERE */
				{
				printf ("[MSQ RB]: un-processing"
					" rcvLP=%d srcLP=%d"
					" RcvTS=%lf\n",
					p->LP, p->SrcLP,
					p->Msg.RcvTS );
				}

			/* remove from free list */
  			RM_FREE(p,p->SrcPE);		

		  	pPrev = p->Prev;
		  	RBDist++;

			/* apply rb function if there is one */
		      	if ( p->RBFunc != NULL ) /* rolling back MsgQ */
				{ /* CurState is set already BType2 */
       				GState[MyPE].CurEvent = p;
				p->RBFunc( statep->state, p->Msg.Data, MyPE );
				}

			p->Processed = FALSE;

		  	/* cancel events scheduled by this event */
 			for ( c=p->CauseList ; c != NULL ; c=NCause )  
				{
				CANCEL;
				}
  			p->CauseList = NULL;

			lastp = p; 

			/* put p back in the calq */
  			EnQueue( p, MyPE );	
  			p->in = 'c';

			ck_EList("MsgQ",p);
	       		} /* end for rolling back evQueue loop */

		/* set up for state restore ---------------- */
		if ( lastp != NULL ) /* IO2 */
			{
	       		if ( (statep->RBSV != NULL)  ) 
				{
				/* a rollback put RBSV in free state list */
	 			statep->RBSV->next = MyState->PState;
	 			MyState->PState = statep->RBSV;
       				}

			if ( lastp->SV == NULL ) 
		 		printf("Straggler:no state found\n");

			 /* save ptr to restore state */
			statep->RBSV = lastp->SV;  
	       		lastp->SV = NULL;

			/* need to do restore */
	       		statep->StateValid = FALSE;	

	       		if ( statep->IncrSave )
				Restore( lastp->LogPtr, statep );
			} 

		if ( RBDist > 0 ) 
			{ /* straggler induced RB */
			(statep->NRBStraggler) += (int) RBDist;
	       		}

	       	/* insert straggler in the calq */
	       	EnQueue ( q, MyPE );
	       	q->in = 'c';

		/* at least 1 processed event remains reset LVT */
	       	statep->Last = p;
	       	statep->LVT = p->Msg.RcvTS;
	       	statep->Last->Next = NULL;

		/* below not in v3_4 */
		if ( q->EventType == BLOCKING )
	       		(statep->NBLStraggler)++;
		else if ( q->EventType == NONBLOCKING )
			(statep->NNBStraggler)++;

		(statep->NStraggler)++;
	       	statep->LongestRB = max(statep->LongestRB, (int) RBDist);

	    	} /* case I */
	else 	
		{ /* case II: no rollback */
		EnQueue(q,MyPE);
	       	q->in = 'c';
	    	}
	 } /* for all events in mp */

   printEX("c2")

   ck4_probe("MsgQ end")
}



/*-----------------------------------------------------------------------------
**  ProcessCanQ()
**
**  DESCRIPTION:  Places AntiMessages in CanQ for PE MyPE.  These are events
**      that have been scheduled by other events.  They are removed and
**      deleted.
**
**----------------------------------------------------------------------------*/
void ProcessCanQ( int MyPE )
{
   int cpe;
   short RBDist;
   struct Event *p=NULL, *q=NULL, *c=NULL, *NCause=NULL, *pNext=NULL;
   struct Event *cp=NULL;
   struct PEState *MyState=NULL;     /* state for this PE */
   struct LPState *statep=NULL;

   MyState = &(GState[MyPE]);

   ck4_probe("canQ begin")

   if ( MyState->CanQ == NULL)
	{
	ck4_probe("canQ end")
	return;
	}

   while ( MyState->CanQ != NULL ) 
	{
	TW_LOCK(MyState->Lck_CanQ);
	   cp = MyState->CanQ;
	   MyState->CanQ = NULL;	/* Reset CanQ */
	TW_UNLOCK(MyState->Lck_CanQ);

	/* for each cancelled event */
	for ( q = cp; q != NULL; q = q->NextCause )
		{

	       	statep = MyState->CList[LPinPE(q->LP)];

		if ( show_gvt >= 2 ) /* PHERE */
			{
			printf ("[CNQ RB]: cancellation"
				" rcvLP=%d srcLP=%d"
				" RcvTS=%lf %s\n",
				q->LP, q->SrcLP,
			 	q->Msg.RcvTS,
			       (q->Processed) ? "IS-PROCESSED" : "UN-PROCESSED" 
				);
			}

	       	/* if processed, rollback and cancel */
	       	if ( q->Processed )
			{
		  	(statep->NCancel)++;	

			ck_FindPrevNULL("CanQ",q) 

	  		statep->Last = q->Prev;
  			statep->LVT = statep->Last->Msg.RcvTS;
  			statep->Last->Next = NULL;

		  	if ( q->Next ) /* mh1: moved from below, dont need */
				q->Next->Prev = q->Prev;

			/* rollback incremental state to this event */
	  		if ( statep->IncrSave )
	     			Restore(q->LogPtr, statep);

	  		if ( statep->RBSV != NULL )
				{
				/* toss earlier time-stamped rb'd state */
	    		    	statep->RBSV->next = MyState->PState;
				MyState->PState = statep->RBSV;
	  		    	}

	  		if( q->SV == NULL )
	    		    	printf("SV is null?\n");

			/* save this state in RBSV for update upon processing */
	  		statep->RBSV = q->SV;
	  		q->SV = NULL;

			/* appl. state wrong */
	  		statep->StateValid = FALSE;	

			/* moved mh1 */
	   		} /*  cancelled processed event  ----*/


               	/* rollback (secondary) cancelled event processed -----------*/

		RBDist = -1;	/* cancelled event not rolled back */
	       	if ( q->Processed ) 
			{  /* need to unprocess events processed after q */

		  	for ( p = q; p != NULL; p = pNext ) 
				{  /* cancels events scheduled by p */
  				RM_FREE(p,p->SrcPE);		

		     		pNext = p->Next;

		     		RBDist++;

				ck_NotProc("CanQ",p)

				if ( p->RBFunc != NULL ) /* rolling back CAN */
					{
               				GState[MyPE].CurState = statep;
       					GState[MyPE].CurEvent = p;
    					p->RBFunc( 
						statep->state, 
					    	p->Msg.Data,
						MyPE );
					}

				p->Processed = FALSE;

		     		/* cancel events scheduled by p */
     				for ( c = p->CauseList; c != NULL; c = NCause ) 
					{
					CANCEL;
		     			}
		     		p->CauseList = NULL;

				EnQueue(p,MyPE);
		     		p->in = 'c';
		  		} /* for each rolled back ev */
			}  /* if processed */

		if ( RBDist > 0 ) 
			{
			/* update stats for secondary roll  back */
  			(statep->NRBSecondary) += (int) RBDist;
  			(statep->NSecondary)++;
  			statep->LongestRB = max(statep->LongestRB,(int)RBDist);
			}

	       	if ( statep->IncrSave )
			q->LogPtr = NULL;

		/* check if cancelling a fossil? */
		ck_GVT("CanQ",q->Msg.RcvTS)

		/* if q was "processed" q is in CalQ, it was enqueued above */
	       	if ( q->InPDS ) 
			{	
			/* delete from calq or skew heap */
			if  ( q->EventType  == NONBLOCKING )
		  		(statep->NCanNBCQ)++;	
			else if  ( q->EventType  == BLOCKING )
		  		(statep->NCanBLCQ)++;	


	       		if ( q->BirthType == -2 ) /*IO4,MH_HERE*/
				{
				if ( LP2PE(q->LP) != MyPE )
					{
					error ("+** CanQ: OP_Cancel BAD\n");
					}
			   	OP_CancelPendingEvent(q->LP,"InPDS");
				q->BirthType == -1;
				}
			else if ( q->BirthType > 0 )
				{
				error ("+** CanQ: give birth event in CalQ\n");
				}
	
		  	Delete_Arbitrary(&q, MyPE);
	       		}

		else if ( (q->InList == INIOPELIST)
			  ||
			  (q->InList == INBLOCKINGLIST)
			  ||
			  (q->InList == INBLOCKINGEVENTLIST)
			 ) 
			{
                       	if ( q->EventType == NONBLOCKING )
                       		{
                                if ( q->InList != INIOPELIST )
                               		{
                                       	printf("**[CanQ]: Deleting a "
					       "PELIST event that is "
					       "NOT in PElist");
                                        }
				(statep->NCanNBQ)++;	
                                NB_IO_Delete(q, MyPE, DO_IN_TW, 0);
                            	}    
             		else  /* regular event or blocking */
                              	{
				if ( q->EventType == REGULAR )
		  		    	(statep->NRCanBQ)++;	
				else if ( q->EventType == BLOCKING )
		  		    	(statep->NBCanBQ)++;	
				else
				        {
					printf ("**[CAN]: %d from BQ\n",
						 q->EventType );
					}
						
               			IOBlockedDelete(q, MyPE);
                               }
                        } /* in an IO list */
		else 
			{
			/* not in CalQ or any IOlist, must be in MsgQ */
			ck1_inMsgQ("canQ1",q)
 			(MyState->NToReturn[q->SrcPE])++;
  			TW_LOCK(MyState->Lck_MsgQ);

	       		if ( q->BirthType == -2 ) /*IO4,MH_HERE*/
				{
			   	OP_CancelPendingEvent(q->LP,"InMsgQ");
				q->BirthType == -1;
				}
			else if ( q->BirthType > 0 )
				{
				error ("+** CanQ: give birth event in MsgQ\n");
				}

  		    	if ( MyState->MsgQ == q ) 
				{	
				/* canned event at front of MsgQ */
		     		if (q->Next != NULL) q->Next->Prev = NULL;

				/* Reset Msg Q */
		     		MyState->MsgQ = q->Next;	
				ck1_inMsgQ("canQ2",q->Next)
				} 
			else 
				{	
				/* rm from middle of MsgQ or Elist the above not
				 *  really right, if rm'ed from MsgQ, but should
				 *  not matter much */

		     		if (q->Prev != NULL)
					{
				    	q->Prev->Next = q->Next;
					ck1_inMsgQ("canQ3",q->Prev)
					}
		     		if ( q->Next != NULL )
					{
				    	q->Next->Prev = q->Prev;
					ck1_inMsgQ("canQ4",q->Next)
					}
				} /* else in middle of MsgQ */

			TW_UNLOCK(MyState->Lck_MsgQ);
	       		} /* end_else not CalQ/IO lists so its in MsgQ */

		q->Msg.RcvTS = -2.0;

		if ( q->BirthType > 0 ) /* IO4 */
                        error (" ** BirthType not reset %d\n", q->BirthType );

		/* put in front of free list */
		PUT_FREE_F(q, q->SrcPE); 
	       	q->in = 'f';	

		} /* for each event in CanQ */
	} /* while CanQ is not empty */

   printEX("c3") 
   ck4_probe("canQ end")

} /* end ProcessCanQ */ 


/*-----------------------------------------------------------------------------
**  ProcessEvents()
**
**
**----------------------------------------------------------------------------*/
void
ProcessEvents( int MyPE )
{
   int cpe;
   static BOOLEAN idling=FALSE;
   struct Event *CurEvent=NULL;      /* current event that */
   struct LPState *CurState=NULL;    /* current LP state */
   int PEStateMem;
   int i = 0 ;
   struct Event *p, *q, *c, *nextq, *NCause, *pNext, *pPrev;
   struct PEState * MyState;     /* state for this PE */
   struct LPState *statep;

   MyState = &(GState[MyPE]);

   ck4_probe("procEv begin")

   /* treat optimistic events like regular events here */
   for ( i = 0; i < batch; i++ ) 
	{
    	/* next event to execute */
       	GState[MyPE].CurEvent = CurEvent = DeQueue(MyPE);

	/* if no pending event, including IOEvents check termination */
	if ( CurEvent == NULL 
                && (GState[MyPE].FirstIoNBList == NULL) 
		&& (GState[MyPE].FirstBlockedLP == NULL) 
            )
		{ /* check for termination */
	      	if ( !idling )  
			{ /* if not idling go in here */
	  		TW_ATOMIC_DCR(NBusy);
	 		idling = TRUE;

	  		if ( (NBusy == 0) 
			    && (MyPE == 0)  /* MHERE */
			    && (MyState->GVT_Progress == NORMAL ) ) 
				{
	               	    	TW_LOCK(Lck_GVTflag);
	    		    	    if ( GVTflag == -(TWnpe+RFthread) ) 
			 		{
		   			GVTflag = TWnpe; 
	       				MyState->GVT_Progress=CALC_GVT;
					}
	                    	TW_UNLOCK(Lck_GVTflag);
		  		} /* NBusy */
	      		} /* of not idling */
		break;
    		} /* CurEvent == NULL */ 

	else if ( CurEvent == NULL )
		{
                break; /* out of batch processing for-loop */
		}
	else 
		{ /* ** found an event to execute or to put in IO LIST */
		ck1_inCalQ(CurEvent)

       		if ( idling ) 
			{ 
			/* retract termination req. */
	  		TW_ATOMIC_INCR(NBusy);
	  		idling = FALSE;
       			}

       		statep = LPST(CurEvent->LP);

    		if ( (statep->IoBList.IONext != NULL) 
		  && (CurEvent->EventType == REGULAR)
		  && (CurEvent->Msg.RcvTS > statep->IoBList.IONext->Msg.RcvTS))
			{
	  		(statep->NBBLEnQ)++;	
                	IO_LpListAppend(CurEvent, statep);
			continue; /* batch loop */
                	}
	
		else if ( CurEvent->EventType == BLOCKING )
               		{
			(statep->NBLEnQ)++;
               		IOBlockedEnQueue(CurEvent, statep);

               		if (CurEvent->Msg.RcvTS == MyState->GVT)
				{
#if 0
			
        			printf("#**BlkTS=%f GVT=%f could process NOW\n",
                                	CurEvent->Msg.RcvTS, MyState->GVT);
				fflush ( stdout );
				/*_TWBlockingProcessing( MyPE,  MyState->GVT);*/
#endif

				}
			else if (CurEvent->Msg.RcvTS <= MyState->GVT)
				{
        			printf("#**BlkTS=%f GVT=%f unable to PROCESS\n",
                                	CurEvent->Msg.RcvTS, MyState->GVT);
				fflush ( stdout );
				}
			else
				{
				}
			continue; /* batch loop */
               		}
 		else if ( CurEvent->EventType == NONBLOCKING )
               		{
	  		(statep->NNBEnQ)++;	
                	IO_NBListEnQueue(CurEvent, MyPE);
			/* _TWNonBlockingProcessing('n', MyPE, current_time);*/
			continue;
                	}
		else

		{ 
		/* normal processing for regular or external event */
		int active_LP;		  				/*IO4*/
		active_LP = OP_ProcessLP(
				CurEvent->LP, 
				CurEvent->Msg.RcvTS, 
				CurEvent->BirthType);

		if ( active_LP < 0 )
			{
			EnQueue( CurEvent, MyPE );
			CurEvent->in = 'c';
			break;
			}

       		statep = LPST(active_LP); 				/*IO4*/

		ck_ProcEvent (CurEvent)
		printEXlast("last") 

               	CurState = GState[MyPE].CurState = statep;
		CurEvent->Processed = IN_PROCESS;

       		/* save state in * this * event, or restore state */
	       	/* save state RBSV if external event */
       		if ( statep->StateValid )
			{
			if ( CurEvent->SV == NULL )
		      		{
	     			/* more state vectors empty pool */
	     		 	if (MyState->PState == NULL) 
			    	   	{
	       		    	   	error ("PE %d ran out of state "
				   		 "vectors, GVT %g\n",
	       			   		MyPE, GVT);
	    		    	   	} 
			      	else 
			    	   	{
	       		    	   	CurEvent->SV = MyState->PState;
	       		    	   	MyState->PState 
						= MyState->PState->next;
	     			   	}
	         		} /* if SV == NULL */
	  		MCpSt(statep->auto_chkpt_state,
				CurEvent->SV->state, 
				statep->CopySize);
			}  /* if valid state and not external  */
		else 
			{  /* not a valid restore it */

			ck_nullRBSV(statep)
		  	MCpSt(statep->RBSV->state,
				statep->auto_chkpt_state,
		 		statep->CopySize);

			statep->StateValid = TRUE;
		  	/* CurEvent already has a SV, save it */
		  	if ( CurEvent->SV != NULL ) 
				{
		     		CurEvent->SV->next = MyState->PState;
		     		MyState->PState    = CurEvent->SV;
		  		}

			/* use the RBSV state */
	  		CurEvent->SV = statep->RBSV;
	  		statep->RBSV = NULL;
       			} /* non valid state */ 

		if (TWdebug & DEBUG_MSG)
			{
	  		printf("RCV EVENT: LP %d, time %f \n",
	 		statep->ID, CurEvent->Msg.RcvTS);
			}

	        /* save pointer Log, before processing an event 
		 * doesnt hurt to save event if an EXTERNAL event 
		 * ext. ev. could change state of LP, but they
		 * dont roll back (may get aborted) */

       		if (statep->IncrSave)
          		CurEvent->LogPtr = statep->LogHead;

		if ( CurEvent->EventType == REGULAR )
			{
        		statep->ProcMsg
				(
				statep->state, 
				CurEvent->Msg.Data,
				MyPE
				);

			}
		else
			{
			printf (" ** WRONG\n");
			fflush (stdout);
			}

		/* number of processed */
       		(statep->NEvent)++;	

       		/* aborted, cancel msgs and put back into queue */
       		if ( CurEvent->Processed == ABORT )
			{ 
  			struct EvState *tmpState;

                        if ( show_gvt >= 2 ) /* PHERE */
                                 {
                                 printf("[ABRT]: * aborted regular event "
                                         "uid=%d rcv_time=%lf\n",
                          	 CurEvent->LP, CurEvent->Msg.RcvTS );
                                 }


       			(statep->NAbort)++;	
       			CurEvent->Processed = FALSE;

			/* ran out of buffers exit */
			if ( same_gvt == 100 )
				{
				error ("+** ABORT at %.4f on %d cnt=%d sz=%d\n",
					CurEvent->Msg.RcvTS, statep->ID,
					same_gvt, MyState->Qsize );
				}
			else
				{
		                if ( show_noprint == 0 )
					{
#if 0
					printf ("ABORT at %.4f on %d\n",
					CurEvent->Msg.RcvTS, 
					statep->ID );
#endif
					}
				}


			/* if it scheduled multiple events */
  			for ( c = CurEvent->CauseList; c != NULL; c = NCause ) 
				{
     				CANCEL;
  				}

  			CurEvent->CauseList = NULL;
       			/* set up for state restore */
			tmpState = statep->RBSV;
	  		statep->RBSV = CurEvent->SV;
	       		CurEvent->SV = tmpState;
	  		statep->StateValid = FALSE;

	  		/* roll back log and incremental state */
               		if ( statep->IncrSave )
	    			Restore(CurEvent->LogPtr, statep);

			if ( CurEvent->RBFunc != NULL ) /* rolling back ABORT */
				{
               			GState[MyPE].CurState = statep;
       				GState[MyPE].CurEvent = CurEvent;
    				CurEvent->RBFunc ( 
					statep->state,
					CurEvent->Msg.Data,
					MyPE );
				}

	  		EnQueue(CurEvent,MyPE);
	  		CurEvent->in = 'c';
#if 0
	  		Terminate(MyPE, 7);
#endif
	  		MyState->TWMsg = NULL;
	  		MyState->NewBuf = NULL;
	  		CurEvent 
				= GState[MyPE].CurEvent 
				= NULL;

	  		statep 
				= CurState 
				= GState[MyPE].CurState 
				= NULL;

	  		break;

       			} /* ABORTED event */
	      	else 
			{  /* successfully processed event */

	       		CurEvent->Processed = TRUE;

       			/* thread into the event list */
 			CurEvent->in = 'e';
  			CurEvent->Prev = statep->Last;
       			CurEvent->Next = NULL;
 			statep->Last->Next = CurEvent;
       			statep->Last = CurEvent;
  			statep->LVT = CurEvent->Msg.RcvTS; 

	  		PUT_FREE_E(CurEvent,CurEvent->SrcPE);
			} /* successfully processed an event */
		} /* normal processing BLOCK -- IO1 continued --  */
		} /* successful dequeue found an event */
	} /* batch loop */

   ck4_probe("procEv end")
}

/*-----------------------------------------------------------------------------
**  
**
**
**----------------------------------------------------------------------------*/
void cntFree( int pe, int MyPE, char *fn )
  {
  struct Event *ev;
  struct Event *lev;
  struct PEState *MyState;             /* state for this PE */
  struct Event **_first;
  struct Event **_last;
  int cnt = 0;

  if ( show_gvt <= 0 )
	return;

  MyState =  &GState[MyPE];

  _first = &(MyState->PFree[(pe)]);
  _last = &(MyState->FLast[(pe)]);

  (ev) = *_first;
  (lev) = *_last;
  while ( (ev != NULL) && (ev != lev) )
        {
	ev = ev->FNext;
        cnt ++;
        }
  printf("PFree events=%3d on MyPE=%d for pe=%d fn=%s\n", cnt, MyPE, pe, fn );
  }


/*-----------------------------------------------------------------------------
**  Scheduler()
**
**  DESCRIPTION: Task scheduler running on each processor in parallel,
**      called from  RunSim()
**
**----------------------------------------------------------------------------*/
#if( defined(SUNMP) || defined(SUN4SOL2))
   void * Scheduler ( void * a )
#endif /* SUNMP || SUN4SOL2 */

#if( defined(SGI5) || defined(SGIMP64) || defined(SGI64) )
   void Scheduler( void )
#endif /* SGI5 || SGIMP64 || SGI64 */

   {
   int cpe;
   int RBDist;
   struct Event *cp;
   BOOLEAN idling;

   int MyPE;
   struct Event *CurEvent=NULL;      /* current event that */
   struct LPState *CurState=NULL;    /* current LP state */
   int PEStateMem;
   int gvtCounter;
   int i = 0;
   int LoopCounter = 0;
   struct Event *p, *q, *c, *nextq, *NCause, *pNext, *pPrev;
   struct Event *mp;
   TWtimer printTimer;		 /* timer to print  GVT */
   struct PEState * MyState;     /* state for this PE */
   TWTime RcvTS;
   struct LPState *statep;


   /* --------- begin.initializations ------------------------------------- */

#if( defined(SUNMP) || defined(SUN4SOL2))
   struct args *a1;
   a1 = (struct args *)a;
   MyPE = a1->id;

   TW_INIT_BARRIER_LOCAL(Go1,MyPE); /* powch:empty,sparc:barrier_init_local()*/
   TW_INIT_BARRIER_LOCAL(RF1,MyPE);
   TW_INIT_BARRIER_LOCAL(RF2,MyPE);

#endif /* SUNMP || SUN4SOL2 */

#if( defined(SGI5) || defined(SGIMP64) || defined(SGI64) )
   MyPE = m_get_myid();
#endif /* SGI5 || SGIMP64 || SGI64 */

   TW_BARRIER_SYNC(Go1, MyPE); 

   if ( MyPE == 0 )
        {
        TWLRnow(mh_init[5]);
        }

   MyState = &GState[MyPE];

   /* initialize PE data structures */

   InitPE ( MyPE );

   TW_BARRIER_SYNC(Go1, MyPE);		/* to print out consistent BufAlloc */

   if ( MyPE == 0 )
        {
        TWLRnow(mh_init[6]);
        }


   if ( MyPE == 0 )
	{
	if( show_noprint == 0 )
		{
      		printf(" Calling IProc procedures ...");
      		fflush(stdout);
		}
	else if( show_noprint == 1 )
		{
      		PrintBufAlloc();
		}
   	}

   /* wait for a connection if any, if not a connection IIOflag will be reset */
   /* after barrier we know if we are connected -- IProc needs to know that */

   TW_BARRIER_SYNC(Go1, MyPE);	/* barrier synch. to start */

   if ( MyPE == 0 )
        {
        TWLRnow(mh_init[7]);
        }


   /* ----------- start each LP ------------------------------------------- */

   /* start un-cloned LPs */
   for (i = 0; i < MyState->LPsOnPE; i++) 
	{
	statep = MyState->CList[i];

      	CurEvent = statep->EList.Next;

      	CurState = statep;
      	GState[MyPE].CurEvent = CurEvent;
      	GState[MyPE].CurState = CurState;

      	/* initial events */
      	if ( statep->IncrSave )
        	CurEvent->LogPtr = statep->LogHead;

      	/* First event should work with the common area for
        ** state. Subsequent event will copy common area to
        **  it's state vector */

      	if ( statep->IProcMsg != NULL && statep->ID < TWnlp )
		{
		statep->IProcMsg(statep->state, MyPE );
		}

      	if ( CurEvent->Processed == ABORT ) /* IProc */
		{
		printf("Insufficient event memory in IProc (LP %d) check "
		"TWMemMap, amount of event memory\n", statep->ID);
		error("Add memory in .tw-config or check TWMemMap\n");
      		}
      	CurEvent->Processed = TRUE;
      	CurEvent->in = 'e';

      	PUT_FREE_E(CurEvent,CurEvent->SrcPE); 

	/* statep->state is still NULL. The application must
	** have set TWLP[].State. Copy from there. */
        statep->state = TWLP[statep->ID].State;
   	} /* endfor start each LP */

   	if ( (MyState->LPsOnPE > 1) && (MyState->CList[1]->EList.Next==NULL) ) 
		{
		printf("**** 1. ELIST is NULL after IPROC i=%d\n",i ); 
	 	error ("Internal error: ELIST\n");
		}

   TW_BARRIER_SYNC(Go1, MyPE);		/* barrier synch. after IProc */

#ifdef MITRE_CUSTOM
   TWStartAppl(MyPE);  /* user-supplied; do AFTER barrier sync */
#endif

   if ( MyPE == 0 )
        {
        TWLRnow(mh_init[8]);
        }

   /* allocate space for state vectors */
   PEStateMem = (MyPE == TWnpe) ? RTGPEStateMem : GPEStateMem ;
   AllocateStateVector(PEStateMem, MyPE);

   if ( MyPE == 0 ) 
	{
     	if( !(TWdebug & DEBUG_SILENT)  && (show_noprint == 0) )
		{
       		printf("(%8.3f MB) ...  done\n\n",
			((float) PEStateMem) / 1000000);
		/*
		** printf("* size( Event, LPState, PEState )->( %d, %d, %d )\n",
	 	** sizeof(struct Event), sizeof(struct LPState), 
		** 	sizeof(struct PEState));
		*/
       		fflush(stdout);
    		}
   	}

   TW_BARRIER_SYNC(Go1, MyPE);		/* last barrier synch. after IProc */

   if ( MyPE == 0 )
        {
        TWLRnow(mh_init[9]);
        }


   /* No state yet for the first event. Get one. */
   for ( i = 0; i < MyState->LPsOnPE; i++ )
	{
      	statep = MyState->CList[i];
      	if (statep->EList.Next->SV != NULL) continue;
      	if (MyState->PState == NULL)
	 	error ("Internal error: too few state vectors\n");
      	statep->EList.Next->SV = MyState->PState;
      	MyState->PState = MyState->PState->next;
      	if (MyState->PState == NULL) error(" state memory too small");
   	}

  /* **** adding a barrier MH_HERE ********  */
  if ( MyPE == 0 )
	{
	if (0)
   		printf("\n" );
	}

  TW_BARRIER_SYNC(Go1, MyPE);          

  if ( gtw_debug && (MyPE == 0) )
	printf( "\tLP: entering scheduling loop\n\n" );

   TWnow(printTimer);

   if ( MyPE == 0 )
        {
        TWLRnow(mh_init[10]);
        }

#ifndef NEWTIMER
   TWnow(MyState->StartTime);
#else
   TWLRnow ( MyState->StartTime );
   TWnow ( MyState->fStartTime );
#endif

   gvtCounter = 0;

   /* ---------------------------------------------------------------
   ** Scheduler Loop Heart of the Kernel
   ** --------------------------------------------------------------*/

   if( !(TWdebug & DEBUG_SILENT)  && (show_noprint == 0) )
	{
	int cnt=0;
	/*
	** for (cnt=0; cnt<TWnpe; cnt++)
	** 	 cntFree( cnt, MyPE, "before scheduling" );
	*/

   	printf(" Scheduling begins on PE=%d ...\n", MyPE );
	}

   while ( TRUE ) 
	{		/* while there are unprocessed events */
	int FailedGVT=0;

        /*****************************************************************/
        /******* GVT computation Check ***********************************/
        /*****************************************************************/
	if ( MyPE == 0 ) 
		{
	 	gvtCounter++;
	 	if ( gvtCounter == GVTPeriod ) 
			{
	   		gvtCounter = 0;

			/* start GVT computation if none in progress */
	   		if ( GVTflag == -TWnpe ) 
				{
               			GVTflag = TWnpe; 
	       			MyState->GVT_Progress = CALC_GVT;
				FailedGVT = 0;
	   			} 
			else 
				{
	     			FailedGVT++;
	     			if (FailedGVT > 100)
	       	   		   error ("GVT failed GVTflag=%d\n", GVTflag );
	   			}
         		} /* endif start gvt computation */
		
      		} /* endif MyPE == 0 */
	else 
		{	
		/* Check if GVT computation has been started */
       		if ( ( GVTflag > 0 ) && ( ( MyState->GVT_Progress ) == NORMAL) )
			{
    			MyState->GVT_Progress = CALC_GVT;
        		}
      		} /* else if */

	/*****************************************************************/
	/* Process Message Queue *****************************************/
	/*****************************************************************/
	if ( GState[MyPE].MsgQ != NULL ) 
		ProcessMsgQ( MyPE );

	/*****************************************************************/
	/* Process Cancel Queue ******************************************/
	/*****************************************************************/
        if ( GState[MyPE].CanQ != NULL ) 
		ProcessCanQ(MyPE);

 	/****************************************************************/
        /* Compute local GVT ********************************************/
        /****************************************************************/
      	if ( (MyState->GVT_Progress) == CALC_GVT ) 
		{
	 	Local_Min[MyPE] = min(Minim_Q(MyPE), MyState->TransMsgTS);
                Local_Min[MyPE] = min(IOBlockedMinim_Q(MyPE), Local_Min[MyPE]);

	 	TW_LOCK(Lck_GVTflag);
	 	if ( GVTflag == 1 ) 		/* if last to compute GVT */
			{	
	    		TWTime tmpGVT;
	    		tmpGVT = Local_Min[0];
	    		for ( i = 1; i < TWnpe; i++ ) 
				{
	       			if ( tmpGVT > Local_Min[i] ) 
					tmpGVT = Local_Min[i];
	    			}
    			GVT = tmpGVT;	/* broadcast new value */
	    		GVTflag = -1;	/* signal new value */
	    		TW_UNLOCK(Lck_GVTflag);	
	    		MyState->TransMsgTS = MAXFLOAT;

			ck_GVTdecrease

			if ( MyState->GVT == GVT )
				{
				/* ring bell */
#ifdef MH_PRINT
			     	fprintf ( stderr, "%c", 7); 
				fprintf ( stderr,  "GVT=%.4f\n", GVT );
				fflush ( stderr );
				printf ( "--*--%.4f\n", GVT );
#endif
				same_gvt++;
				}
			else
				{
				same_gvt=0;
				}
				
			if ( show_gvt > 0 )
				{
				ClockTicks gvt_elapsed;
				TWtimer at_gvt;
				TWLRnow ( at_gvt );
				gvt_elapsed = TWseconds ( at_gvt ) 
					- TWseconds( GState[MyPE].StartTime ) ;
				printf ( "[GTW]: GVT = %.4lf  RT = %.4lf\n", 
					GVT, gvt_elapsed  );
				fflush (stdout);
				}
	    		MyState->GVT = GVT;

	    		if (GVT > ENDTIME) 
				{
				Terminate(MyPE, 1);
				fflush(stdout);
	     			break;	
				}

	    		FossilCollect(MyPE);
	 		MyState->GVT_Progress = NORMAL;
	 		} /* end if last to compute local GVT */

	 	else 
			{ /* not last then have local_gvt but need to wait */
	    		GVTflag--;
	    		TW_UNLOCK(Lck_GVTflag);
	    		(MyState->GVT_Progress) = WAIT_GVT;
	 		}

      		} /* end if in CALC_GVT state */

      else if ( (MyState->GVT_Progress == WAIT_GVT) && (GVTflag < 0))
                { /* if new GVT is available and waiting  */
                ck_GVTdecrease
                MyState->GVT = GVT;
                MyState->TransMsgTS = MAXFLOAT;

                TW_LOCK(Lck_GVTflag);
                    GVTflag--;
                TW_UNLOCK(Lck_GVTflag);

                /* terminate */
                if (GVT > ENDTIME)
                        {
                        Terminate(MyPE,2);
                        fflush(stdout);
                        break;
                        }
                FossilCollect(MyPE);
                MyState->GVT_Progress = NORMAL;
		/* do not set MyState->TransMsgTS  MAXFLOAT here */
                } /* else if waiting for Global GVT */


        /****************************************************************/
        /* Process Events ***********************************************/
        /****************************************************************/
        ProcessEvents ( MyPE );

	} /* outermost while loop (while simulation not done) */

    if ( IO_NBListDeQueue(MyPE, MAXFLOAT) != NULL)
        _TWNonBlockingProcessing( 'x', MyPE, MyState->GVT );
    if ( IOBlockedDeQueue(NULL, MyPE) != NULL)
        {
        printf("**BL list is NOT empty on exiting while loop MyPE = %d\n",MyPE);
        printf("agressively process message PE=%d\n", MyPE);
        _TWBlockingProcessing(MyPE, MyState->GVT );
        }

#ifndef NEWTIMER
   TWnow(MyState->EndTime);
#else
   TWLRnow(MyState->EndTime);
#endif

   if ( MyPE == 0 )
        {
        TWLRnow(mh_init[11]);
        }

   /* get one more last sample */
#if 0
   if ( show_gvt_dump > 0 )
        ipdes_stat_gvt ( MyPE );
#endif

   TW_BARRIER_SYNC(Go1, MyPE);	
   } /* end scheduler function */ 

/*-----------------------------------------------------------------------------
**   checkStraggler ()
**
**  DESCRIPTION: 
**
**----------------------------------------------------------------------------*/
int checkStraggler( int MyPE )
   {
   struct Event *e;

   /* TW_LOCK(GState[MyPE].Lck_MsgQ); */
   for ( e = GState[MyPE].MsgQ; e != NULL; e = e->Next )
      	{
      	if ( e->LP == GState[MyPE].CurState->ID 
		&& e->Msg.RcvTS < GState[MyPE].CurEvent->Msg.RcvTS )
		{
         	/* TW_UNLOCK(GState[MyPE].Lck_MsgQ); */
         	return( TRUE );
         	}
      	}
   /* TW_UNLOCK ( GState[MyPE].Lck_MsgQ ); */
   return ( FALSE );
   }

/*-----------------------------------------------------------------------------
**  print_msg_q ()
**
**  DESCRIPTION: 
**
**----------------------------------------------------------------------------*/
void print_msg_q (struct Event *ev)
{
  while(ev)
	{
    	printf("%x %.2f \n", ev, ev->Msg.RcvTS);
    	ev = ev->Next;
 	}
}
