/******************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.5                             ***
***         Copyright 1997 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/


/*****************************************************************************
**     Various utility procedures, mostly for debugging
******************************************************************************/
#include <stdio.h>
#include <stdarg.h>

#include "gtw_glob.h"
#include "gtw_insim.h"

/* forward reference */
void quit( int );
void PrintState( void );
void PrintEList( struct LPState *st, TWTime upto );


/* ---------------------------------------------------------------------------
** show_date()
**
** DESCRIPTION:  
**
**---------------------------------------------------------------------------*/
void show_date( void )
{
  char day[10];
  char month[10];
  char date[10];
  char the_time[10];
  char zone[4];
  char year[10];

  FILE *the_date = popen("date", "r");
  fscanf( the_date, "%s %s %s %s %s %s\n",  
        day, month, date, the_time, zone, year );
  pclose(the_date);
  printf ( "%s %s %s %s %s %s", day, month, date, the_time, zone, year );
}       


/* ---------------------------------------------------------------------------
** error()
**
** DESCRIPTION:  Prints out an error message and quits -- argument list
**	is of variable length.
**
**---------------------------------------------------------------------------*/
void error( char *fmt, ... )
{
   va_list argp;
   va_start(argp, fmt);
   vfprintf(stdout, fmt, argp);
   va_end(argp);
   fprintf(stdout, "\n\n");
   fflush(stdout);
   fprintf(stderr, "%c %c %c", 7, 7, 7); /* ring the bell */
   fflush(stdout);
   exit(1);
}

/* ---------------------------------------------------------------------------
** warn()
**
** DESCRIPTION:  prints out a warning without exiting
**
**---------------------------------------------------------------------------*/
void warn( char *fmt,... )
{
   va_list argp;
   va_start(argp, fmt);
   vfprintf(stdout, fmt, argp);
   va_end(argp);
   fprintf(stdout, "\n");
   fflush(stdout);
}

/* ---------------------------------------------------------------------------
** quit()
**
** DESCRIPTION:  Bails out.
**
**---------------------------------------------------------------------------*/
void quit( int MyPE )
{
   int i;

   printf("*** GTW QUIT : PE = %d \n", MyPE);

   fflush(stdout); 
   fflush(stderr);

   exit(1);
}

/* =========================================================================== 
**	The following are various debugging utilities
**========================================================================== */

/* ---------------------------------------------------------------------------
** PrintState()
**
** DESCRIPTION:  prints out the state of the system 
**
**---------------------------------------------------------------------------*/
void PrintState( void )
{
   int i, j;
   struct Event *ev;

   printf("\n\n -ve msgs in transit\n");
   for (i = 0; i < TWnpe; i++) {
      for (ev = GState[i].CanQ; ev != NULL; ev = ev->NextCause)
	 printf("%f-%d  ", ev->Msg.RcvTS, i);
   }
   printf("\n +ve msgs in transit\n");
   for (i = 0; i < TWnpe; i++) {
      for (ev = GState[i].MsgQ; ev != NULL; ev = ev->Next)
	 printf("%f-%d  ", ev->Msg.RcvTS, i);
   }
   printf("\n");

   for (i = 0; i < TWnpe; i++) {
      for (j = 0; j < GState[i].LPsOnPE; j++)
	 PrintEList(GState[i].CList[j], -20.0);
   }
   printf("\n");
   printf(" LVTs on PEs\n");
   for (i = 0; i < TWnpe; i++)
      printf("%d:%f  ", i, GState[i].PELVT);
   fflush(stdout);
   printf("\n");
   printf(" NumFCs on PEs\n");
   for (i = 0; i < TWnpe; i++)
      printf("%d:%i  ", i, GState[i].NumFC);
   fflush(stdout);
}


/* ---------------------------------------------------------------------------
** CountMem()
**
** DESCRIPTION:   Counts the amount of memory in use in buffers, no longer
**	needed, but dept for statistics
**
**---------------------------------------------------------------------------*/
int CountMem( void )
{
   int i, cnt = 0;
   struct Event *p;

   for (i = 0; i < TWnlp; i++)
      for (p = LPST(i)->EList.Next; p->Msg.RcvTS != MAXFLOAT; p = p->Next)
	 cnt++;

   /* for (i = 0; i < TWnpe; i++) for (p = GState[i].PFree;
      p != GState[i].FLast; p = p->Next) cnt++; */
   for (p = PFree; p != FLast; p = p->Next)
      cnt++;

   for (i = 0; i < TWnpe; i++)
      for (p = GState[i].MsgQ; p != NULL; p = p->Next)
	 cnt++;
   return (cnt);
}

/* ---------------------------------------------------------------------------
** PrintEList()
**
** DESCRIPTION:  prints out the event list of the specified LP 
**
**---------------------------------------------------------------------------*/
void PrintEList(struct LPState *st, TWTime upto)
{
   struct Event *p;
   int i;
   BOOLEAN flag;

   flag = FALSE;
   i = 0;
   printf("LP %d:\n", st->EList.LP);
   for (p = st->EList.Next; p != NULL; p = p->Next) {
      if (p->Msg.RcvTS >= upto)
	 flag = TRUE;
      if (!flag)
	 continue;
      printf("%f-%d", p->Msg.RcvTS, p->Processed);
      if (p == st->Last)
	 printf("L ");
      else
	 printf("  ");
      i++;
      if (i % 5 == 0)
	 printf("\n");
   }
   printf("\n");
}


/* ---------------------------------------------------------------------------
** system_consistency_check()
**
** DESCRIPTION:  
**
**---------------------------------------------------------------------------*/
void system_consistency_check(void)
{
   int i;
   struct Event *ev;
   int pe, lp;
   int count_free = 0, count_uncommitted = 0;
   int total_free = 0, count_global = 0, total_uncommitted = 0, total_Q = 0;
   int counter_elist = 0;
   BOOLEAN error = 0;

   for (pe = 0; pe < TWnpe;
	pe++) {

      count_free = 0;
      count_uncommitted = 0;
      counter_elist = 0;

      /* Count in free list */
      for (i=0; i<TWnpe; i++)
         for (ev = GState[pe].PFree[i]; ev != NULL; ev = ev->Next)
	    count_free++;
      total_free += count_free;
      if (GState[pe].NumLocalPool != count_free) {
	 error = TRUE;
      }
      printf("PE %d: count free is %d, and counter is %d\n", pe, count_free,
	     GState[pe].NumLocalPool);

      /* Count in uncommitted list */
      for (lp = 0; lp < GState[pe].LPsOnPE; lp++) {
	 for (ev = GState[pe].CList[lp]->EList.Next;
	      ev != GState[pe].CList[lp]->Last;
	      ev = ev->Next)
	    count_uncommitted++;
	 count_uncommitted++;
      }

      total_uncommitted += count_uncommitted;
      /* printf("%d  in queue %d\n", pe, lps->QSize); */
      /* total_Q += lps->QSize; */
   }

   /* check global list */
   for (ev = PFree; ev != FLast; ev = ev->Next)
      count_global++;
   count_global++;
   if (NumGlobalPool != count_global) {
      error = TRUE;
   }
   printf("Global list is %d, counter is %d\n", count_global, NumGlobalPool);

   if (!error) {
      printf("Buffer counters are consistent\n");
   }
   printf("local %d, E List %d, Q %d, global %d; total %d \n",
     total_free, total_uncommitted, total_Q, count_global,
	  total_free + total_uncommitted + total_Q + count_global);

}

/* ---------------------------------------------------------------------------
** print_state()
**
** DESCRIPTION:  
**
**---------------------------------------------------------------------------*/
void print_state( void )
{

   struct PEState *lps;
   int count = 0, count1 = 0, count2 = 0, count_Q = 0, j;

   /* call cancel back here */
   /* printf("%d %.2f Initiate cancelback, or allocate more memory\n",
	  MyPE, GState[0].PELVT); */

   for (j = 0; j < TWnpe; j++) {
      count = 0;
      lps = &GState[j];
      /* lps->QSize */
      count1 += count;
      count2 += lps->NumLocalPool;
      /* count_Q += lps->QSize; */
   }
   printf("Uncomitted %d, Free %d, CalQ %d, Global %d, total %d\n",
	  count1, count2, count_Q, NumGlobalPool,
	  count1 + count2 + count_Q + NumGlobalPool);
}
