/******************************************************************************
*******************************************************************************
***                                                                         ***
***                   Georgia Tech Time Warp Version 2.7                    ***
***         Copyright 1995-2000 by Georgia Institute of Technology          ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/

#include <stdio.h>
#include "gtw_glob.h"
#include "gtw_debug.h"
#include "gtw_insim.h"

extern int largestQsize[]; 	/* calQ.c */
extern int show_noprint;
extern int print_tables[10];
extern int GPEStateMem;

extern TWtimer begin_time;
extern TWtimer end_time;

int NMem;

/* ---------------------------------------------------------------------------
** WrapUpAppl()
**
** DESCRIPTION:  Calls FProc() sequentially, if cloned need to see
**	if they are active
**
**---------------------------------------------------------------------------*/
void WrapUpAppl(void)
{
   int i;
   struct LPState *CurState;

   if ( show_noprint == 0 )
       	printf("\n Calling Wrapup...\n");

   for ( i = 0; i < TWnlp; i++ )
	{
     	GState[LP2PE(i)].CurState = CurState = LPST(i);
      	if ( CurState->FProcMsg != NULL )
		{
		CurState->FProcMsg(CurState->state, LP2PE(i));
		}
	else
		{
		printf( "WrapUpAppl: LP(%d)'s FProcMsg == NULL\n", i);
		fflush(stdout);
		}
    	}
 }

/* ---------------------------------------------------------------------------
** WrapUpSim ()
**
** DESCRIPTION:  Display statistics to wrap up simulation
**
**---------------------------------------------------------------------------*/
void WrapUpSim(void)
{
   int i, j;
   struct LPState *p;
   struct LPState *q;

   int TNEvent, TNCancel, TNStraggler, TNRBStraggler, TNSecondary, TNAbort;
   int TNArt, TNRBArt, TNret, TNtaken;
   int TNRBSecondary, TLongestRB;

   int
	cTNEvent[GTW_MAX_CLONES]={0},
	cTNCancel[GTW_MAX_CLONES]={0},
	cTNStraggler[GTW_MAX_CLONES]={0},
	cTNRBStraggler[GTW_MAX_CLONES]={0},
	cTNSecondary[GTW_MAX_CLONES]={0},
	cTNAbort[GTW_MAX_CLONES]={0},
	cTNArt[GTW_MAX_CLONES]={0},
	cTNRBArt[GTW_MAX_CLONES]={0},
	cTNret[GTW_MAX_CLONES]={0},
	cTNRBSecondary[GTW_MAX_CLONES]={0},
	cTLongestRB[GTW_MAX_CLONES]={0},
	cTNtaken[GTW_MAX_CLONES]={0};

   ClockTicks FCtime[3], MaxFCOverhead, AvgFCOverhead;
   ClockTicks Exectime;

   /* BEGIN IO1 */
   int TNOPEvent, TNEXEvent, TNNBEvent, TNBLEvent;
   int TNOPAbort, TNEXAbort, TNNBAbort, TNBLAbort;
   int TNOPStraggler,TNEXStraggler,TNNBStraggler,TNBLStraggler,TNRBOPStraggler;
   int TNOPCancel;
   int TNRBOPSecondary;
   int TNRBDelays,  TNRBDelaysR;  
   int TNBBLEnQ, TNBLEnQ; 
   int TNCanOPCQ, TNCanEXCQ, TNCanNBCQ, TNCanBLCQ;
   int TNCanNBQ, TNCanOQ,TNRCanBQ, TNBCanBQ, TNOPEnQ, TNNBEnQ;
   /* blockings = NBBLEnQ + NBLEnQ */
   /* END IO1 */

   double Eff, AvgRBDist;
   
   ClockTicks ProcTime, OverHead, SendTime;
   ClockTicks AvgRBTime, AvgCanTime, MaxRBTime, MaxCanTime;

   ClockTicks MaxMsgQTime = 0, MaxDeqTime = 0, MaxInitTime = 0;
   ClockTicks AvgMsgQTime = 0, AvgDeqTime = 0, AvgInitTime = 0;
   ClockTicks AvgTotTime = 0, MaxTotTime = 0;

   /* BEGIN IO1 */
   TNEvent = TNOPEvent = TNEXEvent = TNNBEvent = TNBLEvent = 0;
   TNAbort = TNOPAbort = TNEXAbort = TNNBAbort = TNBLAbort = 0;
   TNCancel = TNOPCancel = 0;
   TNStraggler=TNOPStraggler=TNEXStraggler=TNNBStraggler= TNBLStraggler = 0;
   TNRBStraggler = TNRBOPStraggler= 0;
   TNRBSecondary = TNRBOPSecondary = 0;
   TNRBDelays = TNRBDelaysR = 0;
   TNBBLEnQ = TNBLEnQ = 0;
   /* END IO1 */

   TNSecondary = 0;
   TLongestRB = 0;
   TNArt = 0;
   TNRBArt = 0;
   TNret = TNtaken = 0;

   ProcTime = OverHead = SendTime = MaxFCOverhead = AvgFCOverhead = 0.0;
   AvgRBTime = AvgCanTime = MaxRBTime = MaxCanTime = 0.0;

   for (j = 0; j < 3; j++) FCtime[j] = 0;
   Exectime = 0;

   if (TWstat > 1) 
	{
      	printf("*** Time Warp Statistics\n");
      	printf("%3s %7s %6s %6s %6s %6s %6s %6s %6s %6s %6s %3s %6s \n",
	     	" LP", "Events", "MS/Ev", "Aborts",
		"Cancel", "NStrgl", "RB:Strl", "NSecnd", "RB:Sec",
	     	"NArt", "RB:Art", "LRB", "Net Evt");
   	}

   for (i = 0; i < TWnlp; i++) 
	{
      	p = LPST(i);
      	if (TWstat > 1) 
		{
	      	printf("%3d %7d %6.3f %6d %6d %6d %6d %6d %6d %6d %3d "
			"%6d %6d \n",
		   i, p->NEvent,
		   (p->NEvent == 0) ? 0.0 
		    : (TWseconds(p->EvExecTime) * 1000 / (double) (p->NEvent)),
			p->NAbort, p->NCancel, p->NStraggler, p->NRBStraggler,
			p->NSecondary, p->NRBSecondary, p->NArt, p->NRBArt,
			p->LongestRB,
			p->NEvent - p->NCancel - p->NRBStraggler
			   - p->NRBSecondary - p->NRBArt - p->NAbort);
      		}


/* BEGIN IO1 */
      	TNBBLEnQ += (p->NBBLEnQ);
      	TNBLEnQ  += (p->NBLEnQ);

      	TNRBDelays  +=  (p->NRBDelays);
      	TNRBDelaysR +=  (p->NRBDelaysR);

      	TNOPEvent += (p->NOPEvent);
      	TNEXEvent += (p->NEXEvent);
      	TNNBEvent += (p->NNBEvent);
      	TNBLEvent += (p->NBLEvent);
      
      	TNOPAbort += (p->NOPAbort);
      	TNEXAbort += (p->NEXAbort);
      	TNNBAbort += (p->NNBAbort);
      	TNBLAbort  += (p->NBLAbort);

      	TNOPCancel += (p->NOPCancel);

      	TNOPStraggler += (p->NOPStraggler);
      	TNEXStraggler += (p->NEXStraggler);
      	TNNBStraggler += (p->NNBStraggler);
      	TNBLStraggler += (p->NBLStraggler);

      	TNRBOPStraggler += (p->NRBOPStraggler);

      	TNRBOPSecondary += (p->NRBOPSecondary);
/* END IO1 */

      	TNEvent += (p->NEvent);
      	TNCancel += (p->NCancel);
      	TNRBStraggler += (p->NRBStraggler);
      	TNRBSecondary += (p->NRBSecondary);
      	TNRBArt += (p->NRBArt);
      	TNAbort += (p->NAbort);

      	TNStraggler += (p->NStraggler);
      	TNSecondary += (p->NSecondary);
      	TNArt += (p->NArt);
      	TLongestRB = max(TLongestRB, p->LongestRB);

      	/* TSDist += (p->SDist); TNSDist += (p->NSDist);  */
      	ProcTime += TWseconds(p->EvExecTime);
   	} /* for each lp */

   for (i = 0; i < TWnpe; i++) 
	{

      	Exectime = max(Exectime, TWseconds(GState[i].EndTime)
				- TWseconds(GState[i].StartTime));

      	TNret += GState[i].Nret;
      	TNtaken += GState[i].Ntaken;
   	}

   i = TNEvent - TNCancel - TNRBStraggler - TNRBSecondary - TNRBArt - TNAbort;


   if( (!(TWdebug & DEBUG_SILENT) && (!show_noprint)) || (print_tables[3]==1) )
	{
     	printf("\n*** Summary Time Warp Statistics***\n\n");
	}

   if( (!(TWdebug & DEBUG_SILENT) && (show_noprint == 0) ) || 
	(print_tables[3]==1) )
	{
     	printf("%-23s %9d\n", "No. of PEs", TWnpe);
     	printf("%-23s %9d\n", "No. of LPs", TWnlp );

       /* BEGIN IO1 */
     	printf("%-23s %9d ( RG %d, OP %d, EX %d, NB %d, BL %d )\n", 
		"Events processed", 
		TNEvent, 
		TNEvent - (TNOPEvent + TNEXEvent + TNNBEvent + TNBLEvent),
		TNOPEvent, TNEXEvent, TNNBEvent, TNBLEvent
		);

     	printf("%-23s %9d (%d, %d)\n", "No of Blocked Events", 
		TNBBLEnQ + TNBLEnQ,
		TNBBLEnQ,
		TNBLEnQ
		);

    	printf("%-23s %9d ( RG %d, OP %d, EX %d, NB %d, BL %d )\n",
		"Events Aborted",
		TNAbort,
		TNAbort - (TNOPAbort + TNEXAbort + TNNBAbort + TNBLAbort),
		TNOPAbort, TNEXAbort, TNNBAbort, TNBLAbort
		);

    	printf("%-23s %9d\n",
		"Events RBArt",
		TNRBArt
		);

     	printf("%-23s %9d ( RG %d, OP %d )\n", 
		"Processed Events Cncl'd", 
		TNCancel,
		TNCancel - TNOPCancel,
		TNOPCancel
		);

     	printf("%-23s %9d ( RG %d, OP %d, EX %d, NB %d, BL %d )\n", 
		"Straggler Induced RBs", 
		TNStraggler,
		TNStraggler-(TNOPStraggler+TNEXStraggler
			+TNNBStraggler+TNBLStraggler),
		TNOPStraggler, TNEXStraggler, TNNBStraggler, TNBLStraggler
		);

     	printf("%-23s %9d ( RG %d, OP %d )\n", 
		"  Processed Events RB'd", 
		TNRBStraggler,
		TNRBStraggler - TNRBOPStraggler,
		TNRBOPStraggler
		);

     	printf("%-23s %9d\n", "Cncl. Evnt Induced RBs", TNSecondary);

     	printf("%-23s %9d ( RG %d, OP %d)\n", 
		"  Processed Events RB'd", 
		TNRBSecondary,
		TNRBSecondary - TNRBOPSecondary,
		TNRBOPSecondary
		);

     	printf("%-23s %9d (%d)\n", "  Delayed RB", 
		TNRBDelays,
		TNRBDelaysR
		);
        /* END IO1 */

     	printf("%-23s %9d\n", "Longest rollback", TLongestRB);
     	printf ("%-23s %9d \n",
	     "FossilCollect attempts", GState[0].NumFC);

     	printf ("%-23s %9d %9d\n",
	     "Nreturned and Ntaken", TNret, TNtaken);
     	printf("\n");

	{
        int k,tmp=0;
        for ( k = 0; k < TWnpe ; k++ )
                {
                if (largestQsize[k] > tmp )
                        tmp = largestQsize[k];
                }
        printf ("%-23s %9d\n",
             "Largest Calendar Size", tmp );
        }
     	printf("\n");

     	/* net events */
     	Eff =  ((double) i / (double) TNEvent) * 100.0;
     	printf("%-23s %9d (%6.2f %%)\n", "Net Events processed", i, Eff);
     	printf("%-23s %9d (%6.2f %%)\n", "Aborted Events", TNAbort,
	    ((double) TNAbort / (double) TNEvent) * 100.0);
     	printf("%-23s %9d          \n", "Number of rollbacks",
	    TNStraggler + TNSecondary + TNArt);
     	printf("%-23s %9d (%6.2f %%)\n", "Proc. Evnts RB'd",
	    TNRBStraggler + TNRBSecondary + TNRBArt,
	    ((double) (TNRBStraggler + TNRBSecondary + TNRBArt)
	     / (double) TNEvent) * 100.0);
     	printf("%-23s %9d (%6.2f %%)\n", "Proc. Evnts Cncl'd", TNCancel,
	    ((double) (TNCancel) / (double) TNEvent) * 100.0);
     	if ((TNStraggler + TNSecondary + TNArt) > 0)
		{
       		AvgRBDist = (double)(TNRBStraggler + TNRBSecondary + TNRBArt) 
			    / (double) (TNStraggler + TNSecondary + TNArt);
       		printf("%-23s %9.2f          \n", 
			"Average RB Distance", AvgRBDist);
     		}
     	printf("%-23s %11.2f        \n", "Event Rate (Evnts/Sec)",
		 (double) i / Exectime);
     	printf("\n");
     	printf("%-23s %11.2f        \n\n",
	    "Execution time (Sec)", Exectime);
   	}
   else if ( (show_noprint < 2) || (print_tables[3]==1) )
	 /* makes a table */
	{
     	printf("\n%2s %7s %14s %15s %15s %9s %8s\n",
	   "#PE", "#Events", "#Aborted", "#RB", "#Can", "ExecTime",  "EvRate");

     	printf("%2d",TWnpe); 
     	printf(" %7d(%5.2f)", i, ((double) i / (double) TNEvent) * 100.0);
     	printf(" %5d(%5.2f%%)", TNAbort,
	    ((double) TNAbort / (double) TNEvent) * 100.0);
     	printf("  %5d(%5.2f%%)",   TNRBStraggler + TNRBSecondary + TNRBArt,
	    ((double) (TNRBStraggler + TNRBSecondary + TNRBArt)
	     / (double) TNEvent) * 100.0);
     	printf("  %5d(%5.2f%%)", TNCancel,
	    ((double) (TNCancel) / (double) TNEvent) * 100.0);
     	printf(" %7.2f", Exectime);
     	printf(" %8.2f\n\n", (double) i/ Exectime);
   	}

if ( (print_tables[0] == 1) && (show_noprint != 0) )
   {
   extern int do_headers;

   if( do_headers == 1 )
	{
   	printf("\n# FC_attempts = %4d\n", GState[0].NumFC);
        printf("# -----------------------------------------------"
                        "---------------"
                        "---------------\n");
        printf("#P  #Events  %%Evts  #NtEvts  EvntRate   ExcTime  LPs"
		"      EMem      SMem"
                " pGVT\n");

        printf("#------------------------------------------------"
                        "---------------"
                        "---------------\n");
		
	}
   printf("%2d",TWnpe);
   printf("  %7d", TNEvent);
   printf("  %5.2f", (((double) i) / ( (double) TNEvent)) * 100.0);

   /* number events */
   printf("  %7d",i );

   /* Event rate */ 
   printf("  %8.2f", (double) i/ Exectime );

   printf("  %7.3f", Exectime );

   printf("  %3d", TWnlp);
   NMem = PEEventMem;
   printf("   %6d", NMem);
   printf("   %6d", GPEStateMem);
   printf(" %4d\n", GVTPeriod);
   if ( do_headers == 1 )
   	printf( "\n" );
   }

if( print_tables[4] == 1 )
	{
	extern cs_Cntr_ReportResultGTW( int TWnpe, int TNEvent, int NetEvents );
	cs_Cntr_ReportResultGTW( TWnpe, TNEvent, i );
	}

} /* end Wrapup */
